/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "devdefs.h"
#include "pzdefs.h"
#include "sperror.h"
#include "util.h"
#include "outdata.h"
#include "cktext.h"

#ifdef __STDC__
static int PZinit(CKTcircuit*);
static int PZpost(CKTcircuit*);
#else
static int PZinit();
static int PZpost();
#endif

#define DEBUG if (0)


/* ARGSUSED */
int
PZan(cktp,reset)

GENERIC *cktp;
int reset;
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    PZAN *pzan = (PZAN *) ckt->CKTcurJob;
    int error;

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "PZ analysis not possible with Josephson junctions", NULL);
            return (OK);
    }
    error = PZinit(ckt);
    if (error != OK)
        return error;

    /* Calculate small signal parameters at the operating point */
    error = CKTop(ckt,
        MODEDCOP | MODEINITJCT,
        MODEDCOP | MODEINITFLOAT,
        ckt->CKTdcMaxIter);
    if (error)
        return (error);

    ckt->CKTmode = MODEDCOP | MODEINITSMSIG;
    error = CKTload(ckt);
    if (error)
        return (error);

    if (pzan->PZwhich & PZ_DO_POLES) {
        error = CKTpzSetup(ckt, PZ_DO_POLES);
        if (error != OK)
            return (error);
        error = CKTpzFindZeros(ckt,
            (GENERIC**)&pzan->PZpoleList, &pzan->PZnPoles);
        if (error != OK)
            return (error);
    }

    if (pzan->PZwhich & PZ_DO_ZEROS) {
        error = CKTpzSetup(ckt, PZ_DO_ZEROS);
        if (error != OK)
            return (error);
        error = CKTpzFindZeros(ckt,
            (GENERIC**)&pzan->PZzeroList, &pzan->PZnZeros);
        if (error != OK)
            return (error);
    }

    return PZpost(ckt);
}


static int
PZinit(ckt)

/*
 * Perform error checking
 */
CKTcircuit *ckt;
{
    PZAN *pzan = (PZAN *) ckt->CKTcurJob;
    int i;

    if ((i = CKTtypelook("transmission line")) != -1 &&
        ckt->CKThead[i] != NULL)
        ERROR(E_XMISSIONLINE, "Transmission lines not supported")

    pzan->PZpoleList = (PZtrial *) NULL;
    pzan->PZzeroList = (PZtrial *) NULL;
    pzan->PZnPoles = 0;
    pzan->PZnZeros = 0;

    if (pzan->PZin_pos == pzan->PZin_neg)
        ERROR(E_SHORT, "Input is shorted")

    if (pzan->PZout_pos == pzan->PZout_neg)
        ERROR(E_SHORT, "Output is shorted")

    if (pzan->PZin_pos == pzan->PZout_pos
            && pzan->PZin_neg == pzan->PZout_neg
            && pzan->PZinput_type == PZ_IN_VOL)
        ERROR(E_INISOUT, "Transfer function is unity")
    else
        if (pzan->PZin_pos == pzan->PZout_neg
                && pzan->PZin_neg == pzan->PZout_pos
                && pzan->PZinput_type == PZ_IN_VOL)
            ERROR(E_INISOUT, "Transfer function is -1")

    return (OK);
}


static int
PZpost(ckt)

/*
 * PZpost  Post-processing of the pole-zero analysis results
 */
CKTcircuit *ckt;
{
    PZAN      *pzan = (PZAN *) ckt->CKTcurJob;
    GENERIC   *pzPlotPtr = NULL; /* the plot pointer for front end */
    IFcomplex *out_list;
    IFvalue   outData;    /* output variable (points to out_list) */
    IFuid     *namelist;
    PZtrial   *root;
    char      name[50];
    int       i, j;
    struct sOUTdata outd;

    namelist = (IFuid *) MALLOC((pzan->PZnPoles
        + pzan->PZnZeros)*sizeof(IFuid));
    out_list = (IFcomplex *)MALLOC((pzan->PZnPoles
        + pzan->PZnZeros)*sizeof(IFcomplex));

    j = 0;
    for (i = 0; i < pzan->PZnPoles; i++) {
        sprintf(name, "pole(%-u)", i+1);
        (*(SPfrontEnd->IFnewUid))(ckt,&(namelist[j++]),(IFuid)NULL,
            name,UID_OTHER,(GENERIC **)NULL);
    }
    for (i = 0; i < pzan->PZnZeros; i++) {
        sprintf(name, "zero(%-u)", i+1);
        (*(SPfrontEnd->IFnewUid))(ckt,&(namelist[j++]),(IFuid)NULL,
            name,UID_OTHER,(GENERIC **)NULL);
    }

    outd.circuitPtr = (GENERIC *)ckt;
    outd.analysisPtr = (GENERIC*)pzan;
    outd.analName = pzan->JOBname;
    outd.refName = (IFuid)NULL;
    outd.refType = (int)0;
    outd.numNames = pzan->PZnPoles + pzan->PZnZeros;
    outd.dataNames = namelist;
    outd.dataType = IF_COMPLEX;
    outd.plotPtr = &pzPlotPtr;
    outd.numPts = 1;
    outd.initValue = 0;
    outd.finalValue = 0;
    outd.step = 0;
    (*SPfrontEnd->OUTbeginPlot)((GENERIC*)&outd);

    j = 0;
    if (pzan->PZnPoles > 0) {
        for (root = pzan->PZpoleList; root != NULL; root = root->next) {
            for (i = 0; i < root->multiplicity; i++) {
                out_list[j].real = root->s.real;
                out_list[j].imag = root->s.imag;
                j += 1;
                if (root->s.imag != 0.0) {
                    out_list[j].real = root->s.real;
                    out_list[j].imag = -root->s.imag;
                    j += 1;
                }
            }
            DEBUG printf("LIST pole: (%g,%g) x %d\n",
                root->s.real, root->s.imag, root->multiplicity);
        }
    }

    if (pzan->PZnZeros > 0) {
        for (root = pzan->PZzeroList; root != NULL; root = root->next) {
            for (i = 0; i < root->multiplicity; i++) {
                out_list[j].real = root->s.real;
                out_list[j].imag = root->s.imag;
                j += 1;
                if (root->s.imag != 0.0) {
                    out_list[j].real = root->s.real;
                    out_list[j].imag = -root->s.imag;
                    j += 1;
                }
            }
            DEBUG printf("LIST zero: (%g,%g) x %d\n",
                root->s.real, root->s.imag, root->multiplicity);
        }
    }

    outData.v.numValue = pzan->PZnPoles + pzan->PZnZeros;
    outData.v.vec.cVec = out_list;

    (*SPfrontEnd->OUTdata)(pzPlotPtr, (IFvalue *) 0, &outData);
    (*(SPfrontEnd->OUTendPlot))(pzPlotPtr);

    FREE(out_list);
    FREE(namelist);

    return(OK);
}


int
CKTpzSetup(ckt, type)

/* CKTpzSetup(ckt)
 * iterate through all the various
 * pzSetup functions provided for the circuit elements in the
 * given circuit, setup ...
 */
CKTcircuit *ckt;
int type;
{
    extern SPICEdev *DEVices[];
    PZAN *pzan = (PZAN *) ckt->CKTcurJob;
    SMPmatrix *matrix;
    int error;
    int /*i,*/ temp, solution_col, balance_col;
    int input_pos, input_neg, output_pos, output_neg;
    struct sCKTmodHead *mh;
    int (*func)();

    NIdestroy(ckt);
    error = NIinit(ckt);
    if (error)
        return(error);
    matrix = ckt->CKTmatrix;

    /* Really awful . . . */
    ckt->CKTnumStates = 0;

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if ((func = DEVices[mh->type]->DEVpzSetup) != NULL) {
            error = (*func)(matrix, mh->head,ckt,&ckt->CKTnumStates);
            if (error)
                return(error);
        }
    }

    solution_col = 0;
    balance_col = 0;

    input_pos = pzan->PZin_pos;
    input_neg = pzan->PZin_neg;

    if (type == PZ_DO_ZEROS) {
        /* Vo/Ii in Y */
        output_pos = pzan->PZout_pos;
        output_neg = pzan->PZout_neg;
    }
    else if (pzan->PZinput_type == PZ_IN_VOL) {
        /* Vi/Ii in Y */
        output_pos = pzan->PZin_pos;
        output_neg = pzan->PZin_neg;
    }
    else {
        /* Denominator */
        output_pos = 0;
        output_neg = 0;
        input_pos = 0;
        input_neg = 0;
    }

    if (output_pos) {
        solution_col = output_pos;
        if (output_neg)
            balance_col = output_neg;
    }
    else {
        solution_col = output_neg;
        temp = input_pos;
        input_pos = input_neg;
        input_neg = temp;
    }

    if (input_pos)
        pzan->PZdrive_pptr = spGetElement(matrix, input_pos, solution_col);
    else
        pzan->PZdrive_pptr = NULL;

    if (input_neg)
        pzan->PZdrive_nptr = spGetElement(matrix, input_neg, solution_col);
    else
        pzan->PZdrive_nptr = NULL;

    pzan->PZsolution_col = solution_col;
    pzan->PZbalance_col = balance_col;

    pzan->PZnumswaps = 1;

    error = NIreinit(ckt);
    if (error)
        return (error);

    return (OK);
}
