/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: UCB CAD Group
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "spmatrix.h"
#include "util.h"
#include "srcdefs.h"
#include "outdata.h"
#include "sensdefs.h"
#include "sensgen.h"
#include "uflags.h"


#define NEWN(TYPE,COUNT) ((TYPE *) MALLOC(sizeof(TYPE) * (COUNT)))
#ifndef M_LOG2E
#define M_LOG2E     1.4426950408889634074
#endif

static double Sens_Delta = 0.000001;
static double Sens_Abs_Delta = 0.000001;

#ifdef __STDC__
static int sens_acoperation(CKTcircuit*,struct sOUTdata*,int);
static int sens_dcoperation(CKTcircuit*,struct sOUTdata*,int);
static int sens_loadnew(sgen*,CKTcircuit*,int);
static int sens_setp(sgen*,CKTcircuit*,IFvalue*);
#else
static int sens_acoperation();
static int sens_dcoperation();
static int sens_loadnew();
static int sens_setp();
#endif

extern SPICEdev *DEVices[];

/*
 *    Procedure:
 *
 *        Determine operating point (call CKTop)
 *
 *        For each frequency point:
 *            (for AC) call NIacIter to get base node voltages
 *            For each element/parameter in the test list:
 *                construct the perturbation matrix
 *                Solve for the sensitivities:
 *                    delta_E = Y^-1 (delta_I - delta_Y E)
 *                save results
 */



int
SENSan(cktp, restart)

GENERIC *cktp;
int restart;
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    SENSAN *job = (SENSAN *) ckt->CKTcurJob;
    struct sSENSint *st = &job->ST;
    sgen   *sg;
    static struct sOUTdata outd;
    int    i;
    int    is_dc;
    int    code;
    int    error;
    char   namebuf[513];

    if (ckt->CKTjjPresent) {
        (*(SPfrontEnd->IFerror))(ERR_FATAL,
            "Sensitivity analysis not possible with Josephson junctions",
            NULL);
            return (OK);
    }
    if (restart) {

        /* make sure sources are present */
        code = CKTtypelook("Source");

        if (job->SENSoutSrc) {
            error = CKTfndDev((GENERIC*)ckt,&code,&job->SENSoutSrcDev,
                job->SENSoutSrc, (GENERIC *)NULL,(IFuid)NULL);
            if (error != 0) {
                (*(SPfrontEnd->IFerror))(ERR_FATAL,
                    "Sensitivity Analysis source %s not in circuit",
                    &job->SENSoutSrc);
                return (E_NOTFOUND);
            }
        }
        if (job->DC.eltName[0] != NULL) {
            /* DC source was given */

            for (i = 0; i <= job->DC.nestLevel; i++) {
                SRCinstance *here;

                here = (SRCinstance*)NULL;
                error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
                    job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);
                if (error) {
                    (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                        "DCtrCurv: source %s not in circuit",
                        &(job->DC.eltName[i]));
                    return (E_NOTFOUND);
                }
                job->DC.elt[i] = (GENinstance*)here;
                job->DC.vsave[i] = here->SRCdcValue;
            }
        }

        /* in case these never got freed */
        if (st->dY) {
            spDestroy(st->dY);
            st->dY = NULL;
        }
        FREE(st->dIr);
        FREE(st->dIi);
        FREE(st->dIdYr);
        FREE(st->dIdYi);
        FREE(st->o_values);
        FREE(st->o_cvalues);

        is_dc = (job->AC.stepType == DCSTEP);

        st->size = spGetSize(ckt->CKTmatrix, 1);

        /* Create the perturbation matrix */
        st->dY = spCreate(st->size, 1, &error);
        if (error)
            return (error);

        /* Create extra rhs */
        st->dIr   = NEWN(double, st->size + 1);
        st->dIi   = NEWN(double, st->size + 1);
        st->dIdYr = NEWN(double, st->size + 1);
        st->dIdYi = NEWN(double, st->size + 1);

        outd.numNames = 0;
        for (sg = sgen_init(ckt, is_dc); sg; sgen_next(&sg)) {
            outd.numNames++;
        }
        if (!outd.numNames)
            return (E_NOTFOUND);

        outd.dataNames = NEWN(IFuid,outd.numNames);
        for (i = 0,sg = sgen_init(ckt, is_dc); sg; i++,sgen_next(&sg)) {
            if (!sg->is_instparam) {
                sprintf(namebuf, "%s:%s",
                    sg->instance->GENname,
                    sg->ptable[sg->param].keyword);
            }
            else if ((sg->ptable[sg->param].dataType
                    & IF_PRINCIPAL) && sg->is_principle == 1) {
                sprintf(namebuf, "%s", sg->instance->GENname);
            }
            else {
                sprintf(namebuf, "%s.%s",
                    sg->instance->GENname,
                    sg->ptable[sg->param].keyword);
            }

            (*SPfrontEnd->IFnewUid)((GENERIC *) ckt,
                outd.dataNames + i, NULL,
                namebuf, UID_OTHER, NULL);
        }

        if (is_dc) {
            outd.dataType = IF_REAL;
            (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,
                &outd.refName,(IFuid )NULL,
                "sweep", UID_OTHER, NULL);
        }
        else {
            outd.dataType = IF_COMPLEX;
            (*SPfrontEnd->IFnewUid)((GENERIC *) ckt,
                &outd.refName, NULL,
                "frequency", UID_OTHER, NULL);
        }

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = ckt->CKTcurJob->JOBname;
        outd.refType = IF_REAL;
        outd.plotPtr = &job->SENSplot;
        outd.initValue = 0;
        outd.finalValue = 0;
        outd.step = 0;
        outd.numPts = 1;
        outd.count = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        FREE(outd.dataNames);

        if (is_dc) {
            st->o_values = NEWN(double, outd.numNames);
            st->o_cvalues = NULL;
        }
        else {
            st->o_values = NULL;
            st->o_cvalues = NEWN(IFcomplex, outd.numNames);
        }
    }
    error = DCTloop(sens_dcoperation,ckt,restart,&job->DC,&outd);

    if (error == E_PAUSE)
        return (error);

    (*(SPfrontEnd->OUTendPlot))(*outd.plotPtr);

    spDestroy(st->dY);
    st->dY = NULL;
    FREE(st->dIr);
    FREE(st->dIi);
    FREE(st->dIdYr);
    FREE(st->dIdYi);
    FREE(st->o_values);
    FREE(st->o_cvalues);
    return (error);
}


/* ARGSUSED */
static int
sens_dcoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    SENSAN *job = (SENSAN *) ckt->CKTcurJob;
    struct sSENSint *st = &job->ST;
    sgen   *sg;
    IFvalue value, nvalue;
    double *E;
    SMPmatrix *Y;
    double delta_var;
    int    i, j;
    int    bypass;
    int    error;

    /* find the operating point */

    error = CKTic(ckt);
    if (error)
        return (error);

    error = CKTop(ckt,
        MODEDCOP | MODEINITJCT,
        MODEDCOP | MODEINITFLOAT,
        ckt->CKTdcMaxIter);
    if (error)
        return (error);

    if (job->AC.stepType != DCSTEP) {
        error = ACloop(sens_acoperation,ckt,restart,&job->AC,outd);
        return (error);
    }

    bypass = ckt->CKTbypass;
    ckt->CKTbypass = 0;

    /* The unknown vector of node voltages overwrites rhs */
    E = ckt->CKTrhs;
    ckt->CKTrhsOld = E;
    Y = ckt->CKTmatrix;

    /* Use a different vector & matrix */
    ckt->CKTrhs = st->dIr;
    ckt->CKTmatrix = st->dY;

    /* calc. effect of each param */
    for (i = 0,sg = sgen_init(ckt, TRUE); sg; i++,sgen_next(&sg)) {

        /* clear CKTmatrix, CKTrhs */
        spSetReal(st->dY);
        spClear(st->dY);
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] = 0.0;
        }

        error = sens_loadnew(sg,ckt,TRUE);
        if (error)
            return (error);

        /* Alter the parameter */

        if (sg->value != 0.0)
            delta_var = sg->value * Sens_Delta;
        else
            delta_var = Sens_Abs_Delta;

        nvalue.rValue = sg->value + delta_var;

        sens_setp(sg, ckt, &nvalue);
        if (error)
            return (error);

        /* change sign of CKTmatrix, CKTrhs */
        spConstMult(st->dY, -1.0);
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] = -st->dIr[j];
        }

        error = sens_loadnew(sg,ckt,TRUE);
        if (error)
            return (error);

        /* now have delta_Y = CKTmatrix, delta_I = CKTrhs */

        /* reset parameter */
        value.rValue = sg->value;
        sens_setp(sg, ckt, &value);

        /* delta_Y E */
        spMultiply(st->dY, st->dIdYr, E, NULL, NULL);

        /* delta_I - delta_Y E */
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] -= st->dIdYr[j];
        }

        /* Solve; Y already factored */
        spSolve(Y, st->dIr, st->dIr, NULL, NULL);

        /* delta_I is now equal to delta_E */

        st->dIr[0] = 0.0;
        if (job->SENSoutName)
            st->o_values[i] =
                st->dIr[job->SENSoutPos->number] -
                st->dIr[job->SENSoutNeg->number];
        else {
            st->o_values[i] =
                st->dIr
                    [((SRCinstance*)job->SENSoutSrcDev)->SRCbranch];
        }
        st->o_values[i] /= delta_var;
    }
    nvalue.v.vec.rVec = st->o_values;

    if (job->DC.elt[0])
        value.rValue = ((SRCinstance*)job->DC.elt[0])->SRCdcValue;
    else
        value.rValue = 0;
    OUTdata(job->SENSplot, &value, &nvalue);
    outd->count++;

    ckt->CKTrhs = E;
    ckt->CKTmatrix = Y;
    ckt->CKTbypass = bypass;

    return (OK);
}


/* ARGSUSED */
static int
sens_acoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    SENSAN *job = (SENSAN *) ckt->CKTcurJob;
    struct sSENSint *st = &job->ST;
    sgen   *sg;
    IFvalue value, nvalue;
    double *E, *iE;
    SMPmatrix *Y;
    double delta_var;
    int    i, j;
    int    bypass;
    int    error;
    bypass = ckt->CKTbypass;

    ckt->CKTbypass = 0;

    /* The unknown vector of node voltages overwrites rhs */
    E = ckt->CKTrhs;
    iE = ckt->CKTirhs;
    ckt->CKTrhsOld = E;
    ckt->CKTirhsOld = iE;
    Y = ckt->CKTmatrix;

    for (i = 0; i <= st->size; i++) {
        st->dIr[i] = 0.0;
        st->dIi[i] = 0.0;
    }

    ckt->CKTrhs = E;
    ckt->CKTirhs = iE;
    ckt->CKTmatrix = Y;

    /* This generates Y in LU form */

    /* Yes, all this has to be re-done */
    ckt->CKTpreload = 0;
    error = CKTsetup(ckt);
    if (error)
        return (error);
    error = CKTtemp(ckt);
    if (error)
        return (error);
    ckt->CKTmode = MODEDCOP | MODEINITSMSIG;
    error = CKTload(ckt);
    if (error)
        return (error);
    ckt->CKTmode = MODEAC;
    error = NIacIter(ckt);
    if (error)
        return (error);

    /* Use a different vector & matrix */
    ckt->CKTrhs = st->dIr;
    ckt->CKTirhs = st->dIi;
    ckt->CKTmatrix = st->dY;

    /* calc. effect of each param */
    for (i = 0,sg = sgen_init(ckt, FALSE); sg; i++,sgen_next(&sg)) {

        /* clear CKTmatrix, CKTrhs */
        spSetComplex(st->dY);
        spClear(st->dY);
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] = 0.0;
            st->dIi[j] = 0.0;
        }

        error = sens_loadnew(sg,ckt,FALSE);
        if (error)
            return (error);

        /* Alter the parameter */

        if (sg->value != 0.0)
            delta_var = sg->value * Sens_Delta;
        else
            delta_var = Sens_Abs_Delta;

        nvalue.rValue = sg->value + delta_var;

        sens_setp(sg, ckt, &nvalue);
        if (error)
            return (error);

        /* change sign of CKTmatrix, CKTrhs */
        spConstMult(st->dY, -1.0);
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] = -st->dIr[j];
            st->dIi[j] = -st->dIi[j];
        }

        error = sens_loadnew(sg,ckt,FALSE);
        if (error)
            return (error);

        /* now have delta_Y = CKTmatrix, delta_I = CKTrhs */

        /* reset parameter */
        value.rValue = sg->value;
        sens_setp(sg, ckt, &value);

        /* delta_Y E */
        spMultiply(st->dY, st->dIdYr, E, st->dIdYi, iE);

        /* delta_I - delta_Y E */
        for (j = 0; j <= st->size; j++) {
            st->dIr[j] -= st->dIdYr[j];
            st->dIi[j] -= st->dIdYi[j];
        }

        /* Solve; Y already factored */
        spSolve(Y, st->dIr, st->dIr, st->dIi, st->dIi);

        /* delta_I is now equal to delta_E */

        st->dIr[0] = 0.0;
        st->dIi[0] = 0.0;
        if (job->SENSoutName) {
            st->o_cvalues[i].real =
                st->dIr[job->SENSoutPos->number] -
                st->dIr[job->SENSoutNeg->number];
            st->o_cvalues[i].imag =
                st->dIi[job->SENSoutPos->number] -
                st->dIi[job->SENSoutNeg->number];
        }
        else {
            st->o_cvalues[i].real =
                st->dIr
                    [((SRCinstance*)job->SENSoutSrcDev)->SRCbranch];
            st->o_cvalues[i].imag =
                st->dIi
                    [((SRCinstance*)job->SENSoutSrcDev)->SRCbranch];
        }
        st->o_cvalues[i].real /= delta_var;
        st->o_cvalues[i].imag /= delta_var;
    }
    nvalue.v.vec.cVec = st->o_cvalues;

    value.rValue = ckt->CKTomega/(2*M_PI);
    OUTdata(job->SENSplot, &value, &nvalue);
    outd->count++;

    ckt->CKTrhs = E;
    ckt->CKTirhs = iE;
    ckt->CKTmatrix = Y;
    ckt->CKTbypass = bypass;

    return (OK);
}


/* Get parameter value */
int
sens_getp(sg, ckt, val)

sgen *sg;
CKTcircuit *ckt;
IFvalue *val;
{
    int (*fn)( );
    int pid;
    int error = 0;

    if (sg->is_instparam) {
        fn = DEVices[sg->dev]->DEVask;
        pid = DEVices[sg->dev]->DEVpublic.instanceParms[sg->param].id;
        if (fn)
            error = (*fn)(ckt, sg->instance, pid, val, NULL);
        else
            return (1);
    }
    else {
        fn = DEVices[sg->dev]->DEVmodAsk;
        pid = DEVices[sg->dev]->DEVpublic.modelParms[sg->param].id;
        if (fn)
            error = (*fn)(ckt, sg->model, pid, val);
        else
            return (1);
    }

    if (error) {
        if (sg->is_instparam)
            printf("GET ERROR: %s:%s:%s -> param %s (%d)\n",
                DEVices[sg->dev]->DEVpublic.name,
                sg->model->GENmodName,
                sg->instance->GENname,
                sg->ptable[sg->param].keyword, pid);
        else
            printf("GET ERROR: %s:%s:%s -> mparam %s (%d)\n",
                DEVices[sg->dev]->DEVpublic.name,
                sg->model->GENmodName,
                sg->instance->GENname,
                sg->ptable[sg->param].keyword, pid);
    }

    return (error);
}


/* Set parameter value */
static int
sens_setp(sg, ckt, val)

sgen *sg;
CKTcircuit *ckt;
IFvalue *val;
{
    int (*fn)( );
    int pid;
    int error = 0;

    if (sg->is_instparam) {
        fn = DEVices[sg->dev]->DEVparam;
        pid = DEVices[sg->dev]->DEVpublic.instanceParms[sg->param].id;
        if (fn)
            error = (*fn)(ckt, pid, val, sg->instance, NULL);
        else
            return (1);
    }
    else {
        fn = DEVices[sg->dev]->DEVmodParam;
        pid = DEVices[sg->dev]->DEVpublic.modelParms[sg->param].id;
        if (fn)
            error = (*fn)(pid, val, sg->model);
        else
            return (1);
    }

    if (error) {
        if (sg->is_instparam)
            printf("SET ERROR: %s:%s:%s -> param %s (%d)\n",
                DEVices[sg->dev]->DEVpublic.name,
                sg->model->GENmodName,
                sg->instance->GENname,
                sg->ptable[sg->param].keyword, pid);
        else
            printf("SET ERROR: %s:%s:%s -> mparam %s (%d)\n",
                DEVices[sg->dev]->DEVpublic.name,
                sg->model->GENmodName,
                sg->instance->GENname,
                sg->ptable[sg->param].keyword, pid);
    }
    return (error);
}


static int
sens_loadnew(sg, ckt, is_dc)

sgen *sg;
CKTcircuit *ckt;
int is_dc;
{
    int error = 0;
    int (*fn)();

    if (!is_dc)
        ckt->CKTpreload = 0;
    else
        ckt->CKTpreload = 1;

    /* call setup */
    ckt->CKTnumStates = sg->istate;
    fn = DEVices[sg->dev]->DEVsetup;
    if (fn)
        (*fn)(ckt->CKTmatrix, sg->model, ckt, &ckt->CKTnumStates);

    /* call temp */
    fn = DEVices[sg->dev]->DEVtemperature;
    if (fn)
        (*fn)(sg->model, ckt);

    /* call load */
    if (!is_dc) {
        fn = DEVices[sg->dev]->DEVacLoad;
    }
    else {
        fn = DEVices[sg->dev]->DEVload;
    }
    if (fn)
        error = (*fn)(sg->model, ckt);

    return (error);
}
