/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

/* subroutine to do TRANSIENT analysis */   

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "srcdefs.h"
#include "sperror.h"
#include "outdata.h"
#include "trandefs.h"
#include "optdefs.h"
#include "tskdefs.h"
#include "cpdefs.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"

#ifdef __STDC__
static int  tran_dcoperation(CKTcircuit*,struct sOUTdata*,int);
static void tran_setbp(CKTcircuit*,struct sTRANint*);
static int  tran_accept
       (CKTcircuit*,struct sTRANint*,STATistics*,struct sOUTdata*,bool*,bool*);
static int  tran_step(CKTcircuit*,struct sTRANint*,STATistics*);
static int  interpolate(CKTcircuit*,struct sTRANint*,int*);
static int  is_trunc_dev(CKTcircuit*);
#else
static int  tran_dcoperation();
static void tran_setbp();
static int  tran_accept();
static int  tran_step();
static int  interpolate();
static int  is_trunc_dev();
#endif


int
TRANan(cktp,restart)

GENERIC *cktp;
int restart;    /* forced restart flag */
{
    CKTcircuit *ckt = (CKTcircuit *)cktp;
    TRANAN *job = (TRANAN*)ckt->CKTcurJob;
    struct sTRANint *tran = &job->TS;
    static struct sOUTdata outd;
    int i, error, code;
    double delmax, tt0;
    STATistics *stat = (STATistics*)ckt->CKTstat;
    extern char *kw_nousertp, *kw_hitusertp, *kw_polydegree, *kw_nojjtp;

    tran->startTime  = (*(SPfrontEnd->IFseconds))();
    if (restart) {

        ckt->CKTmode = job->TRANmode;

        if (ckt->CKTjjPresent && !(ckt->CKTmode & MODEUIC)) {
            WARNmsg("Using initial conditions");
            ckt->CKTmode |= MODEUIC;
        }
        ckt->CKTfinalTime = job->TRANfinalTime;
        ckt->CKTstep = job->TRANstep;
        ckt->CKTinitTime = job->TRANinitTime;
        ckt->CKTmaxStep = job->TRANmaxStep;

        if (ckt->CKTstep <= 0 ||
            ckt->CKTinitTime < 0 ||
            ckt->CKTfinalTime <= 0 ||
            ckt->CKTmaxStep < 0 ||
            ckt->CKTfinalTime <= ckt->CKTinitTime) return (E_PARMVAL);

        if (ckt->CKTmaxStep == 0) {
            if (!is_trunc_dev(ckt))
                /* no timestep predicting devices in circuit,
                 * set to user step
                 */ 
                ckt->CKTmaxStep = ckt->CKTstep;
            else
                /* 50 really is too small */
                ckt->CKTmaxStep =
                    (ckt->CKTfinalTime-ckt->CKTinitTime)/50;
        }
        tran->delmax = .02 * (ckt->CKTfinalTime - ckt->CKTinitTime);
        tran->delmax = (1.01)*MIN(tran->delmax,ckt->CKTmaxStep);

        tran->nointerp = false;
        tran->hitusertp = false;
        tran->polydegree = 1;
        cp_getvar(kw_nousertp,VT_BOOL,(char*)&tran->nointerp);
        cp_getvar(kw_hitusertp,VT_BOOL,(char*)&tran->hitusertp);
        cp_getvar(kw_polydegree,VT_NUM,(char*)&tran->polydegree);
        if (job->DC.eltName[0] && tran->nointerp) {
            WARNmsg(
                "Note: DCsource given, \"set nousertp\" ignored.");
            tran->nointerp = false;
        }
        if (tran->nointerp && tran->hitusertp) {
            WARNmsg(
                "Note: nousertp set, \"set hitusertp\" ignored.");
            tran->hitusertp = false;
        }

        if (ckt->CKTjjPresent) {
            tran->nojjtp = false;
            cp_getvar(kw_nojjtp,VT_BOOL,(char*)&tran->nojjtp);
            if (tran->nojjtp)
                ckt->CKTjjPresent = false;
        }
        else
            tran->nojjtp = true;

        if (job->DC.eltName[0] != NULL) {
            /* DC source was given */
            code = CKTtypelook("Source");

            for (i = 0; i <= job->DC.nestLevel; i++) {
                SRCinstance *here;
                here = (SRCinstance*)NULL;
                error = CKTfndDev((GENERIC*)ckt,&code,(GENERIC**)&here,
                    job->DC.eltName[i], (GENERIC *)NULL,(IFuid)NULL);
                if (error) {
                    (*(SPfrontEnd->IFerror))(ERR_FATAL, 
                        "DCtrCurv: source %s not in circuit",
                        &(job->DC.eltName[i]));
                    return (E_NOTFOUND);
                }
                job->DC.elt[i] = (GENinstance*)here;
                job->DC.vsave[i] = here->SRCdcValue;
            }
        }
        error = CKTnames(ckt,&outd.numNames,&outd.dataNames);
        if (error)
            return (error);

        (*(SPfrontEnd->IFnewUid))((GENERIC *)ckt,&outd.refName,(IFuid)NULL,
                "time", UID_OTHER, (GENERIC **)NULL);

        if (tran->nointerp)
            outd.numPts = 1;
        else
            outd.numPts =
                (ckt->CKTfinalTime - ckt->CKTinitTime)/ckt->CKTstep + 2;

        outd.circuitPtr = (GENERIC *)ckt;
        outd.analysisPtr = (GENERIC*)ckt->CKTcurJob;
        outd.analName = ckt->CKTcurJob->JOBname;
        outd.refType = IF_REAL;
        outd.dataType = IF_REAL;
        outd.plotPtr = &job->TRANplot;
        outd.initValue = ckt->CKTinitTime;
        outd.finalValue = ckt->CKTfinalTime;
        outd.step = ckt->CKTstep;
        outd.count = 0;
        (*(SPfrontEnd->OUTbeginPlot))((GENERIC*)&outd);

        FREE(outd.dataNames);
    }
    tran->startIters = stat->STATnumIter;
    tran->startdTime = stat->STATdecompTime;
    tran->startsTime = stat->STATsolveTime;

    if (ckt->CKTminBreak == 0)
        ckt->CKTminBreak = ckt->CKTmaxStep * 5e-5;
    if (ckt->CKTdelmin == 0)
        ckt->CKTdelmin = ckt->CKTmaxStep * 1e-9;
    error = DCTloop(tran_dcoperation,ckt,restart,&job->DC,&outd);
    if (error == E_PAUSE)
        return (error);

    tt0 = (*(SPfrontEnd->IFseconds))();
    (*(SPfrontEnd->OUTendPlot)) (job->TRANplot);
    tran->startTime += (*(SPfrontEnd->IFseconds))() - tt0;
    return (error);
}


/* ARGSUSED */
static int
tran_dcoperation(ckt,outd,restart)

CKTcircuit *ckt;
struct sOUTdata *outd;
int restart;
{
    double *temp, tt;
    STATistics *stat = (STATistics*)ckt->CKTstat;
    struct sTRANint *tran = &((TRANAN*)ckt->CKTcurJob)->TS;
    SRCmodel *model;
    SRCinstance *here;
    bool done = false;
    bool afterpause = false;
    int i, error = 0, type;

    afterpause = !restart;
    if (restart) {

        ckt->CKTtime = 0;
        ckt->CKTdelta = 0;
        ckt->CKTbreak = 1;

        /* set up breakpoint table */
        if (ckt->CKTbreaks)
            FREE(ckt->CKTbreaks);
        ckt->CKTbreaks = (double *)MALLOC(2*sizeof(double));
        if (!ckt->CKTbreaks)
            return(E_NOMEM);
        *(ckt->CKTbreaks) = 0;
        *(ckt->CKTbreaks + 1) = ckt->CKTfinalTime;
        ckt->CKTbreakSize = 2;

        type = CKTtypelook("Source");
        if (type >= 0) {
            model = (SRCmodel*)ckt->CKThead[type];
            for (; model; model = model->SRCnextModel) {
                for (here = model->SRCinstances; here;
                        here = here->SRCnextInstance) {
                    if (here->SRCtree)
                        (*here->SRCtree->IFinit)
                            (here->SRCtree,ckt->CKTstep,ckt->CKTfinalTime);
                }
            }
        }

        tran->dumpit = false;

        tran->tcheck = (ckt->CKTinitTime || !tran->nointerp) ?
            ckt->CKTinitTime : ckt->CKTfinalTime;

        tran->firsttime = true;

        /* set initial conditions */
        error = CKTic(ckt);
        if (error)
            return (error);

        if (!(ckt->CKTmode & MODEUIC)) {
            error = CKTop(ckt,
                    MODETRANOP|MODEINITJCT,
                    MODETRANOP|MODEINITFLOAT,
                    ckt->CKTdcMaxIter);
        }
        else {
            error = CKTsetic(ckt);
            if (error)
                return (error);
            /* niiter() used to just do this and return */
            temp = ckt->CKTrhs;
            ckt->CKTrhs = ckt->CKTrhsOld;
            ckt->CKTrhsOld = temp;
            ckt->CKTmode = MODEUIC|MODETRANOP|MODEINITJCT;
            error = CKTload(ckt);
        }
        if (error)
            return (error);

        stat->STATtimePts++;
        ckt->CKTorder = 1;

        if (ckt->CKTinitDelta > 0)
            ckt->CKTdelta = ckt->CKTinitDelta;
        else
            ckt->CKTdelta = .1 * MIN(tran->delmax,ckt->CKTstep);

        for (i = 0; i < 7; i++) {
            ckt->CKTdeltaOld[i] = tran->delmax;
        }
        ckt->CKTdeltaOld[0] = ckt->CKTdelta;
        ckt->CKTsaveDelta   = ckt->CKTdelta;

        ckt->CKTmode = (ckt->CKTmode&MODEUIC)|MODETRAN | MODEINITTRAN;
        /* modeinittran set here */
        ckt->CKTag[0] = ckt->CKTag[1] = 0;
        bcopy((char *)ckt->CKTstate0,(char *)ckt->CKTstate1,
                ckt->CKTnumStates*sizeof(double));

        if (ckt->CKTtime >= ckt->CKTinitTime)
            tran->dumpit = true;

        tran->tend = ckt->CKTfinalTime - ckt->CKTminBreak;
    }

    for (;;) {
        error = tran_accept(ckt,tran,stat,outd,&done,&afterpause);
        if (error || done) break;

        stat->STAToldIter = stat->STATnumIter;
        if (ckt->CKTdelta > tran->delmax) ckt->CKTdelta = tran->delmax;
        tran_setbp(ckt,tran);

        /* rotate delta vector */
        for (i = 5; i >= 0; i--) {
            ckt->CKTdeltaOld[i+1] = ckt->CKTdeltaOld[i];
        }
        ckt->CKTdeltaOld[0] = ckt->CKTdelta;

        /* rotate states */
        temp = ckt->CKTstates[ckt->CKTmaxOrder+1];
        for (i = ckt->CKTmaxOrder; i >= 0; i--) {
            ckt->CKTstates[i+1] = ckt->CKTstates[i];
        }
        ckt->CKTstates[0] = temp;

        error = tran_step(ckt,tran,stat);
        if (error) break;
    }

    tt  = (*(SPfrontEnd->IFseconds))();
    stat->STATtranTime += tt - tran->startTime;
    stat->STATtranIter += stat->STATnumIter - tran->startIters;
    stat->STATtranDecompTime += stat->STATdecompTime - tran->startdTime;
    stat->STATtranSolveTime += stat->STATsolveTime - tran->startsTime;
    return (error);
}


static void
tran_setbp(ckt,tran)

CKTcircuit *ckt;
struct sTRANint *tran;
{
    double ttmp;

    /*
     * Breakpoint handling scheme:
     * When a timepoint t is accepted (by CKTaccept), clear all previous
     * breakpoints, because they will never be needed again.
     *
     * t may itself be a breakpoint, or indistinguishably close. DON'T
     * clear t itself; recognise it as a breakpoint and act accordingly
     *
     * if t is not a breakpoint, limit the timestep so that the next
     * breakpoint is not crossed
     */

    if (tran->nojjtp) {

        /* are we at a breakpoint, or indistinguishably close? */
        if ((ckt->CKTtime == *(ckt->CKTbreaks)) || (*(ckt->CKTbreaks) -
                        (ckt->CKTtime) <= ckt->CKTdelmin)) {
            /* first timepoint after a breakpoint - cut integration order
             * and limit timestep to .1 times minimum of time to next
             * breakpoint, and previous timestep
             */
            ckt->CKTorder = 1;

            ttmp = *(ckt->CKTbreaks+1) - *(ckt->CKTbreaks);
            if (ckt->CKTsaveDelta < ttmp)
                ttmp = ckt->CKTsaveDelta;
            ttmp *= 0.1;
            if (ttmp < ckt->CKTdelta)
                ckt->CKTdelta = ttmp;

            if (tran->firsttime) {
                ckt->CKTdelta *= 0.1;
            }

            /* don't want to get below delmin for no reason */
            ttmp = 2.0*ckt->CKTdelmin;
            if (ckt->CKTdelta < ttmp)
                ckt->CKTdelta = ttmp;
        }
        else if (ckt->CKTtime + ckt->CKTdelta >= *(ckt->CKTbreaks)) {
            ckt->CKTsaveDelta = ckt->CKTdelta;
            ckt->CKTdelta = *(ckt->CKTbreaks) - ckt->CKTtime;
            ckt->CKTbreak = 1; /* why? the current pt. is not a bkpt. */
        }
    }

    if (tran->hitusertp || tran->nointerp) {
        if (ckt->CKTtime + ckt->CKTdelta >= tran->tcheck) {
            if (ckt->CKTtime < tran->tcheck) {
                ckt->CKTsaveDelta = ckt->CKTdelta;
                ckt->CKTdelta = tran->tcheck - ckt->CKTtime;
                tran->dumpit = true;
            }
            tran->tcheck += ckt->CKTstep;
            if (tran->tcheck > ckt->CKTfinalTime || tran->nointerp)
                tran->tcheck = ckt->CKTfinalTime;
        }
    }
}


static int
tran_accept(ckt,tran,stat,outd,done,afterpause)

CKTcircuit *ckt;
struct sTRANint *tran;
STATistics *stat;
struct sOUTdata *outd;
bool *done, *afterpause;
{
    TRANAN *job = (TRANAN*)ckt->CKTcurJob;
    int error;
    double tt0;

    /* afterpause if true if resuming.  If we paused during
     * interpolation, go back and finish up, otherwise just
     * return.
     */
    if (*afterpause) {
        *afterpause = false;
        if (tran->hitusertp || tran->nointerp)
            return (OK);
    }
    else {
        error = CKTaccept(ckt);
        /* check if current breakpoint is outdated; if so, clear */
        if (ckt->CKTtime > *(ckt->CKTbreaks)) CKTclrBreak(ckt);

        stat->STATaccepted++;
        ckt->CKTbreak = 0;
        if (error) {
            ckt->CKTcurrentAnalysis = DOING_TRAN;
            return (error);
        }
    }

    tt0 = (*(SPfrontEnd->IFseconds))();
    if (tran->hitusertp || tran->nointerp) {
        if (tran->dumpit) {
            CKTdump(ckt,ckt->CKTtime,job->TRANplot);
            outd->count++;
            if (!tran->nointerp)
                tran->dumpit = false;
        }
        if (tran->nointerp) {
            if (ckt->CKTtime >= tran->tend)
                *SPfrontEnd->OUTendit = true;
        }
        else {
            if (tran->tcheck >= tran->tend)
                *SPfrontEnd->OUTendit = true;
        }
    }
    else {
        error = interpolate(ckt,tran,&outd->count);
        if (error == E_NOCHANGE)
            *SPfrontEnd->OUTendit = true;
    }
    tran->startTime += (*(SPfrontEnd->IFseconds))() - tt0;

    if (*SPfrontEnd->OUTendit) {
        *SPfrontEnd->OUTendit = false;
        ckt->CKTcurrentAnalysis = 0;
        *done = true;
        return (OK);
    }

    if ( error == E_PAUSE || (*(SPfrontEnd->IFpauseTest))() ) {
        /* user requested pause... */
        ckt->CKTcurrentAnalysis = DOING_TRAN;
        return (E_PAUSE);
    }
    return (OK);
}


static int
tran_step(ckt,tran,stat)

CKTcircuit *ckt;
struct sTRANint *tran;
STATistics *stat;
{
    double olddelta, new;
    int i, error, converged;

    for (;;) {

        olddelta = ckt->CKTdelta;
        ckt->CKTtime += ckt->CKTdelta;
        ckt->CKTdeltaOld[0] = ckt->CKTdelta;
        NIcomCof(ckt);

        (void)DEVpredNew(ckt);

        converged = NIiter(ckt,ckt->CKTtranMaxIter);
        stat->STATtimePts++;

        ckt->CKTmode = (ckt->CKTmode&MODEUIC)|MODETRAN | MODEINITPRED;

        if (tran->firsttime) {
            for (i = 0; i < ckt->CKTnumStates; i++) {
                *(ckt->CKTstate2+i) = *(ckt->CKTstate1+i);
                *(ckt->CKTstate3+i) = *(ckt->CKTstate1+i);
            }
        }
        if (converged != 0) {
            ckt->CKTtime -= ckt->CKTdelta;
            stat->STATrejected++;
            ckt->CKTdelta *= 0.125;

            if (tran->firsttime) {
                 ckt->CKTmode =
                    (ckt->CKTmode&MODEUIC)|MODETRAN | MODEINITTRAN;
            }
            ckt->CKTorder = 1;
        }
        else {
            if (tran->firsttime) {
                tran->firsttime = false;
                /* no check on first time point */
                break;
            }

            if (!tran->nojjtp) {
                ckt->CKTorder++;
                if (ckt->CKTorder > ckt->CKTmaxOrder)
                    ckt->CKTorder = ckt->CKTmaxOrder;
                break;
            }

            new = ckt->CKTdelta;
            error = CKTtrunc(ckt,&new);
            if (error) {
                ckt->CKTcurrentAnalysis = DOING_TRAN;
                return (error);
            }
            /* spice2g had .5 here, why change?
            if (new > .9*ckt->CKTdelta) {
            */
            if (new > .5*ckt->CKTdelta) {
                if (ckt->CKTorder == 1) {
                    new = ckt->CKTdelta;
                    ckt->CKTorder = 2;
                    error = CKTtrunc(ckt,&new);
                    if (error) {
                        ckt->CKTcurrentAnalysis = DOING_TRAN;
                        return (error);
                    }
                    if (new <= 1.05*ckt->CKTdelta) {
                        ckt->CKTorder = 1;
                    }
                }
                /* time point OK */
                ckt->CKTdelta = new;
                break;
            }
            else {
/*
printf("rejected: time=%g delta=%g new=%g\n",ckt->CKTtime,ckt->CKTdelta,new);
*/
                ckt->CKTtime -= ckt->CKTdelta;
                stat->STATrejected++;
                ckt->CKTdelta = new;
            }
        }
        if (ckt->CKTdelta <= ckt->CKTdelmin) {
            if (olddelta > ckt->CKTdelmin) {
                ckt->CKTdelta = ckt->CKTdelmin;
            }
            else {
                ckt->CKTcurrentAnalysis = DOING_TRAN;
                errMsg = CKTtrouble(ckt, "Timestep too small");
                return (E_TIMESTEP);
                
            }
        }
    }
    return (OK);
}


static int
interpolate(ckt,tran,count)

/* Previously, all raw timepoint data was saved in the output data
 * structures, and interpolation was performed when the analysis was
 * complete.  This can take huge amounts of storage if Josephson junctions
 * are present.  Here, we (by default) save only data at user time points,
 * performing interpolation "on the fly".
 */
CKTcircuit *ckt;
struct sTRANint *tran;
int *count;
{
    TRANAN *job = (TRANAN*)ckt->CKTcurJob;
    double d1,d2,d3,d4,diff[4];
    double c2,c3,c4;
    double finalmax,finalmin;
    double time, delta;
    double *temp;
    int i,size;

    /* use CKTrhs to store the interpolated values, since it is not
     * used for anything - it will be cleared before the next time
     * point.
     */

    if (tran->tcheck > ckt->CKTtime)
        return (OK);
    size = spGetSize(ckt->CKTmatrix,1);
    delta = .5*ckt->CKTdelmin;
    finalmax = ckt->CKTfinalTime + delta;
    finalmin = ckt->CKTfinalTime - delta;
    time = ckt->CKTtime + delta;

    if (tran->polydegree <= 1) {

        c2 = ckt->CKTdeltaOld[0];

        for (; tran->tcheck <= time; tran->tcheck += ckt->CKTstep) {
            if ( (*(SPfrontEnd->IFpauseTest))() )
                return(E_PAUSE);

            d1 = ckt->CKTtime - tran->tcheck;
            d2 = d1 - c2;

            /* first order extrapolation factors */
            diff[0] = d2/(-c2);
            diff[1] = d1/(c2);
 
            for (i = 1; i <= size; i++) {
                ckt->CKTrhs[i] =
                    diff[0]*ckt->CKTsols[0][i] +
                    diff[1]*ckt->CKTsols[1][i];
            }
            if (tran->tcheck < finalmax) {
                temp = ckt->CKTrhsOld;
                ckt->CKTrhsOld = ckt->CKTrhs;
                CKTdump(ckt,tran->tcheck,job->TRANplot);
                ckt->CKTrhsOld = temp;
                (*count)++;
                if (finalmin < tran->tcheck)
                    return (E_NOCHANGE);
            }
            else
                return (E_NOCHANGE);
        }
    }
    else if (tran->polydegree == 2) {

        c2 = ckt->CKTdeltaOld[0];
        c3 = c2 + ckt->CKTdeltaOld[1];

        for (; tran->tcheck <= time; tran->tcheck += ckt->CKTstep) {
            if ( (*(SPfrontEnd->IFpauseTest))() )
                return(E_PAUSE);

            d1 = ckt->CKTtime - tran->tcheck;
            d2 = d1 - c2;
            d3 = d1 - c3;

            /* second order extrapolation factors */
            diff[0] = d2*d3/((c2)*(c3));
            diff[1] = d1*d3/((c2)*(c2-c3));
            diff[2] = d1*d2/((c3)*(c3-c2));
    
            for (i = 1; i <= size; i++) {
                ckt->CKTrhs[i] =
                    diff[0]*ckt->CKTsols[0][i] +
                    diff[1]*ckt->CKTsols[1][i] +
                    diff[2]*ckt->CKTsols[2][i];
            }
            if (tran->tcheck < finalmax) {
                temp = ckt->CKTrhsOld;
                ckt->CKTrhsOld = ckt->CKTrhs;
                CKTdump(ckt,tran->tcheck,job->TRANplot);
                ckt->CKTrhsOld = temp;
                (*count)++;
                if (finalmin < tran->tcheck)
                    return (E_NOCHANGE);
            }
            else
                return (E_NOCHANGE);
        }
    }
    else if (tran->polydegree >= 3) {

        c2 = ckt->CKTdeltaOld[0];
        c3 = c2 + ckt->CKTdeltaOld[1];
        c4 = c3 + ckt->CKTdeltaOld[2];

        for (; tran->tcheck <= time; tran->tcheck += ckt->CKTstep) {
            if ( (*(SPfrontEnd->IFpauseTest))() )
                return(E_PAUSE);

            d1 = ckt->CKTtime - tran->tcheck;
            d2 = d1 - c2;
            d3 = d1 - c3;
            d4 = d1 - c4;

            /* third order extrapolation factors */
            diff[0] = d2*d3*d4/((-c2)*(c3)*(c4));
            diff[1] = d1*d3*d4/((c2)*(c2-c3)*(c2-c4));
            diff[2] = d1*d2*d4/((c3)*(c3-c2)*(c3-c4));
            diff[3] = d1*d2*d3/((c4)*(c4-c2)*(c4-c3));
    
            for (i = 1; i <= size; i++) {
                ckt->CKTrhs[i] =
                    diff[0]*ckt->CKTsols[0][i] +
                    diff[1]*ckt->CKTsols[1][i] +
                    diff[2]*ckt->CKTsols[2][i] +
                    diff[3]*ckt->CKTsols[3][i];
            }
            if (tran->tcheck < finalmax) {
                temp = ckt->CKTrhsOld;
                ckt->CKTrhsOld = ckt->CKTrhs;
                CKTdump(ckt,tran->tcheck,job->TRANplot);
                ckt->CKTrhsOld = temp;
                (*count)++;
                if (finalmin < tran->tcheck)
                    return (E_NOCHANGE);
            }
            else
                return (E_NOCHANGE);
        }
    }
    if (tran->tcheck > finalmin)
        return (E_NOCHANGE);
    return (OK);
}

static int
is_trunc_dev(ckt)
CKTcircuit *ckt;
{
    extern SPICEdev *DEVices[];
    struct sCKTmodHead *mh;

    for (mh = ckt->CKTheadList; mh != NULL; mh = mh->next) {
        if (DEVices[mh->type]->DEVtrunc != NULL)
            return (TRUE);
    }
    return (FALSE);
}
