/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * The main entry point for cshpar.
 */

#include "spice.h"
#include "cpdefs.h"
#define SHORT_SCEDIO
#include "scedio.h"

#ifdef HAVE_SGTTY_H
#include <sgtty.h>
#else
#ifdef HAVE_TERMIO_H
#include <termio.h>
#else
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#endif
#endif

#ifdef HAVE_GETPWUID
#include <pwd.h>
#endif
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif


/* Things go as follows:
 * (1) Read the line and do some initial quoting (by setting the 8th bit),
 *  and command ignoring. Also deal with command completion.
 * (2) Do history substitutions. (!, ^)
 * (3) Do alias substitution.
 * 
 * In front.c these things get done:
 * (4) Do variable substitution. ($varname)
 * (5) Do backquote substitution. (``)
 * (6) Do globbing. (*, ?, [], {}, ~)
 * (7) Do io redirection.
 */

#ifdef __STDC__
static void pwlist(wordlist*,char*);
static bool fileexists(char*);
#else
static void pwlist();
static bool fileexists();
#endif

bool cp_debug = false;

char cp_gt = '>';
char cp_lt = '<';
char cp_amp = '&';

FILE *cp_in;
FILE *cp_out;
FILE *cp_err;

/* These are the fps that cp_ioreset resets the cp_* to.  They are changed
 * by the source routines.
 */

FILE *cp_curin = NULL;
FILE *cp_curout = NULL;
FILE *cp_curerr = NULL;

char cp_chars[128];


void
cp_init()

/* Initialize stuff. */
{
    char *s, *singlec = "<>;&";

    cp_internal_init();
    bzero(cp_chars, 128);
    for (s = singlec; *s; s++)
        cp_chars[*s] = (CPC_BRR | CPC_BRL);
    cp_vset("history", VT_NUM, (char *) &cp_maxhistlength);

    cp_curin = stdin;
    cp_curout = stdout;
    cp_curerr = stderr;

    cp_ioreset();
}


wordlist *
cp_parse(string)

char *string;
{
    wordlist *wlist;

    wlist = cp_lexer(string);

    if (!string)
        cp_event++;

    if (!wlist || !wlist->wl_word)
        return (wlist);

    pwlist(wlist, "Initial parse");

    cp_histsubst(&wlist);
    if (!wlist || !wlist->wl_word)
        return (wlist);

    pwlist(wlist, "After history substitution");

    if (cp_interactive && cp_didhsubst) {
        wl_print(wlist, stdout);
        (void) putc('\n', stdout);
    }

    /* Add the word list to the history. */
    if (!string && *wlist->wl_word)
        cp_addhistent(cp_event - 1, wlist);

    cp_doalias(&wlist);
    pwlist(wlist, "After alias substitution");

    return (wlist);
}


static void
pwlist(wlist, name)

wordlist *wlist;
char *name;
{
    wordlist *wl;

    if (!cp_debug)
        return;
    fprintf(cp_err, "%s : [ ", name);
    for (wl = wlist; wl; wl = wl->wl_next)
        fprintf(cp_err, "%s ", wl->wl_word);
    fprintf(cp_err, "]\n");
    return;
}


#ifdef notdef

/* relocated to newgraf.c for graphics support */

void
com_echo(wlist)

wordlist *wlist;
{
    bool nl = true;

    if (wlist && eq(wlist->wl_word, "-n")) {
        wlist = wlist->wl_next;
        nl = false;
    }

    while (wlist) {
        char *t = copy(wlist->wl_word);
        cp_unquote(t);
        fputs(t, cp_out);
        tfree(t);
        if (wlist->wl_next)
            fputs(" ", cp_out);
        wlist = wlist->wl_next;
    }
    if (nl)
        fputs("\n", cp_out);
}

#endif


/* This routine sets the cp_{in,out,err} pointers and takes the io
 * directions out of the command line.
 */

void
cp_redirect(list)

wordlist **list;
{
    bool gotinput = false, gotoutput = false, goterror = false;
    bool app = false, erralso = false;
    wordlist *w, *wl, *bt, *nw;
    char *s;
    FILE *tmpfp;

    if (list == NULL)
        return;
    wl = *list;
    w = wl->wl_next;    /* Don't consider empty commands. */
    while (w) {
        if (*w->wl_word == cp_lt) {
            bt = w;
            if (gotinput) {
                fprintf(cp_err, 
                "Error: ambiguous input redirect.\n");
                goto error;
            }
            gotinput = true;
            w = w->wl_next;
            if (w == NULL) {
                fprintf(cp_err, 
                "Error: missing name for input.\n");
                goto error;
            }
            if (*w->wl_word == cp_lt) {
                /* Do reasonable stuff here... */
            }
            else {
                char *t = copy(w->wl_word);
                cp_unquote(t);
                tmpfp = fopen(t, "r");
                tfree(t);
                if (!tmpfp) {
                    perror(w->wl_word);
                    goto error;
                }
                else
                    cp_in = tmpfp;
            }
#ifdef CPDEBUG
            if (cp_debug)
                fprintf(cp_err, "Input file is %s...\n",
                    w->wl_word);
#endif
            bt->wl_prev->wl_next = w->wl_next;
            if (w->wl_next)
                w->wl_next->wl_prev = bt->wl_prev;
            nw = w->wl_next;
            w->wl_next = NULL;
            w = nw;
            wl_free(bt);
        }
        else if (*w->wl_word == cp_gt) {
            bt = w;
            if (gotoutput) {
                fprintf(cp_err, 
                "Error: ambiguous output redirect.\n");
                goto error;
            }
            gotoutput = true;
            w = w->wl_next;
            if (w == NULL) {
                fprintf(cp_err, 
                "Error: missing name for output.\n");
                goto error;
            }
            if (*w->wl_word == cp_gt) {
                app = true;
                w = w->wl_next;
                if (w == NULL) {
                    fprintf(cp_err, 
                    "Error: missing name for output.\n");
                    goto error;
                }
            }
            if (*w->wl_word == cp_amp) {
                erralso = true;
                if (goterror) {
                    fprintf(cp_err, 
                "Error: ambiguous error redirect.\n");
                    goto error;
                }
                goterror = true;
                w = w->wl_next;
                if (w == NULL) {
                    fprintf(cp_err, 
                    "Error: missing name for output.\n");
                    goto error;
                }
            }
            s = copy(w->wl_word);
            cp_unquote(s);
            if (cp_noclobber && fileexists(s)) {
                fprintf(stderr, "Error: %s: file exists\n", s);
                tfree(s);
                goto error;
            }
            if (app)
                tmpfp = fopen(s, "a");
            else
                tmpfp = fopen(s, "w+");
            tfree(s);
            if (!tmpfp) {
                perror(w->wl_word);
                goto error;
            }
            else {
                cp_out = tmpfp;
                out_isatty = false;
            }
#ifdef CPDEBUG
            if (cp_debug)
                fprintf(cp_err, "Output file is %s... %s\n",
                    w->wl_word, app ? "(append)" : "");
#endif
            bt->wl_prev->wl_next = w->wl_next;
            if (w->wl_next)
                w->wl_next->wl_prev = bt->wl_prev;
            w = w->wl_next;
            if (w)
                w->wl_prev->wl_next = NULL;
            wl_free(bt);
            if (erralso)
                cp_err = cp_out;
        }
        else
            w = w->wl_next;
    }
    *list = wl;
    return;

error:
    wl_free(wl);
    *list = NULL;
    return;
}


/* Reset the cp_* FILE pointers to the standard ones.  This is tricky, since
 * if we are sourcing a command file, and io has been redirected from inside
 * the file, we have to reset it back to what it was for the source, not for
 * the top level.  That way if you type "foo > bar" where foo is a script,
 * and it has redirections of its own inside of it, none of the output from
 * foo will get sent to stdout...
 */

void
cp_ioreset()
{
    if (cp_in != cp_curin) {
        if (cp_in)
            (void) fclose(cp_in);
        cp_in = cp_curin;
    }
    if (cp_out != cp_curout) {
        if (cp_out)
            (void) fclose(cp_out);
        cp_out = cp_curout;
    }
    if (cp_err != cp_curerr) {
        if (cp_err)
            (void) fclose(cp_err);
        cp_err = cp_curerr;
    }

    /*** Minor bug here... */
    out_isatty = true;
    return;
}


static bool
fileexists(name)

char *name;
{
    if (access(name, 0) == 0)
        return (true);
    return (false);
}


/* Fork a shell. */

void
com_shell(wl)

wordlist *wl;
{
    char *com, *shell = NULL;
    long pid, r;
#ifdef HAVE_SIGNAL
    static RETSIGTYPE (*svint)( ), (*svquit)( ), (*svtstp)( );
#endif

    if (SCEDactive()) {
        ShowPrompt("Can't do this from sced.");
        return;
    }

    shell = getenv("SHELL");
    if (shell == NULL)
        shell = "/bin/csh";
    cp_ccon(false);

#ifdef HAVE_FORK
    pid = fork();
    if (pid == 0) {
        cp_fixdescriptors();
        if (wl == NULL) {
            (void) execl(shell, shell, 0);
            _exit(99);
        }
        else {
            com = wl_flatten(wl);
            (void) execl("/bin/sh", "sh", "-c", com, 0);
        }
    }
    else {
        /* XXX Better have all these signals */
        svint = signal(SIGINT, SIG_DFL);
        svquit = signal(SIGQUIT, SIG_DFL);
        svtstp = signal(SIGTSTP, SIG_DFL);
        /* XXX Sig on proc group */
        do {
            r = wait(NULL);
        } while ((r != pid) && pid != -1);
        (void) signal(SIGINT, svint);
        (void) signal(SIGQUIT, svquit);
        (void) signal(SIGTSTP, svtstp);
    }
#else
    /* Easier to forget about changing the io descriptors. */
    if (wl) {
        com = wl_flatten(wl);
        (void) system(com);
        tfree(com);
    }
    else
        (void) system(shell);
#endif
    return;
}


/* Do this only right before an exec, since we lose the old std*'s. */

void
cp_fixdescriptors()
{
    if (cp_in != stdin)
        (void) dup2(fileno(cp_in), fileno(stdin));
    if (cp_out != stdout)
        (void) dup2(fileno(cp_out), fileno(stdout));
    if (cp_err != stderr)
        (void) dup2(fileno(cp_err), fileno(stderr));
    return;
}


/* ARGSUSED */
void
com_rehash(wl)

wordlist *wl;
{
    char *s;

    if (!cp_dounixcom) {
        fprintf(cp_err, "Error: unixcom not set.\n");
        return;
    }
    s = getenv("PATH");
    if (s)
        cp_rehash(s, !cp_nocc);
    else
        fprintf(cp_err, "Error: no PATH in environment.\n");
    return;
}


#ifdef HAVE_CHDIR
void
com_chdir(wl)

wordlist *wl;
{
    char *s;
    struct passwd *pw;
    char localbuf[257];
    int copied = 0;

    s = NULL;

    if (wl == NULL) {

        s = getenv("HOME");

#ifdef HAVE_GETPWUID
        if (!s) {
            pw = getpwuid(getuid());
            if (pw == NULL) {
                fprintf(cp_err, "Can't get your password entry\n");
                return;
            }           
            s = pw->pw_dir;
        }
#endif
        if (!s) {
            if (s = getcwd(localbuf, sizeof(localbuf))) {
                out_printf("current directory: %s\n", s);
                tfree(s);
            }
            else
                fprintf(cp_err, "Can't get current working directory.\n");
            return;
        }
        if (!s) {
            out_printf("Specify directory.\n");
            return;
        }

    }
    else {
        s = copy(wl->wl_word);
        cp_unquote(s);
        copied = 1;
    }

#ifdef MSDOS
    if (s && s[1] == ':') {
        char buf[8];
        sprintf(buf, "%c:", *s);
        system(buf);
        strcpy(s, s + 2);
    }
#endif

    if (*s && chdir(s) == -1)
        perror(s);

    if (copied)
        tfree(s);

    if (cp_dounixcom)
        cp_rehash(NULL,!cp_nocc);
    return;
}

#else

void
com_chdir(wl)

wordlist *wl;
{
    fprintf(cp_err, "Error: 'chdir' not available.\n");
    return;
}

#endif


/* This is a truly evil thing */

void
com_strcmp(wl)

wordlist *wl;
{
    char *var, *s1, *s2;
    int i;

    var = wl->wl_word;
    s1 = copy(wl->wl_next->wl_word);
    cp_unquote(s1);
    s2 = copy(wl->wl_next->wl_next->wl_word);
    cp_unquote(s2);
    
    i = strcmp(s1, s2);
    tfree(s1);
    tfree(s2);

    cp_vset(var, VT_NUM, (char *) &i);
    return;
}
