/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * The front-end command loop.
 */

#include "spice.h"
#include "cpdefs.h"
#include "suffix.h"

/* Stuff to do control structures. We keep a history (seperate from the
 * cshpar history, for now at least) of commands and their event numbers,
 * with a block considered as a statement. In a goto, the first word in
 * co_text is where to go, likewise for label. For conditional controls,
 * we have to call ft_getpnames and ft_evaluate each time, since the
 * dvec pointers will change... Also we should do variable and backquote
 * substitution each time...
 */

struct control {
    int co_type;                  /* One of CO_* ... */
    wordlist *co_cond;            /* if, while, dowhile */
    char *co_foreachvar;          /* foreach */
    int co_numtimes;              /* repeat, break & continue levels */
    wordlist *co_text;            /* Ordinary text and foreach values. */
    struct control *co_parent;    /* If this is inside a block. */
    struct control *co_children;  /* The contents of this block. */
    struct control *co_elseblock; /* For if-then-else. */
    struct control *co_next;
    struct control *co_prev;
} ;


/* named control blocks */
struct block {
    char *bl_name;
    struct control *bl_cont;
} ;

static void *BlockBase; /* hash header for control blocks */

static bool cp_noexec;  /* supress command execution on source */

#define CO_UNFILLED  0
#define CO_STATEMENT 1
#define CO_WHILE     2
#define CO_DOWHILE   3
#define CO_IF        4
#define CO_FOREACH   5
#define CO_BREAK     6
#define CO_CONTINUE  7
#define CO_LABEL     8
#define CO_GOTO      9
#define CO_REPEAT    10

/* We have to keep the control structures in a stack, so that when we do
 * a 'source', we can push a fresh set onto the top...  Actually there have
 * to be two stacks -- one for the pointer to the list of control structs,
 * and one for the 'current command' pointer...
 */

#define CONTROLSTACKSIZE 256    /* Better be enough. */

static struct control *control[CONTROLSTACKSIZE], *cend[CONTROLSTACKSIZE];
static int stackp = 0;

/* Struct from doblock() */

struct retinfo {
    int type;         /* one of defines below */
    union {
        char *label;
        int num;
    } rt;
};

#define NORMAL    1
#define BROKEN    2
#define CONTINUED 3
#define LABEL     4


#ifdef __STDC__
static void set_block(wordlist*);
static void eval_block(struct control*);
static void doblock(struct control*,struct retinfo*);
static wordlist *quicksub(wordlist*);
static void add_block(char*);
static struct control *findlabel(char*,struct control*);
static void freecontrol(struct control*);
static void docommand(wordlist*);
static struct comm *findcom(char*);
static wordlist *getcommand(char*);
static void pwlist(wordlist*,char*);
static void dodump(struct control*);
#else
static void set_block();
static void eval_block();
static void doblock();
static wordlist *quicksub();
static void add_block();
static struct control *findlabel();
static void freecontrol();
static void docommand();
static struct comm *findcom();
static wordlist *getcommand();
static void pwlist();
static void dodump();
#endif


/* Are we waiting for a command? This lets signal handling be 
more clever. */

bool cp_cwait = false;      
char *cp_csep = ";";

bool cp_dounixcom = false;


/* If there is an argument, give this to cshpar to use instead of stdin. In
 * a few places, we call cp_evloop again if it returns 1 and exit (or close
 * a file) if it returns 0... Because of the way sources are done, we can't
 * allow the control structures to get blown away every time we return --
 * probably every time we type source at the keyboard and every time a
 * source returns to keyboard input is ok though -- use ft_controlreset.
 */

static char *noredirect[] = { "stop", NULL } ;  /* Only one?? */

void
com_codeblock(wl)

/* Manipulation of named control blocks */
wordlist *wl;
{
    char *fname = NULL;
    bool remove = false;
    bool print = false;
    bool add = false;
    bool bind = false;
    bool success;
    wordlist *ww, *wn;
    struct control *c;
    struct block *b;

    if (wl) {
        wl = wl_copy(wl);
        for (ww = wl; wl; wl = wn) {
            wn = wl->wl_next;
            if (*wl->wl_word != '-')
                continue;
            else if (strchr(wl->wl_word,'f') || strchr(wl->wl_word,'d'))
                remove = true;
            else if (strchr(wl->wl_word,'p') || strchr(wl->wl_word,'t'))
                print = true;
            else if (strchr(wl->wl_word,'a'))
                add = true;
            else if (strchr(wl->wl_word,'b'))
                bind = true;
            else
                continue;
            if (wl->wl_prev)
                wl->wl_prev->wl_next = wl->wl_next;
            if (wl->wl_next)
                wl->wl_next->wl_prev = wl->wl_prev;
            if (ww == wl)
                ww = wl->wl_next;
            txfree(wl->wl_word);
            txfree((char*)wl);
        }
        if (ww) {
            fname = copy(ww->wl_word);
            wl_free(ww);
        }
    }

    if (fname == NULL) {
        if (bind) {
            cp_bind(NULL);
            return;
        }
        out_init();

        ww = wn = (wordlist*)htab_wl(BlockBase);
        while (ww) {
            out_printf("%s",ww->wl_word);
            if (print || remove)
                b = (struct block *)htab_get(ww->wl_word,BlockBase);
            if (print) {
                out_send(":\n");
                for (c = b->bl_cont; c; c = c->co_next)
                    dodump(c);
            }
            else
                out_send("\n");
            if (remove) {
                /* keep #_ and ## blocks, associated with circuits */
                if (*b->bl_name != '#') {
                    cp_unbind(b->bl_name);
                    freecontrol(b->bl_cont);
                    tfree(b->bl_name);
                    txfree((char*)b);
                    htab_delete(ww->wl_word,BlockBase);
                }
            }
            ww = ww->wl_next;
        }
        wl_free(wn);
        return;
    }

    if (!print && !remove && !add && !bind)
        add = true;

    if (print && !add) {
        out_init();
        b = (struct block *)htab_get(fname,BlockBase);
        if (b) {
            out_printf("%s:\n",b->bl_name);
            if (print) {
                for (c = b->bl_cont; c; c = c->co_next)
                    dodump(c);
            }
        }
    }
    if (remove) {
        cp_unbind(fname);
        cp_freeblock(fname);
    }
    if (add) {
        cp_noexec = true;
        success = cp_block(fname);
        cp_noexec = false;
        if (success)
            add_block(fname);
        else
            fprintf(cp_err,"Error: %s not found, no block added.\n",fname);
    }
    if (print && add) {
        out_init();
        b = (struct block *)htab_get(fname,BlockBase);
        if (b) {
            out_printf("%s:\n",b->bl_name);
            if (print) {
                for (c = b->bl_cont; c; c = c->co_next)
                    dodump(c);
            }
        }
    }
    if (bind) {
        cp_bind(fname);
    }
    txfree(fname);
}


void
cp_addblock(name,wl)

char *name;
wordlist *wl;
{
    bool inter = cp_interactive;

    cp_interactive = false;
    cp_pushcontrol();
    cp_noexec = true;
    for ( ; wl; wl = wl->wl_next)
        cp_evloop(wl->wl_word);
    cp_noexec = false;
    cp_popcontrol();
    cp_interactive = inter;
    add_block(name);
}


bool
cp_isblockdef(name)

char *name;
{

    if (htab_get(name,BlockBase))
        return (true);
    return (false);
}


void
cp_freeblock(name)

/* delete the named block from the list */
char *name;
{
    struct block *b;

    if (!name || *name == '\0') return;

    b = (struct block *)htab_get(name,BlockBase);
    if (b) {
        tfree(b->bl_name);
        freecontrol(b->bl_cont);
        txfree((char*)b);
        htab_delete(name,BlockBase);
    }
}


int
cp_evloop(string)

char *string;
{
    wordlist *wlist, *ww;
/*
commented stuff is for mtrace
FILE *xfp;
*/

    for (;;) {
        wlist = getcommand(string);
        if (wlist == NULL) {
            /* End of file or end of user input,
               could also be a syntax error */
            if (!string) {
                if (cend[stackp] && cend[stackp]->co_parent)
                    cp_resetcontrol();
                cp_event--;
                continue;
            }
            return (0);
        }
        if ((wlist->wl_word == NULL) || (*wlist->wl_word == '\0')) {
            /* User just typed return. */
            tfree(wlist->wl_word);
            tfree(wlist);
            if (string)
                return (1);
            cp_event--;
            continue;
        }

        /* Just a check... */
        for (ww = wlist; ww; ww = ww->wl_next)
            if (!ww->wl_word) {
                fprintf(cp_err, 
            "cp_evloop: Internal Error: NULL word pointer.\n");
                continue;
            }

        set_block(wlist);
/*
if (!string) {
xfp = fopen("mtemp","a");
wl_print(wlist,xfp);
fprintf(xfp,"\n");
}
*/
        wl_free(wlist);
        if (cend[stackp] && !cend[stackp]->co_parent && !cp_noexec) {
            eval_block(cend[stackp]);
            freecontrol(control[stackp]);
            control[stackp] = cend[stackp] = NULL;
        }
        if (string)
            return (1); /* The return value is irrelevant. */
/*
cp_periodic();
mt_dump(xfp);
fclose(xfp);
*/
    }
}


static void
set_block(wl)

wordlist *wl;
{
    struct control *cur;
    wordlist *ww;
    struct control *x;

    /* Add wl to the control structure list. If cend->co_type
     * is CO_UNFILLED, the last line was the beginning of a 
     * block, and this is the unfilled first statement.
     */

#define newblock    cur->co_children = alloc(struct control); \
            cur->co_children->co_parent = cur; \
            cur = cur->co_children; \
            cur->co_type = CO_UNFILLED;

    cur = cend[stackp];

    if (cur && (cur->co_type != CO_UNFILLED)) {
        cur->co_next = alloc(struct control);
        cur->co_next->co_prev = cur;
        cur->co_next->co_parent = cur->co_parent;
        cur = cur->co_next;
    }
    else if (!cur)
        control[stackp] = cur = alloc(struct control);

    if (eq(wl->wl_word, "while")) {
        cur->co_type = CO_WHILE;
        cur->co_cond = wl_copy(wl->wl_next);
        if (!cur->co_cond) {
            fprintf(stderr,"Error: missing while condition.\n");
        }
        newblock;
    }
    else if (eq(wl->wl_word, "dowhile")) {
        cur->co_type = CO_DOWHILE;
        cur->co_cond = wl_copy(wl->wl_next);
        if (!cur->co_cond) {
            fprintf(stderr,"Error: missing dowhile condition.\n");
        }
        newblock;
    }
    else if (eq(wl->wl_word, "repeat")) {
        cur->co_type = CO_REPEAT;
        if (!wl->wl_next) {
            cur->co_numtimes = -1;
        }
        else {
            char *s = NULL;
            double *dd;

            ww = quicksub(wl);
            if (ww && ww->wl_next)
                s = ww->wl_next->wl_word;

            dd = ft_numparse(&s, false);
            if (dd) {
                if (*dd < 0) {
                    fprintf(cp_err, 
        "Error: can't repeat a negative number of times.\n");
                    *dd = 0.0;
                }
                cur->co_numtimes = (int) *dd;
            }
            else
                fprintf(cp_err, 
                "Error: bad repeat argument %s.\n",s ? s : "");
            if (ww)
                wl_free(ww);
        }
        newblock;
    }
    else if (eq(wl->wl_word, "if")) {
        cur->co_type = CO_IF;
        cur->co_cond = wl_copy(wl->wl_next);
        if (!cur->co_cond) {
            fprintf(stderr,"Error: missing if condition.\n");
        }
        newblock;
    }
    else if (eq(wl->wl_word, "foreach")) {
        cur->co_type = CO_FOREACH;
        ww = wl;
        if (ww->wl_next) {
            ww = ww->wl_next;
            cur->co_foreachvar = copy(ww->wl_word);
            ww = ww->wl_next;
        }
        else
            fprintf(stderr,"Error: missing foreach variable.\n");
        ww = wl_copy(ww);
        cp_doglob(&ww);
        cur->co_text = ww;
        newblock;
    }
    else if (eq(wl->wl_word, "label")) {
        cur->co_type = CO_LABEL;
        if (wl->wl_next) {
            cur->co_text = wl_copy(wl->wl_next);
            /* I think of everything, don't I? */
            cp_addkword(CT_LABEL, wl->wl_next->wl_word);
            if (wl->wl_next->wl_next)
                fprintf(cp_err, 
            "Warning: ignored extra junk after label.\n");
        }
        else
            fprintf(stderr, "Error: missing label.\n");
    }
    else if (eq(wl->wl_word, "goto")) {
        /* Incidentally, this won't work if the values 1 and
         * 2 ever get to be valid character pointers -- I
         * think it's reasonably safe to assume they aren't...
         */
        cur->co_type = CO_GOTO;
        if (wl->wl_next) {
            cur->co_text = wl_copy(wl->wl_next);
            if (wl->wl_next->wl_next)
                fprintf(cp_err, 
            "Warning: ignored extra junk after goto.\n");
        }
        else
            fprintf(stderr, "Error: missing label.\n");
    }
    else if (eq(wl->wl_word, "continue")) {
        cur->co_type = CO_CONTINUE;
        if (wl->wl_next) {
            cur->co_numtimes = scannum(wl->wl_next->wl_word);
            if (wl->wl_next->wl_next)
                fprintf(cp_err, 
        "Warning: ignored extra junk after continue %d.\n",
                    cur->co_numtimes);
        }
        else
            cur->co_numtimes = 1;
    }
    else if (eq(wl->wl_word, "break")) {
        cur->co_type = CO_BREAK;
        if (wl->wl_next) {
            cur->co_numtimes = scannum(wl->wl_next->wl_word);
            if (wl->wl_next->wl_next)
                fprintf(cp_err, 
            "Warning: ignored extra junk after break %d.\n",
                    cur->co_numtimes);
        }
        else
            cur->co_numtimes = 1;
    }
    else if (eq(wl->wl_word, "end")) {
        /* Throw away this thing. */
        if (!cur->co_parent) {
            fprintf(stderr, "Error: no block to end.\n");
            cur->co_type = CO_UNFILLED;
        }
        else if (cur->co_prev) {
            cur->co_prev->co_next = NULL;
            x = cur;
            cur = cur->co_parent;
            tfree(x);
        }
        else {
            x = cur;
            cur = cur->co_parent;
            cur->co_children = NULL;
            tfree(x);
        }
    }
    else if (eq(wl->wl_word, "else")) {
        if (!cur->co_parent ||
                (cur->co_parent->co_type != CO_IF)) {
            fprintf(stderr, "Error: misplaced else.\n");
            cur->co_type = CO_UNFILLED;
        }
        else {
            if (cur->co_prev)
                cur->co_prev->co_next = NULL;
            else
                cur->co_parent->co_children = NULL;
            cur->co_parent->co_elseblock = cur;
            cur->co_prev = NULL;
        }
    }
    else {
        cur->co_type = CO_STATEMENT;
        cur->co_text = wl_copy(wl);
    }
    cend[stackp] = cur;
}


static void
eval_block(x)

struct control *x;
{
    struct retinfo ri;

    /* We have to toss this do-while loop in here so
     * that gotos at the top level will work.
     */
    do {
        doblock(x, &ri);
        switch (ri.type) {

            case NORMAL:
                break;

            case BROKEN:
                fprintf(cp_err, 
"Error: break not in loop or too many break levels given.\n");
                break;

            case CONTINUED:
                fprintf(cp_err, 
"Error: continue not in loop or too many continue levels given.\n");
                break;

            case LABEL:
                x = findlabel(ri.rt.label,control[stackp]);
                if (!x)
                    fprintf(cp_err, 
                "Error: label %s not found.\n",ri.rt.label);
                txfree(ri.rt.label);
        }
        if (x)
            x = x->co_next;
    } while (x);
}


/* Execute a block.  There can be a number of return values from this routine.
 * NORMAL indicates a normal termination
 * BROKEN indicates a break -- if the caller is a breakable loop,
 *      terminate it, otherwise pass the break upwards
 * CONTINUED indicates a continue -- if the caller is a continuable loop,
 *      continue, else pass the continue upwards
 * Any other return code is considered a pointer to a string which is
 * a label somewhere -- if this label is present in the block,
 * goto it, otherwise pass it up. Note that this prevents jumping
 * into a loop, which is good.
 * Note that here is where we expand variables, ``, and globs for controls.
 * The 'num' argument is used by break n and continue n.
 */

static void
doblock(bl, info)

struct control *bl;
struct retinfo *info;
{
    struct control *ch, *cn = NULL;
    wordlist *wl, *ww;
    struct retinfo ri;

    switch (bl->co_type) {

        case CO_WHILE:
            while (bl->co_cond && cp_istrue(bl->co_cond)) {
                for (ch = bl->co_children; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    switch (ri.type) {

                        case NORMAL:
                            break;

                        case BROKEN:    /* Break. */
                            if (ri.rt.num < 2) {
                                info->type = NORMAL;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = BROKEN;
                            }
                            return;

                        case CONTINUED: /* Continue. */
                            if (ri.rt.num < 2) {
                                cn = NULL;
                                break;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = CONTINUED;
                                return;
                            }

                        case LABEL:
                            cn = findlabel(ri.rt.label, bl->co_children);
                            if (!cn) {
                                info->type = LABEL;
                                info->rt.label = ri.rt.label;
                                return;
                            }
                            txfree(ri.rt.label);
                    }
                }
            }
            break;

        case CO_DOWHILE:
            do {
                for (ch = bl->co_children; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    switch (ri.type) {

                        case NORMAL:
                            break;

                        case BROKEN:    /* Break. */
                            if (ri.rt.num < 2) {
                                info->type = NORMAL;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = BROKEN;
                            }
                            return;

                        case CONTINUED: /* Continue. */
                            if (ri.rt.num < 2) {
                                cn = NULL;
                                break;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = CONTINUED;
                                return;
                            }

                        case LABEL:
                            cn = findlabel(ri.rt.label, bl->co_children);
                            if (!cn) {
                                info->type = LABEL;
                                info->rt.label = ri.rt.label;
                                return;
                            }
                            txfree(ri.rt.label);
                    }
                }
            } while (bl->co_cond && cp_istrue(bl->co_cond));
            break;

        case CO_REPEAT:
            while ((bl->co_numtimes > 0) ||
                    (bl->co_numtimes == -1)) {
                if (bl->co_numtimes != -1)
                    bl->co_numtimes--;
                for (ch = bl->co_children; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    switch (ri.type) {

                        case NORMAL:
                            break;

                        case BROKEN:    /* Break. */
                            if (ri.rt.num < 2) {
                                info->type = NORMAL;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = BROKEN;
                            }
                            return;

                        case CONTINUED: /* Continue. */
                            if (ri.rt.num < 2) {
                                cn = NULL;
                                break;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = CONTINUED;
                                return;
                            }

                        case LABEL:
                            cn = findlabel(ri.rt.label, bl->co_children);
                            if (!cn) {
                                info->type = LABEL;
                                info->rt.label = ri.rt.label;
                                return;
                            }
                            txfree(ri.rt.label);
                    }
                }
            }
            break;

        case CO_IF:
            if (bl->co_cond && cp_istrue(bl->co_cond)) {
                for (ch = bl->co_children; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    if (ri.type == LABEL) {
                        cn = findlabel(ri.rt.label,bl->co_children);
                        if (!cn) {
                            info->type = LABEL;
                            info->rt.label = ri.rt.label;
                            return;
                        }
                        txfree(ri.rt.label);
                    }
                    else if (ri.type != NORMAL) {
                        info->type = ri.type;
                        info->rt.num = ri.rt.num;
                        return;
                    }
                }
            }
            else {
                for (ch = bl->co_elseblock; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    if (ri.type == LABEL) {
                        cn = findlabel(ri.rt.label,bl->co_elseblock);
                        if (!cn) {
                            info->type = LABEL;
                            info->rt.label = ri.rt.label;
                            return;
                        }
                        txfree(ri.rt.label);
                    }
                    else if (ri.type != NORMAL) {
                        info->type = ri.type;
                        info->rt.num = ri.rt.num;
                        return;
                    }
                }
            }
            break;

        case CO_FOREACH:
            ww = quicksub(bl->co_text);
            for (wl = ww; wl; wl = wl->wl_next) {
                cp_vset(bl->co_foreachvar, VT_STRING,wl->wl_word);
                for (ch = bl->co_children; ch; ch = cn) {
                    cn = ch->co_next;
                    doblock(ch, &ri);
                    switch (ri.type) {

                        case NORMAL:
                            break;

                        case BROKEN:    /* Break. */
                            if (ri.rt.num < 2) {
                                info->type = NORMAL;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = BROKEN;
                            }
                            return;

                        case CONTINUED: /* Continue. */
                            if (ri.rt.num < 2) {
                                cn = NULL;
                                break;
                            }
                            else {
                                info->rt.num = ri.rt.num - 1;
                                info->type = CONTINUED;
                                return;
                            }

                        case LABEL:
                            cn = findlabel(ri.rt.label, bl->co_children);
                            if (!cn) {
                                info->type = LABEL;
                                info->rt.label = ri.rt.label;
                                return;
                            }
                            txfree(ri.rt.label);
                    }
                }
            }
            wl_free(ww);
            break;

        case CO_BREAK:
            if (bl->co_numtimes > 0) {
                info->rt.num = bl->co_numtimes;
                info->type = BROKEN;
                return;
            }
            fprintf(cp_err, "Warning: break %d a no-op\n",
                    bl->co_numtimes);
            break;

        case CO_CONTINUE:
            if (bl->co_numtimes > 0) {
                info->rt.num = bl->co_numtimes;
                info->type = CONTINUED;
                return;
            }
            fprintf(cp_err, "Warning: continue %d a no-op\n",
                    bl->co_numtimes);
            break;

        case CO_GOTO:
            wl = quicksub(bl->co_text);
            if (wl) {
                if (wl->wl_word) {
                    info->type = LABEL;
                    info->rt.label = copy(wl->wl_word);
                    return;
                }
                wl_free(wl);
            }
            break;

        case CO_LABEL:
            /* Do nothing. */
            break;

        case CO_STATEMENT:
            docommand(wl_copy(bl->co_text));
            break;

        case CO_UNFILLED:
            /* There was probably an error here... */
            fprintf(cp_err, "Warning: ignoring previous error.\n");
            break;

        default:
            fprintf(cp_err, 
                "doblock: Internal Error: bad block type %d.\n",
                bl->co_type);
    }
    info->type = NORMAL;
}


static wordlist *
quicksub(wl)

wordlist *wl;
{
    wordlist *nwl;

    nwl = wl_copy(wl);
    cp_doglob(&nwl);
    cp_bquote(&nwl);
    cp_variablesubst(&nwl);
    return (nwl);
}


static void
add_block(name)

/* add the new block to the list of named blocks */
char *name;
{
    struct block *b;

    if (!name || *name == '\0') return;
    if (!control[stackp+1]) {
        fprintf(cp_err, "Error: block is empty, not added.\n");
        return;
    }

    b = (struct block *)htab_get(name,BlockBase);
    if (b)
        freecontrol(b->bl_cont);
    else {
        b = alloc(struct block);
        b->bl_name = copy(name);
        if (BlockBase == NULL)
            BlockBase = htab_init();
        htab_add(name,(void*)b,BlockBase);
    }
    b->bl_cont = control[stackp+1];
    control[stackp+1] = cend[stackp+1] = NULL;
}


void
cp_execcontrol(name)

/* execute the control structure either named, or the one just above
 * if name is NULL
 */
char *name;
{
    struct control *x;
    struct block *b = NULL;

    ++stackp;
    if (name) {
        b = (struct block *)htab_get(name,BlockBase);
        if (!b) {
            fprintf(cp_err,"Error: named block %s not found.\n",name);
            return;
        }
    }
    if (b) {
        x = control[stackp];
        control[stackp] = b->bl_cont;
        eval_block(control[stackp]);
        control[stackp] = x;
    }
    else
        eval_block(control[stackp]);
    --stackp;
}


static struct control *
findlabel(s, ct)

char *s;
struct control *ct;
{
    while (ct) {
        if ((ct->co_type == CO_LABEL) && eq(s, ct->co_text->wl_word))
            break;
        ct = ct->co_next;
    }
    return (ct);
}


static void
freecontrol(cntrl)

/* free the control structure */
struct control *cntrl;
{
    struct control *cc, *cd;

    for (cc = cntrl; cc; cc = cd) {
        wl_free(cc->co_cond);
        wl_free(cc->co_text);
        tfree(cc->co_foreachvar);
        if (cc->co_children)
            freecontrol(cc->co_children);
        if (cc->co_elseblock)
            freecontrol(cc->co_elseblock);
        cd = cc->co_next;
        tfree(cc);
    }
}


void
cp_resetcontrol()

/* This blows away the control structures... */
{
    if (cend[stackp] && cend[stackp]->co_parent)
        fprintf(cp_err, "Warning: EOF before block terminated\n");
    for (stackp = 0; stackp < CONTROLSTACKSIZE; stackp++) {
        if (!control[stackp]) break;
        freecontrol(control[stackp]);
        control[stackp] = cend[stackp] = NULL;
    }
    stackp = 0;
    (void) cp_kwswitch(CT_LABEL, (char *) NULL);
    return;
}


void
cp_popcontrol()

/* Push or pop a new control structure set... */
{
    if (cp_debug)
        fprintf(cp_err, "pop: stackp: %d -> %d\n", stackp, stackp - 1);
    if (stackp < 1)
        fprintf(cp_err, "cp_popcontrol: Internal Error: stack empty\n");
    else
        stackp--;
    return;
}


void
cp_pushcontrol()
{
    if (cp_debug)
        fprintf(cp_err, "push: stackp: %d -> %d\n", stackp, stackp + 1);
    if (stackp > CONTROLSTACKSIZE - 2) {
        fprintf(cp_err, "Error: stack overflow -- max depth = %d\n",
                CONTROLSTACKSIZE);
        stackp = 0;
    }
    else {
        stackp++;
        freecontrol(control[stackp]);
        control[stackp] = cend[stackp] = NULL;
    }
    return;
}



void
cp_toplevel()

/* And this returns to the top level (for use in the interrupt handlers). */
{
    stackp = 0;
    if (cend[stackp])
        while (cend[stackp]->co_parent)
            cend[stackp] = cend[stackp]->co_parent;
    return;
}


/* Note that we only do io redirection when we get to here - we also
 * postpone some other things until now.
 * NOTE! wlist is freed.
 */

static void
docommand(wlist)

wordlist *wlist;
{
    char *r, *s, *t;
    char *lcom;
    int nargs;
    int i;
    struct comm *command;
    wordlist *wl, *nextc, *ee, *rwlist;
    struct block *b;
    FILE *lastin, *lastout, *lasterr;
    bool tmpintr;

    if (cp_debug) {
        printf("docommand ");
        wl_print(wlist, stdout);
        (void) putc('\n', stdout);
    }

    /* do periodic sorts of things */
    cp_periodic();

    /* Do all the things that used to be done by cshpar when the line
     * was read...
     */
    cp_variablesubst(&wlist);
    pwlist(wlist, "After variable substitution");

    cp_bquote(&wlist);
    pwlist(wlist, "After backquote substitution");

    cp_doglob(&wlist);
    pwlist(wlist, "After globbing");

    if (!wlist || !wlist->wl_word)
        return;

    /* Now loop through all of the commands given. */
    rwlist = wlist;
    do {
        for (nextc = wlist; nextc; nextc = nextc->wl_next)
            if (eq(nextc->wl_word, cp_csep))
                break;

        /* Temporarily hide the rest of the command... */
        if (nextc && nextc->wl_prev)
            nextc->wl_prev->wl_next = NULL;
        ee = wlist->wl_prev;
        if (ee)
            wlist->wl_prev = NULL;

        if (nextc == wlist) {
            /* There was no text... */
            goto out;
        }

        /* And do the redirection. */
        cp_ioreset();
        for (i = 0; noredirect[i]; i++)
            if (eq(wlist->wl_word, noredirect[i]))
                break;
        if (!noredirect[i]) {
            cp_redirect(&wlist);
            if (wlist == NULL) {
                cp_ioreset();
                return;
            }
        }

        /* Get rid of all the 8th bits now... */
        cp_striplist(wlist);

        s = wlist->wl_word;

        /* initialize the more'd output */
        out_init();

        /* First check the named blocks. */
        b = (struct block *)htab_get(s,BlockBase);
        if (b) {
            struct control *x;

            cp_pusharg(wlist);
            lastin = cp_curin;
            lastout = cp_curout;
            lasterr = cp_curerr;
            cp_curin = cp_in;
            cp_curout = cp_out;
            cp_curerr = cp_err;
            tmpintr = cp_interactive;
            cp_interactive = false;

            /* replace call to cp_execcontrol() */
            ++stackp;
            x = control[stackp];
            control[stackp] = b->bl_cont;
            eval_block(control[stackp]);
            control[stackp] = x;
            --stackp;

            cp_interactive = tmpintr;
            cp_curin = lastin;
            cp_curout = lastout;
            cp_curerr = lasterr;
            cp_poparg();
            goto out;
        }
        command = findcom(s);
        if (!command) {
            if (cp_oddcomm(s,wlist))
                goto out;
            if (cp_dounixcom && cp_unixcom(wlist))
                goto out;
            fprintf(cp_err,"%s: no such command available in %s\n",
                s, cp_program);
            goto out;
        }
        else if (ft_nutmeg && command->co_spiceonly) {
            /* If it's there but spiceonly, and this is nutmeg, error. */
            fprintf(cp_err,"%s: command available only in spice\n",s);
            goto out;
        }

        nargs = 0;
        for (wl = wlist->wl_next; wl; wl = wl->wl_next)
            nargs++;
        if (command->co_stringargs) {
            lcom = wl_flatten(wlist->wl_next);
            (*command->co_func) (lcom);
            tfree(lcom);
        }
        else {
            if (nargs < command->co_minargs) {
                if (command->co_argfn) {
                    (*command->co_argfn) (wlist->wl_next, command);
                }
                else {
                    fprintf(cp_err, "%s: too few args.\n", s);
                }
            }
            else if (nargs > command->co_maxargs) {
                fprintf(cp_err, "%s: too many args.\n", s);
            }
            else
                (*command->co_func) (wlist->wl_next);
        }

out:
        /* Now fix the pointers and advance wlist. */
        wlist->wl_prev = ee;
        if (nextc) {
            if (nextc->wl_prev)
                nextc->wl_prev->wl_next = nextc;
            wlist = nextc->wl_next;
        }
    } while (nextc && wlist);

    wl_free(rwlist);
    cp_ioreset();
    return;
}


static struct comm *
findcom(cmd)

char *cmd;
{
    static void *comhash;
    int i;

    if (!comhash) {
        comhash = htab_init();
        for (i = 0; cp_coms[i].co_comname; i++)
            htab_add(cp_coms[i].co_comname,(void*)&cp_coms[i],comhash);
    }
    return (struct comm *)htab_get(cmd,comhash);
}


/* Get a command. This does all the bookkeeping things like turning
 * command completion on and off...
 */

static wordlist *
getcommand(string)

char *string;
{
    wordlist *wlist;
    int i, j;
    static char buf[64];
    struct control *c;

    if (cp_debug)
        fprintf(cp_err, "calling getcommand %s\n", 
                string ? string : "");
    cp_altprompt = NULL;
    if (cend[stackp]) {
        i = 0;
        for (c = cend[stackp]->co_parent; c; c = c->co_parent)
            i++;
        if (i) {
            for (j = 0; j < i; j++)
                buf[j] = '>';
            buf[j] = ' ';
            buf[j + 1] = '\0';
            cp_altprompt = buf;
        }
    }
    cp_cwait = true;
    wlist = cp_parse(string);
    cp_cwait = false;
    if (cp_debug) {
        printf("getcommand ");
        wl_print(wlist, stdout);
        (void) putc('\n', stdout);
    }
    return (wlist);
}


/* This is also in cshpar.c ... */

static void
pwlist(wlist, name)

wordlist *wlist;
char *name;
{
    wordlist *wl;

    if (!cp_debug)
        return;
    fprintf(cp_err, "%s : [ ", name);
    for (wl = wlist; wl; wl = wl->wl_next)
        fprintf(cp_err, "%s ", wl->wl_word);
    fprintf(cp_err, "]\n");
    return;
}


static int indent;

/* ARGSUSED */
void
com_cdump(wl)

wordlist *wl;
{
    struct control *c;

    out_init();
    indent = 0;
    for (c = control[stackp]; c; c = c->co_next)
        dodump(c);
    return;
}


#define tab(num)    for (i = 0; i < num; i++) out_send(" ");

static void
dodump(cc)

struct control *cc;
{
    int i;
    struct control *tc;

    switch (cc->co_type) {
        case CO_UNFILLED:
            tab(indent);
            out_printf( "(unfilled)\n");
            break;
        case CO_STATEMENT:
            tab(indent);
            out_wlprint(cc->co_text);
            out_send("\n");
            break;
        case CO_WHILE:
            tab(indent);
            out_printf( "while ");
            out_wlprint(cc->co_cond);
            out_send("\n");
            indent += 8;
            for (tc = cc->co_children; tc; tc = tc->co_next)
                dodump(tc);
            indent -= 8;
            tab(indent);
            out_printf( "end\n");
            break;
        case CO_REPEAT:
            tab(indent);
            out_printf( "repeat ");
            if (cc->co_numtimes != -1)
                out_printf( "%d\n", cc->co_numtimes);
            else
                out_send("\n");
            indent += 8;
            for (tc = cc->co_children; tc; tc = tc->co_next)
                dodump(tc);
            indent -= 8;
            tab(indent);
            out_printf( "end\n");
            break;
        case CO_DOWHILE:
            tab(indent);
            out_printf( "dowhile ");
            out_wlprint(cc->co_cond);
            out_send("\n");
            indent += 8;
            for (tc = cc->co_children; tc; tc = tc->co_next)
                dodump(tc);
            indent -= 8;
            tab(indent);
            out_printf( "end\n");
            break;
        case CO_IF:
            tab(indent);
            out_printf( "if ");
            out_wlprint(cc->co_cond);
            out_send("\n");
            indent += 8;
            for (tc = cc->co_children; tc; tc = tc->co_next)
                dodump(tc);
            indent -= 8;
            tab(indent);
            out_printf( "end\n");
            break;
        case CO_FOREACH:
            tab(indent);
            out_printf( "foreach %s ", cc->co_foreachvar);
            out_wlprint(cc->co_text);
            out_send("\n");
            indent += 8;
            for (tc = cc->co_children; tc; tc = tc->co_next)
                dodump(tc);
            indent -= 8;
            tab(indent);
            out_printf( "end\n");
            break;
        case CO_BREAK:
            tab(indent);
            if (cc->co_numtimes != 1)
                out_printf( "break %d\n", cc->co_numtimes);
            else
                out_printf( "break\n");
            break;
        case CO_CONTINUE:
            tab(indent);
            if (cc->co_numtimes != 1)
                out_printf( "continue %d\n",
                        cc->co_numtimes);
            else
                out_printf( "continue\n");
            break;
        case CO_LABEL:
            tab(indent);
            out_printf( "label %s\n", cc->co_text->wl_word);
            break;
        case CO_GOTO:
            tab(indent);
            out_printf( "goto %s\n", cc->co_text->wl_word);
            break;
        default:
            tab(indent);
            out_printf( "bad type %d\n", cc->co_type);
            break;
    }
    return;
}

