/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Expand global characters.
 */

#include "spice.h"
#include "misc.h"
#include "cpdefs.h"

#include <sys/types.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#ifndef direct
#define direct dirent
#endif
#else
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#endif
#ifdef HAVE_GETPWUID
#include <pwd.h>
#endif

#ifdef __STDC__
static wordlist *bracexpand(char*);
static wordlist *brac1(char*);
static wordlist *brac2(char*);
static wordlist *globexpand(char*);
static char *pcanon(char*);
static int sortcmp(char**,char**);
static char *tilde_expand(char*);
static bool noglobs(char*);
#else
static wordlist *bracexpand();
static wordlist *brac1();
static wordlist *brac2();
static wordlist *globexpand();
static char *pcanon();
static int sortcmp();
static char *tilde_expand();
static bool noglobs();
#endif

char cp_comma = ',';
char cp_ocurl = '{';
char cp_ccurl = '}';
char cp_huh   = '?';
char cp_star  = '*';
char cp_obrac = '[';
char cp_cbrac = ']';
char cp_til   = '~';

/* For each word, go through two steps: expand the {}'s, and then do ?*[]
 * globbing in them. Sort after the second phase but not the first...
 */

void
cp_doglob(list)

wordlist **list;
{
    wordlist *wl, *w, *nwl, *wlist;
    char *s;

    /* Expand {a,b,c} */

    if (list == NULL) return;

    wlist = *list;

    for (wl = wlist; wl; wl = wl->wl_next) {
        w = bracexpand(wl->wl_word);
        if (!w) {
            wl_free(wlist);
            *list = NULL;
            return;
        }
        nwl = wl_splice(wl, w);
        if (wlist == wl)
            wlist = w;
        wl = nwl;
    }

    /* Do tilde expansion. */

    for (wl = wlist; wl; wl = wl->wl_next)
        if (*wl->wl_word == cp_til) {
            s = cp_tildexpand(wl->wl_word);
            tfree(wl->wl_word);
            if (!s)
                wl->wl_word = copy("");
            else
                wl->wl_word = s;
        }

    /* Now, expand *?[] for each word. unset * and unalias * mean 
     * something special
     */

    if ((cp_noglob == true) || eq(wlist->wl_word, "unset") ||
            eq(wlist->wl_word, "unalias")) {
        *list = wlist;
        return;
    }

    for (wl = wlist; wl; wl = wl->wl_next) {
        if (noglobs(wl->wl_word))
            continue;
        w = globexpand(wl->wl_word);
        if (w == NULL)
            continue;
        nwl = wl_splice(wl, w);
        if (wlist == wl)
            wlist = w;
        wl = nwl;
    }
    *list = wlist;
    return;
}


static wordlist *
bracexpand(string)

char *string;
{
    wordlist *wl, *w;
    char *s;

    if (!string)
        return (NULL);
    wl = brac1(string);
    if (!wl)
        return (NULL);
    for (w = wl; w; w = w->wl_next) {
        s = w->wl_word;
        w->wl_word = copy(s);
        tfree(s);
    }
    return (wl);
}


/* Given a string, returns a wordlist of all the {} expansions. This is
 * called recursively by cp_brac2(). All the words here will be of size
 * BSIZE_SP, so it is a good idea to copy() and free() the old words.
 */

static wordlist *
brac1(string)

char *string;
{
    wordlist *words, *wl, *w, *nw, *nwl, *newwl;
    char *s;
    int nb;

    words = alloc(struct wordlist);
    words->wl_word = tmalloc(BSIZE_SP);
    for (s = string; *s; s++) {
        if (*s == cp_ocurl) {
            nwl = brac2(s);
            nb = 0;
            for (;;) {
                if (*s == cp_ocurl)
                    nb++;
                if (*s == cp_ccurl)
                    nb--;
                if (*s == '\0') {   /* { */
                    fprintf(cp_err, "Error: missing }.\n");
                    wl_free(words);
                    return (NULL);
                }
                if (nb == 0)
                    break;
                s++;
            }
            /* Add nwl to the rest of the strings in words. */
            newwl = NULL;
            for (wl = words; wl; wl = wl->wl_next)
                for (w = nwl; w; w = w->wl_next) {
                    nw = alloc(struct wordlist);
                    nw->wl_word = tmalloc(BSIZE_SP);
                    (void) strcpy(nw->wl_word, wl->wl_word);
                    (void) strcat(nw->wl_word, w->wl_word);
                    newwl = wl_append(newwl, nw);
                }
            wl_free(words);
            words = newwl;
        }
        else
            for (wl = words; wl; wl = wl->wl_next)
                appendc(wl->wl_word, *s);
    }
    return (words);
}


/* Given a string starting with a {, return a wordlist of the expansions
 * for the text until the matching }.
 */

static wordlist *
brac2(string)

char *string;
{
    wordlist *wlist = NULL, *nwl;
    char buf[BSIZE_SP], *s;
    int nb;
    bool eflag = false;

    string++;   /* Get past the first open brace... */
    for (;;) {
        (void) strcpy(buf, string);
        nb = 0;
        s = buf;
        for (;;) {
            if ((*s == cp_ccurl) && (nb == 0)) {
                eflag = true;
                break;
            }
            if ((*s == cp_comma) && (nb == 0))
                break;
            if (*s == cp_ocurl)
                nb++;
            if (*s == cp_ccurl)
                nb--;
            if (*s == '\0') {       /* { */
                fprintf(cp_err, "Error: missing }.\n");
                wl_free(wlist);
                return (NULL);
            }
            s++;
        }
        *s = '\0';
        nwl = brac1(buf);
        wlist = wl_append(wlist, nwl);
        string += s - buf + 1;
        if (eflag)
            return (wlist);
    }
}


/* Return a wordlist, with *?[] expanded and sorted. This is the idea: set
 * up an array with possible matches, and go through each path component
 * and search the appropriate directories for things that match, and add
 * those that do to the array.
 */

static wordlist *
globexpand(string)

char *string;
{
    char *poss[MAXWORDS];
    char buf[BSIZE_SP];
    char *s;
    char *point;    /* Where we are at in the pathname. */
    int i, j, level;
    bool found;
    wordlist *wlist = NULL, *wl, *lwl = NULL;   /* Make lint shut up. */
#if defined(HAVE_DIRENT_H) || defined(HAVE_SYS_DIR_H)
    DIR *wdir;
    struct direct *de;
#endif

    bzero((char *) poss, MAXWORDS * sizeof (char *));
    string = pcanon(string);
    point = string;

    if (*point == DIR_TERM) {
        point++;
        poss[0] = copy(DIR_PATHSEP);
    }
    else if (point[0] == '.' && point[1] == '.' && point[2] == DIR_TERM) {
        poss[0] = copy("..");
        point += 3;
    }
#ifdef MSDOS
    else if (point[1] == ':') {
        point += 2;
        if (point[0] == DIR_TERM) {
            point++;
            poss[0] = copy("?:\\");
        }
        else
            poss[0] = copy("?:.");
        *poss[0] = *string;
    }
#endif
    else
        poss[0] = copy(DIR_CWD);

    level = 0;
nextcomp:
    level++;
    (void) strcpy(buf, point);
    s = strchr(buf, DIR_TERM);
    if (s)
        *s = '\0';
#if defined(HAVE_DIRENT_H) || defined(HAVE_SYS_DIR_H)
    for (i = 0; i < MAXWORDS; i++) {
        if (!poss[i] || (poss[i][0] == '\0'))
            continue;
        found = false;
        wdir = opendir(poss[i]);
        if (wdir == NULL) {
            if (level > 1) {
                tfree(poss[i]);
                continue;
            }
            goto err;
        }
        while ((de = readdir(wdir)) != NULL)
            if (cp_globmatch(buf, de->d_name)) {
                found = true;
                for (j = 0; j < MAXWORDS; j++)
                    if (!poss[j])
                        break;
                if (j == MAXWORDS) {
                    fprintf(cp_err, 
                        "Too many arguments.\n");
                    wl = NULL;
                    goto err1;
                }
                poss[j] = tmalloc(BSIZE_SP);
                (void) strcpy(poss[j] + 1, poss[i]);
                (void) strcat(poss[j] + 1, DIR_PATHSEP);
                (void) strcat(poss[j] + 1, de->d_name);
            }
        tfree(poss[i]);
        poss[i] = NULL;
        (void) closedir(wdir);
        if (!found) {
            if (level > 1) {
                continue;
            }
            goto err;
        }
    }
    /* Hide the newly found words from the globbing process by making 
     * the first byte a '\0'.
     */
    for (i = 0; i < MAXWORDS; i++)
        if (poss[i])
            for (j = 0; poss[i][j] = poss[i][j+1]; j++);
    if (strchr(point, DIR_TERM)) {
        point = strchr(point, DIR_TERM) + 1;
        goto nextcomp;
    }
#endif

    /* Compact everything properly. */

    for (i = j = 0; i < MAXWORDS; i++) {
        if (!poss[i]) continue;

        if (i != j) {
            poss[j] = poss[i];
            poss[i] = NULL;
        }
        j++;
    }
    if (j == 0)
        goto err;

    /* Now, sort the stuff and make it into wordlists. */

    qsort((char *) poss, j, sizeof (char *),
#ifdef __STDC__
        (int(*)(const void*,const void*))sortcmp);
#else
        sortcmp);
#endif

    for (i = 0; i < j; i++) {

        if (wlist == NULL)
            wlist = wl = alloc(wordlist);
        else {
            wl->wl_next = alloc(wordlist);
            wl->wl_next->wl_prev = wl;
            wl = wl->wl_next;
        }
        wl->wl_word = pcanon(poss[i]);
        tfree(poss[i]);
    }
    tfree(string);
    return (wlist);

err:
    if (cp_nonomatch) {
        wl = alloc(wordlist);
        wl->wl_word = copy(string);
    }
    else {
        fprintf(cp_err, "%s: no match.\n", string);
        wl = NULL;
    }

err1:
    for (i = 0; i < MAXWORDS; i++)
        if (poss[i])
            tfree(poss[i]);
    return (NULL);
}


/* Normalize filenames (get rid of extra ///, .///... etc.. ) */

static char *
pcanon(string)

char *string;
{
    char *p, *s;
    
    s = p = tmalloc(strlen(string) + 1);

bcomp:
    if (!strncmp(string, DIR_CWD, sizeof(DIR_CWD) - 1)
        && (*(string + 1) == DIR_TERM)) {
        string += 2;
        goto bcomp;
    }
morew:
    if (*string == DIR_TERM) {
        *s++ = DIR_TERM;
        while (*++string == DIR_TERM);
        goto bcomp;
    }
    if (!*string) {
        if (*(s - 1) == DIR_TERM)
            s--;
        *s = '\0';
        return (p);
    }
    *s++ = *string++;
    goto morew;
}


static int
sortcmp(s1, s2)

char **s1, **s2;
{
    char *a, *b;

    a = *s1;
    b = *s2;
    for (;;) {
        if (*a > *b)
            return (1);
        if (*a < *b)
            return (-1);
        if (*a == '\0')
            return (0);
        a++;
        b++;
    }
}


/* Expand tildes. */

char *
cp_tildexpand(string)

char *string;
{
    char *result;

    result = tilde_expand(string);

    if (!result) {
        if (cp_nonomatch)
            return (copy(string));
        else
            return (NULL);
    }
    return (result);
}


void
cp_pathfix(buf)

char *buf;
{
#ifdef MSDOS
    char *s, *t;
    int bcnt = 0, ecnt = 0;

    s = t = buf;
    while (*t != '\0') {
        if (*t == '/' || *t == '\\') {
            *s++ = '\\';
            t++;
            bcnt = 0;
            ecnt = 0;
        }
        else if (*t == '.') {
            *s++ = *t++;
            ecnt = 1;
        }
        else if (!ecnt) {
            if (bcnt++ < 8) *s++ = *t++;
            else t++;
        }
        else {
            if (ecnt++ < 4) *s++ = *t++;
            else t++;
        }
    }
    *s = '\0';

#else
    char *s;

    if (index(buf,'~')) {
        s = cp_tildexpand(buf);
        if (s) {
            strcpy(buf,s);
            txfree(s);
        }
    }
#endif
}


static char *
tilde_expand(string)

char *string;
{
#ifdef HAVE_GETPWUID
    struct passwd *pw;
    char    *tail;
    char    buf[BSIZE_SP];
    char    *k, c;

    if (!string)
    return (NULL);

    while (*string && isspace(*string))
    string++;

    if (*string != '~')
        return copy(string);

    string += 1;

    if (!*string || *string == '/') {
        pw = getpwuid(getuid());
        *buf = 0;
    }
    else {
        k = buf;
        while ((c = *string) && c != '/')
            *k++ = c, string++;
        *k = 0;
        pw = getpwnam(buf);
    }

    if (pw) {
        strcpy(buf, pw->pw_dir);
        if (*string)
            strcat(buf, string);
    }
    else
        return (NULL);

    return (copy(buf));

#else
    return (copy(string));
#endif
}


/* Say whether the pattern p can match the string s. */

bool
cp_globmatch(p, s)

char *p, *s;
{
    char schar, pchar, bc, fc;
    bool bchar, except;

    if ((*s == '.') && ((*p == cp_huh) || (*p == cp_star)))
        return (false);

    for (;;) {
        schar = strip(*s++);
        pchar = *p++;
        if (pchar == cp_star) {
            if (*p == '\0')
                return (true);
            for (s--; *s != '\0'; s++)
                if (cp_globmatch(p, s))
                    return (true);
            return (false);
        }
        else if (pchar == cp_huh) {
            if (schar == '\0')
                return (false);
            continue;
        }
        else if (pchar == cp_obrac) {
            bchar = false;
            if (*p == '^') {
                except = true;
                p++;
            }
            else
                except = false;
            fc = -1;
            while (bc = *p++) {
                if (bc == cp_cbrac) {
                    if ((bchar && !except) || 
                        (!bchar && except))
                        break;
                    else
                        return (false);
                }
                if (bc == '-') {
                    if (fc <= schar && schar <= *p++)
                        bchar = true;
                }
                else {
                    fc = bc;
                    if (fc == schar)
                        bchar = true;
                }
            }
            if (bc == '\0') {
                fprintf(cp_err, "Error: missing ].\n");
                return (false);
            }
            continue;
        }
        else if (pchar == '\0') {
            if (schar == '\0')
                return (true);
            else
                return (false);
        }
        else {
            if (strip(pchar) != schar)
                return (false);
            continue;
        }
    }
}


static bool
noglobs(string)

char *string;
{
    if (strchr(string, cp_star) || strchr(string, cp_huh) || 
            strchr(string, cp_obrac))
        return (false);
    else
        return (true);
}
