/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Initial lexer.
 */

/***************************************************************************
To implement command completion under DOS, one must define TIOCSTI in
os_msdos.h, and add routines mygetc() and ioctl() which do the right
things, i.e., allow stuffing characters into the input buffer.  In DOS,
^A and ^Z call the abort routine in interactive mode. ^D signals command
match printing and can be first on a line unlike in UNIX.  Beware that
some compilers supply an ioctl() routine which won't work here, if this
is the case, you may have to define ioctl to something else in os_msdos.h.
****************************************************************************/

#include "spice.h"
#include "cpdefs.h"
#include <errno.h>

#include <sys/types.h>

#ifdef HAVE_SGTTY_H
#include <sgtty.h>
#else
#ifdef HAVE_TERMIO_H
#include <termio.h>
#else
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#endif
#endif

#include "ftegraph.h"

#ifdef __STDC__
static void prompt(void);
#else
static void prompt();
#endif

FILE *cp_inp_cur = NULL;
int cp_event = 1;
bool cp_interactive = true;
bool cp_bqflag = false;
char *cp_promptstring = NULL;
char *cp_altprompt = NULL;
char cp_hash = '#';

static int numeofs = 0;

#define ESCAPE  '\033'

/* Return a list of words, with backslash quoting and '' quoting done.
 * Strings en(void) closed in "" or `` are made single words and returned,
 * but with the "" or `` still present. For the \ and '' cases, the
 * 8th bit is turned on (as in csh) to prevent them from being recogized,
 * and stripped off once all processing is done. We also have to deal with
 * command, filename, and keyword completion here.
 * If string is non-NULL, then use it instead of the fp. Escape and EOF
 * have no business being in the string.
 */

#define newword cw->wl_word = copy(buf); \
        cw->wl_next = alloc(struct wordlist); \
        cw->wl_next->wl_prev = cw; \
        cw = cw->wl_next; \
        bzero(buf, BSIZE_SP); \
        i = 0;


wordlist *
cp_lexer(string)

char *string;
{
    int c;
    int i, j;
    wordlist *wlist = NULL, *cw = NULL;
    char buf[BSIZE_SP], linebuf[BSIZE_SP], d;

    if (cp_inp_cur == NULL)
        cp_inp_cur = cp_in;

    if (!string && cp_interactive) {
        cp_ccon(true);
        prompt();
    }
nloop:
    i = 0;
    j = 0;
    bzero(linebuf, BSIZE_SP);
    bzero(buf, BSIZE_SP);
    wlist = cw = alloc(struct wordlist);
    cw->wl_next = cw->wl_prev = NULL;
    for (;;) {
        if (string) {
            c = *string++;
            if (c == '\0')
                c = '\n';
            if (c == ESCAPE)
                c = '[';
        }
        else
            c = DevGetchar(cp_inp_cur);
gotchar:
        if ((c != EOF) && (c != ESCAPE) && (c != 4)) /* SRW - 4 added */
            linebuf[j++] = c;
        if (c != EOF && c != 4) /* SRW - ditto */
            numeofs = 0;
        if (i == BSIZE_SP - 1) {
            fprintf(cp_err, "Warning: word too long.\n");
            c = ' ';
        }
        if (j == BSIZE_SP - 1) {
            fprintf(cp_err, "Warning: line too long.\n");
            if (cp_bqflag)
                c = EOF;
            else
                c = '\n';
        }
        if (c != EOF)
            c = strip(c);   /* Don't need to do this really. */

        /* "backslash" quoting */
        if (c == '\\' && DIR_TERM != '\\' || c == '\026' /* ^V */ ) {
            c = quote(string ? *string++ : DevGetchar(cp_inp_cur));
            linebuf[j++] = strip(c);
        }

        if ((c == '\n') && cp_bqflag)
            c = ' ';
        if ((c == EOF || c == 4) && cp_bqflag) /* SRW - 4 added */
            c = '\n';
        if ((c == cp_hash) && !cp_interactive && (j == 1)) {
            if (string)
                return (NULL);
            while (((c = DevGetchar(cp_inp_cur)) != '\n') &&
                    (c != EOF));
            goto nloop;
        }
        switch (c) {

    case ' ':
    case '\t':
            if (i > 0) {
                newword;
            }
            break;

    case '\n':
            if (i) {
                buf[i] = '\0';
                cw->wl_word = copy(buf);
            }
            else if (cw->wl_prev) {
                cw->wl_prev->wl_next = NULL;
                tfree(cw);
            }
            else {
                cw->wl_word = NULL;
            }
            goto done;

    case '\'':
            while (((c = (string ? *string++ : 
                    DevGetchar(cp_inp_cur))) != '\'')
                    && (i < BSIZE_SP - 1)) {
                if ((c == '\n') || (c == EOF) || (c == ESCAPE))
                    goto gotchar;
                else {
                    buf[i++] = quote(c);
                    linebuf[j++] = c;
                }
            }
            linebuf[j++] = '\'';
            break;

    case '"':
    case '`':
            d = c;
            buf[i++] = d;
            while (((c = (string ? *string++ : DevGetchar(cp_inp_cur)))
                    != d) && (i < BSIZE_SP - 2)) {
                if ((c == '\n') || (c == EOF) || (c == ESCAPE))
                    goto gotchar;
                else if (c == '\\') {
                    linebuf[j++] = c;
                    c = (string ? *string++ :
                            DevGetchar(cp_inp_cur));
                    buf[i++] = quote(c);
                    linebuf[j++] = c;
                }
                else {
                    buf[i++] = c;
                    linebuf[j++] = c;
                }
            }
            buf[i++] = d;
            linebuf[j++] = d;
            break;

    case EOF:
    case 4: /* ^D */

            if (cp_interactive && !cp_nocc && 
                    (string == NULL)) {
#ifdef notdef
/* ^D logout */
                if (j == 0) {
                    if (cp_ignoreeof && (numeofs++
                            < 23)) {
                        fputs(
                    "Use \"quit\" to quit.\n",
                            stdout);
                    }
                    else {
                        fputs("quit\n", stdout);
                        cp_doquit();
                    }
                    goto done;
                }
#endif
                cp_ccom(wlist, buf, false);
                wl_free(wlist);
                (void) fputc('\r', cp_out);
                prompt();
                for (j = 0; linebuf[j]; j++)
#ifdef TIOCSTI
                    /* SRW - below was cp_out, a bug */
                    (void) ioctl(fileno(cp_in), TIOCSTI,
                            linebuf + j);
#else
                fputc(linebuf[j], cp_out);    /* But you can't edit */
#endif
                goto nloop;
            }
            else {   /* EOF during a source */

                if (cp_interactive) {
                    fputs("quit\n", stdout);
                    cp_doquit();
                    goto done;
                }
                else
                    return (NULL);
            }
    case ESCAPE:
            if (cp_interactive && !cp_nocc) {

                fputs("\b\b  \b\b\r", cp_out);
                prompt();
                for (j = 0; linebuf[j]; j++)
#ifdef TIOCSTI
                /* SRW - below was cp_out, a bug */
                (void) ioctl(fileno(cp_in), TIOCSTI,
                        linebuf + j);
#else
                fputc(linebuf[j], cp_out);    /* But you can't edit */
#endif
                cp_ccom(wlist, buf, true);
                wl_free(wlist);
                goto nloop;
            } /* Else fall through */

    default:
            /* We have to remember the special cases $<, $&, $#&, $?& */
            if ((cp_chars[c] & CPC_BRL) && (i > 0)) {
                switch (c) {

                    case '<':
                        if (buf[i-1] == '$')
                            break;
                        { newword; }
                        break;

                    case '&':
                        if (buf[i-1] == '$')
                            break;
                        if (i > 1 && (buf[i-1] == '#' || buf[i-1] == '?') &&
                                buf[i-2] == '$')
                            break;
                        { newword; }
                        break;

                    default:
                        { newword; }
                }
            }

            buf[i++] = c;

            /* have to also catch <=, >=, and <> */

            if (cp_chars[c] & CPC_BRR) {
                switch (c) {

                    case '<':
                        if (i > 1 && buf[i-2] == '$')
                            break;

                    case '>':
                        d = c;
                        if (string) {
                            c = *string++;
                            if (c == '\0')
                                c = '\n';
                            if (c == ESCAPE)
                                c = '[';
                        }
                        else
                            c = DevGetchar(cp_inp_cur);

                        if (c == '=' || (d == '<' && c == '>')) {
                            buf[i++] = c;
                            linebuf[j++] = c;
                            { newword; }
                            break;
                        }
                        { newword; }
                        goto gotchar;

                    case '&':
                        if (i > 1 && buf[i-2] == '$')
                            break;
                        if (i > 2 && (buf[i-2] == '#' || buf[i-2] == '?') &&
                                buf[i-3] == '$')
                            break;
                        { newword; }
                        break;

                    default:
                        { newword; }
                }
            }
        }
    }
done:
    return (wlist);
}


static void
prompt()
{
    char *s;

    if (cp_interactive == false)
        return;
    if (cp_promptstring == NULL)
        s = "-> ";
    else
        s = cp_promptstring;
    if (cp_altprompt)
        s = cp_altprompt;

    while (*s) {
        switch (strip(*s)) {
            case '!':
                fprintf(cp_out, "%d", cp_event);
                break;
            case '\\':
                if (*(s + 1))
                    (void) putc(strip(*++s), cp_out);

            /* SRW  -p prints cwd */
            case '-':
                if (*(s+1) == 'p') {
                    char *c, *getcwd();
                    (void) fprintf(cp_out,"%s",c = getcwd(NULL,128));
                    tfree(c);
                    s++;
                    break;
                }
            default:
                (void) putc(strip(*s), cp_out);
        }
        s++;
    }
    (void) fflush(cp_out);
    return;
}
