/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Do variable substitution.
 */

#include "spice.h"
#include "misc.h"
#include "cpdefs.h"
#include "suffix.h"

#ifdef __STDC__
static wordlist *vareval(char*);
static wordlist *var2wl(struct variable*,int,int);
static char *eval_range(char*);
static char *iv_special(char*);
static int vcmp();
static struct variable *getlist(wordlist**);
#else
static wordlist *vareval();
static wordlist *var2wl();
static char *eval_range();
static char *iv_special();
static int vcmp();
static struct variable *getlist();
#endif

/* A variable substitution is
 * indicated by a $, and the variable name is the following string of
 * non-special characters. All variable values are inserted as a single
 * word, except for lists, which are a list of words.
 * A routine cp_usrset must be supplied by the host program to deal 
 * with variables that aren't used by cshpar -- it should be 
 * cp_usrset(var, isset), where var is a variable *, and isset is 
 * true if the variable is being set, false if unset.
 * Also required is a routine cp_enqvar(name) which returns a struct
 * variable *, which allows the host program to provide values for
 * non-cshpar variables.
 */

bool cp_noglob = true;      /* This is just a bad thing. */
bool cp_nonomatch = false;
bool cp_noclobber = false;
bool cp_ignoreeof = false;

char cp_dol = '$';

static void *variables;

/* Non-alphanumeric characters that may appear in variable names. < is very
 * special...
 */

#define VALIDCHARS "_<#?@.()[]&"

void
cp_variablesubst(list)

wordlist **list;
{
    wordlist *wl, *nwl, *wlist, *twl;
    char *s, *t, buf[BSIZE_SP], wbuf[BSIZE_SP];
    int i, lev1, lev2;


    if (list == NULL)
        return;
    wlist = *list;
    for (wl = wlist; wl; wl = wl->wl_next) {
        t = wl->wl_word;
        i = 0;
        while (s = strchr(t, cp_dol)) {
            while (t < s)
                wbuf[i++] = *t++;
            wbuf[i] = '\0';
            (void) strcpy(buf, ++s);
            s = buf;
            t++;
            /* Get s and t past the end of the var name. */
            if (*s &&
                (isalphanum(*s) || strchr(VALIDCHARS, *s) || *s == cp_dol)) {
                t++;
                s++;
            }
            lev1 = lev2 = 0;
            for ( ; *s; t++,s++) {
                if (isalphanum(*s))
                    continue;
                if (*s == '[') {
                    lev1++;
                    continue;
                }
                if (*s == '(') {
                    lev2++;
                    continue;
                }
                if (*s == ']') {
                    if (lev1) {
                        lev1--;
                        continue;
                    }
                    break;
                }
                if (*s == ')') {
                    if (lev2) {
                        lev2--;
                        continue;
                    }
                    break;
                }
                if (strchr(VALIDCHARS, *s))
                    continue;
                if (lev1 || lev2)
                    continue;
                break;
            }
            if (!*s && (lev1 || lev2)) {
                char *str;
                wordlist *ww;

                /* split token, concatenate them */
                if (wl->wl_next) {
                    ww = wl->wl_next;
                    str = tmalloc(
                        strlen(wl->wl_word) + strlen(ww->wl_word) + 1);
                    strcpy(str,wl->wl_word);
                    strcat(str,ww->wl_word);
                    txfree(wl->wl_word);
                    wl->wl_word = str;
                    txfree(ww->wl_word);
                    wl->wl_next = ww->wl_next;
                    wl->wl_next->wl_prev = wl;
                    txfree((char*)ww);
                    t = wl->wl_word;
                    i = 0;
                    continue;
                }
                else {
                    fprintf(cp_err,
                        "Error: unbalanced parentheses or brackets\n");
                    wl_print(wlist,cp_err);
                    fprintf(cp_err,"\n");
                    *list = NULL;
                    return;
                }
            }

            *s = '\0';
            nwl = vareval(buf);
            if (i) {
                (void) strcpy(buf, wbuf);
                if (nwl) {
                    (void) strcat(buf, nwl->wl_word);
                    tfree(nwl->wl_word);
                }
                else {
                    nwl = alloc(struct wordlist);
                }
                nwl->wl_word = copy(buf);
            }
            if (!(twl = wl_splice(wl, nwl))) {
                wl_free(wlist);
                *list = NULL;
                return;
            }
            if (wl == wlist)
                wlist = nwl;
            wl = twl;
            (void) strcpy(buf, wl->wl_word);
            i = strlen(buf);
            (void) strcat(buf, t);
            tfree(wl->wl_word);
            wl->wl_word = copy(buf);
            t = &wl->wl_word[i];
            s = wl->wl_word;
            for (i = 0; s < t; s++)
                wbuf[i++] = *s;
        }
    }
    *list = wlist;
    return;
}


/* Evaluate a variable. */

static wordlist *
vareval(string)

char *string;
{
    struct variable *v, *vv;
    wordlist *wl;
    char buf[BSIZE_SP], *s;
    char *range = NULL;
    int i, up, low;

    string = copy(string);

    cp_wstrip(string);
    if (s = strchr(string, '[')) {
        *s = '\0';
        range = eval_range(s+1);
    }

    switch (*string) {

        case '$':
            wl = alloc(struct wordlist);
            (void) sprintf(buf, "%d", getpid());
            wl->wl_word = copy(buf);
            goto end;

        case '<':
            (void) fflush(cp_out);
            if (!fgets(buf, BSIZE_SP, cp_in)) {
                clearerr(cp_in);
                (void) strcpy(buf, "EOF");
            }
            for (s = buf; *s && (*s != '\n'); s++) ;
            *s = '\0';
            wl = cp_lexer(buf);
            /* This is a hack. */
            if (!wl->wl_word)
                wl->wl_word = copy("");
            goto end;
    
        case '?':
            wl = alloc(struct wordlist);
            s = string + 1;
            if (htab_get(s,variables))
                wl->wl_word = copy("1");
            else {
                v = cp_enqvar(s);
                if (v) {
                    wl->wl_word = copy("1");
                    va_free(v);
                }
                else
                    wl->wl_word = copy("0");
            }
            goto end;
        
        case '#':
            wl = alloc(struct wordlist);
            s = string + 1;
            v = (struct variable *)htab_get(s,variables);
            if (v) {
                if (v->va_type == VT_LIST)
                    for (v = v->va_vlist, i = 0; v; v = v->va_next,i++);
                else
                    i = (v->va_type != VT_BOOL);
            }
            else {
                /* cp_enqvar() takes care of range */
                if (range) {
                    sprintf(buf,"%s[%s",s,range);
                    txfree(string);
                    string = copy(buf);
                }
                v = cp_enqvar(string);
                if (v) {
                    if (v->va_type == VT_LIST)
                        for (v = v->va_vlist, i = 0; v; v = v->va_next,i++);
                    else
                        i = (v->va_type != VT_BOOL);
                    va_free(v);
                }
                else {
                    fprintf(cp_err, "Error: %s: no such variable.\n",s);
                    wl = (wordlist *)NULL;
                    goto end;
                }
            }
            (void) sprintf(buf, "%d", i);
            wl->wl_word = copy(buf);
            goto end;

        case '\0':
            wl = alloc(struct wordlist);
            wl->wl_word = copy("$");
            goto end;
    }

    /* The notation var[stuff] has two meanings...  If this is a real
     * variable, then the [] denotes range, but if this is a strange
     * (e.g, device parameter) variable, it could be anything...
     */
    v = (struct variable *)htab_get(string,variables);
    if (!v && isdigit(*string)) {
        v = (struct variable *)htab_get("arvg",variables);
        if (range)
            txfree(range);
        range = copy(string);
    }

    if (!v) {
        /* cp_enqvar() takes care of range */
        if (range) {
            sprintf(buf,"%s[%s",string,range);
            txfree(string);
            string = copy(buf);
        }
        /* take care of forms like v($something) */
        if (*string == '&' && strchr("vViI",*(string+1)) &&
                *(string+2) == '(' && *(string+3) == '$')
            string = iv_special(string);

        v = cp_enqvar(string);
        if (v) {
            wl = cp_varwl(v);
            va_free(v);
            goto end;
        }
    }
    if (!v) {
        if (s = getenv(string)) {
            wl = alloc(struct wordlist);
            wl->wl_word = copy(s);
            goto end;
        }
    }
    if (!v) {
        fprintf(cp_err, "Error: %s: no such variable.\n", string);
        wl = (wordlist *)NULL;
        goto end;
    }

    /* Now parse and deal with 'range' ... */
    if (range) {
        s = range;
        if (!isdigit(*s) && *s != '-')
            fprintf(cp_err,
                "Warning: nonparseable range specified, %s[%s\n",string,s);

        for (low = 0; isdigit(*s); s++)
            low = low * 10 + *s - '0';
        if ((*s == '-') && isdigit(s[1]))
            for (up = 0, s++; isdigit(*s); s++)
                up = up * 10 + *s - '0';
        else if (*s == '-') {
            if (v->va_type != VT_LIST)
                up = 0;
            else {
                for (up = 0, vv = v->va_vlist; vv; vv = vv->va_next,up++) ;
                if (!up) {
                    wl = (wordlist*)NULL;
                    goto end;
                }
                up--;
            }
        }
        else
            up = low;
        wl = var2wl(v, low, up);
    }
    else
        wl = cp_varwl(v);

end:
    if (range)
        txfree(range);
    txfree(string);
    return (wl);
}


static struct wordlist *
var2wl(v,lo,hi)

struct variable *v;
int lo, hi;
{
    struct variable *vx, *tx, *vn, *vlo, *vl0, *vl1;
    wordlist *wl;
    int i;
    bool rev = false;

    if (v->va_type != VT_LIST) {
        if (lo != 0 && hi != 0)
            return (NULL);
        else
            return (cp_varwl(v));
    }
    if (hi < lo) {
        i = hi;
        hi = lo;
        lo = i;
        rev = true;
    }

    vl0 = v->va_vlist;
    vx = vl0;
    for (i = 0; i < lo; i++) {
        vx = vx->va_next;
        if (vx == NULL)
            return (NULL);
    }
    vlo = vx;
    for (i = lo; i < hi; i++) {
        vx = vx->va_next;
        if (vx == NULL)
            break;
    }
    if (vx) {
        vl1 = vx->va_next;
        vx->va_next = NULL;
    }

    if (rev) {
        tx = NULL;
        for ( ; vlo; vlo = vn) {
            vn = vlo->va_next;
            vlo->va_next = tx;
            tx = vlo;
        }
        vlo = tx;
    }

    v->va_vlist = vlo;
    wl = cp_varwl(v);

    if (rev) {
        tx = NULL;
        for ( ; vlo; vlo = vn) {
            vn = vlo->va_next;
            vlo->va_next = tx;
            tx = vlo;
        }
        vlo = tx;
    }

    if (vx)
        vx->va_next = vl1;
    v->va_vlist = vl0;
    return (wl);
}
        

static char *
eval_range(string)

/* Parse and evaluate the range specification, replacing any shell
 * variables by integer equivalents.
 */
char *string;
{
    char *s, *t, *tok1, *tok2, *tokd = NULL, *tokb = NULL;
    char buf[128];
    int i, lev1, lev2;
    wordlist *wl;

    /* get rid of any white space */
    t = string;
    s = tmalloc(strlen(string)+1);
    string = s;
    for ( ; *t; t++) {
        if (isspace(*t))
            continue;
        *s++ = *t;
    }
    *s = '\0';

    if (!strchr(string,'$'))
        return (string);

    s = tok1 = string;
    lev1 = lev2 = 0;
    while (*s) {
        if (*s == '[')
            lev1++;
        else if (*s == '(')
            lev2++;
        else if (*s == ']') {
            if (lev1)
                lev1--;
            else if (!lev2)
                /* [tok1] */
                break;
        }
        else if (*s == ')') {
            if (lev2)
                lev2--;
        }
        else if (*s == '-' && !lev1 && !lev2) {
            if (s == tok1)
                /* [-tok2] */
                tok1 = NULL;
            /* [tok1-...] */
            break;
        }
        s++;
    }
    if (*s == '-') {
        tokd = s;
        *s = '\0';
        s++;
        tok2 = s;
        lev1 = lev2 = 0;
        while (*s) {
            if (*s == '[')
                lev1++;
            else if (*s == '(')
                lev2++;
            else if (*s == ')') {
                if (lev2)
                    lev2--;
            }
            else if (*s == ']') {
                if (lev1)
                    lev1--;
                else if (!lev2) {
                    if (tok2 == s) {
                        /* [...-] */
                        tok2 == NULL;
                        break;
                    }
                    /* [...-tok2] */
                    tokb = s;
                    *s = '\0';
                    break;
                }
            }
            s++;
        }
        if (!*s && tok2 == s) {
            /* no last ], not to worry */
            tok2 = NULL;
        }
    }
    else {
        if (*s == ']') {
            tokb = s;
            *s = '\0';
        }
        tok2 = NULL;
    }

    *buf = '\0';
    if (tok1) {
        if (*tok1 == '$') {
            tok1++;
            wl = vareval(tok1);
            if (wl) {
                tok1 = wl_flatten(wl);
                wl_free(wl);
                i = atof(tok1) + .5;
                txfree(tok1);
            }
            else
                i = 0;
            sprintf(buf,"%d",i);
        }
        else
            sprintf(buf,"%s",tok1);
    }
    if (tokd)
        strcat(buf,"-");
    if (tok2) {
        if (*tok2 == '$') {
            tok2++;
            wl = vareval(tok2);
            if (wl) {
                tok2 = wl_flatten(wl);
                wl_free(wl);
                i = atof(tok2) + .5;
                txfree(tok2);
            }
            else
                i = 0;
            sprintf(buf + strlen(buf),"%d",i);
        }
        else
            sprintf(buf + strlen(buf),"%s",tok2);

    }
    strcat(buf,"]");
    txfree(string);
    return (copy(buf));
}


static char *
iv_special(str)

/* this substitutes in forms like v($something) */
char *str;
{
    char *s, *t, *u;
    wordlist *wl;
    int cnt;

    cnt = 1;
    for (s = str+4; *s; s++) {
        if (*s == '(')
            cnt++;
        else if (*s == ')')
            cnt--;
        if (!cnt)
            break;
    }
    if (*s != ')')
        /* no closing paren, error */
        return (str);
    *s = '\0';
    wl = vareval(str+4);
    *s = ')';
    t = wl_flatten(wl);
    wl_free(wl);
    if (!t)
        /* some error occurred */
        return (str);
    u = tmalloc(strlen(t) + strlen(s) + 4);
    u[0] = '&';
    u[1] = str[1];
    u[2] = '(';
    strcpy(u+3,t);
    strcat(u,s);
    txfree(str);
    txfree(t);
    return (u);
}


/* Print the values of currently defined variables. */

struct xxx {
    struct variable *x_v;
    char x_char;
} ;

void
cp_vprint()
{
    struct variable *v;
    struct variable *uv1, *uv2;
    wordlist *wl, *wl0;
    int i, j;
    char *s;
    struct xxx *vars;

    cp_usrvars(&uv1, &uv2);

    for (v = uv1, i = 0; v; v = v->va_next)
        i++;
    for (v = uv2; v; v = v->va_next)
        i++;
    wl0 = (wordlist*)htab_list(variables);
    i += wl_length(wl0);
    
    vars = (struct xxx *) tmalloc(sizeof (struct xxx) * i);

    for (wl = wl0, i = 0; wl; wl = wl->wl_next, i++) {
        vars[i].x_v = (struct variable*)((char**)wl->wl_word)[1];
        vars[i].x_char = ' ';
    }
    for (v = uv1; v; v = v->va_next, i++) {
        vars[i].x_v = v;
        vars[i].x_char = '*';
    }
    for (v = uv2; v; v = v->va_next, i++) {
        vars[i].x_v = v;
        vars[i].x_char = '+';
    }

    qsort((char *) vars, i, sizeof (struct xxx), vcmp);

    out_send("\n");
    for (j = 0; j < i; j++) {
        if (j && eq(vars[j].x_v->va_name, vars[j - 1].x_v->va_name))
            continue;
        v = vars[j].x_v;
        if (v->va_type == VT_BOOL) {
             out_printf("%c %-18s\n", vars[j].x_char, v->va_name);
        }
        else {
            out_printf("%c %-18s", vars[j].x_char, v->va_name);
            wl = vareval(v->va_name);
            s = wl_flatten(wl);
            wl_free(wl);
            if (v->va_type == VT_LIST)
                out_printf("( %s )\n", s);
            else
                out_printf("%s\n", s);
            txfree(s);
        }
    }
    
    wl_free(wl0);
    va_free(uv1);
    va_free(uv2);
    txfree((char*)vars);
    out_send("\n");
    return;
}


static int
vcmp(v1, v2)

struct xxx *v1, *v2;
{
    int i;

    if (i = strcmp(v1->x_v->va_name, v2->x_v->va_name))
        return (i);
    else
        return (v1->x_char - v2->x_char);
}


wordlist *
cp_varwl(var)

struct variable *var;
{
    wordlist *wl = NULL, *w, *wx = NULL;
    char buf[BSIZE_SP];
    struct variable *vt;

    switch(var->va_type) {
        case VT_BOOL:
            /* Can't ever be false. */
            (void) sprintf(buf, "%s", var->va_bool ? "true" : "false");
            break;
        case VT_NUM:
            (void) sprintf(buf, "%d", var->va_num);
            break;
        case VT_REAL:
            /* This is a case where printnum isn't too good... */
            (void) sprintf(buf, "%g", var->va_real);
            (void) fixxp2(buf);
            break;
        case VT_STRING:
            if (var->va_string == NULL)
                (void)strcpy(buf,"(null)");
            else {
                char *t = copy(var->va_string);
                cp_unquote(t);
                (void) strcpy(buf, t);
                tfree(t);
            }
            break;
        case VT_LIST:   /* The tricky case. */
            for (vt = var->va_vlist; vt; vt = vt->va_next) {
                if ((w = cp_varwl(vt)) == NULL)
                    continue;
                if (wl == NULL) {
                    wl = wx = w;
                }
                else {
                    wx->wl_next = w;
                    w->wl_prev = wx;
                }
                while (wx->wl_next)
                    wx = wx->wl_next;
            }
            return (wl);
        default:
            fprintf(cp_err, 
            "cp_varwl: Internal Error: bad variable type %d\n", 
                var->va_type);
            return (NULL);
    }
    wl = alloc(struct wordlist);
    wl->wl_next = wl->wl_prev = NULL;
    wl->wl_word = copy(buf);
    return (wl);
}


/* Set a variable. */
/* This now handles forms like "set var[n] = whatever", where var must be
 * a previously defined list, and n (index 0) must be in bounds, and
 * whatever must be of the same type.  n must be an explicit integer.
 */

void
cp_vset(varname, type, value)

char *varname;
char type;
char *value;
{
    struct variable *v, *vv;
    int i, rnum;
    bool alreadythere = false, rangefound = false;
    char *t, *range;
    char vname[BSIZE_SP];

    strcpy(vname,varname);
    cp_unquote(vname);

    if (eq(vname,"argv") || eq(vname,"argc"))
        return;

    range = strchr(vname,'[');
    if (range) {
        rangefound = true;
        *range = '\0';
        range++;
        range = eval_range(range);
        for (rnum = 0,t = range; isdigit(*t); t++)
            rnum = rnum * 10 + *t - '0';
        tfree(range);
    }
    if ((v = (struct variable *)htab_get(vname,variables)) == NULL) {
        if (range) {
            fprintf(cp_err,
                "Error: %s is not set, length unknown.\n",vname);
            return;
        }
        v = alloc(struct variable);
        v->va_name = copy(vname);
    }
    else {
        alreadythere = true;
        if (range) {
            if (v->va_type != VT_LIST) {
                fprintf(cp_err,"Error: %s is not a list.\n",vname);
                return;
            }
            for (vv = v->va_vlist,i = rnum; vv && i; vv = vv->va_next,i--);
            if (vv == NULL) {
                fprintf(cp_err,
                    "Error: %d out of range for %s.\n",rnum,vname);
                return;
            }
            if (type != vv->va_type) {
                fprintf(cp_err,"Error: type mismatch for %s.\n",vname);
                return;
            }
            switch (type) {
                case VT_BOOL:
                    vv->va_bool = *(int*)value;
                    break;
                case VT_NUM:
                    vv->va_num = *(int*)value;
                    break;
                case VT_REAL:
                    vv->va_real = *(double*)value;
                    break;
                case VT_STRING:
                    txfree(vv->va_string);
                    vv->va_string = copy(value);
                    break;
                case VT_LIST:
                    va_free(vv->va_vlist);
                    vv->va_vlist = (struct variable*)value;
                    break;
                default:
                    fprintf(cp_err, 
                        "cp_vset: Internal Error: bad variable type %d.\n",
                        type);
            }
            return;
        }
        if (v->va_type == VT_STRING)
            txfree(v->va_string);
        else if (v->va_type == VT_LIST)
            va_free(v->va_vlist);
    }

    switch (type) {
        case VT_BOOL:
            if (* ((bool *) value) == false) {
                cp_remvar(vname);
                return;
            }
            else
                v->va_bool = true;
            break;

        case VT_NUM:
            v->va_num = * (int *) value;
            break;

        case VT_REAL:
            v->va_real = * (double *) value;
            break;

        case VT_STRING:
            v->va_string = copy(value);
            break;

        case VT_LIST:
            v->va_vlist = (struct variable *) value;
            break;

        default:
            fprintf(cp_err, 
                "cp_vset: Internal Error: bad variable type %d.\n",type);
            return;
    }
    v->va_type = type;

    i = cp_internalset(v,true);

    switch (i) {

        case US_OK:
            /* Normal case. */
            if (!alreadythere) {
                if (!variables)
                    variables = htab_init();
                htab_add(v->va_name,(void*)v,variables);
            }
            break;

        case US_DONTRECORD:
            /* Do nothing... */
            if (alreadythere) {
                fprintf(cp_err, 
        "cp_vset: Internal Error: %s already there, but 'dont record'\n",
                        v->va_name);
            }
            break;
    
        case US_READONLY:
            /* Badness... */
            fprintf(cp_err, "Error: %s is a read-only variable.\n",
                    v->va_name);
            if (alreadythere)
                fprintf(cp_err, 
            "cp_vset: Internal Error: it was already there too!!\n");
            break;

        default:
            fprintf(cp_err, "cp_vset: Internal Error: bad US val %d\n", i);
            break;
    }
    return;
}


void
cp_remvar(varname)

char *varname;
{
    struct variable *v;
    bool found = true;
    int i;

    if (eq(varname,"argv") || eq(varname,"argc")) {
        return;
    }
    v = (struct variable *)htab_get(varname,variables);
    if (!v) {
        /* Gotta make up a var struct for cp_internalset()... */
        v = alloc(struct variable);
        v->va_name = varname;
        found = false;
    }

    i = cp_internalset(v,false);

    switch (i) {

        case US_OK:
            /* Normal case. */
            if (found) {
                htab_delete(varname,variables);
                va_free(v);
            }
            break;

        case US_DONTRECORD:
            /* Do nothing... */
            if (found)
                fprintf(cp_err,
            "cp_remvar: Internal Error: var %d in list but shouldn't be\n",
                        varname);
            break;
    
        case US_READONLY:
            /* Badness... */
            fprintf(cp_err, "Error: %s is a read-only variable.\n",
                    v->va_name);
            if (found)
                fprintf(cp_err,
            "cp_remvar: Internal Error: var %d in list but shouldn't be\n",
                        varname);
            break;

        default:
            fprintf(cp_err, "cp_remvar: Internal Error: bad US val %d\n",i);
            break;
    }
    if (!found)
        txfree((char*)v);
    return;
}


/* The set command. Syntax is 
 * set [opt ...] [opt = val ...]. Val may be a string, an int, a float,
 * or a list of the form (elt1 elt2 ...).
 */

void
com_set(wl)

wordlist *wl;
{
    struct variable *vars, *v0;
    char *s;

    if (wl == NULL) {
        cp_vprint();
        return;
    }
    v0 = vars = cp_setparse(wl);

    /* This is sort of a hassle... */
    while (vars) { 
        switch (vars->va_type) {
            case VT_BOOL:
                s = (char *) &vars->va_bool;
                break;
            case VT_NUM:
                s = (char *) &vars->va_num;
                break;
            case VT_REAL:
                s = (char *) &vars->va_real;
                break;
            case VT_STRING:
                s = vars->va_string;
                break;
            case VT_LIST:
                s = (char *) va_copy(vars->va_vlist);
                break;
        }
        cp_vset(vars->va_name, vars->va_type, s);
        vars = vars->va_next;
    }
    va_free(v0);
    return;
}


struct variable *
cp_setparse(wl)

wordlist *wl;
{
    char *name, *val, *s, *st;
    double *td;
    struct variable *vv, *lv;
    struct variable *vars = NULL;

    while (wl) {
        name = copy(wl->wl_word);
        cp_unquote(name);
        wl = wl->wl_next;
        if (((wl == NULL) || (*wl->wl_word != '=')) && 
                strchr(name, '=') == NULL) {
            vv = alloc(struct variable);
            vv->va_name = name;
            vv->va_type = VT_BOOL;
            vv->va_bool = true;
            vv->va_next = vars;
            vars = vv;
            continue;
        }
        if (wl && eq(wl->wl_word, "=")) {
            wl = wl->wl_next;
            if (wl == NULL || strchr(name,'=') != NULL)
                goto bad;
            val = wl->wl_word;
            wl = wl->wl_next;
        }
        else if (wl && (*wl->wl_word == '=')) {
            if (strchr(name, '=') != NULL)
                goto bad;
            val = wl->wl_word + 1;
            wl = wl->wl_next;
        }
        else if (s = strchr(name, '=')) {
            if (s == name) goto bad;
            val = s + 1;
            *s = '\0';
            if (*val == '\0') {
                if (wl == NULL) {
                    fprintf(cp_err,
                        "Error:  %s equals what?.\n",name);
                    goto bad;
                }
                val = wl->wl_word;
                wl = wl->wl_next;
            }
        }
        else {
            goto bad;
        }
        val = copy(val);
        cp_unquote(val);
        if (eq(val, "(")) {
            tfree(val);
            lv = getlist(&wl);
            if (!lv)
                goto bad;
            vv = alloc(struct variable);
            vv->va_name = name;
            vv->va_type = VT_LIST;
            vv->va_vlist = lv;
            vv->va_next = vars;
            vars = vv;

            continue;
        }

        st = val;
        td = ft_numparse(&st, false);
        vv = alloc(struct variable);
        vv->va_name = name;
        vv->va_next = vars;
        vars = vv;
        if (td && *st == '\0') {
            vv->va_type = VT_REAL;
            vv->va_real = *td;
            tfree(val);
        }
        else {
            vv->va_type = VT_STRING;
            vv->va_string = val;
        }
    }
    return (vars);
bad:
    fprintf(cp_err, "Error: bad set form.\n");
    tfree(name);
    if (vars) va_free(vars);
    return (NULL);
}


static struct variable *
getlist(wlist)

wordlist **wlist;
{
    wordlist *wl = *wlist;
    struct variable *listv = NULL, *vv;
    char *ss, *st;
    double *td;

    while (wl && wl->wl_word) {
        if (eq(wl->wl_word, ")")) {
            wl = wl->wl_next;
            *wlist = wl;
            return (listv);
        }
        if (!listv)
            listv = vv = alloc(struct variable);
        else {
            vv->va_next = alloc(struct variable);
            vv = vv->va_next;
        }
        if (eq(wl->wl_word, "(")) {
            wl = wl->wl_next;
            vv->va_type = VT_LIST;
            vv->va_vlist = getlist(&wl);
            if (vv->va_vlist == NULL) {
                va_free(listv);
                *wlist = wl;
                return (NULL);
            }
            continue;
        }
        st = ss = copy(wl->wl_word);
        cp_unquote(ss);
        td = ft_numparse(&ss, false);
        if (td && *ss == '\0') {
            vv->va_type = VT_REAL;
            vv->va_real = *td;
            tfree(st);
        }
        else {
            vv->va_type = VT_STRING;
            vv->va_string = st;
        }
        wl = wl->wl_next;
    }
    *wlist = wl;
    va_free(listv);
    return (NULL);
}


void
com_unset(wl)

wordlist *wl;
{
    char *name;
    struct variable *var, *nv;
    wordlist *tl, *wl0;

    if (eq(wl->wl_word, "*")) {
        wl0 = (wordlist*)htab_wl(variables);
        for (tl = wl0; tl; tl = tl->wl_next)
            cp_remvar(tl->wl_word);
        wl_free(wl0);
        return;
    }
    while (wl != NULL) {
        name = wl->wl_word;
        cp_remvar(name);
        wl = wl->wl_next;
    }
    return;
}


/* Shift a list variable, by default argv, one to the left (or more if a
 * second argument is given.
 */

void
com_shift(wl)

wordlist *wl;
{
    struct variable *v, *vv, *lv;
    char *n = "argv";
    int num = 1;

    if (wl) {
        n = wl->wl_word;
        wl = wl->wl_next;
    }
    if (wl)
        num = scannum(wl->wl_word);
    
    v = (struct variable *)htab_get(n,variables);
    if (!v) {
        fprintf(cp_err, "Error: %s: no such variable\n", n);
        return;
    }
    if (v->va_type != VT_LIST) {
        fprintf(cp_err, "Error: %s not of type list\n", n);
        return;
    }

    lv = NULL;
    for (vv = v->va_vlist; vv && (num > 0); num--) {
        lv = vv;
        vv = vv->va_next;
    }

    if (num) {
        fprintf(cp_err, "Error: variable %s not long enough\n", n);
        return;
    }

    if (lv) {
        lv->va_next = NULL;
        va_free(lv);
    }
    v->va_vlist = vv;
    return;
}


/* Determine the value of a variable.  Fail if the variable is unset,
 * and if the type doesn't match, try and make it work...
 */

bool
cp_getvar(name, type, retval)

int type;
char *name, *retval;
{
    struct variable *v;

    v = (struct variable *)htab_get(name,variables);
    if (v == NULL) {
        if (type == VT_BOOL)
            * (bool *) retval = false; 
        return (false);
    }
    if (v->va_type == type) {
        switch (type) {
            case VT_BOOL:
                * (bool *) retval = true; 
                break;
            case VT_NUM: {
                int *i;
                i = (int *) retval;
                *i = v->va_num;
                break;
            }
            case VT_REAL: {
                double *d;
                d = (double *) retval;
                *d = v->va_real;
                break;
            }
            case VT_STRING: { /* Gotta be careful to have room. */
                char *s;
                s = copy(v->va_string);
                cp_unquote(s);
                cp_wstrip(s);
                (void) strcpy(retval, s);
                txfree(s);
                break;
            }
            case VT_LIST: { /* Funny case... */
                struct variable **tv;
                tv = (struct variable **) retval;
                *tv = v->va_vlist;
                break;
            }
            default:
                fprintf(cp_err, 
                "cp_getvar: Internal Error: bad var type %d.\n", type);
                break;
        }
        return (true);
    }
    else {
        /* Try to coerce it.. */
        if ((type == VT_NUM) && (v->va_type == VT_REAL)) {
            int *i;
            i = (int *) retval;
            *i = (int) v->va_real;
            return (true);
        }
        else if ((type == VT_REAL) && (v->va_type == VT_NUM)) {
            double *d;
            d = (double *) retval;
            *d = (double) v->va_num;
            return (true);
        }
        else if ((type == VT_STRING) && (v->va_type == VT_NUM)) {
            (void) sprintf(retval, "%d", v->va_num);
            return (true);
        }
        else if ((type == VT_STRING) && (v->va_type == VT_REAL)) {
            (void) sprintf(retval, "%f", v->va_real);
            return (true);
        }
        return (false);
    }
}


/* free a variable structure */
void
va_free(v)

struct variable *v;
{
    struct variable *vv;

    for ( ; v; v = vv) {
        vv = v->va_next;
        txfree(v->va_name);
        if (v->va_type == VT_STRING)
            txfree(v->va_string);
        if (v->va_type == VT_LIST)
            va_free(v->va_vlist);
        txfree((char*)v);
    }
}


/* copy a variable structure */
struct variable *
va_copy(v)

struct variable *v;
{
    struct variable *nv, *n0 = NULL;

    while (v) {
        if (n0 == NULL)
            n0 = nv = alloc(struct variable);
        else {
            nv->va_next = alloc(struct variable);
            nv = nv->va_next;
        }
        nv->va_name = copy(v->va_name);
        nv->va_type = v->va_type;
        if (v->va_type == VT_BOOL)
            nv->va_bool = v->va_bool;
        else if (v->va_type == VT_REAL)
            nv->va_real = v->va_real;
        else if (v->va_type == VT_NUM)
            nv->va_num = v->va_num;
        else if (v->va_type == VT_STRING)
            nv->va_string = copy(v->va_string);
        else if (v->va_type == VT_LIST)
            nv->va_vlist = va_copy(v->va_vlist);
        v = v->va_next;
    }
    return (n0);
}


/* Implelentation of a stack for argc, argv[] variables.
 * These variables are set and cleared through these routines
 * only, not through com_set(), com_unset(), cp_remvar().
 */

struct avstack {
    int ac;
    struct variable *av;
};

#define DEPTH 100

static struct avstack av_stack[DEPTH];
static int stackp;

void
cp_pusharg(wl)

wordlist *wl;
{
    struct variable *listv = NULL, *lv;
    struct variable *v, **pv = NULL;
    int *pc = NULL;
    int n;
    char *st, *ss;
    double *td;

    if (stackp == 0) {

        v = alloc(struct variable);
        v->va_name = copy("argc");
        v->va_type = VT_NUM;
        pc = &v->va_num;
        htab_add("argc",(void*)v,variables);

        v = alloc(struct variable);
        v->va_name = copy("argv");
        v->va_type = VT_LIST;
        pv = &v->va_vlist;
        htab_add("argv",(void*)v,variables);
    }
    else {
        v = (struct variable *)htab_get("argv",variables);
        if (v)
            pv = &v->va_vlist;
        v = (struct variable *)htab_get("argc",variables);
        if (v)
            pc = &v->va_num;
    }
    if (pv && pc) {
        if (stackp) {
            av_stack[stackp].av = *pv;
            av_stack[stackp].ac = *pc;
        }
        stackp++;
        if (stackp == DEPTH) {
            fprintf(cp_err, "Warning: stack overflow\n");
            stackp--;
            return;
        }

        n = 0;
        while (wl && wl->wl_word) {
            v = alloc(struct variable);
            st = ss = copy(wl->wl_word);
            cp_unquote(ss);
            td = ft_numparse(&ss, false);
            if (td && *ss == '\0') {
                v->va_type = VT_REAL;
                v->va_real = *td;
                txfree(st);
            }
            else {
                v->va_type = VT_STRING;
                v->va_string = st;
            }
            if (listv) {
                lv->va_next = v;
                lv = v;
            }
            else
                listv = lv = v;
            n++;
            wl = wl->wl_next;
        }
        *pv = listv;
        *pc = n;
    }
}


void
cp_poparg()

{
    struct variable *vv, *vc, **pv = NULL;
    int *pc = NULL;

    if (stackp == 0)
        return;

    vv = (struct variable *)htab_get("argv",variables);
    if (vv)
        pv = &vv->va_vlist;
    vc = (struct variable *)htab_get("argc",variables);
    if (vc)
        pc = &vc->va_num;

    if (pc && pv) {
        stackp--;
        if (stackp == 0) {

            htab_delete("argv",variables);
            htab_delete("argc",variables);
            va_free(vv);
            va_free(vc);
        }
        else {
            va_free(*pv);
            *pv = av_stack[stackp].av;
            *pc = av_stack[stackp].ac;
        }
    }
}
