/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * The 'compose' command.  This is a more powerful and convenient form of the
 * 'let' command.
 */

#include "spice.h"
#include "ftedefs.h"

struct sCompose {
    double start;
    double stop;
    double step;
    double lin;
    double center;
    double span;
    double mean;
    double sd;
    int log;
    int dec;
    int gauss;
    int randm;
    char *pool;
    bool startgiven;
    bool stopgiven;
    bool stepgiven;
    bool lingiven;
    bool centergiven;
    bool spangiven;
    bool meangiven;
    bool sdgiven;
    bool loggiven;
    bool decgiven;
    bool gaussgiven;
    bool randmgiven;
    bool poolgiven;
    bool reverse;
};

#ifdef __STDC__
static void dimxpand(struct dvec*,int*,double*);
static bool cmp_values(wordlist*,int*,double**,complex**,bool*);
static bool cmp_parse(wordlist*,struct sCompose*);
static bool cmp_linsweep(struct sCompose*,int*,double**);
static bool cmp_logsweep(struct sCompose*,int*,double**);
static bool cmp_random(struct sCompose*,int*,double**);
static bool cmp_gauss(struct sCompose*,int*,double**);
#else
static void dimxpand();
static bool cmp_values();
static bool cmp_parse();
static bool cmp_linsweep();
static bool cmp_logsweep();
static bool cmp_random();
static bool cmp_gauss();
#endif

/* The general syntax is 'compose name parm = val ...'
 * The possible parms are:
 *  start       The value at which the vector should start.
 *  stop        The value at which the vector should end.
 *  step        The difference between sucessive elements.
 *  lin         The number of points, linearly spaced.
 *  log         The number of points, logarithmically spaced.
 *  dec         The number of points per decade, logarithmically spaced.
 *  center      Where to center the range of points.
 *  span        The size of the range of points.
 *  gauss       The number of points in the gaussian distribution.
 *  mean        The mean value for the gass. dist.
 *  sd          The standard deviation for the gauss. dist.
 *  random      The number of randomly selected points.
 *  pool        The name of a vector (must be already defined) to get
 *              random values -- default is 'unitvec(npoints)'
 *
 * The case 'compose name values val val ...' takes the values and creates a
 * new vector -- the vals may be arbitrary expressions.
 *
 * NOTE: most of this doesn't work -- there will be plenty of unused variable
 * lint messages...
 */


void
com_compose(wl)

wordlist *wl;
{
    struct sCompose sc;
    struct dvec *result;
    complex *cdata;
    double *data;
    double tt;
    int length;
    bool realflag = true;
    char *resname;

    bzero(&sc,sizeof(struct sCompose));

    resname = wl->wl_word;
    vec_remove(resname);
    wl = wl->wl_next;

    if (eq(wl->wl_word, "values")) {
        if (cmp_values(wl->wl_next,&length,&data,&cdata,&realflag))
            return;
    }
    else {
        if (cmp_parse(wl,&sc))
            return;

        /* Now see what we have... start and stop are pretty much
         * compatible with everything...
         */
        if (sc.stepgiven && (sc.step == 0.0)) {
            fprintf(cp_err, "Error: step cannot = 0.0\n");
            return;
        }
        if (sc.startgiven && sc.stopgiven && (sc.start > sc.stop)) {
            tt = sc.start;
            sc.start = sc.stop;
            sc.stop = tt;
            sc.reverse = true;
        }
        if (sc.lingiven + sc.loggiven + sc.decgiven +
            sc.randmgiven + sc.gaussgiven > 1) {
            fprintf(cp_err,
    "Error: can have at most one of (lin, log, dec, random, gauss)\n");
            return;
        }
        else if (sc.lingiven + sc.loggiven + sc.decgiven + sc.randmgiven +
                sc.gaussgiven == 0) {
            /* Hmm, if we have a start, stop, and step we're ok. */
            if (sc.startgiven && sc.stopgiven && sc.stepgiven) {
                sc.lingiven = true;
                sc.lin = (sc.stop - sc.start) / sc.step + 1;
                sc.stepgiven = false;  /* Problems below... */
            }
            else {
                fprintf(cp_err,
"Error: either one of (lin, log, dec, random, gauss) must be given, or all\n");
                fprintf(cp_err,
                    "\tof (start, stop, and step) must be given.\n");
                return;
            }
        }

        if (sc.lingiven) {
            if (cmp_linsweep(&sc,&length,&data))
                return;
        }
        else if (sc.loggiven || sc.decgiven) {
            if (cmp_logsweep(&sc,&length,&data))
                return;
        }
        else if (sc.randmgiven) {
            if (cmp_random(&sc,&length,&data))
                return;
        }
        else if (sc.gaussgiven) {
            if (cmp_gauss(&sc,&length,&data))
                return;
        }
    }
    result = alloc(struct dvec);
    result->v_name = copy(resname);
    result->v_type = SV_NOTYPE;
    result->v_flags = (realflag ? VF_REAL : VF_COMPLEX);
    if (realflag)
        result->v_realdata = data;
    else
        result->v_compdata = cdata;
    result->v_length = length;
    result->v_numdims = 1;
    vec_newperm(result);
    return;
}


static void
dimxpand(v, newdims, data)

/* Copy the data from a vector into a buffer with larger dimensions. */
struct dvec *v;
int *newdims;
double *data;
{
    complex *cdata = (complex *) data;
    bool realflag = isreal(v);
    int i, j, o, n, t, u;
    int ncount[MAXDIMS], ocount[MAXDIMS];
    int dims;

    dims = v->v_numdims;
    if (dims == 0)
        dims = 1;

    for (i = 0; i < MAXDIMS; i++)
        ncount[i] = ocount[i] = 0;
    
    for (;;) {
        for (o = n = i = 0; i < dims; i++) {
            for (j = i, t = u = 1; j < dims; j++) {
                t *= v->v_dims[j];
                u *= newdims[j];
            }
            o += ocount[i] * t;
            n += ncount[i] * u;
        }

        if (realflag) {
            data[n] = v->v_realdata[o];
        }
        else {
            realpart(&cdata[n]) = realpart(&v->v_compdata[o]);
            imagpart(&cdata[n]) = imagpart(&v->v_compdata[o]);
        }
        /* Now find the next index element... */
        for (i = dims - 1; i >= 0; i--) {
            if ((ocount[i] < v->v_dims[i] - 1) &&
                    (ncount[i] < newdims[i] - 1)) {
                ocount[i]++;
                ncount[i]++;
                break;
            }
            else
                ocount[i] = ncount[i] = 0;
        }
        if (i < 0)
            break;
    }
}


static bool
cmp_values(wl,length,datap,cdatap,realflag)

wordlist *wl;
int *length;
double **datap;
complex **cdatap;
bool *realflag;
{
    struct pnode *pn;
    struct dvec *vec, *v;
    int i, j, len, dim, blocksize, dims[MAXDIMS];
    bool rflg = true;
    double *data;
    complex *cdata;
    struct dvlist *dl, *dl0;

    if ((pn = ft_getpnames(wl, true)) == NULL)
        return (true);
    if ((dl0 = ft_dvlist(pn)) == NULL)
        return (true);

    /* Now make sure these are all of the same dimensionality.  We
     * can coerce the sizes...
     */
    dim = dl0->dl_dvec->v_numdims;
    if (dim < 2)
        dim = 0;
    if (dim >= MAXDIMS) {
        fprintf(cp_err, "Error: max dimensionality is %d\n", MAXDIMS);
        vec_dlfree(dl0);
        return (true);
    }
    len = dim ? 1 : dl0->dl_dvec->v_length;
    if (iscomplex(dl0->dl_dvec))
        rflg = false;

    for (dl = dl0->dl_next; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        i = v->v_numdims;
        if (i < 2) {
            i = 0;
            len += v->v_length;
        }
        else
            len++;
        if (i != dim) {
            fprintf(cp_err,
            "Error: all vectors must be of the same dimensionality\n");
            vec_dlfree(dl0);
            return (true);
        }
        if (iscomplex(v))
            rflg = false;
    }
    for (i = 0; i < dim; i++) {
        dims[i] = dl0->dl_dvec->v_dims[i];
        for (dl = dl0->dl_next; dl; dl = dl->dl_next) {
            v = dl->dl_dvec;
            if (v->v_dims[i] > dims[i])
                dims[i] = v->v_dims[i];
        }
    }
    dim++;
    dims[dim - 1] = len;
    for (i = 0, blocksize = 1; i < dim - 1; i++)
        blocksize *= dims[i];
    if (rflg)
        data = (double *) tmalloc(sizeof (double) * len *  blocksize);
    else
        cdata = (complex *) tmalloc(sizeof (complex) * len * blocksize);

    /* Now copy all the data over... If the sizes are too small
     * then the extra elements are left as 0.
     */
    for (i = 0, dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        if (dim == 1) {
            for (j = 0; j < v->v_length; j++) {
                if (rflg && isreal(v))
                    data[i] = v->v_realdata[j];
                else if (isreal(v)) {
                    realpart(&cdata[i]) = v->v_realdata[j];
                    imagpart(&cdata[i]) = 0.0;
                }
                else {
                    realpart(&cdata[i]) =
                        realpart(&v->v_compdata[j]);
                    imagpart(&cdata[i]) =
                        imagpart(&v->v_compdata[j]);
                }
                i++;
            }
            continue;
        }
        dimxpand(v, dims, (rflg ? (data + i * blocksize) : 
                (double *) (cdata + i * blocksize)));
        i++;
    }
    len *= blocksize;
    *length = len;
    *realflag = rflg;
    if (rflg)
        *datap = data;
    else
        *cdatap = cdata;
    vec_dlfree(dl0);
    return (false);
}


static bool
cmp_parse(wl,sc)

wordlist *wl;
struct sCompose *sc;
{
    char *s, *var, *val;
    double *td;

    /* Parse the line... */
    while (wl) {
        if ((s = strchr(wl->wl_word, '=')) && s[1]) {
            /* This is var=val. */
            *s = '\0';
            var = wl->wl_word;
            val = s + 1;
            wl = wl->wl_next;
        }
        else if (strchr(wl->wl_word, '=')) {
            /* This is var= val. */
            *s = '\0';
            var = wl->wl_word;
            wl = wl->wl_next;
            if (wl) {
                val = wl->wl_word;
                wl = wl->wl_next;
            }
            else {
                fprintf(cp_err, "Error: bad syntax\n");
                return (true);
            }
        }
        else {
            /* This is var =val or var = val. */
            var = wl->wl_word;
            wl = wl->wl_next;
            if (wl) {
                val = wl->wl_word;
                if (*val != '=') {
                    fprintf(cp_err,"Error: bad syntax\n");
                    return (true);
                }
                val++;
                if (!*val) {
                    wl = wl->wl_next;
                    if (wl) {
                        val = wl->wl_word;
                    }
                    else {
                        fprintf(cp_err,"Error: bad syntax\n");
                        return (true);
                    }
                }
                wl = wl->wl_next;
            }
            else {
                fprintf(cp_err, "Error: bad syntax\n");
                return (true);
            }
        }

        if (cieq(var, "start")) {
            sc->startgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->start = *td;
        }
        else if (cieq(var, "stop")) {
            sc->stopgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->stop = *td;
        }
        else if (cieq(var, "step")) {
            sc->stepgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->step = *td;
        }
        else if (cieq(var, "center")) {
            sc->centergiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->center = *td;
        }
        else if (cieq(var, "span")) {
            sc->spangiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->span = *td;
        }
        else if (cieq(var, "mean")) {
            sc->meangiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->mean = *td;
        }
        else if (cieq(var, "sd")) {
            sc->sdgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->sd = *td;
        }
        else if (cieq(var, "lin")) {
            sc->lingiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->lin = *td;
        }
        else if (cieq(var, "log")) {
            sc->loggiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->log = *td;
        }
        else if (cieq(var, "dec")) {
            sc->decgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->dec = *td;
        }
        else if (cieq(var, "gauss")) {
            sc->gaussgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->gauss = *td;
        }
        else if (cieq(var, "random")) {
            sc->randmgiven = true;
            if (!(td = ft_numparse(&val, false))) {
                goto bad;
            }
            sc->randm = *td;
        }
        else if (cieq(var, "pool")) {
            sc->poolgiven = true;
            sc->pool = val;
        }
    }
    return (false);
bad:
    fprintf(cp_err,"Error: bad parm %s = %s\n", var, val);
    return (true);
}


static bool
cmp_linsweep(sc,length,datap)

/* Create a linear sweep... */
struct sCompose *sc;
int *length;
double **datap;
{
    double *data, tt;
    int i;


    data = (double *) tmalloc(sizeof (double) * sc->lin);

    if (sc->stepgiven && sc->startgiven && sc->stopgiven) {
        if (sc->step !=
            (sc->stop - sc->start) / sc->lin * (sc->reverse ? -1 : 1)) {
            fprintf(cp_err, "Warning: bad step -- should be %lg\n",
                (sc->stop - sc->start) / sc->lin * (sc->reverse ? -1 : 1));
            sc->stepgiven = false;
        }
    } 
    if (!sc->startgiven) {
        if (sc->stopgiven && sc->stepgiven) {
            sc->start = sc->stop - sc->step * sc->lin;
        }
        else if (sc->stopgiven) {
            sc->start = sc->stop - sc->lin;
        }
        else {
            sc->start = 0;
        }
        sc->startgiven = true;
    }
    if (!sc->stopgiven) {
        if (sc->stepgiven)
            sc->stop = sc->start + sc->lin * sc->step;
        else
            sc->stop = sc->start + sc->lin;
        sc->stopgiven = true;
    } 
    if (!sc->stepgiven) {
        sc->step = (sc->stop - sc->start) / sc->lin;
    }
    if (sc->reverse)
        for (i = 0, tt = sc->stop; i < sc->lin; i++, tt -= sc->step)
            data[i] = tt;
    else
        for (i = 0, tt = sc->start; i < sc->lin; i++, tt += sc->step)
            data[i] = tt;
    *length = sc->lin;
    *datap = data;
    return (false);
}


static bool
cmp_logsweep(sc,length,datap)

/* Create a log sweep... */
struct sCompose *sc;
int *length;
double **datap;
{
    return (false);
}


static bool
cmp_random(sc,length,datap)

/* Create a set of random values... */
struct sCompose *sc;
int *length;
double **datap;
{
    double *data;
    int i;

    if (!sc->centergiven) {
        sc->center = 0.0;
        sc->centergiven = true;
    }
    if (!sc->spangiven) {
        sc->span = 2.0;
        sc->spangiven = true;
    }
    data = (double *) tmalloc(sizeof (double) * sc->randm);
    for (i = 0; i < sc->randm; i++)
        data[i] = (xrandom() - 0.5)*sc->span + sc->center;
    *length = sc->randm;
    *datap = data;
    return (false);
}


static bool
cmp_gauss(sc,length,datap)

/* Create a gaussian distribution... */
struct sCompose *sc;
int *length;
double **datap;
{
    double *data;
    int i;

    if (!sc->meangiven) {
        sc->mean = 0;
        sc->meangiven = true;
    }
    if (!sc->sdgiven) {
        sc->sd = 1.0;
        sc->sdgiven = true;
    }

    data = (double *) tmalloc(sizeof (double) * sc->gauss);
    for (i = 0; i < sc->gauss; i++)
        data[i] = xgauss()*sc->sd + sc->mean;
    *length = sc->gauss;
    *datap = data;
    return (false);
}


double
xrandom()

/* return random number between 0 and 1 */
{
    return ((random() & 0x7fffffff)/(0x7fffffff + 1.0));
}


double
xgauss()

/* Based on gasdev() from Numerical Recipes in C.  Use the Box-Muller
 * transformation to get two normal deviates, return one and store the
 * other for the next call.
 */
{
    static bool iset;
    static double gset;
    double fac, r, v1, v2;
    /* xrandom() returns uniform deviate [0-1] */

    if (!iset) {
        do {
            v1 = 2.0*xrandom() - 1.0;
            v2 = 2.0*xrandom() - 1.0;
            r = v1*v1 + v2*v2;
        } while (r >= 1.0);
        fac = sqrt(-2.0*log(r)/r);
        gset = v1*fac;
        iset = true;
        return (v2*fac); 
    }
    iset = false;
    return (gset);
}



