/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Do a 'diff' of two plots.
 */

#include "spice.h"
#include "ftedefs.h"
#include "ftecmath.h"
#include "util.h"

#ifdef __STDC__
static bool nameeq(char*,char*);
static bool inwlist(char*,wordlist*);
#else
static bool nameeq();
static bool inwlist();
#endif


struct dvlist2 {
    char *d2_name;
    struct dvec *d2_dv1;
    struct dvec *d2_dv2;
    struct dvlist2 *d2_next;
};



void
com_diff(wl)
wordlist *wl;
{
    struct plot *p1, *p2 = NULL;
    struct dvec *v1, *v2;
    struct dvlist2 *dl, *dl0;
    wordlist *tw, *wn, *wl1, *wl10, *wl2, *wl20;
    complex c1, c2, c3;
    double vntol, abstol, reltol, tol, cmax, cm1, cm2;
    double d1, d2;
    int i, j;
    extern char *kw_vntol, *kw_abstol, *kw_reltol;

    if (!cp_getvar(kw_vntol, VT_REAL, (char *) &vntol))
        vntol = 1.0e-6;
    if (!cp_getvar(kw_abstol, VT_REAL, (char *) &abstol))
        abstol = 1.0e-12;
    if (!cp_getvar(kw_reltol, VT_REAL, (char *) &reltol))
        reltol = 0.001;

    /* Let's try to be clever about defaults. */
    if (!wl || !wl->wl_next) {
        p1 = plot_cur;
        if (eq(p1->pl_typename,"constants")) {
            fprintf(cp_err, "Error: invalid current plot (constants)\n");
            return;
        }
    }

    if (!wl) {
        /* try the current and previous plot of same name */
        for (p2 = p1->pl_next; p2 && !eq(p1->pl_name,p2->pl_name);
            p2 = p2->pl_next) ;
        if (!p2) {
            fprintf(cp_err, "Error: no previous similar plot\n");
            return;
        }
        fprintf(cp_err, "Plots are \"%s\" and \"%s\"\n",
            p1->pl_typename, p2->pl_typename);
    }

    else if (!wl->wl_next) {
        /* try current and named plot */
        for (p2 = plot_list; p2 && !eq(p2->pl_typename,wl->wl_word);
                p2 = p2->pl_next) ;

        if (!p2 || eq(p2->pl_typename,"constants")) {
            fprintf(cp_err, "Error: invalid plot \"%s\"\n", wl->wl_word);
            return;
        }
        fprintf(cp_err, "Plots are \"%s\" and \"%s\"\n",
                p1->pl_typename, p2->pl_typename);
        wl = NULL;
    }
    else {
        for (p1 = plot_list; p1; p1 = p1->pl_next)
            if (eq(wl->wl_word, p1->pl_typename))
                break;
        if (!p1) {
            fprintf(cp_err, "Error: no such plot %s\n", wl->wl_word);
            return;
        }
        wl = wl->wl_next;

        for (p2 = plot_list; p2; p2 = p2->pl_next)
            if (eq(wl->wl_word, p2->pl_typename))
                break;
        if (!p2) {
            fprintf(cp_err, "Error: no such plot %s\n", wl->wl_word);
            return;
        }
        wl = wl->wl_next;
    }

    if (p1 == p2)
        return;

    /* Now do some tests to make sure these plots are really the
     * same type, etc.
     */
    if (!eq(p1->pl_name, p2->pl_name))
        fprintf(cp_err, 
        "Warning: plots %s and %s seem to be of different types\n",
            p1->pl_typename, p2->pl_typename);
    if (!eq(p1->pl_title, p2->pl_title))
        fprintf(cp_err, 
        "Warning: plots %s and %s seem to be from different circuits\n",
            p1->pl_typename, p2->pl_typename);

    wl10 = (wordlist*)htab_wl(p1->pl_hashtab);
    wl20 = (wordlist*)htab_wl(p2->pl_hashtab);

    /* Throw out the ones that aren't in the arg list */
    if (wl && !eq(wl->wl_word, "all")) {    /* Just in case */
        for (wl1 = wl10; wl1; wl1 = wn) {
            wn = wl1->wl_next;
            for (tw = wl; tw; tw = tw->wl_next)
                if (nameeq(wl1->wl_word, tw->wl_word))
                    break;
            if (!tw) {
                if (!wl1->wl_prev) {
                    wl10 = wn;
                    wl10->wl_prev = NULL;
                }
                else {
                    wl1->wl_prev->wl_next = wn;
                    if (wn)
                        wn->wl_prev = wl1->wl_prev;
                }
                txfree(wl1->wl_word);
                txfree((char*)wl1);
            }
        }
        for (wl2 = wl20; wl2; wl2 = wn) {
            wn = wl2->wl_next;
            for (tw = wl; tw; tw = tw->wl_next)
                if (nameeq(wl2->wl_word, tw->wl_word))
                    break;
            if (!tw) {
                if (!wl2->wl_prev) {
                    wl20 = wn;
                    wl20->wl_prev = NULL;
                }
                else {
                    wl2->wl_prev->wl_next = wn;
                    if (wn)
                        wn->wl_prev = wl2->wl_prev;
                }
                txfree(wl2->wl_word);
                txfree((char*)wl2);
            }
        }
    }

    /* rid anything in 1 not in 2 */
    for (wl1 = wl10; wl1; wl1 = wn) {
        wn = wl1->wl_next;
        if (!inwlist(wl1->wl_word,wl20)) {
            fprintf(cp_err, 
            ">>> vector %s in %s not in %s\n",
                wl1->wl_word, p1->pl_typename,
                p2->pl_typename);
            if (!wl1->wl_prev) {
                wl10 = wn;
                wl10->wl_prev = NULL;
            }
            else {
                wl1->wl_prev->wl_next = wn;
                if (wn)
                    wn->wl_prev = wl1->wl_prev;
            }
            txfree(wl1->wl_word);
            txfree((char*)wl1);
        }
    }

    /* rid 2, warn first if not in 1 */
    for (wl2 = wl20; wl2; wl2 = wl2->wl_next) {
        if (!inwlist(wl2->wl_word,wl10)) {
            fprintf(cp_err, 
            ">>> vector %s in %s not in %s\n",
                wl2->wl_word, p2->pl_typename,
                p1->pl_typename);
        }
    }
    wl_free(wl20);

    dl0 = NULL;
    for (wl1 = wl10; wl1; wl1 = wl1->wl_next) {
        v1 = (struct dvec*)htab_get(wl1->wl_word,p1->pl_hashtab);
        v2 = (struct dvec*)htab_get(wl1->wl_word,p2->pl_hashtab);
        if (((v1->v_flags & (VF_REAL | VF_COMPLEX)) ==
                (v2->v_flags & (VF_REAL | VF_COMPLEX))) &&
                (v1->v_type == v2->v_type)) {
            if (!dl0)
                dl0 = dl = alloc(struct dvlist2);
            else {
                dl->d2_next = alloc(struct dvlist2);
                dl = dl->d2_next;
            }
            dl->d2_dv1 = v1;
            dl->d2_dv2 = v2;
        }
        else {
            fprintf(cp_err, 
                ">>> vector %s type mismatch\n", wl1->wl_word);
        }
    }
    wl_free(wl10);

    
    /* Now we have all the vectors linked to their twins.  Travel
     * down each one and print values that differ enough.
     */
    for (dl = dl0; dl; dl = dl->d2_next) {
        v1 = dl->d2_dv1;
        v2 = dl->d2_dv2;

        if (v1->v_type == SV_VOLTAGE)
            tol = vntol;
        else
            tol = abstol;
        j = MAX(v1->v_length, v2->v_length);
        for (i = 0; i < j; i++) {
            if (v1->v_length <= i) {
                out_printf( 
                    ">>> %s is %d long in %s and %d long in %s\n",
                    v1->v_name, v1->v_length,
                    p1->pl_typename, v2->v_length,
                    p2->pl_typename);
                break;
            }
            else if (v2->v_length <= i) {
                out_printf( 
                    ">>> %s is %d long in %s and %d long in %s\n",
                    v2->v_name, v2->v_length,
                    p2->pl_typename, v1->v_length,
                    p1->pl_typename);
                break;
            }
            else {
                if (isreal(v1)) {
                    d1 = v1->v_realdata[i];
                    d2 = v2->v_realdata[i];
                    if (MAX(FABS(d1), FABS(d2)) * reltol +
                            tol < FABS(d1 - d2)) {
                        out_printf(
                            "%s.%s[%d] = %-15s ",
                            p1->pl_typename, v1->v_name, i,
                            printnum(d1));
                        out_printf(
                            "%s.%s[%d] = %s\n",
                            p2->pl_typename, v2->v_name, i,
                            printnum(d2));
                    }
                }
                else {
                    c1 = v1->v_compdata[i];
                    c2 = v2->v_compdata[i];
                    realpart(&c3) = realpart(&c1) - realpart(&c2);
                    imagpart(&c3) = imagpart(&c1) - imagpart(&c2);
                    /* Stupid evil PC compilers */
                    cm1 = cmag(&c1);
                    cm2 = cmag(&c2);
                    cmax = MAX(cm1, cm2);
                    if (cmax * reltol +
                            tol < cmag(&c3)) {
                        out_printf(
                            "%s.%s[%d] = %-10s, ",
                            p1->pl_typename, v1->v_name, i,
                            printnum(realpart(&c1)));
                        out_printf("%-10s ",printnum(imagpart(&c1)));
                        out_printf(
                            "%s.%s[%d] = %-10s, ",
                            p2->pl_typename, v2->v_name, i,
                            printnum(realpart(&c2)));
                        out_printf("%-10s\n",printnum(imagpart(&c2)));
                    }
                }
            }
        }
    }
    for (; dl0; dl0 = dl) {
        dl = dl0->d2_next;
        txfree((char*)dl0);
    }
    return;
}


static bool
nameeq(n1, n2)

/* Determine if two vectors have the 'same' name. */
char *n1, *n2;
{
    char buf1[BSIZE_SP], buf2[BSIZE_SP];
    int i;

    if (eq(n1, n2))
        return (true);
    if (isdigit(*n1))
        (void) sprintf(buf1, "v(%s)", n1);
    else
        (void) strcpy(buf1, n1);
    if (isdigit(*n2))
        (void) sprintf(buf2, "v(%s)", n2);
    else
        (void) strcpy(buf2, n2);
    for (i = 0; buf1[i]; i++)
        if (isupper(buf1[i]))
            buf1[i] = tolower(buf1[i]);
    for (i = 0; buf2[i]; i++)
        if (isupper(buf2[i]))
            buf2[i] = tolower(buf2[i]);
    return (eq(buf1, buf2) ? true : false);
}


static bool
inwlist(word,wl)

char *word;
wordlist *wl;
{
    while (wl) {
        if (!strcmp(word,wl->wl_word))
            return (true);
        wl = wl->wl_next;
    }
    return (false);
}
