/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * The user-supplied routine to deal with variables. Most variables we
 * don't use often, so just call cp_getvar when they are needed. Spice
 * variables, though, and a few commonly used ones are dealt with here.
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"

/* used by Josephson junction model,
 * default max phase change is 2*PI/10
 */
double JJdphi = 2.07e-16;

bool ft_nopage = false;
bool ft_parsedb = false;
bool ft_vecdb = false;
bool ft_simdb = false;
bool ft_evdb = false;
bool ft_grdb = false;
bool ft_gidb = false;
bool ft_controldb = false;
bool ft_asyncdb = false;

#ifdef __STDC__
static struct variable *vec2var(struct dvec*,int,int);
static struct variable *getlist(struct dvec*,int,char**);
#else
static struct variable *vec2var();
static struct variable *getlist();
#endif

/* The following routines take care of managing internal variables.
 * Some of the functionality was previously found in variable.c. Here,
 * all variables are included in a single table below, many of which
 * have dispatch functions.  This code should be much faster.
 */

/* These are the dispatch functions... */

#define def_fn(name) \
static int name(v,flag) \
struct variable *v; \
bool flag;

/* "shell" variables */

def_fn(_iv_cpdebug)
{
    cp_debug = flag;
#ifndef CPDEBUG
    if(flag)
        fprintf(cp_err, 
        "Warning: program not compiled with cshpar debug messages\n");
#endif
    return (US_OK);
}

def_fn(_iv_history)
{
    if (!flag) return;
    if (v->va_type == VT_NUM)
        cp_maxhistlength = v->va_num;
    else if (v->va_type == VT_REAL)
        cp_maxhistlength = v->va_real;
    return (US_OK);
}

def_fn(_iv_ignoreeof) {cp_ignoreeof=flag; return (US_OK);}
def_fn(_iv_noclobber) {cp_noclobber=flag; return (US_OK);}
def_fn(_iv_noglob)    {cp_noglob=flag;    return (US_OK);}
def_fn(_iv_nonomatch) {cp_nonomatch=flag; return (US_OK);}

def_fn(_iv_prompt)
{
    if (!flag)
        cp_promptstring = "";
    else if (v->va_type == VT_STRING)
        cp_promptstring = copy(v->va_string);
    return (US_OK);
}


/* spice and nutmeg variables */

static char *huhmsg = "Excuse me??\n";
static char *pltmsg = "Error: can't set plot";

def_fn(_iv_acct)    {ft_acctprint=flag;  return (US_OK);}

def_fn(_iv_curplot)
{
    if (v->va_type == VT_STRING)
        plot_setcur(v->va_string);
    else if (flag)
        fprintf(cp_err, "Error: plot name not a string\n");
    return (US_DONTRECORD);
}

def_fn(_iv_curplotdate)
{
    if (plot_cur && (v->va_type == VT_STRING))
        plot_cur->pl_date = copy(v->va_string);
    else if (flag)
        fprintf(cp_err,"%s %s\n",pltmsg,"date");
    return (US_DONTRECORD);
}

def_fn(_iv_curplotname)
{
    if (plot_cur && (v->va_type == VT_STRING))
        plot_cur->pl_name = copy(v->va_string);
    else if (flag)
        fprintf(cp_err,"%s %s\n",pltmsg,"name");
    return (US_DONTRECORD);
}

def_fn(_iv_curplottitle)
{
    if (plot_cur && (v->va_type == VT_STRING))
        plot_cur->pl_title = copy(v->va_string);
    else
        fprintf(cp_err,"%s %s\n",pltmsg,"title");
    return (US_DONTRECORD);
}

static void
setdb(word)

char *word;
{
    if (eq(word,"async"))
        ft_asyncdb = true;
    else if (eq(word,"control"))
        ft_controldb = true;
    else if (eq(word,"cshpar"))
        cp_debug = true;
    else if (eq(word,"eval"))
        ft_evdb = true;
    else if (eq(word,"ginterface"))
        ft_gidb = true;
    else if (eq(word,"graf"))
        ft_grdb = true;
    else if (eq(word,"parser"))
        ft_parsedb = true;
    else if (eq(word,"siminterface"))
        ft_simdb = true;
    else if (eq(word,"vecdb"))
        ft_vecdb = true;
}

def_fn(_iv_debug)
{
    struct variable *tv;
    char *dbgmsg = "Error: bad type for debug var\n";

    if (v->va_type == VT_BOOL) {
        cp_debug = ft_simdb = ft_parsedb = ft_evdb = ft_vecdb =
            ft_grdb = ft_gidb = ft_controldb = flag;
    }
    else if (v->va_type == VT_LIST) {
        for (tv = v->va_vlist; tv; tv = tv->va_next) {
            if (v->va_type != VT_STRING)
                fprintf(cp_err,dbgmsg);
            else
                setdb(tv->va_string);
        }
    }
    else if (v->va_type == VT_STRING) {
        setdb(v->va_string);
    }
    else if (flag)
        fprintf(cp_err,dbgmsg);
#ifndef FTEDEBUG
    if (flag)
        fprintf(cp_err, "Warning: %s compiled without debug messages\n",
        cp_program);
#endif
    return (US_OK);
}

def_fn(_iv_display)  { return (US_READONLY);}

def_fn(_iv_height)
{
    if ((v->va_type == VT_BOOL) && (flag == false))
        out_height = 0;
    else if (v->va_type == VT_REAL)
        out_height = v->va_real;
    else if (v->va_type == VT_NUM)
        out_height = v->va_num;
    else
        fprintf(cp_err, "Bad 'height' \"%s\"\n", v->va_name);
    return (US_OK);
}

def_fn(_iv_jjdphimax)
{
    if (flag == false)
        JJdphi = 2.07e-16;
    else if (v->va_type == VT_REAL) {
        if (v->va_real < M_PI/50 || v->va_real > M_PI/2)
            goto errm;
        JJdphi = v->va_real*3.3e-16; /* phi0 /2*pi */
    }
    else if (v->va_type == VT_NUM) {
        if (v->va_num < M_PI/50 || v->va_num > M_PI/2)
            goto errm;
        JJdphi = v->va_num*3.3e-16; /* phi0 /2*pi */
    }
    else
        fprintf(cp_err,huhmsg);
    return (US_OK);
errm:
    fprintf(cp_err, "Error: must be between pi/50 and pi/2\n");
    return (US_OK);
}

def_fn(_iv_list)       {ft_listprint=flag;  return (US_OK);}
def_fn(_iv_nocc)       {cp_nocc=flag;       return (US_OK);}
def_fn(_iv_node)       {ft_nodesprint=flag; return (US_OK);}
def_fn(_iv_noiter)     {return (US_OK);}
def_fn(_iv_nomoremode) {out_moremode=!flag; return (US_OK);}
def_fn(_iv_nopage)     {ft_nopage=flag;     return (US_OK);}

def_fn(_iv_numdgt)
{
    if ((v->va_type == VT_BOOL) && (flag == false))
        cp_numdgt = -1;
    else if (v->va_type == VT_REAL)
        cp_numdgt = v->va_real;
    else if (v->va_type == VT_NUM)
        cp_numdgt = v->va_num;
    else
        fprintf(cp_err,huhmsg);
    return (US_OK);
}

def_fn(_iv_opts)    {ft_optsprint=flag;  return (US_OK);}
def_fn(_iv_plots)   {return (US_READONLY);}

def_fn(_iv_program)
{
    if (!flag)
        cp_program = "";
    else if (v->va_type == VT_STRING)
        cp_program = copy(v->va_string);
    return (US_OK);
}

def_fn(_iv_rawfile)
{
    if (!flag)
        ft_rawfile = "";
    else if (v->va_type == VT_STRING)
        ft_rawfile = copy(v->va_string);
    return (US_OK);
}

def_fn(_iv_rawfileprec)
{
    if ((v->va_type == VT_BOOL) && (flag == false))
        raw_prec = -1;
    else if (v->va_type == VT_REAL)
        raw_prec = v->va_real;
    else if (v->va_type == VT_NUM)
        raw_prec = v->va_num;
    else
        fprintf(cp_err, "Bad 'rawfileprec' \"%s\"\n", v->va_name);
    return (US_OK);
}

def_fn(_iv_strictnumparse) {ft_strictnumparse=flag; return (US_OK);}

def_fn(_iv_units)
{
    if (false && ((*v->va_string == 'd') ||
            (*v->va_string == 'D')))
        cx_degrees = true;
    else
        cx_degrees = false;
    return (US_OK);
}

def_fn(_iv_unixcom)
{
    char *s;

    cp_dounixcom = flag;
    if (flag) {
        s = getenv("PATH");
        if (s)
            cp_rehash(s, !cp_nocc);
        else
            fprintf(cp_err, 
                "Warning: no PATH in environment.\n");
    }
    else {
        struct comm *c;
        cp_ccrestart(false);
        for (c = cp_coms; c->co_func; c++)
            cp_addcomm(c->co_comname, c->co_cctypes[0],
                c->co_cctypes[1], c->co_cctypes[2],
                c->co_cctypes[3]);
    }
    return (US_OK);
}

def_fn(_iv_width)
{
    if ((v->va_type == VT_BOOL) && (flag == false))
        out_width = 0;
    else if (v->va_type == VT_REAL)
        out_width = v->va_real;
    else if (v->va_type == VT_NUM)
        out_width = v->va_num;
    else
        fprintf(cp_err, "Bad 'width' \"%s\"\n", v->va_name);
    return (US_OK);
}



/* The table below includes all variables with significance to the
 * program.
 */
char *kw_abstol           = "abstol";
char *kw_acct             = "acct";
char *kw_appendwrite      = "appendwrite";
char *kw_checkiterate     = "checkiterate";
char *kw_chgtol           = "chgtol";
char *kw_color            = "color";
char *kw_cpdebug          = "cpdebug";
char *kw_curplot          = "curplot";
char *kw_curplotname      = "curplotname";
char *kw_curplottitle     = "curplottitle";
char *kw_curplotdate      = "curplotdate";
char *kw_debug            = "debug";
char *kw_defad            = "defad";
char *kw_defas            = "defas";
char *kw_defl             = "defl";
char *kw_defw             = "defw";
char *kw_device           = "device";
char *kw_display          = "display";
char *kw_dontplot         = "dontplot";
char *kw_editor           = "editor";
char *kw_filetype         = "filetype";
char *kw_fourgridsize     = "fourgridsize";
char *kw_gmin             = "gmin";
char *kw_gridsize         = "gridsize";
char *kw_gridstyle        = "gridstyle";
char *kw_hcopydev         = "hcopydev";
char *kw_hcopydevtype     = "hcopydevtype";
char *kw_height           = "height";
char *kw_history          = "history";
char *kw_hitusertp        = "hitusertp";
char *kw_ignoreeof        = "ignoreeof";
char *kw_itl1             = "itl1";
char *kw_itl2             = "itl2";
char *kw_itl3             = "itl3";
char *kw_itl4             = "itl4";
char *kw_itl5             = "itl5";
char *kw_jjdphimax        = "jjdphimax";
char *kw_level            = "level";
char *kw_list             = "list";
char *kw_maxwins          = "maxwins";
char *kw_modelcard        = "modelcard";
char *kw_mplot_cur        = "mplot_cur";
char *kw_nfreqs           = "nfreqs";
char *kw_noasciiplotvalue = "noasciiplotvalue";
char *kw_noaskquit        = "noaskquit";
char *kw_nobjthack        = "nobjthack";
char *kw_nobreak          = "nobreak";
char *kw_nocc             = "nocc";
char *kw_noclobber        = "noclobber";
char *kw_node             = "node";
char *kw_noglob           = "noglob";
char *kw_nogrid           = "nogrid";
char *kw_noiter           = "noiter";
char *kw_nojjtp           = "nojjtp";
char *kw_nomoremode       = "nomoremode";
char *kw_nonomatch        = "nonomatch";
char *kw_nopage           = "nopage";
char *kw_noparse          = "noparse";
char *kw_noprintscale     = "noprintscale";
char *kw_nosort           = "nosort";
char *kw_nosubckt         = "nosubckt";
char *kw_nousertp         = "nousertp";
char *kw_numdgt           = "numdgt";
char *kw_opts             = "opts";
char *kw_pivrel           = "pivrel";
char *kw_pivtol           = "pivtol";
char *kw_plots            = "plots";
char *kw_plotstyle        = "plotstyle";
char *kw_pointchars       = "pointchars";
char *kw_polydegree       = "polydegree";
char *kw_polysteps        = "polysteps";
char *kw_printinfo        = "printinfo";
char *kw_program          = "program";
char *kw_prompt           = "prompt";
char *kw_rawfile          = "rawfile";
char *kw_rawfileprec      = "rawfileprec";
char *kw_reltol           = "reltol";
char *kw_renumber         = "renumber";
char *kw_rhost            = "rhost";
char *kw_rprogram         = "rprogram";
char *kw_scedfont         = "scedfont";
char *kw_slowplot         = "slowplot";
char *kw_sourcepath       = "sourcepath";
char *kw_spicepath        = "spicepath";
char *kw_strictnumparse   = "strictnumparse";
char *kw_subend           = "subend";
char *kw_subinvoke        = "subinvoke";
char *kw_substart         = "substart";
char *kw_term             = "term";
char *kw_ticmarks         = "ticmarks";
char *kw_tnom             = "tnom";
char *kw_trtol            = "trtol";
char *kw_units            = "units";
char *kw_unixcom          = "unixcom";
char *kw_vntol            = "vntol";
char *kw_width            = "width";
char *kw_wpboxh           = "wpboxh";
char *kw_wpboxw           = "wpboxw";
char *kw_xglinewidth      = "xglinewidth";
char *kw_xgmarkers        = "xgmarkers";
char *kw_xfont            = "xfont";


struct intvar {
    char **name;
    char *descr;
#ifdef __STDC__
    int (*func)(struct variable*,bool);
#else
    int (*func)();
#endif
};
static struct intvar int_vars[] = {
{&kw_abstol, "real   Absolute error tolerance, default 1e-12.", NULL},
{&kw_acct, "bool   Print accounting summary.", _iv_acct},
{&kw_appendwrite, "bool   Append to file with write command, no overwrite.", NULL},
{&kw_checkiterate,"int    Binary search iterations in range analysis.", NULL},
{&kw_chgtol,"real   Charge tolerance, default 1e-14.", NULL},
{&kw_color, "list   X window colors, see help.", NULL},
{&kw_cpdebug, "bool   Enable command processor debugging.", _iv_cpdebug},
{&kw_curplot, "str    Current plot.", _iv_curplot},
{&kw_curplotname, "str    Current plot name, read only.", _iv_curplotname},
{&kw_curplottitle, "str    Current plot title, read only.", _iv_curplottitle},
{&kw_curplotdate, "str    Current plot date, read only", _iv_curplotdate},
{&kw_debug, "list   Enable debugging, see help for keywords.", _iv_debug},
{&kw_defad, "real   MOS drain diffusion area, default 0.", NULL},
{&kw_defas, "real   MOS source diffusion area, defalut 0", NULL},
{&kw_defl, "real   MOS channel length, default 100 microns.", NULL},
{&kw_defw, "real   MOS channel width, default 100 microns.", NULL},
{&kw_device, "str    UNIX /dev/tty<??> device for output.", NULL},
{&kw_display, "str    X uses this display, read only.", _iv_display},
{&kw_dontplot, "bool   Don't send graphics code to output device.", NULL},
{&kw_editor, "str    Editor invocation string.", NULL},
{&kw_filetype, "str    Rawfile type: ascii or binary.", NULL},
{&kw_fourgridsize, "int    Number of interpolation points in fourier analysis.", NULL},
{&kw_gmin, "real   Minimum conductance allowed, default 1e-12.", NULL},
{&kw_gridsize, "real   Number of y-axis grid lines.", NULL},
{&kw_gridstyle, "str    Plot default grid type (lingrid, loglog, nogrid, \n\t\t   xlog, ylog, polar, smith).", NULL},
{&kw_hcopydev, "str    UNIX: send hardcopy using LPR -P??. DOS: prn,lpt?.", NULL},
{&kw_hcopydevtype, "str    Hardcopy device type: postscript, etc.", NULL},
{&kw_height, "real   Screen height in characters.", _iv_height},
{&kw_history, "int    Number of remembered commands, default 1000.", _iv_history},
{&kw_hitusertp, "bool   Force transient analysis at user time points.", NULL},
{&kw_ignoreeof,"bool   Ignore end of file (^D) in UNIX.", _iv_ignoreeof},
{&kw_itl1, "int    DC iteration limit, default 100.", NULL},
{&kw_itl2, "int    DC transfer curve iteration limit, default 50", NULL},
{&kw_itl3, "int    Lower transient iteration limit, default 4.", NULL},
{&kw_itl4, "int    Upper transient iteration limit, default 10.", NULL},
{&kw_itl5, "int    Unused.", NULL},
{&kw_jjdphimax, "real   Max Josephson phase delta per step, default pi/10.", _iv_jjdphimax},
{&kw_level, "str    Level for newhelp, (b, i, or a).", NULL},
{&kw_list, "bool   Causes the input to be listed.", _iv_list},
{&kw_maxwins, "int    Maximum number of windows in X.", NULL},
{&kw_modelcard, "str    Name of model card, default .model.", NULL},
{&kw_mplot_cur, "str    Name of current output file for check command", NULL},
{&kw_nfreqs, "int    Number of frequencies in fourier command, default 10.", NULL},
{&kw_noasciiplotvalue, "bool   Don't print scale value in asciiplot.", NULL},
{&kw_noaskquit, "bool   Don't verify before termination.", NULL},
{&kw_nobjthack, "bool   Assume BJT's have 4 nodes.", NULL},
{&kw_nobreak, "bool   No break between pages in asciiplot.", NULL},
{&kw_nocc, "bool   No command completion.", _iv_nocc},
{&kw_noclobber, "bool   Don't overwrite files when redirecting output.", _iv_noclobber},
{&kw_node, "bool   Print node table.", _iv_node},
{&kw_noglob, "bool   Don't expand wildcard characters.", _iv_noglob},
{&kw_nogrid, "bool   Don't plot a grid.", NULL},
{&kw_noiter, "bool   No transient iterations past predictor.", _iv_noiter},
{&kw_nojjtp, "bool   No Josephson phase change timestep, use trunc error.", NULL},
{&kw_nomoremode, "bool   Turn off more mode.", _iv_nomoremode},
{&kw_nonomatch, "bool   Use wildcard characters literally if no match.", _iv_nonomatch},
{&kw_nopage, "bool   Supress page ejects.", _iv_nopage},
{&kw_noparse, "bool   Don't parse circuit.", NULL},
{&kw_noprintscale, "bool   Don't print scale in print command.", NULL},
{&kw_nosort, "bool   Supress sorting of variable names.", NULL},
{&kw_nosubckt, "bool   No subcircuit expansion.", NULL},
{&kw_nousertp, "bool   Save raw time point values in transient analysis.", NULL},
{&kw_numdgt, "int    Number of significant digits to print, default 4.", _iv_numdgt},
{&kw_opts, "bool   Print options.", _iv_opts},
{&kw_pivrel, "real   Relative pivot value, default 1e-3.", NULL},
{&kw_pivtol, "real   Minimum pivot value, default 1e-13.", NULL},
{&kw_plots, "list   Read only, list of plots.", _iv_plots},
{&kw_plotstyle, "str    Default plot type (linplot, combplot, pointplot).", NULL},
{&kw_pointchars, "str    Point characters for plotting.", NULL},
{&kw_polydegree, "int    Degree of interpolating polynomial.", NULL},
{&kw_polysteps, "int    Number of interpolating points, default 10.", NULL},
{&kw_printinfo, "bool   Print debugging info when plot starts.", NULL},
{&kw_program, "str    Program name.", _iv_program},
{&kw_prompt, "str    Prompt string.", _iv_prompt},
{&kw_rawfile, "str    Rawfile path.", _iv_rawfile},
{&kw_rawfileprec, "int    Significant digits in rawfile.", _iv_rawfileprec},
{&kw_reltol, "real   Relative error tolerance, default 1e-3.", NULL},
{&kw_renumber, "bool   Renumber source lines after subcircuit expansion.", NULL},
{&kw_rhost, "str    Remote host name, for UNIX remote simulations.", NULL},
{&kw_rprogram, "str    Remote simulation program name, for UNIX.", NULL},
{&kw_scedfont, "str    Name of X font used in SCED.", NULL},
{&kw_slowplot, "bool   Wait for input between plots.", NULL},
{&kw_sourcepath, "str    Path to search for source command.", NULL},
{&kw_spicepath, "str    Path to use in aspice command.", NULL},
{&kw_strictnumparse, "bool   Don't allow trailing characters after number.", _iv_strictnumparse},
{&kw_subend, "str    End subcircuits, default .ends.", NULL},
{&kw_subinvoke, "str    Prefix to invoke subcircuits, default x.", NULL},
{&kw_substart, "str    Start subcircuit definition, default .subckt.", NULL},
{&kw_term, "str    Name of current terminal, for MFB.", NULL},
{&kw_ticmarks, "bool   Plot tic marks, if int, set tic mark separation.", NULL},
{&kw_tnom, "real   Nominal temperature, default 27C.", NULL},
{&kw_trtol, "real   Truncation error tolerance, default 7.", NULL},
{&kw_units, "str    If \"degrees\", trig functions don't use radians.", _iv_units},
{&kw_unixcom, "bool   Execute operating system commands.", _iv_unixcom},
{&kw_vntol, "real   Voltage error tolerance, default 1e-6.", NULL},
{&kw_width, "int    Width of screen in characters.", _iv_width},
{&kw_wpboxh, "real   Graphics box height for WordPerfect plot, (4\").", NULL},
{&kw_wpboxw, "real   Graphics box width for WordPerfect plot (4\").", NULL},
{&kw_xglinewidth, "int    Pixel linewidth used in xgraph, default 1.", NULL},
{&kw_xgmarkers, "bool   Use markers in xgraph pointplot, else big pixels.", NULL},
{&kw_xfont, "str    Name of X font for graphics, default \"fixed\".", NULL}
} ;


static void *internalvars; /* hash table for dispatch functions */

void
cp_internal_init()

/* This has to be called before anything, it sets up the hash table */
{
    int i;

    if (internalvars == NULL)
        internalvars = htab_init();
    for (i = 0; i < sizeof(int_vars)/sizeof(struct intvar); i++) {
        if (int_vars[i].func) {
            htab_add(*(int_vars[i].name),(void*)&int_vars[i],internalvars);
            cp_addkword(CT_VARIABLES, *(int_vars[i].name));
        }
    }
}


void
com_usrset(wl)

/* print a listing of option keywords and descriptions */
wordlist *wl;
{
    int i;

    out_init();
    if (wl == NULL) {
        for (i = 0; i < sizeof(int_vars)/sizeof(struct intvar); i++) {
            out_printf("%-18s %s\n",*(int_vars[i].name),int_vars[i].descr);
        }
        return;
    }
    while (wl) {
        for (i = 0; i < sizeof(int_vars)/sizeof(struct intvar); i++) {
            if (eq(wl->wl_word,*(int_vars[i].name))) {
                out_printf("%-18s %s\n",*(int_vars[i].name),int_vars[i].descr);
                break;
            }
        }
        if (i == sizeof(int_vars)/sizeof(struct intvar))
            out_printf("%-18s %s\n",wl->wl_word,"not an internal variable");
        wl = wl->wl_next;
    }
}


int
cp_internalset(v,flag)

/* The one variable that we consider read-only so far is plots.  The ones
 * that are 'dontrecord' are curplottitle, curplotname, and curplotdate.
 * Also things already in the plot env are 'dontrecord'.
 */
struct variable *v;
bool flag;
{
    struct intvar *x;
    struct variable *tv;
    int i;
    char val[BSIZE_SP];
    char *vv;
    bool bv;
    double dv;
    int iv;

    x = (struct intvar *)htab_get(v->va_name,internalvars);
    if (x)
        i = (*x->func)(v,flag);
    else
        return (US_OK);

    if (i != US_OK)
        return (i);

    if (plot_cur)
        for (tv = plot_cur->pl_env; tv; tv = tv->va_next)
            if (eq(tv->va_name, v->va_name))
                return (US_READONLY);

    if (!flag || ft_nutmeg)
        return (US_OK);

    /* Now call the interface option routine. */
    switch (v->va_type) {
        case VT_BOOL:
            if (v->va_bool) {
                val[0] = '\0';
                bv = true;
                vv = (char *) &bv;
                break;
            }
            else {
                bv = false;
                vv = (char *) &bv;
            }
        case VT_STRING:
            (void) strcpy(val, v->va_string);
            vv = val;
            break;
        case VT_NUM:
            (void) sprintf(val, "%d", v->va_num);
            iv = v->va_num;
            vv = (char *) &iv;
            break;
        case VT_REAL:
            (void) strcpy(val, printnum(v->va_real));
            dv = v->va_real;
            vv = (char *) &dv;
            break;
        case VT_LIST:
            /* if_option can't handle lists anyway. */
            break;
        default:
            fprintf(cp_err, 
            "cp_usrset: Internal Error: Bad var type %d\n",
                    v->va_type);
    }

    if (ft_curckt)
        if_option(ft_curckt->ci_ckt, v->va_name, v->va_type, vv);

    return (US_OK);
}


/* The user-supplied routine to query the values of variables. This
 * recognises the $&varname notation, and also searches the values of
 * plot and circuit environment variables.
 * This function now accepts a range extension for $& variables.
 * It will also evaluate vector expressions encased in ( ).
 */

struct variable *
cp_enqvar(word)

char *word;
{
    struct dvec *d;
    struct variable *vv = NULL, *tv;
    struct plot *pl;
    int i;
    char **x;
    char *range, *r;
    char *tt = NULL;
    int low, up;

    if (*word == '&') {
        word++;

        low = 0;
        up = -1;

        if (*word == '(') {
            /* an expression */
            tt = strrchr(word,')');
            if (tt == NULL) {
                fprintf(cp_err,"Error: closing ')' not found.\n");
                return (NULL);
            }
            r = range = strchr(tt,'[');
        }
        else
            r = range = strchr(word,'[');

        if (r) {
            *r = '\0';
            r++;
            if (!isdigit(*r) && *r != '-')
                fprintf(cp_err,
                    "Warning: nonparseable range specified, %s[%s\n",
                    word,r);
            for (low = 0; isdigit(*r); r++)
                low = low * 10 + *r - '0';
            if ((*r == '-') && isdigit(r[1]))
                for (up = 0, r++; isdigit(*r); r++)
                    up = up * 10 + *r - '0';
            else if (*r != '-')
                up = low;
        }
        if (tt) {
            /* evaluate the expression */
            wordlist wl;
            struct pnode *pn;

            wl.wl_word = word;
            wl.wl_next = wl.wl_prev = NULL;
            pn = ft_getpnames(&wl,true);
            if (pn == NULL) {
                fprintf(cp_err,"Error: parse failed.\n");
                return (NULL);
            }
            d = ft_evaluate(pn);
            inp_pnfree(pn);
        }
        else
            d = vec_get(word);

        if (d) {
            if (d->v_link2) {
                fprintf(cp_err, 
    "Warning: only one vector may be accessed with the $& notation.\n");
                d = d->v_link2->dl_dvec;
            }
            if (d->v_numdims <= 1) {
                d->v_numdims = 1;
                d->v_dims[0] = d->v_length;
            }
            if (up == -1)
                up = d->v_dims[0] - 1;
            vv = vec2var(d,low,up);
        }
        if (range)
            *range = '[';
        return (vv);
    } 
    if (eq(word, kw_display)) {
        vv = alloc(struct variable);
        vv->va_name = copy(word);
        vv->va_type = VT_STRING;
        vv->va_string = copy(cp_display);
        return (vv);
    }
    if (plot_cur) {
        for (vv = plot_cur->pl_env; vv; vv = vv->va_next)
            if (eq(vv->va_name, word))
                break;
        if (vv)
            vv =  va_copy(vv);
        else if (eq(word, kw_curplotname)) {
            vv = alloc(struct variable);
            vv->va_name = copy(word);
            vv->va_type = VT_STRING;
            vv->va_string = copy(plot_cur->pl_name);
        }
        else if (eq(word, kw_curplottitle)) {
            vv = alloc(struct variable);
            vv->va_name = copy(word);
            vv->va_type = VT_STRING;
            vv->va_string = copy(plot_cur->pl_title);
        }
        else if (eq(word, kw_curplotdate)) {
            vv = alloc(struct variable);
            vv->va_name = copy(word);
            vv->va_type = VT_STRING;
            vv->va_string = copy(plot_cur->pl_date);
        }
        else if (eq(word, kw_curplot)) {
            vv = alloc(struct variable);
            vv->va_name = copy(word);
            vv->va_type = VT_STRING;
            vv->va_string = copy(plot_cur->pl_typename);
        }
        else if (eq(word, kw_plots)) {
            vv = alloc(struct variable);
            vv->va_name = copy(word);
            vv->va_type = VT_LIST;
            for (pl = plot_list; pl; pl = pl->pl_next) {
                tv = alloc(struct variable);
                tv->va_type = VT_STRING;
                tv->va_string = copy(pl->pl_typename);
                tv->va_next = vv->va_vlist;
                vv->va_vlist = tv;
            }
        }
        if (vv)
            return (vv);
    }
    if (ft_curckt) {
        for (vv = ft_curckt->ci_vars; vv; vv = vv->va_next)
            if (eq(vv->va_name, word))
                break;
        if (vv)
            return va_copy(vv);
    }
    return (NULL);
}


static struct variable *
vec2var(d,lo,hi)

struct dvec *d;
int lo, hi;
{
    int i, sz, sn;
    struct variable *vv, *tv = NULL;
    char *x;

    sn = ((hi < lo) ? -1 : 1);
    sz = d->v_length/d->v_dims[0];

    for (i = lo; ((sn == 1) ? (i <= hi) : (i >= hi)); i += sn) {
        if (i >= d->v_dims[0])
            continue;
        if (!tv)
            tv = vv = alloc(struct variable);
        else {
            vv->va_next = alloc(struct variable);
            vv = vv->va_next;
        }
        if (d->v_numdims == 1) {
            if (isreal(d))
                vv->va_real = d->v_realdata[i];
            else
                vv->va_real = realpart(&d->v_compdata[i]);
            vv->va_type = VT_REAL;
        }
        else {
            if (isreal(d))
                x = (char*)&d->v_realdata[i*sz];
            else
                x = (char*)&d->v_compdata[i*sz];
            vv->va_type = VT_LIST;
            vv->va_vlist = getlist(d,1,&x);
        }
    }
    if (tv) {
        if (lo == hi)
            return (tv);
        vv = alloc(struct variable);
        vv->va_type = VT_LIST;
        vv->va_vlist = tv;
        return (vv);
    }
    return (NULL);
}


static struct variable *
getlist(d,dim,x)

struct dvec *d;
int dim;
char **x;
{
    struct variable *vv, *tv = NULL;
    int i;

    for (i = d->v_dims[dim] - 1; i >= 0; i--) {
        if (!tv)
            tv = vv = alloc(struct variable);
        else {
            vv->va_next = alloc(struct variable);
            vv = vv->va_next;
        }
        if (dim == d->v_numdims-1) {
            if (isreal(d)) {
                vv->va_real = *(*(double**)x);
                *x += sizeof(double);
            }
            else {
                vv->va_real = (*(complex**)x)->cx_real;
                *x += sizeof(complex);
            }
            vv->va_type = VT_REAL;
        }
        else {
            vv->va_vlist = getlist(d,dim+1,x);
            vv->va_type = VT_LIST;
        }
    }
    return (tv);
}


/* Return the plot and ckt env vars, $plots, and $curplot{name,title,date,} */

void
cp_usrvars(v1, v2)

struct variable **v1, **v2;
{
    struct variable *v, *tv;

    if (plot_cur)
        v =  va_copy(plot_cur->pl_env);
    else
        v = NULL;
    if (tv = cp_enqvar(kw_plots)) {
        tv->va_next = v;
        v = tv;
    }
    if (tv = cp_enqvar(kw_curplot)) {
        tv->va_next = v;
        v = tv;
    }
    if (tv = cp_enqvar(kw_curplottitle)) {
        tv->va_next = v;
        v = tv;
    }
    if (tv = cp_enqvar(kw_curplotname)) {
        tv->va_next = v;
        v = tv;
    }
    if (tv = cp_enqvar(kw_curplotdate)) {
        tv->va_next = v;
        v = tv;
    }
    if (tv = cp_enqvar(kw_display)) {
        tv->va_next = v;
        v = tv;
    }
    *v1 = v;
    if (ft_curckt)
        *v2 = va_copy(ft_curckt->ci_vars);
    else
        *v2 = NULL;
    return;
}


/* Extract the .option cards from the deck... */

struct line *
inp_getopts(deck)

struct line *deck;
{
    struct line *last, *opts = NULL, *dd, *next;

    last = deck;
    for (dd = deck->li_next; dd; dd = next) {
        next = dd->li_next;

        if (ciprefix(".opt", dd->li_line)) {
            inp_casefix(dd->li_line);
            last->li_next = next;
            dd->li_next = opts;
            opts = dd;
            continue;
        }
        last = dd;
    }

    return (opts);
}
