/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1986 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Read and write the ascii and binary rawfile formats.
 */

#include "spice.h"
#include "ftedefs.h"


#ifdef __STDC__
static void fixdims(struct dvec*,char*);
static char *dimstring(int*,int,char*);
static int  atodims(char*,int*,int*);
#else
static void fixdims();
static char *dimstring();
static int  atodims();
#endif

int raw_prec = -1;  /* How many sigfigs to use, default 15 (max).  */

#define DEFPREC 15


/* Write a raw file.  We write everything in the plot pointed to. */

void
raw_write(name, pl, app, binary)

char *name;
struct plot *pl;
bool app, binary;
{
    FILE *fp;
    bool realflag = true;
    bool raw_padding;
    int length, numdims, dims[MAXDIMS];
    int nvars, i, j, prec;
    char buf[BSIZE_SP];
    struct dvec *v;
    struct variable *vv;
    struct dvlist *dl0, *dl, *tl;
    wordlist *wl;
    double dd;

    if (!cp_getvar("nopadding", VT_BOOL, (char *) &raw_padding))
        raw_padding = false;
    /* Invert since we want to know if we should pad. */
    raw_padding = !raw_padding;

    /* Why bother printing out an empty plot? */
    if (htab_empty(pl->pl_hashtab)) {
        fprintf(cp_err, "Error: plot is empty, nothing written.\n");
        return;
    }

    if (raw_prec != -1)
        prec = raw_prec;
    else
        prec = DEFPREC;

    if (!(fp = fopen(name, app ? "a" : "w"))) {
        perror(name);
        return;
    }

    numdims = nvars = length = 0;

    v = vec_fromplot("all",pl);
    vec_sort(v);
    dl0 = v->v_link2;

    for (nvars = 0, dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        if (iscomplex(v))
            realflag = false;
        nvars++;
        /* Find the length and dimensions of the longest vector
         * in the plot.
         * Be paranoid and assume somewhere we may have
         * forgotten to set the dimensions of 1-D vectors.
         */
        if (v->v_numdims <= 1) {
            v->v_numdims = 1;
            v->v_dims[0] = v->v_length;
        }
        if (v->v_length > length) {
            length = v->v_length;
            numdims = v->v_numdims;
            for (j = 0; j < numdims; j++) {
                dims[j] = v->v_dims[j];
            }
        }
    }

    fprintf(fp, "Title: %s\n", pl->pl_title);
    fprintf(fp, "Date: %s\n", pl->pl_date);
    fprintf(fp, "Plotname: %s\n", pl->pl_name);
    fprintf(fp, "Flags: %s%s\n",
        realflag ? "real" : "complex", raw_padding ? "" : " unpadded" );
    fprintf(fp, "No. Variables: %d\n", nvars);
    fprintf(fp, "No. Points: %d\n", length);
    if (numdims > 1) {
        fprintf(fp, "Dimensions: %s\n", dimstring(dims, numdims, buf));
    }

    for (wl = pl->pl_commands; wl; wl = wl->wl_next)
        fprintf(fp, "Command: %s\n", wl->wl_word);

    for (vv = pl->pl_env; vv; vv = vv->va_next) {
        wl = cp_varwl(vv);
        if (vv->va_type == VT_BOOL) {
            fprintf(fp, "Option: %s\n", vv->va_name);
        }
        else {
            fprintf(fp, "Option: %s = ", vv->va_name);
            if (vv->va_type == VT_LIST)
                fprintf(fp, "( ");
            wl_print(wl, fp);
            if (vv->va_type == VT_LIST)
                fprintf(fp, " )");
            (void) putc('\n', fp);
        }
    }

    /* Before we write the stuff out, make sure that the scale is the first
     * in the list.
     */

    for (tl = NULL,dl = dl0; dl; tl = dl, dl = dl->dl_next) {
        if (dl->dl_dvec == pl->pl_scale) {
            if (tl) {
                tl->dl_next = dl->dl_next;
                dl->dl_next = dl0;
                dl0 = dl;
            }
            break;
        }
    }

    fprintf(fp, "Variables:");
    for (i = 0, dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        fprintf(fp, " %d %s %s", i++, v->v_name,
                ft_typenames(v->v_type));
        if (v->v_flags & VF_MINGIVEN)
            fprintf(fp, " min=%e", v->v_minsignal);
        if (v->v_flags & VF_MAXGIVEN)
            fprintf(fp, " max=%e", v->v_maxsignal);
        if (v->v_defcolor)
            fprintf(fp, " color=%s", v->v_defcolor);
        if (v->v_gridtype)
            fprintf(fp, " grid=%d", v->v_gridtype);
        if (v->v_plottype)
            fprintf(fp, " plot=%d", v->v_gridtype);

        /* Only write dims if they are different from default. */
        j = 0;
        if (v->v_numdims == numdims) {
            for ( ; j < numdims; j++)
                if (dims[j] != v->v_dims[j])
                    break;
        }
        if (j < numdims) {
            fprintf(fp, " dims=%s", dimstring(v->v_dims, v->v_numdims, buf));
        }
        (void) putc('\n', fp);
    }

    if (binary) {
        fprintf(fp, "Binary:\n");
        for (i = 0; i < length; i++) {
            for (dl = dl0; dl; dl = dl->dl_next) {
                v = dl->dl_dvec;
                /* Don't run off the end of this vector's data. */
                if (i < v->v_length) {
                    if (realflag) {
                        dd = (isreal(v) ? v->v_realdata[i] :
                            realpart(&v->v_compdata[i]));
                        (void) fwrite((char *) &dd, sizeof
                                (double), 1, fp);
                    }
                    else if (isreal(v)) {
                        dd = v->v_realdata[i];
                        (void) fwrite((char *) &dd, sizeof
                                (double), 1, fp);
                        dd = 0.0;
                        (void) fwrite((char *) &dd, sizeof
                                (double), 1, fp);
                    }
                    else {
                        dd = realpart(&v->v_compdata[i]);
                        (void) fwrite((char *) &dd, sizeof
                                (double), 1, fp);
                        dd = imagpart(&v->v_compdata[i]);
                        (void) fwrite((char *) &dd, sizeof
                                (double), 1, fp);
                    }
                }
                else if (raw_padding) {
                    dd = 0.0;
                    if (realflag) {
                        (void) fwrite((char *) &dd, sizeof
                            (double), 1, fp);
                    }
                    else {
                        (void) fwrite((char *) &dd, sizeof
                            (double), 1, fp);
                        (void) fwrite((char *) &dd, sizeof
                            (double), 1, fp);
                    }
                }
            }
        }
    }
    else {
        fprintf(fp, "Values:\n");
        for (i = 0; i < length; i++) {
            fprintf(fp, " %d", i);
            for (dl = dl0; dl; dl = dl->dl_next) {
                v = dl->dl_dvec;
                if (i < v->v_length) {
                    if (realflag) {
                        fprintf(fp, "\t%.*e\n", prec,
                            isreal(v) ? v->v_realdata[i] :
                            realpart(&v->v_compdata[i]));
                    }
                    else if (isreal(v)) {
                        fprintf(fp, "\t%.*e,0.0\n", prec,
                            v->v_realdata[i]);
                    }
                    else {
                        fprintf(fp, "\t%.*e,%.*e\n", prec,
                            realpart(&v->v_compdata[i]),
                            prec,
                            imagpart(&v->v_compdata[i]));
                    }
                }
                else if (raw_padding) {
                    if (realflag) {
                        fprintf(fp, "\t%.*e\n", prec, 0.0);
                    }
                    else {
                        fprintf(fp, "\t%.*e,%.*e\n",
                            prec, 0.0, prec, 0.0);
                    }
                }
            }
        }
    }
    (void) fclose(fp);
    return;
}


/* Read a raw file. Returns a list of plot structures. This routine should be
 * very flexible about what it expects to see in the rawfile.  Really all we
 * require is that there be one variables and one values section per plot
 * and that the variables precede the values.
 */

#define skip(s) while (*(s) && !isspace(*(s)))(s)++; while (isspace(*(s)))(s)++
#define nonl(s) r = (s); while (*r && (*r != '\n')) r++; *r = '\0'

struct plot *
raw_read(name)

char *name;
{
    char *title = "default title";
    char *date = 0;
    struct plot *plots = NULL, *curpl = NULL;
    char buf[BSIZE_SP], buf2[BSIZE_SP], *s, *t, *r;
    int flags, nvars, npoints, i, j;
    int numdims = 0, dims[MAXDIMS];
    struct dvec *v, *nv;
    struct variable *vv;
    wordlist *wl, *nwl;
    FILE *fp, *lastin, *lastout, *lasterr;
    bool raw_padded = true;
    double junk;
    char *errbr = "Error: bad rawfile\n";

    if (!(fp = fopen(name, "r"))) {
        perror(name);
        return (NULL);
    }

    /* Since we call cp_evloop() from here, we have to do this junk. */
    lastin = cp_curin;
    lastout = cp_curout;
    lasterr = cp_curerr;
    cp_curin = cp_in;
    cp_curout = cp_out;
    cp_curerr = cp_err;

    cp_pushcontrol();

    while (fgets(buf, BSIZE_SP, fp)) {
        /* Figure out what this line is... */
        if (ciprefix("title:", buf)) {
            s = buf;
            skip(s);
            nonl(s);
            title = copy(s);
        }
        else if (ciprefix("date:", buf)) {
            s = buf;
            skip(s);
            nonl(s);
            date = copy(s);
        }
        else if (ciprefix("plotname:", buf)) {
            s = buf;
            skip(s);
            nonl(s);
            if (curpl) {    /* reverse commands list */
                for (wl=curpl->pl_commands,
                        curpl->pl_commands=NULL; wl &&
                        wl->wl_next; wl=nwl) {
                    nwl = wl->wl_next;
                    wl->wl_next = curpl->pl_commands;
                    curpl->pl_commands = wl;
                }
            }
            curpl = alloc(struct plot);
            curpl->pl_next = plots;
            plots = curpl;
            curpl->pl_name = copy(s);
            if (!date)
                date = copy(datestring( ));
            curpl->pl_date = date;
            curpl->pl_title = title;
            flags = 0;
            nvars = npoints = 0;
        }
        else if (ciprefix("flags:", buf)) {
            s = buf;
            skip(s);
            while (copytok(buf2,&s)) {
                if (cieq(buf2, "real"))
                    flags |= VF_REAL;
                else if (cieq(buf2, "complex"))
                    flags |= VF_COMPLEX;
                else if (cieq(buf2, "unpadded"))
                    raw_padded = false;
                else if (cieq(buf2, "padded"))
                    raw_padded = true;
                else
                    fprintf(cp_err, 
                        "Warning: unknown flag %s\n",buf2);
            }
        }
        else if (ciprefix("no. variables:", buf)) {
            s = buf;
            skip(s);
            skip(s);
            nvars = scannum(s);
        }
        else if (ciprefix("no. points:", buf)) {
            s = buf;
            skip(s);
            skip(s);
            npoints = scannum(s);
        }
        else if (ciprefix("dimensions:", buf)) {
            if (npoints == 0) {
                fprintf(cp_err, 
                    "Error: misplaced Dimensions: line\n");
                continue;
            }
            s = buf;
            skip(s);
            if (atodims(s, dims, &numdims)) { /* Something's wrong. */
                fprintf(cp_err, 
                    "Warning: syntax error in dimensions, ignored.\n");
                numdims = 0;
                continue;
            }
            if (numdims > MAXDIMS) {
                numdims = 0;
                continue;
            }
            /* Let's just make sure that the no. of points
             * and the dimensions are consistent.
             */
            for (j = 0, i = 1; j < numdims; j++) {
                i *= dims[j];
            }

            if (i != npoints) {
                fprintf(cp_err,
        "Warning: dimensions inconsistent with no. of points, ignored.\n");
                numdims = 0;
            }
        }
        else if (ciprefix("command:", buf)) {
            /* Note that we reverse these commands eventually... */
            s = buf;
            skip(s);
            nonl(s);
            if (curpl) {
                wl = alloc(struct wordlist);
                wl->wl_word = copy(s);
                wl->wl_next = curpl->pl_commands;
                if (curpl->pl_commands)
                    curpl->pl_commands->wl_prev = wl;
                curpl->pl_commands = wl;
            }
            else
                fprintf(cp_err, 
                    "Error: misplaced Command: line\n");
            /* Now execute the command if we can. */
            (void) cp_evloop(s);
        }
        else if (ciprefix("option:", buf)) {
            s = buf;
            skip(s);
            nonl(s);
            if (curpl) {
                wl = cp_lexer(s);
                for (vv = curpl->pl_env; vv && vv->va_next;
                        vv = vv->va_next)
                    ;
                if (vv)
                    vv->va_next = cp_setparse(wl);
                else
                    curpl->pl_env = cp_setparse(wl);
                wl_free(wl);
            }
            else
                fprintf(cp_err, 
                    "Error: misplaced Command: line\n");
        }
        else if (ciprefix("variables:", buf)) {
            /* We reverse the dvec list eventually... */
            if (!curpl) {
                fprintf(cp_err, "Error: no plot name given\n");
                plots = NULL;
                break;
            }
            s = buf;
            skip(s);
            if (!*s) {
                (void) fgets(buf, BSIZE_SP, fp);
                s = buf;
            }
            if (numdims == 0) {
                numdims = 1;
                dims[0] = npoints;
            }
            /* Now read all the variable lines in. */
            for (i = 0; i < nvars; i++) {
                v = alloc(struct dvec);
                v->v_next = curpl->pl_dvecs;
                curpl->pl_dvecs = v;
                if (!curpl->pl_scale)
                    curpl->pl_scale = v;
                v->v_flags = flags;
                v->v_plot = curpl;

                v->v_length = npoints;
                v->v_numdims = 0;
                /* Length and dims might be changed by options. */

                if (!i)
                    curpl->pl_scale = v;
                else {
                    (void) fgets(buf, BSIZE_SP, fp);
                    s = buf;
                }
                advtok(&s);  /* The index field. */
                if (t = gettok(&s))
                    v->v_name = t;
                else
                    fprintf(cp_err, 
                        "Error: bad var line %s\n", buf);
                t = gettok(&s); /* The type name. */
                if (t) {
                    v->v_type = ft_typnum(t);
                    txfree(t);
                }
                else
                    fprintf(cp_err, 
                        "Error: bad var line %s\n", buf);
                
                /* Fix the name... */
                if (isdigit(*v->v_name) &&
                        (r = ft_typabbrev(v->v_type))) {
                    (void) sprintf(buf2, "%s(%s)", r, v->v_name);
                    v->v_name = copy(buf2);
                }
                /* Now come the strange options... */
                while (copytok(buf2,&s)) {
                    if (ciprefix("min=", buf2)) {
                        if (sscanf(buf2 + 4, "%lf",&v->v_minsignal) != 1)
                            fprintf(cp_err, "Error: bad arg %s\n", buf2);
                        v->v_flags |= VF_MINGIVEN;
                    }
                    else if (ciprefix("max=", buf2)) {
                        if (sscanf(buf2 + 4, "%lf",&v->v_maxsignal) != 1)
                            fprintf(cp_err, "Error: bad arg %s\n", buf2);
                        v->v_flags |= VF_MAXGIVEN;
                    }
                    else if (ciprefix("color=", buf2)) {
                        v->v_defcolor = copy(buf2 + 6);
                    }
                    else if (ciprefix("scale=", buf2)) {
                        /* This is bad, but... */
                        v->v_scale = (struct dvec *) copy(buf2 + 6);
                    }
                    else if (ciprefix("grid=", buf2)) {
                        v->v_gridtype = (GRIDTYPE) scannum(buf2 + 5);
                    }
                    else if (ciprefix("plot=", buf2)) {
                        v->v_plottype = (PLOTTYPE) scannum(buf2 + 5);
                    }
                    else if (ciprefix("dims=", buf2)) {
                        fixdims(v, buf2 + 5);
                    }
                    else {
                        fprintf(cp_err,"Warning: bad var param %s\n", buf2);
                    }
                }

                /* Now we default any missing dimensions. */
                if (!v->v_numdims) {
                    v->v_numdims = numdims;
                    for (j = 0; j < numdims; j++)
                        v->v_dims[j] = dims[j];
                }
                /* And allocate the data array. We would use
                 * the desired vector length, but this would
                 * be dangerous if the file is invalid.
                 */

                if (isreal(v))
                    v->v_realdata =
                        (double *) tmalloc(npoints * sizeof (double));
                else
                    v->v_compdata =
                    (complex *) tmalloc(npoints * sizeof (complex));
            }
        }
        else if (ciprefix("values:", buf) || 
                ciprefix("binary:", buf)) {
            if (!curpl) {
                fprintf(cp_err, "Error: no plot name given\n");
                plots = NULL;
                break;
            }
            /* We'd better reverse the dvec list now... */
            for (v = curpl->pl_dvecs, curpl->pl_dvecs = NULL; v;
                    v = nv) {
                nv = v->v_next;
                v->v_next = curpl->pl_dvecs;
                curpl->pl_dvecs = v;
            }
            /* And fix the scale pointers. */
            for (v = curpl->pl_dvecs; v; v = v->v_next) {
                if (v->v_scale) {
                    for (nv = curpl->pl_dvecs; nv; nv = nv->v_next)
                        if (cieq((char *) v->v_scale, nv->v_name)) {
                            v->v_scale = nv;
                            break;
                        }
                    if (!nv) {
                        fprintf(cp_err,"Error: no such vector %s\n",
                            (char *) v->v_scale);
                        v->v_scale = NULL;
                    }
                }
            }
            for (i = 0; i < npoints; i++) {
                if ((*buf == 'v') || (*buf == 'V')) {
                    /* It's an ASCII file. */
                    (void) fscanf(fp, " %d", &j);
                    for (v = curpl->pl_dvecs; v; v = v->v_next) {
                        if (i < v->v_length) {
                            if (flags & VF_REAL) {
                                if (fscanf(fp, " %lf",
                                    &v->v_realdata[i]) != 1)
                                    fprintf(cp_err, errbr);
                            }
                            else {
                                if (fscanf(fp, " %lf, %lf",
                                    &realpart(&v->v_compdata[i]),
                                    &imagpart(&v->v_compdata[i])) != 2)
                                    fprintf(cp_err, errbr);
                            }
                        }
                        else if (raw_padded) {
                            if (flags & VF_REAL) {
                                if (fscanf(fp, " %lf", &junk) != 1)
                                    fprintf(cp_err, errbr);
                            }
                            else {
                                if (fscanf(fp, " %lf, %lf",
                                    &junk, &junk) != 2)
                                    fprintf(cp_err, errbr);
                            }
                        }
                    }
                }
                else {
                    /* It's a Binary file. */
                    for (v = curpl->pl_dvecs; v; v = v->v_next) {
                        if (i < v->v_length) {
                            if (flags & VF_REAL) {
                                if (fread((char *) &v->v_realdata[i],
                                    sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                            }
                            else {
                                if (fread((char *) &v->v_compdata[i].
                                    cx_real, sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                                if (fread((char *) &v->v_compdata[i].
                                    cx_imag, sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                            }
                        }
                        else if (raw_padded) {
                            if (flags & VF_REAL) {
                                if (fread((char *) &junk,
                                    sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                            }
                            else {
                                if (fread((char *) &junk,
                                    sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                                if (fread((char *) &junk,
                                    sizeof (double), 1, fp) != 1)
                                    fprintf(cp_err, errbr);
                            }
                        }
                    }
                }
            }
        }
        else {
            s = buf;
            skip(s);
            if (*s) {
                fprintf(cp_err, 
            "Error: strange line in rawfile -- load aborted\n");
                return (NULL);
            }
        }
    }

    if (curpl) {    /* reverse commands list */
        for (wl=curpl->pl_commands,
                curpl->pl_commands=NULL; wl &&
                wl->wl_next; wl=nwl) {
            nwl = wl->wl_next;
            wl->wl_next = curpl->pl_commands;
            curpl->pl_commands = wl;
        }
    }

    /* Fix everything up nicely again. */
    cp_popcontrol();
    cp_curin = lastin;
    cp_curout = lastout;
    cp_curerr = lasterr;
    (void) fclose(fp);

    /* make the vectors permanent */
    for (curpl = plots; curpl; curpl = curpl->pl_next) {
        curpl->pl_hashtab = htab_init();
        for (v = curpl->pl_dvecs; v; v = nv) {
            nv = v->v_next;
            v->v_next = NULL;
            htab_add(v->v_name,(void*)v,curpl->pl_hashtab);
        }
        curpl->pl_dvecs = NULL;
    }
    return (plots);
}


static void
fixdims(v, s)

/* s is a string of the form d1,d2,d3... */
struct dvec *v;
char *s;
{
    if (atodims(s, v->v_dims, &(v->v_numdims)))
        /* Something's wrong. */
        fprintf(cp_err, 
            "Warning: syntax error in dimensions, ignored.\n");
    return;
}


static char *
dimstring(data, length, buf)

/*
 * Create a string of the form "12,1,10".
 */
int *data;
int length;
char *buf;
{
    int i;

    buf[0] = '\0';

    if (!data || length < 1)
        return (buf);

    for (i = 0; i < length; i++) {
        sprintf(buf + strlen(buf), "%d%s",  data[i],
            (i < length - 1) ? "," : "");
    }
    return (buf);
}


static int
atodims( p, data, outlength )

/*
 * Read a string of one of the following forms into a dimensions array:
 *  [12][1][10]
 *  [12,1,10]
 *  12,1,10
 *  12, 1, 10
 *  12 , 1 , 10
 *  Basically, we require that all brackets be matched, that all numbers
 *  be separated by commas or by "][", that all whitespace is ignored, and
 *  the beginning [ and end ] are ignored if they exist.  The only valid
 *  characters in the string are brackets, commas, spaces, and digits.
 *  If any dimension is blank, its entry in the array is set to 0.
 *  
 *  Return 0 on success, 1 on failure.
 */
char *p;
int *data;
int *outlength;
{
    int length = 0;
    int state = 0;
    int err = 0;
    int needbracket = 0;
    char sep = '\0';

    if (!data || !outlength)
        return (1);

    if (!p) {
        *outlength = 0;
        return (0);
    }

    while (*p && isspace(*p))
        p++;

    if (*p == '[') {
        p++;
        while (*p && isspace(*p))
            p++;
        needbracket = 1;
    }

    while (*p && state != 3) {
        switch (state) {
            case 0: /* p just at or before a number */
                if (length >= MAXDIMS) {
                    if (length == MAXDIMS)
                        printf("Error: maximum of %d dimensions allowed.\n",
                            MAXDIMS);
                    length += 1;
                }
                else if (!isdigit(*p)) {
                    data[length++] = 0;    /* This position was empty. */
                }
                else {
                    data[length++] = atoi(p);
                    while (isdigit(*p))
                        p++;
                }
                state = 1;
                break;

            case 1: /* p after a number, looking for ',' or ']' */
                if (sep == '\0') {
                    sep = *p;
                }
                if (*p == ']' && *p == sep) {
                    p++;
                    state = 2;
                }
                else if (*p == ',' && *p == sep) {
                    p++;
                    state = 0;
                }
                else  /* Funny character following a # */
                    break;
                break;

            case 2: /* p after a ']', either at the end or looking for '[' */
                if (*p == '[') {
                    p++;
                    state = 0;
                }
                else {
                    state = 3;
                }
                break;
        }

        while (*p && isspace(*p))
            p++;
    }

    *outlength = length;
    if (length > MAXDIMS) {
        return (1);
    }
    if (state == 3) {  /* We finished with a closing bracket */
        err = !needbracket;
    }
    else if (*p) {   /* We finished by hitting a bad character after a # */
        err = 1;
    }
    else {           /* We finished by exhausting the string */
        err = needbracket;
    }
    if (err) {
        *outlength = 0;
    }
    return (err);
}
