/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Routines for dealing with the vector database.
 */

#include "spice.h"
#include "ftedefs.h"

static double boltz;
static double c;
static double e;
static double echarge;
static complex ii;
static double kelvin;
static double pi;
static double planck;

#ifdef __STDC__
static void vecfree(struct dvec*);
static struct dvec *findvec(char*,struct plot*);
static int veccmp(struct dvec**,struct dvec**);
static int namecmp(char*,char*);
#else
static void vecfree();
static struct dvec *findvec();
static int veccmp();
static int namecmp();
#endif

/* Where 'constants' go when defined on initialization. */

static struct plot constantplot = {
    "Constant values", "Sat Aug 16 10:55:15 PDT 1986", "constants",
    "constants", NULL, NULL, NULL, NULL, NULL, NULL, NULL, true
} ;

struct plot *plot_cur = &constantplot;
struct plot *plot_list = &constantplot;


void
ft_loadfile(file)

/* Load in a rawfile. */
char *file;
{
    struct plot *pl, *np, *pp;

    out_printf("Loading raw data file (\"%s\") . . . ", file);
        pl = raw_read(file);
    if (pl)
        out_printf("done.\n");
    else
        out_printf("no data read.\n");

    /* This is a minor annoyance -- we should reverse the plot list so
     * they get numbered in the correct order.
     */
    for (pp = pl, pl = NULL; pp; pp = np) {
        np = pp->pl_next;
        pp->pl_next = pl;
        pl = pp;
    }
    for (; pl; pl = np) {
        np = pl->pl_next;
        plot_add(pl);
        /* Don't want to get too many "plot not written" messages. */
        pl->pl_written = true;
    }
}


void
plot_add(pl)

struct plot *pl;
{
    struct dvec *v;
    struct plot *tp;
    char *s, buf[BSIZE_SP];
    wordlist *wl, *wl0;
    int plot_num, n;

    out_printf("Title:  %s\nName: %s\nDate: %s\n\n", pl->pl_title,
        pl->pl_name, pl->pl_date);
    
    if (plot_cur)
        plot_cur->pl_ccom = cp_kwswitch(CT_VECTOR, pl->pl_ccom);
    
    wl = wl0 = (wordlist*)htab_wl(pl->pl_hashtab);
    while (wl) {
        cp_addkword(CT_VECTOR, wl->wl_word);
        wl = wl->wl_next;
    }
    wl_free(wl0);
    cp_addkword(CT_VECTOR, "all");
    plot_num = 1;
    if (!(s = ft_plotabbrev(pl->pl_name)))
        s = "unknown";
    for (tp = plot_list; tp; tp = tp->pl_next) {
        if (ciprefix(s,tp->pl_typename)) {
            n = atoi(tp->pl_typename + strlen(s));
            if (n >= plot_num)
                plot_num = n+1;
        }
    }
    (void) sprintf(buf, "%s%d", s, plot_num);
    pl->pl_typename = copy(buf);
    plot_new(pl);
    cp_addkword(CT_PLOT, buf);
    pl->pl_ccom = cp_kwswitch(CT_VECTOR, (char *) NULL);
    plot_setcur(pl->pl_typename);
}


void
plot_new(pl)

/* Add a plot to the plot list. This is different from plot_add() in that
 * all this does is update the list and the variable $plots.
 */
struct plot *pl;
{
    pl->pl_next = plot_list;
    plot_list = pl;

    return;
}


void
plot_delete(pl)

struct plot *pl;
{
    struct dvec *v, *nv;
    struct plot *op;
    wordlist *wl0, *wl;

    if (eq(pl->pl_typename, "constants")) {
        fprintf(cp_err, "Error: can't destroy the constant plot\n");
        return;
    }

    /* clear temporary dvecs */
    for (v = pl->pl_dvecs; v; v = nv) {
        nv = v->v_next;
        vecfree(v);
    }

    /* clear permanent dvecs */
    wl0 = (wordlist*)htab_list(pl->pl_hashtab);
    for (wl = wl0; wl; wl = wl->wl_next) {
        v = (struct dvec *)((char**)wl->wl_word)[1];
        vecfree(v);
    }
    wl_free(wl0);
    htab_free(pl->pl_hashtab,false);
    txfree((char*)pl->pl_hashtab);

    va_free(pl->pl_env);
    if (pl == plot_list) {
        plot_list = pl->pl_next;
        if (pl == plot_cur) {
            plot_cur = plot_list;
            pl->pl_ccom = cp_kwswitch(CT_VECTOR,plot_cur->pl_ccom);
        }
    }
    else {
        for (op = plot_list; op; op = op->pl_next)
            if (op->pl_next == pl)
                break;
        if (!op) {
            fprintf(cp_err,
                "Internal Error: kill plot -- not in list\n");
            goto xxx;
        }
        op->pl_next = pl->pl_next;
        if (pl == plot_cur) {
            plot_cur = op;
            pl->pl_ccom = cp_kwswitch(CT_VECTOR,plot_cur->pl_ccom);
        }
    }
xxx:
    cp_ccfreetrie(pl->pl_ccom);
    txfree(pl->pl_title);
    txfree(pl->pl_date);
    txfree(pl->pl_name);
    txfree(pl->pl_typename);
    wl_free(pl->pl_commands);
    txfree((char*)pl);

    return;
}


void
plot_setcur(name)

/* Make a plot the current one.  This gets called by cp_usrset() when one
 * does a 'set curplot = name'.
 */
char *name;
{
    struct plot *pl;

    if (cieq(name, "new")) {
        pl = plot_alloc("unknown");
        pl->pl_title = copy("Anonymous");
        pl->pl_name = copy("unknown");
        pl->pl_date = copy(datestring( ));
        plot_new(pl);
        plot_cur = pl;
        return;
    }
    for (pl = plot_list; pl; pl = pl->pl_next)
        if (plot_prefix(name, pl->pl_typename))
            break;
    if (!pl) {
        fprintf(cp_err, "Error: no such plot named %s\n", name);
        return;
    }
    if (plot_cur)
        plot_cur->pl_ccom = cp_kwswitch(CT_VECTOR, pl->pl_ccom);
    plot_cur = pl;
    return;
}


void
plot_docoms(wl)

/* Execute the commands for a plot. This is done whenever a plot becomes
 * the current plot.
 */
wordlist *wl;
{
    bool inter;

    inter = cp_interactive;
    cp_interactive = false;
    cp_pushcontrol();
    while (wl) {
        (void) cp_evloop(wl->wl_word);
        wl = wl->wl_next;
    }
    cp_popcontrol();
    cp_interactive = inter;
    return;
}


struct plot *
plot_alloc(name)

/* Create a new plot structure. This just fills in the typename and sets up
 * the ccom struct.
 */
char *name;
{
    struct plot *pl = alloc(struct plot), *tp;
    char *s;
    char buf[BSIZE_SP];
    int plot_num, n;

    if (!(s = ft_plotabbrev(name)))
        s = "unknown";
    plot_num = 1;
    for (tp = plot_list; tp; tp = tp->pl_next) {
        if (ciprefix(s,tp->pl_typename)) {
            n = atoi(tp->pl_typename + strlen(s));
            if (n >= plot_num)
                plot_num = n+1;
        }
    }
    (void) sprintf(buf, "%s%d", s, plot_num);
    pl->pl_typename = copy(buf);
    cp_addkword(CT_PLOT, buf);
    s = cp_kwswitch(CT_VECTOR, (char *) NULL);
    cp_addkword(CT_VECTOR, "all");
    pl->pl_ccom = cp_kwswitch(CT_VECTOR, s);
    return (pl);
}


bool
plot_prefix(pre, str)

/* This function will match "op" with "op1", but not "op1" with "op12". */
char *pre, *str;
{
    if (!*pre)
        return (true);
    while (*pre && *str) {
        if (*pre != *str)
            break;
        pre++; str++;
    }
    if (*pre || (*str && isdigit(pre[-1])))
        return (false);
    else
        return (true);
}


void
vec_newtemp(d)

struct dvec *d;
{
#ifdef FTEDEBUG
    if (ft_vecdb)
        fprintf(cp_err, "new temporary vector %s\n", d->v_name);
#endif
    d->v_flags &= ~VF_PERMANENT;
    if (!d->v_plot)
        d->v_plot = plot_cur;
    d->v_next = d->v_plot->pl_dvecs;
    d->v_plot->pl_dvecs = d;
    return;
}


void
vec_newperm(d)

struct dvec *d;
{
#ifdef FTEDEBUG
    if (ft_vecdb)
        fprintf(cp_err, "new permanent vector %s\n", d->v_name);
#endif
    d->v_flags |= VF_PERMANENT;
    if (plot_cur->pl_scale == NULL)
        plot_cur->pl_scale = d;
    if (!d->v_plot)
        d->v_plot = plot_cur;
    if (plot_cur->pl_hashtab == NULL)
        plot_cur->pl_hashtab = htab_init();
    htab_add(d->v_name,(void*)d,plot_cur->pl_hashtab);
    return;
}


void
vec_remove(name)

/* Remove a vector from the current plot, if it is there. */
char *name;
{
    struct dvec *v;

    v = (struct dvec *)htab_get(name,plot_cur->pl_hashtab);
    if (!v)
        return;
    htab_delete(name,plot_cur->pl_hashtab);
    if (v->v_plot && v->v_plot->pl_scale == v)
        v->v_plot->pl_scale = NULL;
    vecfree(v);

    /* Remove from the keyword list. */
    cp_remkword(CT_VECTOR, name);
    return;
}


static void
vecfree(v)

struct dvec *v;
{
    txfree(v->v_name);
    vec_dlfree(v->v_link2);
    if (isreal(v)) {
        txfree((char*)v->v_realdata);
    }
    else {
        txfree((char*)v->v_compdata);
    }
    txfree((char*)v);
}


void
vec_dlfree(dl)

struct dvlist *dl;
{
    struct dvlist *dn;

    for (; dl; dl = dn) {
        dn = dl->dl_next;
        txfree((char*)dl);
    }
}


struct dvec *
vec_fromplot(word, plot)

/* Get a vector by name. This deals with v(1), etc. almost properly. Also,
 * it checks for pre-defined vectors.
 */
char *word;
struct plot *plot;
{
    struct dvec *d;
    char buf[BSIZE_SP], buf2[BSIZE_SP], cc, *s;

    d = findvec(word, plot);

    /* scanf("%c(%s)" doesn't do what it should do. ) */
    if (!d && (sscanf(word, "%c(%s", /* ) */ &cc, buf) == 2) &&
        /* ( */ ((s = strrchr(buf, ')')) != NULL) &&
            (*(s + 1) == '\0')) {
        *s = '\0';
        if (prefix("i(", /* ) */ word)) {
            /* Spice dependency... */
            (void) sprintf(buf2, "%s#branch", buf);
            (void) strcpy(buf, buf2);
        }
        d = findvec(buf, plot);
    }
    return (d);
}


#define SPECCHAR '@'


struct dvec *
vec_get(word)

/* This is the main lookup routine for names. The possible types of names are:
 *  name           An ordinary vector.
 *  plot.name      A vector from a particular plot.
 *  @device[parm]  A device parameter.
 *  @model[parm]   A model parameter.
 *  @param         A circuit parameter.
 * For the @ cases, we construct a dvec with length 1 to hold the value.
 * In the other two cases, either the plot or the name can be "all", a
 * wildcard.
 * The vector name may have imbedded dots -- if the first component is a plot
 * name, it is considered the plot, otherwise the current plot is used.
 */
char *word;
{
    struct dvec *d;
    struct plot *pl;
    char buf[BSIZE_SP], *s, *wd, *name, *param;
    int i = 0;
    struct variable *vv;
    struct dvlist *dl0, *dl, *dll;

    wd = word = copy(word);     /* Gets mangled below... */

    if (strchr(word, '.')) {
        /* Snag the plot... */
        for (i = 0, s = word; *s != '.'; i++, s++)
            buf[i] = *s;
        buf[i] = '\0';
        if (cieq(buf, "all")) {
            word = ++s;
            pl = NULL;  /* NULL pl signifies a wildcard. */
        }
        else {
            for (pl = plot_list; pl && !plot_prefix(buf, 
                    pl->pl_typename); pl = pl->pl_next)
                ;
            if (pl) {
                word = ++s;
            }
            else {
                /* This used to be an error... */
                pl = plot_cur;
            }
        }
    }
    else
        pl = plot_cur;

    if (pl) {
        d = vec_fromplot(word, pl);
        if (!d)
            d = vec_fromplot(word, &constantplot);
    }
    else {
        dl0 = NULL;
        for (pl = plot_list; pl; pl = pl->pl_next) {
            if (cieq(pl->pl_typename, "constants"))
                continue;
            d = vec_fromplot(word, pl);
            if (d) {
                if (dl0 == NULL)
                    dl0 = dl = alloc(struct dvlist);
                else {
                    dl->dl_next = alloc(struct dvlist);
                    dl = dl->dl_next;
                }
                if (d->v_link2) {
                    for (dll = d->v_link2; dll; dll = dll->dl_next) {
                        dl->dl_dvec = dll->dl_dvec;
                        if (dll->dl_next) {
                            dl->dl_next = alloc(struct dvlist);
                            dl = dl->dl_next;
                        }
                    }
                }
                else
                    dl->dl_dvec = d;
            }
        }
        if (!dl0) {
            fprintf(cp_err, 
            "Error: plot wildcard (name %s) matches nothing\n", word);
            txfree(wd);
            return (NULL);
        }
        d = alloc(struct dvec);
        vec_newtemp(d);
        d->v_link2 = dl0;
        d->v_name = copy("list");
        vec_sort(d);
        txfree(wd);
        return (d);
    }
    if (!d && (*word == SPECCHAR)) {
        /* This is a special quantity... */
        if (ft_nutmeg) {
            fprintf(cp_err,
        "Error: circuit parameters only available with spice\n");
            txfree(wd);
            return (false);
        }
        name = ++word;
        for (param = name; *param && (*param != '['); param++)
            ;
        if (*param) {
            *param++ = '\0';
            for (s = param; *s && *s != ']'; s++)
                ;
            *s = '\0';
        }
        else
            param = NULL;
        vv = (*if_getparam)(ft_curckt->ci_ckt, &name,
            param, 0,(wordlist**)NULL);
        if (!vv) {
            txfree(wd);
            return (NULL);
        }
        d = alloc(struct dvec);
        d->v_name = copy(word);
        d->v_type = SV_NOTYPE;
        d->v_flags |= VF_REAL;  /* No complex values yet... */
        d->v_realdata = (double *) tmalloc(sizeof (double));
        d->v_length = 1;
        *d->v_realdata = vv->va_real;
        vec_newtemp(d);
        va_free(vv);
        txfree(wd);
        return (d);
    }
    txfree(wd);
    return (d);
}


static struct dvec *
findvec(word, pl)

/* Find a named vector in a plot. */
char *word;
struct plot *pl;
{
    struct dvec *d;
    char buf[BSIZE_SP], *s;
    wordlist *wl, *wl0;
    struct dvlist *dl, *dl0;

    if (pl == NULL)
        return (NULL);

    if (word && *word) {

        if (cieq(word, "all")) {

            /* create a dummy vector with link2 pointing at dvlist */
            wl0 = (wordlist*)htab_list(pl->pl_hashtab);
            if (wl0 == NULL)
                return (NULL);
            dl0 = NULL;
            for (wl = wl0; wl; wl = wl->wl_next) {
                d = (struct dvec*)((char**)wl->wl_word)[1];
                if (!dl0)
                    dl = dl0 = alloc(struct dvlist);
                else {
                    dl->dl_next = alloc(struct dvlist);
                    dl = dl->dl_next;
                }
                dl->dl_dvec = d;
            }
            if (dl0) {
                d = alloc(struct dvec);
                vec_newtemp(d);
                d->v_name = copy("list");
                d->v_link2 = dl0;
                wl_free(wl0);
                return (d);
            }
            return (NULL);
        }

        d = (struct dvec *)htab_get(word,pl->pl_hashtab);
        if (!d) {
            buf[0] = 'v';
            buf[1] = '(';
            s = buf+2;
            while (*word)
                *s++ = *word++;
            *s++ = ')';
            *s = '\0';
            d = (struct dvec *)htab_get(buf,pl->pl_hashtab);
        }
        return (d);
    }
    return (NULL);
}


/* Create a copy of a vector. */

struct dvec *
vec_copy(v)

struct dvec *v;
{
    struct dvec *nv;
    int i;
    
    if (!v)
        return (NULL);
    nv = alloc(struct dvec);
    nv->v_name = copy(v->v_name);
    nv->v_type = v->v_type;
    nv->v_flags = v->v_flags;
    nv->v_length = v->v_length;
    nv->v_plot = v->v_plot;
    nv->v_minsignal = v->v_minsignal;
    nv->v_maxsignal = v->v_maxsignal;
    nv->v_defcolor = v->v_defcolor;
    nv->v_gridtype = v->v_gridtype;
    nv->v_plottype = v->v_plottype;
    nv->v_scale = v->v_scale;
    nv->v_numdims = v->v_numdims;
    for (i = 0; i < v->v_numdims; i++)
        nv->v_dims[i] = v->v_dims[i];
    if (v->v_length) {
        if (isreal(v)) {
            nv->v_realdata =
                (double *) tmalloc(sizeof(double) * v->v_length);
            DCOPY(v->v_realdata, nv->v_realdata, v->v_length);
        }
        else {
            nv->v_compdata =
                (complex *) tmalloc(sizeof(complex) * v->v_length);
            CCOPY(v->v_compdata, nv->v_compdata, v->v_length);
        }
    }
    return (nv);
}


void
vec_gc()

/* Clear out the temporary vectors linked into the plot */
{
    struct plot *pl;
    struct dvec *d, *nd;

    for (pl = plot_list; pl; pl = pl->pl_next) {
        for (d = pl->pl_dvecs; d; d = nd) {
            nd = d->v_next;

            if (ft_vecdb)
                fprintf(cp_err,"vec_gc: throwing away %s.%s\n",
                    pl->pl_typename, d->v_name);
            if (pl->pl_scale == d) /* this shouldn't happen */
                pl->pl_scale = NULL;
            vecfree(d);
        }
        pl->pl_dvecs = NULL;
    }
}


bool
vec_eq(v1, v2)

/* This is something we do in a few places...  Since vectors get copied a
 * lot, we can't just compare pointers to tell if two vectors are 'really'
 * the same.
 */
struct dvec *v1, *v2;
{
    char *s1, *s2;
    bool i;

    if (v1->v_plot != v2->v_plot)
        return (false);
    
    s1 = vec_basename(v1);
    s2 = vec_basename(v2);

    i = cieq(s1, s2);
    txfree(s1);
    txfree(s2);

    return (i);
}


char *
vec_basename(v)

/* Return the name of the vector with the plot prefix stripped off.  This
 * is no longer trivial since '.' doesn't always mean 'plot prefix'.
 */
struct dvec *v;
{
    char buf[BSIZE_SP], *t, *s;
    int i;

    if (strchr(v->v_name, '.')) {
        for (t = v->v_name, i = 0; *t && *t != '.'; t++)
            buf[i++] = *t;
        buf[i] = '\0';
        if (cieq(v->v_plot->pl_typename, buf))
            (void) strcpy(buf, t + 1);
        else
            (void) strcpy(buf, v->v_name);
    }
    else
        (void) strcpy(buf, v->v_name);
    
    for (t = buf; isspace(*t); t++)
        ;
    s = t;
    for (t = s; *t; t++)
        ;
    while ((t > s) && isspace(t[-1]))
        *--t = '\0';
    return (copy(s));
}


void
vec_sort(d)

/* Sort all the vectors in d, first by plot name and then by vector name.
 * Do the right thing with numbers.
 */
struct dvec *d;
{
    struct dvec **array;
    struct dvlist *dl;
    int i;

    for (i = 0,dl = d->v_link2; dl; i++,dl = dl->dl_next) ;
    if (i < 2)
        return;
    array = (struct dvec **) tmalloc(i * sizeof (struct dvec *));
    for (i = 0,dl = d->v_link2; dl; i++,dl = dl->dl_next)
        array[i] = dl->dl_dvec;
    
    qsort((char *) array, i, sizeof (struct dvec *),
#ifdef __STDC__
        (int(*)(const void*,const void*))veccmp);
#else
        veccmp);
#endif

    /* Now string everything back together... */
    for (i = 0,dl = d->v_link2; dl; i++,dl = dl->dl_next)
        dl->dl_dvec = array[i];
    txfree((char*)array);
    return;
}


static int
veccmp(d1, d2)

struct dvec **d1, **d2;
{
    int i;

    if ((i = namecmp((*d1)->v_plot->pl_typename,
            (*d2)->v_plot->pl_typename)) != 0)
        return (i);
    return (namecmp((*d1)->v_name, (*d2)->v_name));
}


static int
namecmp(s, t)

/* If there are imbedded numeric strings, compare them numerically, not
 * alphabetically.
 */
char *s, *t;
{
    int i, j;

    for (;;) {
        while ((*s == *t) && !isdigit(*s) && *s)
            s++, t++;
        if (!*s)
            return (0);
        if ((*s != *t) && (!isdigit(*s) || !isdigit(*t)))
            return (*s - *t);
        
        /* The beginning of a number... Grab the two numbers 
         * and then compare them...
         */
        for (i = 0; isdigit(*s); s++)
            i = i * 10 + *s - '0';
        for (j = 0; isdigit(*t); t++)
            j = j * 10 + *t - '0';
        
        if (i != j)
            return (i - j);
    }
}


struct dvec *
vec_mkfamily(v)

/* This routine takes a multi-dimensional vector and turns it into a family
 * of 1-d vectors, linked together with v_link2.  It is here so that plot
 * can do intelligent things.
 */
struct dvec *v;
{
    int size, numvecs, i, j, count[MAXDIMS];
    struct dvec *vecs, *d;
    struct dvlist *dl, *dl0;
    char buf[BSIZE_SP];

    if (v->v_numdims < 2)
        return (v);

    size = v->v_dims[v->v_numdims - 1];
    numvecs = v->v_length/size;

    for (i = 0, dl = dl0 = NULL; i < numvecs; i++) {
        if (!dl0)
            dl0 = dl = alloc(struct dvlist);
        else {
            dl->dl_next = alloc(struct dvlist);
            dl = dl->dl_next;
        }
        dl->dl_dvec = alloc(struct dvec);
        vec_newtemp(dl->dl_dvec);
    }
    for (i = 0; i < MAXDIMS; i++)
        count[i] = 0;
    for (dl = dl0, j = 0; j < numvecs; j++, dl = dl->dl_next) {
        d = dl->dl_dvec;
        (void) strcpy(buf, v->v_name);
        for (i = 0; i < v->v_numdims - 1; i++)
            (void) sprintf(buf + strlen(buf), "[%d]", count[i]);
        d->v_name = copy(buf);
        d->v_type = v->v_type;
        d->v_flags = v->v_flags;
        d->v_minsignal = v->v_minsignal;
        d->v_maxsignal = v->v_maxsignal;
        d->v_gridtype = v->v_gridtype;
        d->v_plottype = v->v_plottype;
        d->v_scale = v->v_scale;
        /* Don't copy the default color, since there will be many
         * of these things...
         */
        d->v_numdims = 1;
        d->v_length = size;

        if (isreal(v)) {
            d->v_realdata =
                (double *) tmalloc(sizeof(double) * size);
            DCOPY(v->v_realdata + size * j, d->v_realdata, size);
        }
        else {
            d->v_compdata =
                (complex *) tmalloc(sizeof(complex) * size);
            CCOPY(v->v_compdata + size * j, d->v_compdata, size);
        }

        for (i = v->v_numdims - 2; i >= 0; i--) {
            if (count[i]++ < v->v_dims[i])
                break;
            else
                count[i] = 0;
        }
        if (i < 0)
            break;
    }
    if (dl0) {
        vecs = alloc(struct dvec);
        vec_newtemp(vecs);
        vecs->v_name = copy("list");
        vecs->v_link2 = dl0;
        return (vecs);
    }
    return (NULL);
}

