/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include "inpdefs.h"
#include "iferrmsg.h"
#include "misc.h"

static IFvalue *values = NULL;
static int *types;
static int numvalues;
static GENERIC *circuit;
static INPtables *tables;

#ifdef __STDC__
static void PTalias(char**);
static INPparseNode *PTdifferentiate(INPparseNode*,int);
static INPparseNode *copy_node(INPparseNode*);
static INPparseNode *mkcon(double);
static INPparseNode *mkb(int,INPparseNode*,INPparseNode*);
static INPparseNode *mkf(int,INPparseNode*);
static int PTcheck(INPparseNode*);
static INPparseNode *PTparse(char**);
static INPparseNode *makepnode(PTelement*);
static INPparseNode *mkbnode(int,INPparseNode*,INPparseNode*);
static INPparseNode *mkfnode(char*,INPparseNode*);
static INPparseNode *mknnode(double);
static INPparseNode *mksnode(char*);
static PTelement *PTlexer(char**);
static char *get_stringvar(char**,char*);
static int tran_hack(char*,char*);
static int is_tranfunc(char*,INPparseNode*);
static INPparseNode *allocate_pnode(int);
static int pt_const(INPparseNode*,double*,double*,double*);
static int pt_var(INPparseNode*,double*,double*,double*);
static int pt_parm(INPparseNode*,double*,double*,double*);
static int pt_fcn(INPparseNode*,double*,double*,double*);
static int pt_table(INPparseNode*,double*,double*,double*);
static int pt_tran(INPparseNode*,double*,double*,double*);
static int pt_op(INPparseNode*,double*,double*,double*);
#else
static void PTalias();
static INPparseNode *PTdifferentiate();
static INPparseNode *copy_node();
static INPparseNode *mkcon();
static INPparseNode *mkb();
static INPparseNode *mkf();
static int PTcheck();
static INPparseNode *PTparse();
static INPparseNode *makepnode();
static INPparseNode *mkbnode();
static INPparseNode *mkfnode();
static INPparseNode *mknnode();
static INPparseNode *mksnode();
static PTelement *PTlexer();
static char *get_stringvar();
static int tran_hack();
static int is_tranfunc();
static INPparseNode *allocate_pnode();
static int pt_const();
static int pt_var();
static int pt_parm();
static int pt_fcn();
static int pt_table();
static int pt_tran();
static int pt_op();
#endif

extern IFsimulator *ft_sim; /* XXX */

#define P_ALLOC 0
#define P_CLEAR 1
#define P_RESET 2


/* These are all the constants any sane person needs. */

static struct constant {
    char *name;
    double value;
} constants[] = {
    { "e",  M_E },
    { "pi", M_PI },
    { NULL, 0 }
} ;

#define issep(c) ((c)==' '||(c)=='\t'||(c)=='='||(c)=='('||(c)==')'||\
(c)==','||(c)=='-'||(c)=='+'||(c)=='*'||(c)=='/'||(c)=='^')


void
INPgetTree(line, pt, ckt, tab)

/* Parse the expression in *line as far as possible, and return the parse
 * tree obtained.  If there is an error, *pt will be set to NULL and an error
 * message will be printed.
 */
char **line;
INPparseTree **pt;
GENERIC *ckt;
INPtables *tab;
{
    INPparseNode *p;
    int i;
    char *tl, *tl0;

    values = NULL;
    types = NULL;
    numvalues = 0;

    circuit = ckt;
    tables = tab;

    tl = *line;
    PTalias(&tl);
    tl0 = tl;
    p = PTparse(&tl);
    *line += strlen(*line) - strlen(tl);
    *--tl = '\0';

    if (!p || !PTcheck(p)) {
        *pt = NULL;
        (void)allocate_pnode(P_CLEAR);
        txfree(tl0);
        return;
    }

    (*pt) = alloc(INPparseTree);

    (*pt)->p.numVars = numvalues;
    (*pt)->p.varTypes = types;
    (*pt)->p.vars = values;
    (*pt)->p.IFinit = IFinit;
    (*pt)->p.IFeval = IFeval;
    (*pt)->p.IFfree = IFfree;
    (*pt)->p.line = tl0;
    (*pt)->tree = p;

    if (numvalues) {
        (*pt)->derivs = (INPparseNode **) 
            tmalloc(numvalues * sizeof (INPparseNode *));

        for (i = 0; i < numvalues; i++)
            (*pt)->derivs[i] = copy_node(PTdifferentiate(p, i));
    }

/*
INPptPrint("tree:",*pt);
*/
    (void)allocate_pnode(P_RESET);
    return;
}


static void
PTalias(line)

/* remove any aliasing in line */
char **line;
{
    char *s, *buf, *tok, c, cc;
    char *newline, *nl, *tl;
    int len, olen, count = 0;
    IFvalue *v;

    tl = *line;
    newline = copy(tl);
    nl = newline;
    len = strlen(newline);

    while (*tl) {

        while (*tl && issep(*tl)) {
            if ((olen = (nl - newline)) >= len) {
                len += 16;
                newline = trealloc(newline,len+1);
                nl = newline + olen;
            }
            *nl++ = *tl++;
        }

        if (!*tl)
            break;

        tok = tl;

        while (*tl && !issep(*tl))
            tl++;

        c = *tl;
        *tl = '\0';

        buf = NULL;

        /* is it a reference to a branch? */
        if ((s = strchr(tok,'#')) != NULL && !strcmp(s+1,"branch")) {
            cc = *s;
            *s = '\0';
            buf = tmalloc(strlen(tok) + 4);
            sprintf(buf,"i(%s)",tok);
            *s = cc;
        }

        /* is it 'x' ? */
        if (*tok == 'x' && !*(tok+1) && ft_sim->ptXalias) {
            buf = copy(ft_sim->ptXalias);
        }

        if (buf)
            tok = buf;

        while (*tok) {
            if ((olen = (nl - newline)) >= len) {
                len += 16;
                newline = trealloc(newline,len);
                nl = newline + olen;
            }
            *nl++ = *tok++;
        }

        if (buf)
            txfree(buf);

        *tl = c;
    }
    *nl = '\0';
    *line = newline;
}


static INPparseNode *
PTdifferentiate(p, varnum)

/* This routine takes the partial derivative of the parse tree with respect
 * to the i'th variable.  We try to do optimizations like getting rid of
 * 0-valued terms.
 */

INPparseNode *p;
int varnum;
{
    INPparseNode *arg1, *arg2, *newp;

/* printf("differentiating: "); printTree(p); printf(" wrt var %d\n", varnum);*/

    switch (p->type) {

        case PT_CONSTANT:
        case PT_PARAM:
        case PT_TFUNC:
            newp = mkcon((double) 0);
            break;

        case PT_VAR:
            /* Is this the variable we're differentiating wrt? */
            if (p->valueIndex == varnum)
                newp = mkcon((double) 1);
            else
                newp = mkcon((double) 0);
            break;

        case PT_PLUS:
        case PT_MINUS:
            arg1 = PTdifferentiate(p->left, varnum);
            arg2 = PTdifferentiate(p->right, varnum);
            if (arg1->type == PT_CONSTANT && arg1->constant == 0) {
                txfree((char*)arg1);
                if (p->type == PT_PLUS)
                    newp = arg2;
                else
                    newp = mkf(PTF_UMINUS,arg2);
            }
            else if (arg2->type == PT_CONSTANT && arg2->constant == 0) {
                txfree((char*)arg2);
                newp = arg1;
            }
            else
                newp = mkb(p->type, arg1, arg2);
            break;

        case PT_TIMES:
            /* d(a * b) = d(a) * b + d(b) * a */
            arg1 = PTdifferentiate(p->left, varnum);
            arg2 = PTdifferentiate(p->right, varnum);
            if (arg1->type == PT_CONSTANT && arg1->constant == 0) {
                txfree((char*)arg1);
                newp = mkb(PT_TIMES,p->left,arg2);
            }
            else if (arg2->type == PT_CONSTANT && arg2->constant == 0) {
                txfree((char*)arg2);
                newp = mkb(PT_TIMES,arg1,p->right);
            }
            else
                newp = mkb(PT_PLUS, mkb(PT_TIMES, arg1, p->right),
                        mkb(PT_TIMES, p->left, arg2));
            break;

        case PT_DIVIDE:
            /* d(a / b) = (d(a) * b - d(b) * a) / b^2 */
            arg1 = PTdifferentiate(p->left, varnum);
            arg2 = PTdifferentiate(p->right, varnum);
            if (arg1->type == PT_CONSTANT && arg1->constant == 0) {
                txfree((char*)arg1);
                newp = mkb(PT_DIVIDE, mkf(PTF_UMINUS,
                    mkb(PT_TIMES, p->left, arg2)),
                    mkb(PT_POWER, p->right, mkcon((double) 2)));
            }
            else if (arg2->type == PT_CONSTANT && arg2->constant == 0) {
                txfree((char*)arg2);
                newp = mkb(PT_DIVIDE, arg1, p->right);
            }
            else
                newp = mkb(PT_DIVIDE, mkb(PT_MINUS, mkb(PT_TIMES, arg1,
                    p->right), mkb(PT_TIMES, p->left, arg2)),
                    mkb(PT_POWER, p->right, mkcon((double) 2)));
            break;

        case PT_POWER:
            /* Two cases... If the power is a constant then we're cool.
             * Otherwise we have to be tricky.
             */

            if (p->right->type == PT_CONSTANT) {

                arg1 = PTdifferentiate(p->left, varnum);
                newp = mkb(PT_TIMES, mkb(PT_TIMES,
                        mkcon(p->right->constant),
                        mkb(PT_POWER, p->left,
                        mkcon(p->right->constant - 1))),
                        arg1);
            }
            else {
                /* This is complicated.  f(x) ^ g(x) ->
                 * exp(y(x) * ln(f(x)) ...
                 */
                arg1 = PTdifferentiate(p->left, varnum);
                arg2 = PTdifferentiate(p->right, varnum);
                newp = mkb(PT_TIMES, mkf(PTF_EXP, mkb(PT_TIMES,
                        p->right, mkf(PTF_LN, p->left))),
                        mkb(PT_PLUS, mkb(PT_TIMES, p->right,
                        mkb(PT_DIVIDE, arg1, p->left)),
                        mkb(PT_TIMES, arg2,
                        mkf(PTF_LN, arg1))));

            }
            break;

        case PT_FUNCTION:
            /* Many cases.  Set arg1 to the derivative of the function,
             * and arg2 to the derivative of the argument.
             */
            arg2 = PTdifferentiate(p->left, varnum);
            if (arg2->type == PT_CONSTANT && arg2->constant == 0) {
                newp = arg2;
                break;
            }

            switch (p->funcnum) {

                case PTF_ABS:  /* sgn(u) */
                    arg1 = mkf(PTF_SGN, p->left);
                    break;

                case PTF_SGN:  /* u -> a hack */
                    arg1 = p->left;
                    break;

                case PTF_ACOS:  /* - 1 / sqrt(1 - u^2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) -1), mkf(PTF_SQRT,
                            mkb(PT_MINUS, mkcon((double) 1),
                            mkb(PT_POWER, p->left, mkcon((double)
                            2)))));
                    break;

                case PTF_ACOSH: /* 1 / sqrt(u^2 - 1) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkf(PTF_SQRT,
                            mkb(PT_MINUS, mkb(PT_POWER, p->left,
                            mkcon((double) 2)),
                            mkcon((double) 1))));

                    break;

                case PTF_ASIN:  /* 1 / sqrt(1 - u^2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkf(PTF_SQRT,
                            mkb(PT_MINUS, mkcon((double) 1),
                            mkb(PT_POWER, p->left, mkcon((double)
                            2)))));
                    break;

                case PTF_ASINH: /* 1 / sqrt(u^2 + 1) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkf(PTF_SQRT,
                            mkb(PT_PLUS, mkb(PT_POWER, p->left,
                            mkcon((double) 2)),
                            mkcon((double) 1))));
                    break;

                case PTF_ATAN:  /* 1 / (1 + u^2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkb(PT_PLUS,
                            mkb(PT_POWER, p->left, mkcon((double)
                            2)), mkcon((double) 1)));
                    break;

                case PTF_ATANH: /* 1 / (1 - u^2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkb(PT_MINUS,
                            mkcon((double) 1), mkb(PT_POWER,
                            p->left, mkcon((double) 2))));
                    break;

                case PTF_COS:   /* - sin(u) */
                    arg1 = mkf(PTF_UMINUS, mkf(PTF_SIN, p->left));
                    break;

                case PTF_COSH:  /* sinh(u) */
                    arg1 = mkf(PTF_SINH, p->left);
                    break;

                case PTF_EXP:   /* exp(u) */
                    arg1 = mkf(PTF_EXP, p->left);
                    break;

                case PTF_LN:    /* 1 / u */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), p->left);
                    break;

                case PTF_LOG:   /* log(e) / u */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) M_LOG10E), p->left);
                    break;

                case PTF_SIN:   /* cos(u) */
                    arg1 = mkf(PTF_COS, p->left);
                    break;

                case PTF_SINH:  /* cosh(u) */
                    arg1 = mkf(PTF_COSH, p->left);
                    break;

                case PTF_SQRT:  /* 1 / (2 * sqrt(u)) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkb(PT_TIMES,
                            mkcon((double) 2), mkf(PTF_SQRT,
                            p->left)));
                    break;

                case PTF_TAN:   /* 1 / (cos(u) ^ 2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkb(PT_POWER,
                            mkf(PTF_COS, p->left), mkcon((double)
                            2)));
                    break;

                case PTF_TANH:  /* 1 / (cosh(u) ^ 2) */
                    arg1 = mkb(PT_DIVIDE, mkcon((double) 1), mkb(PT_POWER,
                            mkf(PTF_COSH, p->left), mkcon((double)
                            2)));
                    break;

                case PTF_UMINUS:    /* - 1 ; like a constant (was 0 !) */
                    arg1 = mkcon((double) - 1.0);
                    break;

                default:
                    fprintf(stderr, "Internal Error: bad function # %d\n",
                            p->funcnum);
                    newp = NULL;
                    break;
            }
            newp = mkb(PT_TIMES, arg1, arg2);
            break;

        case PT_TABLE:
            arg2 = PTdifferentiate(p->left, varnum);
            if (arg2->type == PT_CONSTANT && arg2->constant == 0) {
                newp = arg2;
                break;
            }
            arg1 = allocate_pnode(P_ALLOC);
            arg1->new_deriv = 1;
            arg1->type = PT_TABLE;
            arg1->left = p->left;
            arg1->funcnum = 1; /* return derivative if 1 */
            arg1->function = p->function;
            arg1->funcname = p->funcname;
            arg1->evfunc = pt_table;
            newp = mkb(PT_TIMES, arg1, arg2);
            break;

        default:
            fprintf(stderr, "Internal error: bad node type %d\n", p->type);
            newp = NULL;
            break;
    }

/* printf("result is: "); printTree(newp); printf("\n"); */
    return (newp);
}


static INPparseNode *
copy_node(p)
INPparseNode *p;
{
    INPparseNode *new;
    int num;

    if (p == NULL)
        return (NULL);

    if (p->new_deriv) {
        p->new_deriv = 0;
        p->left = copy_node(p->left);
        p->right = copy_node(p->right);
        return (p);
    }

    new = allocate_pnode(P_ALLOC);

    *new = *p;
    if (new->type == PT_TFUNC) {
        new->trancoeffs =
            (double*)tmalloc(new->numtrancoeffs*sizeof(double));
        memcpy(new->trancoeffs,p->trancoeffs,new->numtrancoeffs*sizeof(double));
        if (new->funcnum == PTF_tPULSE) {
            new->trancache = (double*)tmalloc(2*sizeof(double));
        }
        else if (new->funcnum == PTF_tPWL) {
            num = new->numtrancoeffs/2 - 1;
            new->trancache = (double*)tmalloc(num*sizeof(double));
            memcpy(new->trancache,p->trancache,num*sizeof(double));
        }
    }
    new->left = copy_node(p->left);
    new->right = copy_node(p->right);
    return (new);
}


static INPparseNode *
mkcon(value)

double value;
{
    INPparseNode *p = allocate_pnode(P_ALLOC);

    p->type = PT_CONSTANT;
    p->constant = value;
    p->evfunc = pt_const;
    p->new_deriv = 1;

    return (p);
}


static INPparseNode *
mkb(type, left, right)

int type;
INPparseNode *left, *right;
{
    INPparseNode *p;
    int i;

    if ((right->type == PT_CONSTANT) && (left->type == PT_CONSTANT)) {
        switch (type) {
            case PT_TIMES:
                return (mkcon(left->constant * right->constant));

            case PT_DIVIDE:
                return (mkcon(left->constant / right->constant));

            case PT_PLUS:
                return (mkcon(left->constant + right->constant));

            case PT_MINUS:
                return (mkcon(left->constant - right->constant));

            case PT_POWER:
                return (mkcon(pow(left->constant, right->constant)));
        }
    }
    switch (type) {
        case PT_TIMES:
            if ((left->type == PT_CONSTANT) && (left->constant == 0))
                return (left);
            else if ((right->type == PT_CONSTANT) && (right->constant == 0))
                return (right);
            else if ((left->type == PT_CONSTANT) && (left->constant == 1))
                return (right);
            else if ((right->type == PT_CONSTANT) && (right->constant == 1))
                return (left);
            break;
    
        case PT_DIVIDE:
            if ((left->type == PT_CONSTANT) && (left->constant == 0))
                return (left);
            else if ((right->type == PT_CONSTANT) && (right->constant == 1))
                return (left);
            break;

        case PT_PLUS:
            if ((left->type == PT_CONSTANT) && (left->constant == 0))
                return (right);
            else if ((right->type == PT_CONSTANT) && (right->constant == 0))
                return (left);
            break;

        case PT_MINUS:
            if ((right->type == PT_CONSTANT) && (right->constant == 0))
                return (left);
            else if ((left->type == PT_CONSTANT) && (left->constant == 0))
                return (mkf(PTF_UMINUS, right));
            break;
    
        case PT_POWER:
            if (right->type == PT_CONSTANT) {
                if (right->constant == 0)
                    return (mkcon(1.0));
                else if (right->constant == 1)
                    return (left);
            }
            break;
    }

    for (i = 0; PTops[i].number != PT_PLACEHOLDER; i++)
        if (PTops[i].number == type)
            break;
    if (PTops[i].number == PT_PLACEHOLDER) {
        fprintf(stderr, "Internal Error: bad type %d\n", type);
        return (NULL);
    }
    p = allocate_pnode(P_ALLOC);
    p->new_deriv = 1;
    p->type = type;
    p->left = left;
    p->right = right;
    p->function = PTops[i].funcptr;
    p->funcname = PTops[i].name;
    p->evfunc = pt_op;

    return (p);
}


static INPparseNode *
mkf(type, arg)

int type;
INPparseNode *arg;
{
    INPparseNode *p;
    int i;
    double constval;

    for (i = 0; PTfuncs[i].name; i++)
        if (PTfuncs[i].number == type)
            break;
    if (!PTfuncs[i].name) {
        fprintf(stderr, "Internal Error: bad type %d\n", type);
        return (NULL);
    }

    if (arg->type == PT_CONSTANT) {
        constval = ((*PTfuncs[i].funcptr) (arg->constant));
        return (mkcon(constval));
    }
    p = allocate_pnode(P_ALLOC);
    p->new_deriv = 1;
    p->type = PT_FUNCTION;
    p->left = arg;
    p->funcnum = i;
    p->function = PTfuncs[i].funcptr;
    p->funcname = PTfuncs[i].name;
    p->evfunc = pt_fcn;

    return (p);
}


static int
PTcheck(p)

/* Check for remaining PT_PLACEHOLDERs in the parse tree.
 * Returns 1 if ok.
 */
INPparseNode *p;
{
    switch (p->type) {
        case PT_PLACEHOLDER:
            return (0);

        case PT_CONSTANT:
        case PT_VAR:
        case PT_PARAM:
        case PT_TFUNC:
            return (1);

        case PT_FUNCTION:
            return (PTcheck(p->left));

        case PT_PLUS:
        case PT_MINUS:
        case PT_TIMES:
        case PT_DIVIDE:
        case PT_POWER:
            return (PTcheck(p->left) && PTcheck(p->right));
    
        default:
            fprintf(stderr, "Internal error: bad node type %d\n", p->type);
            return (0);
    }
}


/* The operator-precedence table for the parser. */

#define G 1 /* Greater than. */
#define L 2 /* Less than. */
#define E 3 /* Equal. */
#define R 4 /* Error. */

static char prectable[11][11] = {
       /* $  +  -  *  /  ^  u- (  )  v  , */
/* $ */ { R, L, L, L, L, L, L, L, R, L, R },
/* + */ { G, G, G, L, L, L, L, L, G, L, G },
/* - */ { G, G, G, L, L, L, L, L, G, L, G },
/* * */ { G, G, G, G, G, L, L, L, G, L, G },
/* / */ { G, G, G, G, G, L, L, L, G, L, G },
/* ^ */ { G, G, G, G, G, L, L, L, G, L, G },
/* u-*/ { G, G, G, G, G, G, G, L, G, L, R },
/* ( */ { R, L, L, L, L, L, L, L, E, L, L },
/* ) */ { G, G, G, G, G, G, G, R, G, R, G },
/* v */ { G, G, G, G, G, G, G, G, G, R, G },
/* , */ { G, L, L, L, L, L, L, L, G, L, G }

} ;


static INPparseNode *
PTparse(line)

/* Return an expr. */
char **line;
{
    PTelement stack[PT_STACKSIZE];
    int sp = 0, st, i;
    PTelement *top, *next;
    INPparseNode *pn, *lpn, *rpn;

    stack[0].token = TOK_END;
    next = PTlexer(line);

    while ((sp > 1) || (next->token != TOK_END)) {
        /* Find the top-most terminal. */
        i = sp;
        do {
            top = &stack[i--];
        } while (top->token == TOK_VALUE);


        switch (prectable[top->token][next->token]) {
        case L:
        case E:
            /* Push the token read. */
            if (sp == (PT_STACKSIZE - 1)) {
                errMsg = copy("stack overflow");
                return (NULL);
            }
            bcopy((char *) next, (char *) &stack[++sp],
                    sizeof (PTelement));
            next = PTlexer(line);
            continue;

        case R:
            return (NULL);

        case G:
            /* Reduce. Make st and sp point to the elts on the
             * stack at the end and beginning of the junk to
             * reduce, then try and do some stuff. When scanning
             * back for a <, ignore VALUES.
             */
            st = sp;
            if (stack[sp].token == TOK_VALUE)
                sp--;
            while (sp > 0) {
                if (stack[sp - 1].token == TOK_VALUE)
                    i = 2;  /* No 2 pnodes together... */
                else
                    i = 1;
                if (prectable[stack[sp - i].token]
                         [stack[sp].token] == L)
                    break;
                else
                    sp = sp - i;
            }
            if (stack[sp - 1].token == TOK_VALUE)
                sp--;
            /* Now try and see what we can make of this.
             * The possibilities are: - node
             *            node op node
             *            ( node )
             *            func ( node )
             *            func ( node, node, node, ... )    <- new
             *            node
             */
            if (st == sp) {
                pn = makepnode(&stack[st]);
                if (pn == NULL)
                    goto err;
            }
            else if ((stack[sp].token == TOK_UMINUS) &&
                    (st == sp + 1)) {
                lpn = makepnode(&stack[st]);
                if (lpn == NULL)
                        goto err;
                pn = mkfnode(copy("-"), lpn);
            }
            else if ((stack[sp].token == TOK_LPAREN) &&
                       (stack[st].token == TOK_RPAREN)) {
                pn = makepnode(&stack[sp + 1]);
                if (pn == NULL)
                    goto err;
            }
            else if ((stack[sp + 1].token == TOK_LPAREN) &&
                       (stack[st].token == TOK_RPAREN)) {
                lpn = makepnode(&stack[sp + 2]);
                if ((lpn == NULL) || (stack[sp].type != TYP_STRING))
                    goto err;
                if (!(pn = mkfnode(stack[sp].value.string, lpn)))
                    return (NULL);
            }
            else { /* node op node */
                lpn = makepnode(&stack[sp]);
                rpn = makepnode(&stack[st]);
                if ((lpn == NULL) || (rpn == NULL))
                    goto err;
                pn = mkbnode(stack[sp + 1].token, lpn, rpn);
            }
            stack[sp].token = TOK_VALUE;
            stack[sp].type = TYP_PNODE;
            stack[sp].value.pnode = pn;
            continue;
        }
    }
    pn = makepnode(&stack[1]);
    if (pn)
        return (pn);
err:
    return (NULL);
}


static INPparseNode *
makepnode(elem)

/* Given a pointer to an element, make a pnode out of it (if it already
 * is one, return a pointer to it). If it isn't of type VALUE, then return
 * NULL.
 */
PTelement *elem;
{
    if (elem->token != TOK_VALUE)
        return (NULL);

    switch (elem->type) {
        case TYP_STRING:
            return (mksnode(elem->value.string));

        case TYP_NUM:
            return (mknnode(elem->value.real));

        case TYP_PNODE:
            return (elem->value.pnode);

        default:
            fprintf(stderr, "Internal Error: bad token type\n");
            return (NULL);
    }   
}


static INPparseNode *
mkbnode(opnum, arg1, arg2)

/* Binop node. */
int opnum;
INPparseNode *arg1, *arg2;
{
    INPparseNode *p;
    int i;

    for (i = 0; PTops[i].name != PT_PLACEHOLDER; i++)
        if (PTops[i].number == opnum)
            break;

    if (PTops[i].number == PT_PLACEHOLDER) {
        fprintf(stderr, "Internal Error: no such op num %d\n", opnum);
        return (NULL);
    }
    p = allocate_pnode(P_ALLOC);

    p->type = opnum;
    p->funcname = PTops[i].name;
    p->function = PTops[i].funcptr;
    p->evfunc = pt_op;
    p->left = arg1;
    p->right = arg2;

    return (p);
}


static INPparseNode *
mkfnode(fname, arg)

/* fname freed */
char *fname;
INPparseNode *arg;
{
    int i;
    INPparseNode *p;
    char *name, *s;
    IFvalue temp;
    GENERIC *table;

    p = allocate_pnode(P_ALLOC);

    if (!strcmp(fname, "v")) {
        name = tmalloc(128);
        if (arg->type == PT_PLACEHOLDER) {
            strcpy(name, arg->funcname);
            txfree(arg->funcname);
        }
        else if (arg->type == PT_CONSTANT) {
            (void) sprintf(name, "%d", (int) arg->constant);
        }
        else if (arg->type != PT_COMMA) {
            errMsg = copy("badly formed node voltage");
            txfree(fname);
            return (NULL);
        }

        if (arg->type == PT_COMMA) {
            /* Change v(a,b) into v(a) - v(b) */
            p = mkb(PT_MINUS, mkfnode(copy(fname), arg->left),
            mkfnode(copy(fname), arg->right));
        }
        else {
            /* printf("getting a node called '%s'\n", name); */
            INPtermInsert(circuit, &name, tables, &(temp.nValue));
            for (i = 0; i < numvalues; i++)
                if ((types[i] == IF_NODE) && (values[i].nValue ==
                        temp.nValue))
                    break;
                if (i == numvalues) {
                    if (numvalues) {
                        values = (IFvalue *) 
                            trealloc((char *) values,
                                (numvalues + 1) * sizeof (IFvalue));
                        types = (int *) 
                            trealloc((char *) types,
                                (numvalues + 1) * sizeof (int));
                    }
                    else {
                        values = alloc(IFvalue);
                        types = (int *) tmalloc(sizeof (int));
                }
                values[i] = temp;
                types[i] = IF_NODE;
                numvalues++;
            }
            p->valueIndex = i;
            p->type = PT_VAR;
            p->evfunc = pt_var;
        }
    }
    else if (!strcmp(fname, "i")) {
        name = tmalloc(128);
        if (arg->type == PT_PLACEHOLDER) {
            strcpy(name, arg->funcname);
            txfree(arg->funcname);
        }
        else if (arg->type == PT_CONSTANT) {
            (void) sprintf(name, "%d", (int) arg->constant);
        }
        else {
            errMsg = copy("badly formed branch current");
            txfree(fname);
            return (NULL);
        }
/* printf("getting a device called '%s'\n", name); */
        INPinsert(&name, tables);
        for (i = 0; i < numvalues; i++)
            if ((types[i] == IF_INSTANCE) && (values[i].uValue ==
                    temp.uValue))
                break;
        if (i == numvalues) {
            if (numvalues) {
                values = (IFvalue *) 
                    trealloc((char *) values,
                        (numvalues + 1) * sizeof (IFvalue));
                types = (int *) 
                    trealloc((char *) types,
                        (numvalues + 1) * sizeof (int));
            }
            else {
                values = alloc(IFvalue);
                types = (int *) tmalloc(sizeof (int));
            }
            values[i].uValue = (IFuid) name;
            types[i] = IF_INSTANCE;
            numvalues++;
        }
        p->valueIndex = i;
        p->type = PT_VAR;
        p->evfunc = pt_var;
    }
    else {
        for (i = 0; PTfuncs[i].name; i++)
            if (!strcmp(PTfuncs[i].name, fname))
            break;
    
        if (!PTfuncs[i].name) {
            if (INPtablFind(fname,circuit,&table)) {
                errMsg = tmalloc(strlen(fname) + 100);
                sprintf(errMsg,"no such function '%s'",fname);
                txfree(fname);
                return (NULL);
            }
            p->type = PT_TABLE;
            p->left = arg;
            p->funcname = fname;
            p->funcnum = 0;
            p->function = (double(*)()) table;
            p->evfunc = pt_table;
            return (p);
        }

        p->type = PT_FUNCTION;
        p->left = arg;
        p->funcname = PTfuncs[i].name;
        p->funcnum = PTfuncs[i].number;
        p->function = PTfuncs[i].funcptr;
        p->evfunc = pt_fcn;
    }

    txfree(fname);
    return (p);
}


static INPparseNode *
mknnode(number)

/* Number node. */
double number;
{
    struct INPparseNode *p;

    p = allocate_pnode(P_ALLOC);

    p->type = PT_CONSTANT;
    p->constant = number;
    p->evfunc = pt_const;

    return (p);
}


static INPparseNode *
mksnode(string)

/* String node, string is freed */
char *string;
{
    int i, j;
    
    INPparseNode *p;

    p = allocate_pnode(P_ALLOC);

    /* First check it it is really a tran function */
    if (is_tranfunc(string,p)) {
        txfree(string);
        return (p);
    }

    /* Next see if it's a special parameter */
    for (i = 0; i < ft_sim->numSpecSigs; i++)
        if (!strcmp(ft_sim->specSigs[i], string))
            break;
    if (i < ft_sim->numSpecSigs) {
        p->valueIndex = i;
        p->type = PT_PARAM;
        p->evfunc = pt_parm;
        txfree(string);
        return (p);
    }

    for (i = 0; constants[i].name; i++)
        if (!strcmp(constants[i].name, string))
            break;
    
    if (!constants[i].name) {
        /* We'd better save this in case it's part of i(something). */
        p->type = PT_PLACEHOLDER;
        p->funcname = string;
    }
    else {
        p->type = PT_CONSTANT;
        p->constant = constants[i].value;
        p->evfunc = pt_const;
        txfree(string);
    }

    return (p);
}



static PTelement *
PTlexer(line)

/* The lexical analysis routine. */
char **line;
{
    double td;
    int err;
    static PTelement el;
    static char *specials = " \t()^+-*/,";
    static int lasttoken = TOK_END, lasttype;
    char *sbuf, *s;

    sbuf = *line;
/*
printf("entering lexer, sbuf = '%s', lastoken = %d, lasttype = %d\n", 
sbuf, lasttoken, lasttype);
*/
    while ((*sbuf == ' ') || (*sbuf == '\t') || (*sbuf == '='))
        sbuf++;

    switch (*sbuf) {
        case '\0':
            el.token = TOK_END;
            break;

        case ',':
            el.token = TOK_COMMA;
            sbuf++;
            break;

        case '-':
            if ((lasttoken == TOK_VALUE) || (lasttoken == TOK_RPAREN))
                el.token = TOK_MINUS;
            else
                el.token = TOK_UMINUS;
            sbuf++;
            break;

        case '+':
            el.token = TOK_PLUS; 
            sbuf++;
            break;

        case '*':
            el.token = TOK_TIMES; 
            sbuf++;
            break;

        case '/':
            el.token = TOK_DIVIDE; 
            sbuf++;
            break;

        case '^':
            el.token = TOK_POWER; 
            sbuf++;
            break;

        case '(':
            if (((lasttoken == TOK_VALUE) && ((lasttype == TYP_NUM))) ||
                    (lasttoken == TOK_RPAREN)) {
                el.token = TOK_END;
            }
            else {
                el.token = TOK_LPAREN; 
                sbuf++;
            }
            break;

        case ')':
            el.token = TOK_RPAREN; 
            sbuf++;
            break;
    
        default:
            if ((lasttoken == TOK_VALUE) || (lasttoken == TOK_RPAREN)) {
                el.token = TOK_END;
                break;
            }

            td = INPevaluate(&sbuf, &err, 0);
            if (err == OK) {
                while (isalpha(*sbuf))
                    sbuf++;
                el.token = TOK_VALUE;
                el.type = TYP_NUM;
                el.value.real = td;
            }
            else {
                el.token = TOK_VALUE;
                el.type = TYP_STRING;
                el.value.string = get_stringvar(&sbuf,specials);
            }
    }

    lasttoken = el.token;
    lasttype = el.type;

    *line = sbuf;

/*
printf("PTlexer: token = %d, type = %d, left = '%s'\n", 
el.token, el.type, sbuf);
*/

    return (&el);
}


static char *
get_stringvar(s,specials)

char **s, *specials;
{
    char *t;
    char *buf;
    int i, j;

    for (t = *s, i = 0; *t; t++, i++) {
        if (strchr(specials,*t))
            break;
    }
    buf = tmalloc(i+1);
    memcpy(buf,*s,i);
    buf[i] = '\0';


    for (j = 0; PTtFuncs[j].name; j++)
        if (!strcmp(PTtFuncs[j].name, buf))
            break;
    
    if (tran_hack(PTtFuncs[j].name,t)) {
        /* found a match, copy parameters and () */
        for ( ; *t; t++, i++) {
            if (*t == ')') {
                t++;
                i++;
                break;
            }
        }
        txfree(buf);
        buf = tmalloc(i+1);
        memcpy(buf,*s,i);
        buf[i] = '\0';
    }
    *s = t;

    /* set to lower case */
    for (t = buf; *t; t++)
        if (isupper(*t)) *t = tolower(*t);

    return (buf);
}


static int
tran_hack(name,str)

char *name,*str;
/* since tranfuncs sin, exp have the same name as math functions, we
 * figure out from the context whether to interpret as a tran func
 * or not.
 */
{
    char *s, *t, *tok;
    char *buf;
    int n, error;

    if (name == NULL)
        return (false);

    if (strcmp(name,"sin") && strcmp(name,"exp"))
        return (true);

    for (n = 0, s = str; *s != ')'; n++,s++) ;

    buf = tmalloc(n+1);
    memcpy(buf,str,n);
    buf[n] = '\0';

    n = 0;
    t = buf;
    while (*t) {
        error = INPgetTok(&t,&tok,1);
        if (error) break;
        s = tok;
        (void)INPevaluate(&tok,&error,1);
        if (error) {
            txfree(s);
            break;
        }
        txfree(s);
        n++;
    }
    txfree(buf);
    if (!error && n >= 2)
        return (true);
    return (false);
}


static int
is_tranfunc(s,p)

char *s;
INPparseNode *p;
{
    int i, num, error;
    char *t, *tt;
    double tmp, *list;

    /* the tran function string includes parameters enclosed in () */
    if (!strchr(s,')'))
        return (false); /* can't be a tran function */

    t = strchr(s,'(');
    if (t == NULL)
        return (false); /* syntax error */

    *t = '\0';
    t++;

    /* get rid of intermediate space */
    tt = s;
    while (*tt && !isspace(*tt)) tt++;
    *tt = '\0';

    for (i = 0; PTtFuncs[i].name; i++)
        if (!strcmp(PTtFuncs[i].name, s))
        break;
    
    if (!PTtFuncs[i].name) {
        fprintf(stderr, "Internal Error: no such tran function '%s'", s);
        return (false);
    }
    if (p) {
        p->type = PT_TFUNC;

        num = 0;
        list = (double *)tmalloc(sizeof(double));
        tmp = INPevaluate(&t,&error,1);
        while (error == 0) {
            num++;
            list = (double *)
                trealloc((char *)list,num*sizeof(double));
            *(list+num-1) = tmp;
            tmp = INPevaluate(&t,&error,1);
        }
        p->trancoeffs = list;
        p->numtrancoeffs = num;
        p->funcname = PTtFuncs[i].name;
        p->funcnum = PTtFuncs[i].number;
        p->function = PTtFuncs[i].funcptr;
        p->evfunc = pt_tran;
        if (p->funcnum != PTF_tPWL)
            p->tranparms = (double*)tmalloc(8*sizeof(double));

        if (p->funcnum == PTF_tPULSE)
            p->trancache = (double*)tmalloc(2*sizeof(double));
        else if (p->funcnum == PTF_tPWL) {
            num = p->numtrancoeffs/2 - 1;
            if (num) {
                p->trancache = (double*)tmalloc(num*sizeof(double));
                for (i = 0; i < num; i++) {
                    p->trancache[i] =
                        (*(p->trancoeffs+2*i+3) - *(p->trancoeffs+2*i+1))/
                        (*(p->trancoeffs+2*i+2) - *(p->trancoeffs+2*i));
                }
            }
            else {
                p->trancache = (double*)tmalloc(sizeof(double));
                p->trancache[0] = 0;
            }
        }
    }
    return (true);
}


/* memory management */

static INPparseNode *
allocate_pnode(mode)

int mode;
{
    static nodes, size;
    static INPparseNode **pnodes;
    int i;

    if (mode == P_ALLOC) {
        nodes++;
        if (nodes > size) {
            size += 10;
            if (pnodes)
                pnodes = (INPparseNode**)
                    trealloc((char*)pnodes,size * sizeof(INPparseNode*));
            else
                pnodes = (INPparseNode**)
                    tmalloc(size * sizeof(INPparseNode*));
        }
        pnodes[nodes-1] = alloc(INPparseNode);
        return (pnodes[nodes-1]);
    }
    if (mode == P_CLEAR) {
        for (i = 0; i < nodes; i++) {
            if (pnodes[i]->type == PT_PLACEHOLDER) continue;
            if (pnodes[i]->type == PT_TFUNC) {
                if (pnodes[i]->trancoeffs)
                    txfree((char*)pnodes[i]->trancoeffs);
                if (pnodes[i]->tranparms)
                    txfree((char*)pnodes[i]->tranparms);
                if (pnodes[i]->trancache)
                    txfree((char*)pnodes[i]->trancache);
            }
            txfree((char*)pnodes[i]);
        }
    }
    for (i = 0; i < nodes; i++) {
        if (pnodes[i]->type != PT_PLACEHOLDER) continue;
        txfree((char*)pnodes[i]);
    }
    txfree((char*)pnodes);
    pnodes = NULL;
    size = 0;
    nodes = 0;
    return (NULL);
}


/* evaluation functions */

static int
pt_const(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    *res = tree->constant;
    return (OK);
}


static int
pt_var(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    *res = vals[tree->valueIndex];
    return (OK);
}


static int
pt_parm(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    *res = parms[tree->valueIndex];
    return (OK);
}


static int
pt_fcn(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    int err;
    double r1;

    err = (*tree->left->evfunc)(tree->left, &r1, vals, parms);
    if (err != OK)
        return (err);
    *res = (*tree->function)(r1);
    /*
    if (*res == HUGE) {
        fprintf(stderr, "Error: %lg out of range for %s\n",
                r1, tree->funcname);
        return (E_PARMVAL);
    }
    */
    return (OK);
}


static int
pt_table(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    int err;
    double r1;

    err = (*tree->left->evfunc)(tree->left, &r1, vals, parms);
    if (err != OK)
        return (err);

    if (tree->funcnum == 0)
        *res = INPtablEval((GENERIC*)tree->function,r1);
    else
        *res = INPtablEvalDeriv((GENERIC*)tree->function,r1);

    return (OK);
}


static int
pt_tran(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    *res = (*tree->function)(tree,parms);
    return (OK);
}


static int
pt_op(tree, res, vals, parms)

INPparseNode *tree;
double *res;
double *vals;
double *parms;
{
    int err;
    double r1, r2;

    err = (*tree->left->evfunc)(tree->left, &r1, vals, parms);
    if (err != OK)
        return (err);
    err = (*tree->right->evfunc)(tree->right, &r2, vals, parms);
    if (err != OK)
        return (err);
    *res = (*tree->function)(r1, r2);
    /*
    if (*res == HUGE) {
        fprintf(stderr, "Error: %lg, %lg out of range for %s\n",
                r1, r2, tree->funcname);
        return (E_PARMVAL);
    }
    */
    return (OK);
}


/* Debugging stuff. */

#ifdef __STDC__
static void printTree(INPparseNode*);
#else /* stdc */
static void printTree();
#endif /* stdc */


void
INPptPrint(str, ptree)

char *str;
IFparseTree *ptree;
{
    int i;

    printf("%s\n\t", str);
    printTree(((INPparseTree *) ptree)->tree);
    printf("\n");
    for (i = 0; i < ptree->numVars; i++) {
        printf("d / d v%d : ", i);
        printTree(((INPparseTree *) ptree)->derivs[i]);
        printf("\n");
    }
    return;
}


static void
printTree(pt)

INPparseNode *pt;
{
    int i;

    switch (pt->type) {
        case PT_CONSTANT:
            printf("%g", pt->constant);
            break;

        case PT_VAR:
            printf("v%d", pt->valueIndex);
            break;

        case PT_PARAM:
            printf("%s", ft_sim->specSigs[pt->valueIndex]);
            break;

        case PT_PLUS:
            printf("(");
            printTree(pt->left);
            printf(") + (");
            printTree(pt->right);
            printf(")");
            break;

        case PT_MINUS:
            printf("(");
            printTree(pt->left);
            printf(") - (");
            printTree(pt->right);
            printf(")");
            break;

        case PT_TIMES:
            printf("(");
            printTree(pt->left);
            printf(") * (");
            printTree(pt->right);
            printf(")");
            break;

        case PT_DIVIDE:
            printf("(");
            printTree(pt->left);
            printf(") / (");
            printTree(pt->right);
            printf(")");
            break;

        case PT_POWER:
            printf("(");
            printTree(pt->left);
            printf(") ^ (");
            printTree(pt->right);
            printf(")");
            break;

        case PT_FUNCTION:
            printf("%s (", pt->funcname);
            printTree(pt->left);
            printf(")");
            break;

        case PT_TFUNC:
            printf("%s (", pt->funcname);
            for (i = 0; i < pt->numtrancoeffs; i++)
                printf(" %g",pt->trancoeffs[i]);
            printf(" )");
            break;

        default:
            printf("oops");
            break;
    }
    return;
}

