/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1992 Stephen R. Whiteley
****************************************************************************/

/* 
 * Routines for parsing and checking data tables input by the user.
 * Note: these tables are completely different from those manipulated
 * with the INPsymTab routines.
 */

#include "spice.h"
#include <stdio.h>
#include "util.h"
#include "cpdefs.h"
#include "inpdefs.h"
#include "tabdefs.h"
#include "iferrmsg.h"

/*
#define debug
*/

#ifdef __STDC__
static void tablfix(struct sCKTtable *);
static void testprnt(GENERIC*);
static int  tabl_add(GENERIC*,struct sCKTtable*);
#else
static void tablfix();
static void testprnt();
static int  tabl_add();
#endif

char *
INPtablParse(line,ckt)

/* Parse a line containing a table description of the form
 * .table tabname [val1] {stuff1} val2 {stuff2} ... valN [{stuffN}]
 * where {stuffI} can be:
 *  a real number, or
 *  function function_string, or
 *  table table_name.
 * If the first val is omitted the first {stuff} must be a function or table.
 * the last {stuff} can be omitted.
 * The function_string must not contain but be surrounded by white space.
 */

char **line;     /* the line to parse */
GENERIC *ckt;    /* the circuit this device is a member of */
{
    char *t, *errbuf;
    int error;
    struct sCKTtable *table;
    struct sCKTentry *entry, *etmp;
    double ord, val;

    /* throw out .table */
    error = INPgetTok(line,&t,1);
    if (error) return(INPerror(error));
    txfree(t);

    if (*line == NULL) return NULL;

    error = INPgetTok(line,&t,1);
    if (error) return(INPerror(error));
    if (INPtablFind(t,ckt,(GENERIC **)NULL) == 0) {
        /* already there - error */
        errbuf = MALLOC(strlen(t) + 40);
        (void) sprintf(errbuf," Table %d already defined - ignored\n",t);
        txfree(t);
        return errbuf;
    }

    table = (struct sCKTtable *) MALLOC(sizeof(struct sCKTtable));
    if (table == NULL) return(INPerror(E_NOMEM));
    table->tab_name = t;

    while (**line != (char)0) {

        etmp = (struct sCKTentry *) MALLOC(sizeof(struct sCKTentry));
        if (etmp == NULL) return(INPerror(E_NOMEM));

        if (table->tab_entry == NULL) {
            entry = table->tab_entry = etmp;
            ord = INPevaluate(line,&error,1);
            if (error == 0)
                entry->ent_val = ord;
            else
                entry->ent_val = 0;
        }
        else {
            entry->ent_next = etmp;
            entry = entry->ent_next;
            ord = INPevaluate(line,&error,1);
            if (error == 0)
                entry->ent_val = ord;
            else {
                t = gettok(line);
                errbuf = MALLOC(strlen(t) + 40);
                (void) sprintf(errbuf,
                    " Syntax error: expected value, got %s\n",t);
                txfree(t);
                return errbuf;
            }
        }
        if (**line == (char)0) {
            /* last entry omitted */
            entry->ent_type = ENT_OMITTED;
            break;
        }

        val = INPevaluate(line,&error,1);
        if (error == 0) {
            /* entry is a real */
            entry->ent_type = ENT_NUMBER;
            entry->ent_real = val;
            continue;
        }
        error = INPgetTok(line,&t,1);
        if (error) return(INPerror(error));

        if (strcmp(t,"table") == 0) {
            /* table specified */
            txfree(t);
            entry->ent_type = ENT_TABLE;
            error = INPgetTok(line,&t,1);
            if (error) return(INPerror(error));
            entry->ent_string = t;
            continue;
        }
        /* uh-oh */
        errbuf = MALLOC(strlen(t) + 40);
        (void) sprintf(errbuf," Syntax error: unknown keyword %s\n",t);
        txfree(t);
        return errbuf;
    }
    tabl_add(ckt,table);
    return NULL;
}

/************************************************************************/


char *
INPtablCheck(line,ckt)

/* Check a .table line for sub-table existence.  Returns NULL except if
 * an undefined table reference is found.  A syntax error will return NULL
 * immediately, such errors are already flagged.  If an unknown table
 * is found, an error message is returned.
 * If no syntx errors, fix up the corresponding table structure so that
 * sub-table references are pointers to the sub-table, not to the name.
 */

char *line;     /* the .table line to check (".table" stripped) */
GENERIC *ckt;
{
    char *t, *errbuf = NULL;
    int error;
    struct sCKTtable *table, *tb, *tt;
    struct sCKTentry *e;

    if (line == NULL) return NULL;

    error = INPgetTok(&line,&t,1); /* table name */
    if (error) return(INPerror(error));

    (void)INPtablFind(NULL,ckt,(GENERIC**)&table);
    if (table == NULL) {
        /* no tables saved (syntax error) */
        txfree(t);
        return NULL;
    }

    for (tb = table; tb; tb = tb->tab_next)
        if (eq(t,tb->tab_name)) break;

    txfree(t);
    if (tb == NULL) /* can't find own symbol (syntax error) */
        return NULL;

    while (*line != (char)0) {

        (void) INPevaluate(&line,&error,1);
        if (*line == (char)0)  /* last entry was omitted */
            break;
        INPevaluate(&line,&error,1);
        if (error == 0) continue; /* entry is a real */
        error = INPgetTok(&line,&t,1);
        if (error) return(INPerror(error));
        if (strcmp(t,"table") == 0) {
            /* table specified */
            txfree(t);
            error = INPgetTok(&line,&t,1); /* table reference */
            if (error) return(INPerror(error));
            if (INPtablFind(t,ckt,(GENERIC **)NULL) != 0) {
                if (errbuf) {
                    char *temp = errbuf;
                    errbuf = MALLOC(strlen(temp) + strlen(t) + 40);
                    (void) sprintf(errbuf,
                        "%s Unresolved table reference: %s\n",temp,t);
                }
                else {
                    errbuf = MALLOC(strlen(t) + 40);
                    (void) sprintf(errbuf,
                        " Unresolved table reference: %s\n",t);
                }
                /* find the reference and null it */
                for (e = tb->tab_entry; e; e = e->ent_next) {
                    if (e->ent_type != ENT_TABLE) continue;
                    if (e->ent_string && eq(e->ent_string,t)) {
                        tfree(e->ent_string); /* tfree NULLs it */
                        break;
                    }
                }
            }
            txfree(t);
            continue;
        }
        return NULL;  /* syntax error */
    }
    /* no syntax errors */
    /* replace known table name references with pointers to tables */

    for (e = tb->tab_entry; e; e = e->ent_next) {
        if (e->ent_type == ENT_TABLE && e->ent_string != NULL) {
            for (tt = table; tt; tt = tt->tab_next)
                if (eq(e->ent_string,tt->tab_name)) {
                    /* replace with pointer to table */
                    txfree(e->ent_string);
                    e->ent_string = (char *)tt;
                    break;
                }
        }
    }
    return errbuf;
}

/************************************************************************/


void
INPtablFix(ckt)

/*
 * Fix the actual values for the omitted final entries, and
 * set the ent_real components to the function or table evaluated
 * with the ent_val component.
 */

GENERIC *ckt;
{
    struct sCKTtable *t;

    for (t = *DEVtableHead(ckt); t; t = t->tab_next)
        tablfix(t);

#ifdef debug
    testprnt(ckt);
#endif
}

/************************************************************************/


static void
tablfix(table)

/*
 * If the final entry argument is omitted, the final entry will be taken
 * as a number equal to the previous entry evaluated at the final value.
 * The ent_real component of each function or table entry is set to the
 * function or table evaluated at the entry value.
 */

struct sCKTtable *table;
{
    struct sCKTentry *e, *ee;

    if (table == NULL) return;

    for (e = table->tab_entry; e; e = ee) {
        ee = e->ent_next;

        if (ee && ee->ent_type == ENT_OMITTED) {
            if (e->ent_type == ENT_NUMBER)
                ee->ent_real = e->ent_real;
            else if (e->ent_type == ENT_TABLE) {
                tablfix((struct sCKTtable *)e->ent_string);
                ee->ent_real =
                    INPtablEval((struct sCKTtable *)e->ent_string,
                    ee->ent_val);
            }
            ee->ent_type = ENT_NUMBER;
            break;
        }

        if (e->ent_type == ENT_NUMBER)
            continue;
        if (e->ent_type == ENT_TABLE) {
            tablfix((struct sCKTtable *)e->ent_string);
            e->ent_real =
                INPtablEval((struct sCKTtable *)e->ent_string,
                ee->ent_val);
            continue;
        }
        /* should never get here */
        continue;
    }
}

/************************************************************************/

#ifdef debug

static void
testprnt(ckt)
GENERIC *ckt;
{
    struct sCKTtable *table, *t;
    struct sCKTentry *e, *ee;

    (void)INPtablFind(NULL,ckt,&table);

    for (t = table; t; t = t->tab_next) {
        printf("%s\n",t->tab_name);
        for (e = t->tab_entry; e; e = e->ent_next) {
            switch(e->ent_type) {
                case ENT_NUMBER:
                    printf("  type=%d val=%g real=%g\n",e->ent_type,
                        e->ent_val,e->ent_real);
                    continue;
                case ENT_TABLE:
                    printf("  type=%d val=%g real=%g (table %s)\n",
                        e->ent_type,e->ent_val,e->ent_real,
                        e->ent_string ?
                        ((struct sCKTtable *)e->ent_string)->tab_name : "(null)");
                    continue;
                case ENT_OMITTED:
                    printf("  type=%d val=%g omitted\n",e->ent_type,
                        e->ent_val);
                    continue;
                default:
                    printf("  type=%d error\n",e->ent_type);
            }
        }
    }
}

#endif

/************************************************************************/


static int
tabl_add(ckt,table)

/* Add table to the list kept in the ckt structure.
 */

GENERIC *ckt;
struct sCKTtable *table;
{
    if (!table) return (0);
    if (!ckt) return (1);
    table->tab_next = *DEVtableHead(ckt);
    *DEVtableHead(ckt) = table;
}

/************************************************************************/


int
INPtablFind(str,ckt,table)

/* Find the list head of the table named in str which is kept in ckt
 * return 0 if found, 1 if not found.
 */

char *str;
GENERIC *ckt;
GENERIC **table;
{
    struct sCKTtable *t;

    if (!ckt) return (1);
    if (!str) {
        if (table)
            *table = (GENERIC*) *DEVtableHead(ckt);
        return (0);
    }
    for (t = *DEVtableHead(ckt); t; t = t->tab_next)
        if (!strcmp(str,t->tab_name)) break;

    if (!t) return (1);
    if (table)
        *table = (GENERIC*)t;
    return (0);
}

/************************************************************************/


double
INPtablEval(tab,x)

GENERIC *tab;
double x;
{
    struct sCKTentry *ee, *e;

    if (tab == NULL) return (0);

    e = ((struct sCKTtable *)tab)->tab_entry;

    if (x <= e->ent_val) {
        return (e->ent_real);
    }

    for ( ; e; e = ee) {
        ee = e->ent_next;
        if (ee == NULL) break;
        if (x >= e->ent_val && x < ee->ent_val) {
            if (e->ent_type == ENT_TABLE)
                return (INPtablEval((struct sCKTtable*)e->ent_string,x));
            else
                return (e->ent_real + (x - e->ent_val)*
                    (ee->ent_real - e->ent_real)/(ee->ent_val - e->ent_val));
        }
    }
    if (e->ent_type == ENT_TABLE)
        return (INPtablEval((struct sCKTtable*)e->ent_string,x));
    return (e->ent_real);
}

/************************************************************************/


double
INPtablEvalDeriv(tab,x)

GENERIC *tab;
double x;
{
    struct sCKTentry *ee, *e;

    if (tab == NULL) return (0);

    e = ((struct sCKTtable *)tab)->tab_entry;

    if (x <= e->ent_val) {
        return (0);
    }

    for ( ; e; e = ee) {
        ee = e->ent_next;
        if (ee == NULL) break;
        if (x >= e->ent_val && x < ee->ent_val) {
            if (e->ent_type == ENT_TABLE)
                return
                    (INPtablEvalDeriv((struct sCKTtable*)e->ent_string),x);
            else
                return
                    ((ee->ent_real - e->ent_real)/(ee->ent_val - e->ent_val));
        }
    }
    return (0);
}

/************************************************************************/


void
INPtablFree(ckt)

/* 
 * Free the storage used for tables.
 */

GENERIC *ckt;
{
    struct sCKTtable *t, *tn;
    struct sCKTentry *e, *en;

    for (t = *DEVtableHead(ckt); t; t = tn) {
        tn = t->tab_next;
        txfree(t->tab_name);
        for (e = t->tab_entry; e; e = en) {
            en = e->ent_next;
            txfree((char*)e);
        }
        txfree((char*)t);
    }
}
