/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * All the functions used in the parse tree.  These functions return HUGE
 * if their argument is out of range.
 */

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "inpptree.h"

/* These should be in math.h */
extern double asinh(), acosh(), atanh();

#ifdef __STDC__
static double PTtPulse(INPparseNode*,double*);
static double PTtPwl(INPparseNode*,double*);
static double PTtSin(INPparseNode*,double*);
static double PTtSpulse(INPparseNode*,double*);
static double PTtExp(INPparseNode*,double*);
static double PTtSffm(INPparseNode*,double*);
static double PTabs(double);
static double PTsgn(double);
static double PTplus(double,double);
static double PTminus(double,double);
static double PTtimes(double,double);
static double PTdivide(double,double);
static double PTpower(double,double);
static double PTacos(double);
static double PTasin(double);
static double PTatan(double);
#ifdef HAVE_ACOSH
static double PTacosh(double);
static double PTasinh(double);
static double PTatanh(double);
#endif /* HAVE_ACOSH */
static double PTcos(double);
static double PTcosh(double);
static double PTexp(double);
static double PTln(double);
static double PTlog(double);
static double PTsin(double);
static double PTsinh(double);
static double PTsqrt(double);
static double PTtan(double);
static double PTtanh(double);
static double PTuminus(double);
#else
static double PTtPulse();
static double PTtPwl();
static double PTtSin();
static double PTtSpulse();
static double PTtExp();
static double PTtSffm();
static double PTabs();
static double PTsgn();
static double PTplus();
static double PTminus();
static double PTtimes();
static double PTdivide();
static double PTpower();
static double PTacos();
static double PTasin();
static double PTatan();
#ifdef HAVE_ACOSH
static double PTacosh();
static double PTasinh();
static double PTatanh();
#endif /* HAVE_ACOSH */
static double PTcos();
static double PTcosh();
static double PTexp();
static double PTln();
static double PTlog();
static double PTsin();
static double PTsinh();
static double PTsqrt();
static double PTtan();
static double PTtanh();
static double PTuminus();
#endif


double PTfudge_factor;

#define MODULUS(NUM,LIMIT)  ((NUM) - ((int) ((NUM) / (LIMIT))) * (LIMIT))


/* Tables that the parser uses. */

struct sPTop PTops[] = { 
    { PT_COMMA,       ",",    NULL } ,
    { PT_PLUS,        "+",    PTplus } ,
    { PT_MINUS,       "-",    PTminus } ,
    { PT_TIMES,       "*",    PTtimes } ,
    { PT_DIVIDE,      "/",    PTdivide },
    { PT_POWER,       "^",    PTpower },
    { PT_PLACEHOLDER, NULL,   (double(*)())NULL }
};

struct sPTfunc PTfuncs[] = {
    { "abs",    PTF_ABS,    PTabs } ,
    { "acos",   PTF_ACOS,   PTacos } ,
    { "asin",   PTF_ASIN,   PTasin } ,
    { "atan",   PTF_ATAN,   PTatan } ,
#ifdef HAVE_ACOSH
    { "acosh",  PTF_ACOSH,  PTacosh } ,
    { "asinh",  PTF_ASINH,  PTasinh } ,
    { "atanh",  PTF_ATANH,  PTatanh } ,
#endif /* HAVE_ACOSH */
    { "cos",    PTF_COS,    PTcos } ,
    { "cosh",   PTF_COSH,   PTcosh } ,
    { "exp",    PTF_EXP,    PTexp } ,
    { "ln",     PTF_LN,     PTln } ,
    { "log",    PTF_LOG,    PTlog } ,
    { "sgn",    PTF_SGN,    PTsgn } ,
    { "sin",    PTF_SIN,    PTsin } ,
    { "sinh",   PTF_SINH,   PTsinh } ,
    { "sqrt",   PTF_SQRT,   PTsqrt } ,
    { "tan",    PTF_TAN,    PTtan } ,
    { "tanh",   PTF_TANH,   PTtanh } ,
    { "-",      PTF_UMINUS, PTuminus },
    { NULL,     0,          (double(*)())NULL }
};

struct sPTfunc PTtFuncs[] = {
    { "pulse",  PTF_tPULSE, PTtPulse } ,
    { "pwl",    PTF_tPWL,   PTtPwl } ,
    { "sin",    PTF_tSIN,   PTtSin } ,
    { "spulse", PTF_tSPULSE,PTtSpulse } ,
    { "exp",    PTF_tEXP,   PTtExp } ,
    { "sffm",   PTF_tSFFM,  PTtSffm } ,
    { NULL,     0,          (double(*)())NULL }
};

#define V1    p->tranparms[0]
#define V2    p->tranparms[1]
#define TD    p->tranparms[2]
#define TR    p->tranparms[3]
#define TF    p->tranparms[4]
#define PW    p->tranparms[5]
#define PER   p->tranparms[6]

#define SPER  p->tranparms[2]
#define SDEL  p->tranparms[3]

#define VO    p->tranparms[0]
#define VA    p->tranparms[1]
#define FREQ  p->tranparms[2]
#define TDL   p->tranparms[3]
#define THETA p->tranparms[4]

#define TD1   p->tranparms[2]
#define TAU1  p->tranparms[3]
#define TD2   p->tranparms[4]
#define TAU2  p->tranparms[5]

#define FC    p->tranparms[2]
#define MDI   p->tranparms[3]
#define FS    p->tranparms[4]

/* evaluate the tran functions */

static double
PTtPulse(p,t)

INPparseNode *p;
double *t;
{
    double time, basetime, value;
    double pw = TR + PW;
    int i;

    time = *t - TD;
    if (time > PER) {
        /* repeating signal - figure out where we are in period */
        basetime = PER * (int)(time/PER);
        time -= basetime;
    }
    if (time < TR || time >= pw + TF) {
        value = V1;
        if (time > 0 && time < TR)
            value += time* *p->trancache;
    }
    else {
        value = V2;
        if (time > pw)
            value += (time - pw)* *(p->trancache+1);
    }

    /* new feature: additional entries are added  as in
     * pulse(V1 V2 TD ...) + pulse(0 V2-V1 TD1 ...) + ...
     * extra entries have offset subtracted
     */
    for (i = 7; i < p->numtrancoeffs; i++) {
        time = *t - p->trancoeffs[i];
        if (time > PER) {
            /* repeating signal - figure out where we are in period */
            basetime = PER * (int)(time/PER);
            time -= basetime;
        }
        if (time < TR || time >= pw + TF) {
            if (time > 0 && time < TR)
                value += time* *p->trancache;
        }
        else {
            value += (V2-V1);
            if (time > pw)
                value += (time - pw)* *(p->trancache+1);
        }
    }
    return (value);
}


static double
PTtPwl(p,t)

INPparseNode *p;
double *t;
{
    if (*t >= (*p->trancoeffs+p->numtrancoeffs-2))
        return (*(p->trancoeffs+p->numtrancoeffs-1));
    if (p->pwlindex + 2 < p->numtrancoeffs) {
        for (;;) {
            if (*t <= *(p->trancoeffs+p->pwlindex+2)) break;
            p->pwlindex += 2;
            if (p->pwlindex >= p->numtrancoeffs) break;
        }
    }
    for (;;) {
        if (*t < *(p->trancoeffs+p->pwlindex) && p->pwlindex)
            p->pwlindex -= 2;
        else break;
    }
    if (!p->pwlindex && *t <= *(p->trancoeffs)) {
        return (*(p->trancoeffs+1));
    }
    if (p->pwlindex >= p->numtrancoeffs) {
        return (*(p->trancoeffs+p->numtrancoeffs-1));
    }
    return ( *(p->trancoeffs+p->pwlindex+1) +
        *(p->trancache + (p->pwlindex >> 1)) *
        (*t - *(p->trancoeffs+p->pwlindex)) );
}


static double
PTtSin(p,t)

INPparseNode *p;
double *t;
{
    double time, value;

    time = *t - TDL;
    if (time <= 0)
        return (VO);
    return (VO + VA*sin(FREQ * time * 2*M_PI)*exp(-time*THETA));
}


static double
PTtSpulse(p,t)

INPparseNode *p;
double *t;
{
    double time;

    time = *t - SDEL;
    if (time <= 0)
        return (V1);
    return (V1 +
        (V2-V1)*( 1 - cos(2*M_PI*time/SPER)*exp(-time*THETA) )/2);
}


static double
PTtExp(p,t)

INPparseNode *p;
double *t;
{
    double value;

    value = V1;
    if (*t <= TD1)
        return (value);
    value += (V2-V1)*(1-exp(-(*t-TD1)/TAU1));
    if (*t <= TD2)
        return (value);
    value += (V1-V2)*(1-exp(-(*t-TD2)/TAU2));
    return (value);
}


static double
PTtSffm(p,t)

INPparseNode *p;
double *t;
{
    return (VO + VA*sin((2*M_PI * FC * *t) +
        MDI * sin(2*M_PI * FS * *t)));
}


/* evaluate the math functions */

static double
PTabs(arg)

double arg;
{   return arg >= 0.0 ? arg : -arg; }


static double
PTsgn(arg)

double arg;
{   return arg > 0.0 ? 1.0 : arg < 0.0 ? -1.0 : 0.0; }


static double
PTplus(arg1, arg2)

double arg1, arg2;
{   return (arg1 + arg2); }


static double
PTminus(arg1, arg2)

double arg1, arg2;
{   return (arg1 - arg2); }


static double
PTtimes(arg1, arg2)

double arg1, arg2;
{   return (arg1 * arg2); }


static double 
PTdivide(arg1, arg2)

double arg1, arg2;
{
    if (arg2 >= 0.0)
        arg2 += PTfudge_factor;
    else
        arg2 -= PTfudge_factor;

    if (arg2 == 0.0)
        return (HUGE);

    return (arg1 / arg2);
}


static double
PTpower(arg1, arg2)

double arg1, arg2;
{   return (pow(arg1, arg2)); }


static double
PTacos(arg)

double arg;
{   return (acos(arg)); }


static double
PTasin(arg)

double arg;
{   return (asin(arg)); }


static double
PTatan(arg)

double arg;
{   return (atan(arg)); }


#ifdef HAVE_ACOSH

static double
PTacosh(arg)

double arg;
{   return (acosh(arg)); }


static double
PTasinh(arg)

double arg;
{   return (asinh(arg)); }


static double
PTatanh(arg)

double arg;
{   return (atanh(arg)); }

#endif /* HAVE_ACOSH */


static double
PTcos(arg)

double arg;
{   return (cos(MODULUS(arg, 2 * M_PI))); }


static double
PTcosh(arg)

double arg;
{   return (cosh(arg)); }


static double
PTexp(arg)

double arg;
{   return (exp(arg)); }


static double
PTln(arg)

double arg;
{
    if (arg < 0.0)
        arg = -arg;
    return (log(arg));
}


static double
PTlog(arg)

double arg;
{
    if (arg < 0.0)
        arg = -arg;
    return (log10(arg));
}


static double
PTsin(arg)

double arg;
{   return (sin(MODULUS(arg, 2 * M_PI))); }


static double
PTsinh(arg)

double arg;
{   return (sinh(arg)); }


static double
PTsqrt(arg)

double arg;
{
    if (arg < 0.0)
        arg = -arg;
    return (sqrt(arg));
}


static double
PTtan(arg)

double arg;
{   return (tan(MODULUS(arg, M_PI))); }


static double
PTtanh(arg)

double arg;
{   return (tanh(arg)); }


static double
PTuminus(arg)

double arg;
{   return (- arg); }

