/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Thomas L. Quarles
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * NIiter(ckt,maxIter)
 *
 *  This subroutine performs the actual numerical iteration.
 *  It uses the sparse matrix stored in the circuit struct
 *  along with the matrix loading program, the load data, the
 *  convergence test function, and the convergence parameters
 */

#include "spice.h"
#include <stdio.h>
#include "cktdefs.h"
#include "optdefs.h"
#include "sperror.h"
#include "util.h"
#include "cktext.h"
#include "niext.h"

/* NIiter() - return value is non-zero for convergence failure */

int
NIiter(ckt,maxIter)

CKTcircuit * ckt;
int maxIter;
{
    int iterno;
    int ipass;
    int error;
    int i,j; /* temporaries for finding error location */
    char *message;  /* temporary message buffer */
    double *temp;
    double startTime;
    static char *msg = "Too many iterations without convergence";
    STATistics *stat = (STATistics*)ckt->CKTstat;

    /* assume that NIreinit() has been called */
    iterno = 0;
    ipass = 0;
    ckt->CKTnoncon = 0;

    error = CKTload(ckt);
    if (error) {
#ifdef STEPDEBUG
        printf("load (1) returned error\n");
#endif
        return (error);
    }

    if (!(ckt->CKTniState & NIDIDPREORDER)) {
        spMNA_Preorder(ckt->CKTmatrix);
        error = spError(ckt->CKTmatrix);
        if (error) {
            /* badly formed matrix */
#ifdef STEPDEBUG
            printf("pre-order returned error \n");
#endif
            return (error);
        }
        ckt->CKTniState |= NIDIDPREORDER;
    }

    if ((ckt->CKTmode & MODEINITJCT) || 
            (ckt->CKTmode & MODEINITTRAN)) {
        ckt->CKTniState |= NISHOULDREORDER;
    }

    for (;;) {

        if (ckt->CKTmode & MODEINITFLOAT) {
            if (ckt->CKTnoncon == 0 && ckt->CKTjjPresent) {
                error = OK;
                break;
            }
        }
        iterno++;

        if (ckt->CKTniState & NISHOULDREORDER) {
            startTime = (*(SPfrontEnd->IFseconds))();
            spSetReal(ckt->CKTmatrix);
            spLoadGmin(ckt->CKTmatrix,ckt->CKTdiagGmin);
            error = spOrderAndFactor(ckt->CKTmatrix,NULL,
                ckt->CKTpivotRelTol,ckt->CKTpivotAbsTol,1);
            stat->STATreorderTime += 
                    (*(SPfrontEnd->IFseconds))() - startTime;
            if (error) {
                /* new feature - we can now find out something about what is
                 * wrong - so we ask for the troublesome entry 
                 */
                spWhereSingular(ckt->CKTmatrix,&i,&j);
                message = (char *)MALLOC(1000); /* should be enough */
                (void)sprintf(message,
                        "singular matrix:  check nodes %s and %s\n",
                        NODENAME(ckt,i),NODENAME(ckt,j));
                (*(SPfrontEnd->IFerror))(ERR_WARNING,message,(IFuid *)NULL);
                FREE(message);
                /* can't handle these errors - pass up! */
#ifdef STEPDEBUG
                printf("reorder returned error ");
                if (ckt->CKTmode & MODETRAN)
                    printf("at time=%g",ckt->CKTtime);
                printf("\n");
#endif
                break;
            }
            ckt->CKTniState &= ~NISHOULDREORDER;
            spGetStat(ckt->CKTmatrix,
                &stat->STATmatSize,
                &stat->STATnonZero,
                &stat->STATfillIn);
        }
        else {
            startTime = (*(SPfrontEnd->IFseconds))();
            spSetReal(ckt->CKTmatrix);
            spLoadGmin(ckt->CKTmatrix,ckt->CKTdiagGmin);
            error = spFactor(ckt->CKTmatrix);
            stat->STATdecompTime += 
                    (*(SPfrontEnd->IFseconds))() - startTime;
            if (error) {
                if (error == E_SINGULAR) {
                    ckt->CKTniState |= NISHOULDREORDER;
                    DEBUGMSG(" forced reordering....\n");
                    error = CKTload(ckt);
                    if (error) {
#ifdef STEPDEBUG
                        printf("load (3) returned error \n");
#endif
                        break;
                    }
                    spMNA_Preorder(ckt->CKTmatrix);
                    error = spError(ckt->CKTmatrix);
                    if (error) {
#ifdef STEPDEBUG
                        printf("preorder returned error\n");
#endif
                        break;
                    }
                    ckt->CKTniState |= NIDIDPREORDER;
                    continue;
                }
                /* seems to be singular - pass the bad news up */
#ifdef STEPDEBUG
                printf("lufac returned error \n");
#endif
                break;
            }
        } 

        startTime = (*(SPfrontEnd->IFseconds))();
        spSolve(ckt->CKTmatrix,ckt->CKTrhs,ckt->CKTrhs,NULL,NULL);
        stat->STATsolveTime += (*(SPfrontEnd->IFseconds))() - startTime;

        temp = ckt->CKTrhsOld;
        ckt->CKTrhsOld = ckt->CKTrhs;
        ckt->CKTrhs = temp;

        *ckt->CKTrhs = 0;
        *ckt->CKTrhsSpare = 0;
        *ckt->CKTrhsOld = 0;

        if (iterno > maxIter) {
            error = E_ITERLIM;
            errMsg = MALLOC(strlen(msg)+1);
            strcpy(errMsg,msg);
#ifdef STEPDEBUG
            printf("iterlim exceeded, time = %g\n",ckt->CKTtime);
#endif
            break;
        }
        if (!ckt->CKTjjPresent) {
            if (ckt->CKTnoncon == 0 && iterno != 1) {
                ckt->CKTnoncon = NIconvTest(ckt);
            }
            else {
                ckt->CKTnoncon = 1;
            }
        }

        if (ckt->CKTmode & MODEINITFLOAT) {
            if ((ckt->CKTmode & MODEDC) && ckt->CKThadNodeset) {
                if (ipass) {
                    ckt->CKTnoncon = ipass;
                }
                ipass = 0;
            }
            if (ckt->CKTnoncon == 0) {
                error = OK;
                break;
            }
        }
        else if (ckt->CKTmode & MODEINITPRED) {
            ckt->CKTmode = (ckt->CKTmode&(~INITF))|MODEINITFLOAT;
        }
        else if (ckt->CKTmode & MODEINITTRAN) {
            if (iterno <= 1) ckt->CKTniState |= NISHOULDREORDER;
            ckt->CKTmode = (ckt->CKTmode&(~INITF))|MODEINITFLOAT;
        }
        else if (ckt->CKTmode & MODEINITSMSIG) {
            ckt->CKTmode = (ckt->CKTmode&(~INITF))|MODEINITFLOAT;
        }
        else if(ckt->CKTmode & MODEINITJCT) {
            ckt->CKTmode = (ckt->CKTmode&(~INITF))|MODEINITFIX;
            ckt->CKTniState |= NISHOULDREORDER;
        }
        else if (ckt->CKTmode & MODEINITFIX) {
            if (ckt->CKTnoncon == 0)
                ckt->CKTmode = (ckt->CKTmode&(~INITF))|MODEINITFLOAT;
            ipass = 1;
        }
        else {
            error = E_INTERN;
#ifdef STEPDEBUG
            printf("bad initf state \n");
#endif
            break;
        }
    
        ckt->CKTnoncon = 0;
        error = CKTload(ckt);

        if (error) {
#ifdef STEPDEBUG
            printf("load (2) returned error \n");
#endif
            break;
        }
    }
    stat->STATnumIter += iterno;
#ifdef STEPDEBUG
    printf("time=%15g, %d iters, %s\n", ckt->CKTtime,iterno,
        (error==OK) ? "ok" : "fail");
#endif
    return (error);
}
