/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher
         1992 David A. Gates ( ft_xgraph() )
         1993 Stephen R. Whiteley
****************************************************************************/

/*
 * Plotting routines
 */

#include "spice.h"
#include "ftedefs.h"
#include "plotdefs.h"

#ifdef __STDC__
static bool plotit(wordlist*,char*,char*);
static bool asciilin(struct dvlist*,bool*);
static void parsewl(wordlist**,bool*,bool*,bool*);
static bool fixlimits(struct dvlist*,double*,double*,bool);
static double *getlims(wordlist**,char*,int);
static char *getword(wordlist**,char*);
static bool getflag(wordlist**,char*);
static void compress(struct dvec*,double*,double*);
#else
static bool plotit();
static bool asciilin();
static void parsewl();
static bool fixlimits();
static double *getlims();
static char *getword();
static bool getflag();
static void compress();
#endif

static bool sameflag;

void
com_asciiplot(wl)

/* asciiplot file name ... [xlimit] xhi xlo] [vs xname] */
wordlist *wl;
{
    (void) plotit(wl, (char *) NULL, "lpr");
    return;
}


/* xgraph file plotargs
 */

void
com_xgraph(wl)

wordlist *wl;
{
    char *fname;
    char buf[BSIZE_SP];
    bool tempf = false;
    char *devtype;

    if (wl) {
        fname = wl->wl_word;
        wl = wl->wl_next;
    }
    if (!wl) {
        return;
    }
    if (cieq(fname, "temp") || cieq(fname, "tmp")) {
        fname = smktemp("xg");
        tempf = true;
    }

    (void) plotit(wl, fname, "xgraph");
    return;
}


void
com_hardcopy(wl)

/* hardcopy file plotargs, or 'hardcopy file' -- with no other args this
 * prompts the user for a window to dump to a plot file.
 */
wordlist *wl;
{
    ft_hardcopy(wl, NULL, true, NULL, false);
}


void
ft_hardcopy(wl, graph, copygraph, mesg, fileonly)

/* Makes hardcopy of graph if not null.  If copygraph is true, it is
 * assumed that graph is being displayed in a window, and a copy is
 * made.  Otherwise, graph is assumed to have been created specifically
 * for the copy, and is destroyed.  No effect if graph is null.
 */
wordlist *wl;
GRAPH *graph;
bool copygraph;
char *mesg;
bool fileonly;
{
    char *fname, *devtype, *s;
    char buf[BSIZE_SP], device[BSIZE_SP];
    bool tempf = false;
    GRAPH *tempgraph;
    extern char *kw_hcopydev, *kw_hcopydevtype;

    *device = '\0';
    if (!fileonly) {
        if (!cp_getvar(kw_hcopydev, VT_STRING, device)) {
            s = getenv("PRINTER");
            if (s)
                strcpy(device, s);
        }
        if (!*device) {
            device[0] = 'l'; device[1] = 'p'; device[2] = '\0';
        }
    }

    if (wl) {
        fname = wl->wl_word;
        wl = wl->wl_next;
    }
    else {
        fname = smktemp("hc");
        tempf = true;
    }

    if (!cp_getvar(kw_hcopydevtype, VT_STRING, buf)) {
        devtype = "postscript";
    }
    else {
        devtype = buf;
    }

    if (graph) {
        
        if (DevSwitch(devtype)) return;
        if (copygraph)
            tempgraph = CopyGraph(graph);
        else
            tempgraph = graph;
        tempgraph->devdep = copy(fname);
        if (DevNewViewport(tempgraph)) {
            DestroyGraph(tempgraph->graphid);
            DevSwitch(NULL);
            return;
        }
        (*tempgraph->redraw)(tempgraph);
        PushGraphContext(tempgraph);
        DevHalt();
        PopGraphContext();
        DestroyGraph(tempgraph->graphid);
        DevSwitch(NULL);
    }
    else {
        if (!wl) {
            wl = outmenuprompt("which variable ? ");
            if (!wl)
                return;
        }
        if (DevSwitch(devtype)) return;

        if (!plotit(wl, fname, (char *) NULL)) {
            DevSwitch(NULL);    /* remember to switch back */
            return;
        }

        DevSwitch(NULL);
    }

    ft_makehardcopy(copygraph ? graph : NULL, fname, devtype, device, mesg);

    if (tempf) {
        if (*device)
            (void) unlink(fname);
        txfree(fname);
    }

    return;
}


void
com_plot(wl)

/* plot name ... [xl[imit]] xlo xhi] [yl[imit ylo yhi] [vs xname] */
wordlist *wl;
{
    (void) plotit(wl, (char *) NULL, (char *) NULL);
    return;
}


struct grstuff {
    double xlim[2];
    double ylim[2];
    double xcompress;
    double xindices[2];
    double xdelta;
    double ydelta;
    bool xlimset;
    bool ylimset;
    bool xcompset;
    bool xindset;
    bool xdeltaset;
    bool ydeltaset;
    bool nointerp;
    int gtype;
    int ptype;
    char *xlabel;
    char *ylabel;
    char *title;
    char *commandline;
};
static struct grstuff grs;


static bool
plotit(wl, hcopy, devname)

/* The common routine for all plotting commands. This does hardcopy
 * and graphics plotting.
 */
wordlist *wl;
char *hcopy, *devname;
{
    GRAPH *graph;
    struct pnode *n, *names;
    struct dvec *d, *lv;
    struct dvlist *dl0, *dl, *tl, *dn, *lastv;
    struct grinit gr;
    double ylims[2], xlims[2];
    int i;
    bool sameflag, gfound, pfound, oneval = false;
    char buf[BSIZE_SP];
    wordlist *ww;

    parsewl(&wl,&sameflag,&gfound,&pfound);

    if (!wl) {
        fprintf(cp_err, "Error: no vectors given\n");
        return (false);
    }

    /* If we have forms like v(1) vs -v(2), force a unary - by converting
     * to v(1) vs --v(2)
     */

    for (ww = wl; ww; ww = ww->wl_next) {
        if (eq(ww->wl_word,"vs")) {
            ww = ww->wl_next;
            if (ww && *ww->wl_word == '-' && *(ww->wl_word+1) != '-') {
                sprintf(buf,"-%s",ww->wl_word);
                txfree(ww->wl_word);
                ww->wl_word = copy(buf);
            }
        }
    }

    /* Now parse the vectors.  We have a list of the form
     * "a b vs c d e vs f g h".  Since it's a bit of a hassle for
     * us to parse the vector boundaries here, we do this -- call
     * ft_getpnames() without the check flag, and then look for 0-length
     * vectors with the name "vs"...  This is a sort of a gross hack,
     * since we have to check for 0-length vectors ourselves after
     * evaulating the pnodes...
     */

    names = ft_getpnames(wl, false);
    if (names == NULL)
        return (false);

    /* hack! have to fool ft_evaluate() */
    for (n = names; n; n = n->pn_next) {
        d = n->pn_value;
        if (d && d->v_length == 0 && eq(d->v_name,"vs"))
            d->v_length = -1;
    }

    dl0 = ft_dvlist(names);
    if (dl0 == NULL)
        return (false);

    tl = NULL;
    lastv = NULL;
    for (dl = dl0; dl; dl = dn) {
        dn = dl->dl_next;
        d = dl->dl_dvec;
        if (d->v_length == -1 && eq(d->v_name,"vs")) {
            d->v_length = 0; /* just in case */
            if (!tl || !dn) {
                fprintf(cp_err, "Error: misplaced vs arg\n");
                vec_dlfree(dl0);
                return (false);
            }
            if (!lastv)
                tl = dl0;
            else
                tl = lastv->dl_next;
            for ( ; tl != dl; lastv = tl, tl = tl->dl_next)
                tl->dl_dvec->v_scale = dn->dl_dvec;
            dn = dn->dl_next;
            lastv->dl_next = dn;
            tl = NULL;
            txfree((char*)dl);
            txfree((char*)dn);
        }
        else
            tl = dl;
    }

    /* Now check for 0-length vectors. */
    for (dl = dl0; dl; dl = dl->dl_next) {
        if (!dl->dl_dvec->v_length) {
            fprintf(cp_err, "Error: %s: no such vector\n",
                    dl->dl_dvec->v_name);
            vec_dlfree(dl0);
            return (false);
        }
    }
    
    if (devname && (eq(devname,"xgraph") || eq(devname,"lpr"))) {
        /* If there are higher dimensional vectors, transform them into a
         * family of vectors.
         */
        tl = NULL;
        for (dl = dl0; dl; dl = dn) {
            dn = dl->dl_next;
            d = dl->dl_dvec;
            if (d->v_numdims > 1) {
                lv = vec_mkfamily(d);
                if (lv && lv->v_link2) {
                    if (!tl)
                        dl0 = lv->v_link2;
                    else
                        tl->dl_next = lv->v_link2;
                    for (dl = lv->v_link2; dl->dl_next; dl = dl->dl_next);
                    dl->dl_next = dn;
                    tl = dl;
                }
                else {
                    if (!tl)
                        dl0 = dn;
                    else
                        tl->dl_next = dn;
                }
                txfree((char*)dl);
            }
            else
                tl = dl;
        }
    }

    /* Now fill in the scales for vectors who aren't already fixed up. */
    for (dl = dl0; dl; dl = dl->dl_next) {
        d = dl->dl_dvec;
        if (!d->v_scale) {
            if (d->v_plot->pl_scale)
                d->v_scale = d->v_plot->pl_scale;
            else
                d->v_scale = d;
        }
    }

    /* See if the log flag is set anywhere... */
    if (!gfound) {
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_scale && (d->v_scale->v_gridtype == GRID_XLOG))
                grs.gtype = GRID_XLOG;
        }
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_gridtype == GRID_YLOG) {
                if ((grs.gtype == GRID_XLOG) ||
                        (grs.gtype == GRID_LOGLOG))
                    grs.gtype = GRID_LOGLOG;
                else
                    grs.gtype = GRID_YLOG;
            }
        }
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_gridtype == GRID_SMITH) {
                grs.gtype = GRID_SMITH;
                break;
            }
            else if (d->v_gridtype == GRID_POLAR) {
                grs.gtype = GRID_POLAR;
                break;
            }
        }
    }

    /* See if there are any default plot types...  Here, like above, we
     * don't do entirely the best thing when there is a mixed set of
     * default plot types...
     */
    if (!sameflag && !pfound) {
        grs.ptype = PLOT_LIN;
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_plottype != PLOT_LIN) {
                grs.ptype = d->v_plottype;
                break;
            }
        }
    }

    /* Check and see if this is pole zero stuff. */
    if ((dl0->dl_dvec->v_type == SV_POLE) ||
            (dl0->dl_dvec->v_type == SV_ZERO))
        oneval = true;
    
    for (dl = dl0; dl; dl = dl->dl_next) {
        d = dl->dl_dvec;
        if (((d->v_type == SV_POLE) || (d->v_type == SV_ZERO)) !=
                oneval ? 1 : 0) {
            fprintf(cp_err,
"Error: plot must be either all pole-zero or contain no poles or zeros\n");
            return (false);
        }
    }
    
    if ((grs.gtype == GRID_POLAR) || (grs.gtype == GRID_SMITH))
        oneval = true;

    /* If we are plotting scalars, make sure there is enough
     * data to fit on the screen.
     */
    
    for (dl = dl0; dl; dl = dl->dl_next) {
        d = dl->dl_dvec;
        if (d->v_length != d->v_scale->v_length)
            plot_extend(d, d->v_scale->v_length);
    }

    /* Now patch up each vector with the compression and the index
     * selection.
     */
    if (grs.xcompset || grs.xindset) {
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            compress(d, &grs.xcompress, grs.xindices);
            d->v_scale = vec_copy(d->v_scale);
            vec_newtemp(d->v_scale);
            compress(d->v_scale, &grs.xcompress, grs.xindices);
        }
    }

    if (!fixlimits(dl0,xlims,ylims,oneval)) {
        vec_dlfree(dl0);
        return (false);
    }

    gr.xlims = xlims;
    gr.ylims = ylims;
    gr.plotname = grs.title ? grs.title : dl0->dl_dvec->v_plot->pl_title;
    gr.hcopy = hcopy;
    gr.xdelta = grs.xdeltaset ? grs.xdelta : 0.0;
    gr.ydelta = grs.ydeltaset ? grs.ydelta : 0.0;
    gr.gridtype = grs.gtype;
    gr.plottype = grs.ptype;
    gr.xlabel = grs.xlabel;
    gr.ylabel = grs.ylabel;
    gr.pname = plot_cur->pl_typename;
    gr.commandline = grs.commandline;
    gr.nointerp = grs.nointerp;

    /* See if there is one type we can give for the y scale... */
    gr.ytype = dl0->dl_dvec->v_type;
    for (dl = dl0->dl_next; dl; dl = dl->dl_next) {
        if (dl->dl_dvec->v_type != gr.ytype) {
            gr.ytype = SV_NOTYPE;
            break;
        }
    }
    for (dl = dl0, i = 0; dl; dl = dl->dl_next, i++) ;
    gr.nplots = i;

    /* Figure out the X name and the X type.  This is sort of bad... */
    gr.xname = oneval ? NULL : dl0->dl_dvec->v_scale->v_name;
    gr.xtype = dl0->dl_dvec->v_scale->v_type;

    if (devname) {
        if (eq(devname, "xgraph")) {
            /* Interface to XGraph-11 Plot Program */
            ft_xgraph(dl0,(char*)&gr);
            vec_dlfree(dl0);
            return (true);
        }
        if (eq(devname, "lpr")) {
            /* yecch, line printer plot */
            if (!asciilin(dl0,&gr.nointerp)) {
                vec_dlfree(dl0);
                return (false);
            }
            ft_agraf(dl0, (char*)&gr);
            vec_dlfree(dl0);
            return (true);
        }
    }

    /* The graph->plotdata dveclist as returned from gr_init contains
     * pointers to the actual dvecs, for iplot.  Thus, we must copy
     * them before destroying the graph.
     */
    graph = gr_init(dl0,(char*)&gr);
    if (graph == NULL)
        return (false);

    /* copy the dveclist, then free the old structures */
    graph->plotdata =
        (*graph->copydata)(graph->plotdata);
    vec_dlfree(dl0);

    if (!dispdev->windows) {
        PushGraphContext(graph);
        gr_end();
        DevHalt();
        PopGraphContext();
        DestroyGraph(graph->graphid);
    }
    return (true);
}


static bool
asciilin(dl0,nointerp)

struct dvlist *dl0;
bool *nointerp;
/* check if we should (can) linearize */
{
    struct plot *p = plot_cur;
    struct dvec *d, *newv_scale;
    struct dvlist *dl;
    double *newdata, *newscale;
    double ttime;
    double dummy;
    int i, newlen;

    if (!( !p || !p->pl_scale || !isreal(p->pl_scale) ||
            !ciprefix("tran", p->pl_typename) ||
            ((p->pl_stop - p->pl_start)*p->pl_step <= 0.0) ||
            ((p->pl_stop - p->pl_start) < p->pl_step) )) {

        newlen = (p->pl_stop - p->pl_start) / p->pl_step + 1.5;

        newscale = (double *) tmalloc(newlen * sizeof(double));
        d = dl0->dl_dvec;

        newv_scale = alloc(struct dvec);
        newv_scale->v_flags = d->v_scale->v_flags;
        newv_scale->v_type = d->v_scale->v_type;
        newv_scale->v_gridtype = d->v_scale->v_gridtype;
        newv_scale->v_length = newlen;
        newv_scale->v_name = copy(d->v_scale->v_name);
        newv_scale->v_realdata = newscale;
        vec_newtemp(newv_scale);

        for (i = 0, ttime = p->pl_start; i < newlen;
                i++, ttime += p->pl_step)
            newscale[i] = ttime;

        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            newdata = (double *) tmalloc(newlen * sizeof (double));

            if (!ft_interpolate(d->v_realdata, newdata,
                d->v_scale->v_realdata, d->v_scale->v_length,
                newscale, newlen, 1)) {
                fprintf(cp_err,
                    "Error: can't interpolate %s\n", d->v_name);
                return(false);
            }

            tfree(d->v_realdata);
            d->v_realdata = newdata;

            /* Why go to all this trouble if agraf ignores it? */
            *nointerp = true;
        }
        dl0->dl_dvec->v_scale = newv_scale;
    }
    return (true);
}


struct sPlotFlags {
    char *name;
    int type;
};
static struct sPlotFlags gtypes[] = {
    {"lingrid",GRID_LIN},
    {"loglog",GRID_LOGLOG},
    {"nogrid",GRID_NONE},
    {"xlog",GRID_XLOG},
    {"ylog",GRID_YLOG},
    {"polar",GRID_POLAR},
    {"smith",GRID_SMITH}
};
#define NumGrTypes sizeof(gtypes)/sizeof(struct sPlotFlags)

static struct sPlotFlags ptypes[] = {
    {"linplot",PLOT_LIN},
    {"combplot",PLOT_COMB},
    {"pointplot",PLOT_POINT}
};
#define NumPlTypes sizeof(ptypes)/sizeof(struct sPlotFlags)



static void
parsewl(pwl,sameflag,gfound,pfound)

/* Parse the plot command wordlist. */
wordlist **pwl;
bool *sameflag, *gfound,*pfound;
{
    double *dd;
    int i;
    wordlist *wwl;
    char buf[BSIZE_SP], *s;
    extern char *kw_gridstyle, *kw_plotstyle;

    *gfound = false;
    *pfound = false;

    /* First get the command line, without the limits. */
    wwl = wl_copy(*pwl);
    (void) getlims(&wwl, "xl", 2);
    (void) getlims(&wwl, "xlimit", 2);
    (void) getlims(&wwl, "yl", 2);
    (void) getlims(&wwl, "ylimit", 2);
    s = wl_flatten(wwl);
    (void) sprintf(buf, "plot %s", s);
    txfree(grs.commandline);
    grs.commandline = copy(buf);
    wl_free(wwl);
    txfree(s);

    /* Now extract all the parameters. */

    *pwl = wl_copy(*pwl);

    *sameflag = getflag(pwl, "samep");

    if (!*sameflag || !grs.xlimset) {
        dd = getlims(pwl, "xl", 2);
        if (!dd)
            dd = getlims(pwl, "xlimit", 2);
        if (dd) {
            grs.xlim[0] = dd[0];
            grs.xlim[1] = dd[1];
            grs.xlimset = true;
        }
        else
            grs.xlimset = false;
    }
    else {
        (void) getlims(pwl, "xl", 2);
        (void) getlims(pwl, "xlimit", 2);
    }

    if (!*sameflag || !grs.ylimset) {
        dd = getlims(pwl, "yl", 2);
        if (!dd)
            dd = getlims(pwl, "ylimit", 2);
        if (dd) {
            grs.ylim[0] = dd[0];
            grs.ylim[1] = dd[1];
            grs.ylimset = true;
        }
        else
            grs.ylimset = false;
    }
    else {
        (void) getlims(pwl, "yl", 2);
        (void) getlims(pwl, "ylimit", 2);
    }

    if (!*sameflag || !grs.xcompset) {
        dd = getlims(pwl, "xcompress", 1);
        if (!dd)
            dd = getlims(pwl, "xcomp", 1);
        if (dd) {
            grs.xcompress = dd[0];
            grs.xcompset = true;
        }
        else
            grs.xcompset = false;
    }
    else {
        (void) getlims(pwl, "xcompress", 1);
        (void) getlims(pwl, "xcomp", 1);
    }

    if (!*sameflag || !grs.xindset) {
        dd = getlims(pwl, "xindices", 2);
        if (!dd)
            dd = getlims(pwl, "xind", 2);
        if (dd) {
            grs.xindices[0] = dd[0];
            grs.xindices[1] = dd[1];
            grs.xindset = true;
        }
        else
            grs.xindset = false;
    }
    else {
        (void) getlims(pwl, "xindices", 2);
        (void) getlims(pwl, "xind", 2);
    }

    if (!*sameflag || !grs.xdeltaset) {
        dd = getlims(pwl, "xdelta", 1);
        if (!dd)
            dd = getlims(pwl, "xdel", 1);
        if (dd) {
            grs.xdelta = dd[0];
            grs.xdeltaset = true;
        }
        else
            grs.xdeltaset = false;
    }
    else {
        (void) getlims(pwl, "xdelta", 1);
        (void) getlims(pwl, "xdel", 1);
    }
    if (!*sameflag || !grs.ydeltaset) {
        dd = getlims(pwl, "ydelta", 1);
        if (!dd)
            dd = getlims(pwl, "ydel", 1);
        if (dd) {
            grs.ydelta = dd[0];
            grs.ydeltaset = true;
        }
        else
            grs.ydeltaset = false;
    }
    else {
        (void) getlims(pwl, "ydelta", 1);
        (void) getlims(pwl, "ydel", 1);
    }
    
    /* Get the grid type and the point type.  Note we can't do if-else
     * here because we want to catch all the grid types.
     */
    for (i = 0; i < NumGrTypes; i++) {
        if (getflag(pwl, gtypes[i].name)) {
            if (*gfound)
                fprintf(cp_err, 
                    "Warning: too many grid types given\n");
            else {
                grs.gtype = gtypes[i].type;
                *gfound = true;
            }
        }
    }

    if (!*sameflag && !*gfound) {
        if (cp_getvar(kw_gridstyle, VT_STRING, buf)) {
            for (i = 0; i < NumGrTypes; i++) {
                if (eq(buf, gtypes[i].name)) {
                    grs.gtype = gtypes[i].type;
                    break;
                }
            }
            if (i == NumGrTypes) {
                fprintf(cp_err, "Warning: strange grid type %s\n", buf);
                grs.gtype = GRID_LIN;
            }
            *gfound = true;
        }
        else
            grs.gtype = GRID_LIN;
    }

    /* Now get the point type.  */

    for (i = 0; i < NumPlTypes; i++) {
        if (getflag(pwl, ptypes[i].name)) {
            if (*pfound)
                fprintf(cp_err, 
                    "Warning: too many plot types given\n");
            else {
                grs.ptype = ptypes[i].type;
                *pfound = true;
            }
        }
    }

    if (!*sameflag && !*pfound) {
        if (cp_getvar(kw_plotstyle, VT_STRING, buf)) {
            for (i = 0; i < NumGrTypes; i++) {
                if (eq(buf, ptypes[i].name)) {
                    grs.ptype = ptypes[i].type;
                    break;
                }
            }
            if (i == NumPlTypes) {
                fprintf(cp_err, "Warning: strange plot type %s\n", buf);
                grs.ptype = PLOT_LIN;
            }
            *pfound = true;
        }
        else
            grs.ptype = PLOT_LIN;
    }

    s = getword(pwl, "xlabel");
    if (!*sameflag || (!grs.xlabel && s)) {
        txfree(grs.xlabel);
        grs.xlabel = s;
    }
    else
        txfree(s);

    s = getword(pwl, "ylabel");
    if (!*sameflag || (!grs.ylabel && s)) {
        txfree(grs.ylabel);
        grs.ylabel = s;
    }
    else
        txfree(s);

    s = getword(pwl, "title");
    if (!*sameflag || (!grs.title && s)) {
        txfree(grs.title);
        grs.title = s;
    }
    else
        txfree(s);


    if (!*sameflag)
        grs.nointerp = getflag(pwl, "nointerp");
    else if (getflag(pwl, "nointerp"))
        grs.nointerp = true;
}


static bool
fixlimits(dl0,xlims,ylims,oneval)

struct dvlist *dl0;
double *xlims,*ylims;
bool oneval;
{
    double *dd, tt, mx, my, rad;
    struct dvlist *dl;
    struct dvec *d;
    
    /* Figure out the proper x- and y-axis limits. */
    if (grs.ylimset) {
        if (grs.gtype == GRID_SMITH) {
            if (grs.xlimset) {
                SMITH_tfm(grs.xlim[0], grs.ylim[0], &tt, &ylims[0]);
                SMITH_tfm(grs.xlim[1], grs.ylim[1], &tt, &ylims[1]);
            }
            else {
                SMITH_tfm(0.0, grs.ylim[0], &tt, &ylims[0]);
                SMITH_tfm(0.0, grs.ylim[1], &tt, &ylims[1]);
            }
        }
        else {
            ylims[0] = grs.ylim[0];
            ylims[1] = grs.ylim[1];
        }
    }
    else if (oneval) {
        ylims[0] = 1;
        ylims[1] = -1;
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (grs.gtype == GRID_SMITH) {
                dd = ft_SMITHminmax(d, true);
                if (dd[0] < 0.0)
                    dd[0] *= 1.1;
                else
                    dd[0] *= 0.9;
                if (dd[1] >= 0.0)
                    dd[1] *= 1.1;
                else
                    dd[1] *= 0.9;
            }
            else
                dd = ft_minmax(d, true);

            if (ylims[0] > ylims[1]) {
                ylims[0] = dd[0];
                ylims[1] = dd[1];
            }
            else {
                if (dd[0] < ylims[0]) ylims[0] = dd[0];
                if (dd[1] > ylims[1]) ylims[1] = dd[1];
            }
        }
    }
    else {
        ylims[0] = 1;
        ylims[1] = -1;
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            dd = ft_minmax(d, true);
            if (ylims[0] > ylims[1]) {
                ylims[0] = dd[0];
                ylims[1] = dd[1];
            }
            else {
                if (dd[0] < ylims[0]) ylims[0] = dd[0];
                if (dd[1] > ylims[1]) ylims[1] = dd[1];
            }
        }

        /* XXX */
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_flags & VF_MINGIVEN)
            if (ylims[0] < d->v_minsignal)
                ylims[0] = d->v_minsignal;
            if (d->v_flags & VF_MAXGIVEN)
            if (ylims[1] > d->v_maxsignal)
                ylims[1] = d->v_maxsignal;
        }
    }

    if (grs.xlimset) {
        if (grs.gtype == GRID_SMITH) {
            if (grs.ylimset) {
                SMITH_tfm(grs.xlim[0], grs.ylim[0], &xlims[0], &tt);
                SMITH_tfm(grs.xlim[1], grs.ylim[1], &xlims[1], &tt);
            }
            else {
                SMITH_tfm(grs.xlim[0], 0.0, &xlims[0], &tt);
                SMITH_tfm(grs.xlim[1], 0.0, &xlims[1], &tt);
            }
        }
        else {
            xlims[0] = grs.xlim[0];
            xlims[1] = grs.xlim[1];
        }
    }
    else if (oneval) {
        xlims[0] = 1;
        xlims[1] = -1;
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (grs.gtype == GRID_SMITH) {
                dd = ft_SMITHminmax(d, false);
                if (dd[0] < 0.0)
                    dd[0] *= 1.1;
                else
                    dd[0] *= 0.9;
                if (dd[1] >= 0.0)
                    dd[1] *= 1.1;
                else
                    dd[1] *= 0.9;
            }
            else
                dd = ft_minmax(d, false);

            if (xlims[0] > xlims[1]) {
                xlims[0] = dd[0];
                xlims[1] = dd[1];
            }
            else {
                if (dd[0] < xlims[0]) xlims[0] = dd[0];
                if (dd[1] > xlims[1]) xlims[1] = dd[1];
            }
        }
    }
    else {
        xlims[0] = 1;
        xlims[1] = -1;
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            dd = ft_minmax(d->v_scale, true);
            if (xlims[0] > xlims[1]) {
                xlims[0] = dd[0];
                xlims[1] = dd[1];
            }
            else {
                if (dd[0] < xlims[0]) xlims[0] = dd[0];
                if (dd[1] > xlims[1]) xlims[1] = dd[1];
            }
        }
        for (dl = dl0; dl; dl = dl->dl_next) {
            d = dl->dl_dvec;
            if (d->v_scale->v_flags & VF_MINGIVEN)
                if (xlims[0] < d->v_scale->v_minsignal)
                    xlims[0] = d->v_scale->v_minsignal;
            if (d->v_scale->v_flags & VF_MAXGIVEN)
                if (xlims[1] > d->v_scale->v_maxsignal)
                    xlims[1] = d->v_scale->v_maxsignal;
        }
    }

    /* Do some coercion of the limits to make them reasonable. */
    if ((xlims[0] == 0) && (xlims[1] == 0)) {
        xlims[0] = -1.0;
        xlims[1] = 1.0;
    }
    if ((ylims[0] == 0) && (ylims[1] == 0)) {
        ylims[0] = -1.0;
        ylims[1] = 1.0;
    }
    if (xlims[0] > xlims[1]) {
        tt = xlims[1];
        xlims[1] = xlims[0];
        xlims[0] = tt;
    }
    if (ylims[0] > ylims[1]) {
        tt = ylims[1];
        ylims[1] = ylims[0];
        ylims[0] = tt;
    }
    if (xlims[0] == xlims[1]) {
        xlims[0] *= (xlims[0] > 0) ? 0.9 : 1.1;
        xlims[1] *= (xlims[1] > 0) ? 1.1 : 0.9;
    }
    if (ylims[0] == ylims[1]
        || fabs(ylims[0])/(ylims[1]-ylims[0]) > 1.0e9
        || fabs(ylims[1])/(ylims[1]-ylims[0]) > 1.0e9) {
        ylims[0] *= (ylims[0] > 0) ? 0.9 : 1.1;
        ylims[1] *= (ylims[1] > 0) ? 1.1 : 0.9;
    }

    /* Now shrink the limits very slightly -- this helps prevent round-off
     * error from doing bad things.
     */
    if (grs.gtype != GRID_LOGLOG && grs.gtype != GRID_XLOG
            && grs.gtype != GRID_POLAR && grs.gtype != GRID_SMITH) {
        tt = xlims[1] - xlims[0];
        xlims[0] += tt * 0.001;
        xlims[1] -= tt * 0.001;
    }
    if (grs.gtype != GRID_LOGLOG && grs.gtype != GRID_YLOG
            && grs.gtype != GRID_POLAR && grs.gtype != GRID_SMITH) {
        tt = ylims[1] - ylims[0];
        ylims[0] += tt * 0.001;
        ylims[1] -= tt * 0.001;
    }

    if ((xlims[0] < 0.0) && ((grs.gtype == GRID_XLOG) ||
            (grs.gtype == GRID_LOGLOG))) {
        fprintf(cp_err, 
            "Error: X values must be >= 0 for log scale\n");
        return (false);
    }
    if ((ylims[0] < 0.0) && ((grs.gtype == GRID_YLOG) ||
            (grs.gtype == GRID_LOGLOG))) {
        fprintf(cp_err, 
            "Error: Y values must be >= 0 for log scale\n");
        return (false);
    }

    /* Fix the plot limits for smith and polar grids. */
    if (!grs.xlimset && !grs.ylimset && (grs.gtype == GRID_POLAR)) {
        /* (0,0) must be in the center of the screen. */
        mx = (fabs(xlims[0]) > fabs(xlims[1])) ? fabs(xlims[0]) :
                fabs(xlims[1]);
        my = (fabs(ylims[0]) > fabs(ylims[1])) ? fabs(ylims[0]) :
                fabs(ylims[1]);
        rad = (mx > my) ? mx : my;
        /* rad = sqrt(mx * mx + my * my); */
        xlims[0] = - rad;
        xlims[1] = rad;
        ylims[0] = - rad;
        ylims[1] = rad;
    }
    else if (!grs.xlimset && !grs.ylimset && (grs.gtype == GRID_SMITH)) {
        mx = (fabs(xlims[0]) > fabs(xlims[1])) ? fabs(xlims[0]) :
                fabs(xlims[1]);
        my = (fabs(ylims[0]) > fabs(ylims[1])) ? fabs(ylims[0]) :
                fabs(ylims[1]);
        rad = (mx > my) ? mx : my;
#ifdef notdef
        /* XXX */
        xlims[0] = - rad;
        xlims[1] = rad;
        ylims[0] = - rad;
        ylims[1] = rad;
#endif
        xlims[0] = -1.0;
        xlims[1] = 1.0;
        ylims[0] = -1.0;
        ylims[1] = 1.0;
    }
    return (true);
}


static double *
getlims(wl, name, number)

/* This routine gets parameters from the command line, which are of the
 * form "name number ..."  It returns static data.  number <= 2.
 */
wordlist **wl;
char *name;
int number;
{
    double *td;
    wordlist *beg, *wk;
    char *ss;
    int n;
    static double d[2];

    for (beg = *wl; beg; beg = beg->wl_next) {
        if (eq(beg->wl_word, name)) {
            wk = beg;
            if (number) {
                for (n = 0; n < number; n++) {
                    wk = wk->wl_next;
                    if (!wk) {
                        fprintf(cp_err,
                            "Syntax error.\n");
                        return (NULL);
                    }
                    ss = wk->wl_word;
                    td = ft_numparse(&ss, false);
                    if (td == NULL)
                        goto bad;
                    d[n] = *td;
                }
            }
            else
                d[0] = 1.0;

            if (beg->wl_prev)
                beg->wl_prev->wl_next = wk->wl_next;
            if (wk->wl_next)
                wk->wl_next->wl_prev = beg->wl_prev;
            if (beg == *wl)
                *wl = wk->wl_next;
            wk->wl_next = NULL;
            wl_free(beg);
            return (d);
        }
    }
    return (NULL);
bad:
    fprintf(cp_err, "Error: bad %s parameters.\n", name);
    return (NULL);
}


static char *
getword(wl, name)

/* Return a parameter of the form "xlabel foo" */
wordlist **wl;
char *name;
{
    wordlist *beg;
    char *s;

    for (beg = *wl; beg; beg = beg->wl_next) {
        if (eq(beg->wl_word, name)) {
            if (!beg->wl_next) {
                fprintf(cp_err, "Syntax error.\n");
                return (NULL);
            }
            s = copy(beg->wl_next->wl_word);
            if (beg->wl_prev)
                beg->wl_prev->wl_next = beg->wl_next->wl_next;
            if (beg->wl_next->wl_next)
                beg->wl_next->wl_next->wl_prev = beg->wl_prev;
            if (beg == *wl)
                *wl = beg->wl_next->wl_next;
            beg->wl_next->wl_next = NULL;
            wl_free(beg);
            return (s);
        }
    }
    return (NULL);
}


static bool
getflag(wl, name)

/* Check for and remove a one-word keyword. */
wordlist **wl;
char *name;
{
    wordlist *beg;

    for (beg = *wl; beg; beg = beg->wl_next) {
        if (eq(beg->wl_word, name)) {
            if (beg->wl_prev)
                beg->wl_prev->wl_next = beg->wl_next;
            if (beg->wl_next)
                beg->wl_next->wl_prev = beg->wl_prev;
            if (beg = *wl)
                *wl = beg->wl_next;
            txfree(beg->wl_word);
            txfree((char*)beg);
            return (true);
        }
    }
    return (false);
}


void
plot_extend(v, length)

/* Extend a data vector to length by replicating the
 * last element, or truncate it if it is too long.
 */
struct dvec *v;
int length;
{
    int i;
    complex c, *oc;
    double d, *od;

    if (v->v_length == length)
        return;
    if (v->v_length > length) {
        v->v_length = length;
        return;
    }
    if (isreal(v)) {
        od = v->v_realdata;
        v->v_realdata = (double *) tmalloc(length * sizeof (double));
        for (i = 0; i < v->v_length; i++)
            v->v_realdata[i] = od[i];
        d = od[--i];
        while (i < length)
            v->v_realdata[i++] = d;
        tfree(od);
    }
    else {
        oc = v->v_compdata;
        v->v_compdata = (complex *) tmalloc(length * sizeof (complex));
        for (i = 0; i < v->v_length; i++) {
            realpart(&v->v_compdata[i]) = realpart(&oc[i]);
            imagpart(&v->v_compdata[i]) = imagpart(&oc[i]);
        }
        realpart(&c) = realpart(&oc[--i]);
        imagpart(&c) = imagpart(&oc[i]);
        while (i < length) {
            realpart(&v->v_compdata[i]) = realpart(&c);
            imagpart(&v->v_compdata[i++]) = imagpart(&c);
            tfree(oc);
        }
    }
    v->v_length = length;
    return;
}


static void
compress(d, xcomp, xind)

/* Collapse every *xcomp elements into one, and use only the elements
 * between xind[0] and xind[1].  Do this for each block of a multi-
 * dimensional vector.
 */
struct dvec *d;
double *xcomp, *xind;
{
    int cfac, ihi, ilo, newlen, i, j, blocks, bsize;
    double *dd;
    complex *cc;

    if (d->v_numdims > 1) {
        bsize = d->v_dims[d->v_numdims-1];
        blocks = d->v_length/bsize;
    }
    else {
        bsize = d->v_length;
        blocks = 1;
    }

    if (xind) {
        ilo = (int) xind[0];
        ihi = (int) xind[1];

        if ((ilo <= ihi) && (ilo > 0) && (ilo < bsize) &&
                (ihi > 1) && (ihi <= bsize)) {

            i = d->v_length - blocks*bsize;
            if (i > ihi)
                blocks++;
            if (blocks) {
                newlen = ihi - ilo;
                if (isreal(d)) {
                    double *src, *dst;

                    dd = (double *) tmalloc(newlen*blocks * sizeof(double));
                    dst = dd;
                    src = d->v_realdata + ilo;
                    for (i = 0; i < blocks; i++) {
                        DCOPY(src, dst, newlen);
                        src += bsize;
                        dst += newlen;
                    }
                    txfree((char*)d->v_realdata);
                    d->v_realdata = dd;
                }
                else {
                    complex *src, *dst;

                    cc = (complex *) tmalloc(newlen*blocks * sizeof(complex));
                    dst = cc;
                    src = d->v_compdata + ilo;
                    for (i = 0; i < blocks; i++) {
                        CCOPY(src, dst, newlen);
                        src += bsize;
                        dst += newlen;
                    }
                    txfree((char*)d->v_compdata);
                    d->v_compdata = cc;
                }
                d->v_length = newlen*blocks;
                if (d->v_numdims <= 1) {
                    d->v_numdims = 1;
                    d->v_dims[0] = newlen;
                }
                else
                    d->v_dims[d->v_numdims-1] = newlen;
            }
        }
    }

    if (xcomp) {
        cfac = (int) *xcomp;
        if ((cfac > 1) && (cfac < bsize)) {
            newlen = bsize/cfac;
            for (j = 0; j < blocks; j++) {
                for (i = 0; i < newlen; i++)
                    if (isreal(d))
                        d->v_realdata[i + j*newlen] = 
                            d->v_realdata[i*cfac + j*bsize];
                    else
                        d->v_compdata[i + j*newlen] = 
                            d->v_compdata[i*cfac + j*bsize];
            }
            d->v_length = blocks*newlen;
            if (d->v_numdims <= 1)
                d->v_dims[0] = d->v_length;
            else
                d->v_dims[d->v_numdims-1] = d->v_length;
        }
    }
    return;
}



#define XG_MAXVECTORS 64


void
ft_xgraph(dl0,grp)

/*
 * Xgraph-11 plots.
 */
struct dvlist *dl0;
char *grp;
{
    struct grinit *gr = (struct grinit*)grp;
    FILE *file;
    struct dvec *v, *scale;
    struct dvlist *dl;
    double xval, yval;
    int i, numVecs, linewidth;
    bool xlog, ylog, nogrid, markers;
    char buf[BSIZE_SP], *text;
    extern char *kw_xglinewidth, *kw_xgmarkers;

    /* Sanity checking. */
    for (dl = dl0, numVecs = 0; dl; dl = dl->dl_next) {
        numVecs++;
    }
    if (numVecs == 0) {
        return;
    }
    else if (numVecs > XG_MAXVECTORS) {
        fprintf( cp_err, "Error: too many vectors for Xgraph.\n" );
        return;
    }
    if (!cp_getvar(kw_xglinewidth, VT_NUM, (char*)&linewidth))
        linewidth = 1;
    if (linewidth < 1) linewidth = 1;

    if (!cp_getvar(kw_xgmarkers, VT_BOOL, (char*)&markers)) {
        markers = false;
    }

    /* Make sure the gridtype is supported. */
    switch (gr->gridtype) {
    case GRID_LIN:
        nogrid = xlog = ylog = false;
        break;
    case GRID_XLOG:
        xlog = true;
        nogrid = ylog = false;
        break;
    case GRID_YLOG:
        ylog = true;
        nogrid = xlog = false;
        break;
    case GRID_LOGLOG:
        xlog = ylog = true;
        nogrid = false;
        break;
    case GRID_NONE:
    for (dl = dl0, numVecs = 0; dl; dl = dl->dl_next)
        nogrid = true;
        xlog = ylog = false;
        break;
    default:
        fprintf( cp_err, "Error: grid type unsupported by Xgraph.\n" );
        return;
    }

    /* Open the output file. */
    if (!(file = fopen(gr->hcopy, "w"))) {
        perror(gr->hcopy);
        return;
    }

    /* Set up the file header. */
    if (gr->plotname)
        fprintf( file, "TitleText: %s\n", gr->plotname );
    if (gr->xlabel)
        fprintf( file, "XUnitText: %s\n", gr->xlabel );
    if (gr->ylabel) {
        fprintf( file, "YUnitText: %s\n", gr->ylabel );
    }
    if (nogrid) {
        fprintf( file, "Ticks: True\n" );
    }
    if (xlog) {
        fprintf( file, "LogX: True\n" );
        if (gr->xlims) {
            fprintf( file, "XLowLimit:  % e\n", log10(gr->xlims[0]) );
            fprintf( file, "XHighLimit: % e\n", log10(gr->xlims[1]) );
        }
    }
    else {
        if (gr->xlims) {
            fprintf( file, "XLowLimit:  % e\n", gr->xlims[0] );
            fprintf( file, "XHighLimit: % e\n", gr->xlims[1] );
        }
    }
    if (ylog) {
        fprintf( file, "LogY: True\n" );
        if (gr->ylims) {
            fprintf( file, "YLowLimit:  % e\n", log10(gr->ylims[0]) );
            fprintf( file, "YHighLimit: % e\n", log10(gr->ylims[1]) );
        }
    }
    else {
        if (gr->ylims) {
            fprintf( file, "YLowLimit:  % e\n", gr->ylims[0] );
            fprintf( file, "YHighLimit: % e\n", gr->ylims[1] );
        }
    }
    fprintf( file, "LineWidth: %d\n", linewidth );
    fprintf( file, "BoundBox: True\n" );
    if (gr->plottype == PLOT_COMB) {
        fprintf( file, "BarGraph: True\n" );
        fprintf( file, "NoLines: True\n" );
    }
    else if (gr->plottype == PLOT_POINT) {
        if (markers) {
            fprintf( file, "Markers: True\n" );
        }
        else {
            fprintf( file, "LargePixels: True\n" );
        }
        fprintf( file, "NoLines: True\n" );
    }

    /* Write out the data. */
    for (dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;
        scale = v->v_scale;
        if (v->v_name) {
            fprintf( file, "\"%s\"\n", v->v_name );
        }
        for ( i = 0; i < scale->v_length; i++ ) {
            xval = isreal(scale) ?
                scale->v_realdata[i] : realpart(&scale->v_compdata[i]);
            yval = isreal(v) ?
                v->v_realdata[i] : realpart(&v->v_compdata[i]);
            fprintf( file, "% e % e\n", xval, yval );
        }
        fprintf( file, "\n" );
    }
    (void) fclose( file );
    (void) sprintf( buf, "xgraph %s &", gr->hcopy );
    (void) system( buf );
    return;
}
