/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jeffrey M. Hsu
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 *  "plot" command graphics.
 */

#include "spice.h"
#include "ftedefs.h"
#include "plotdefs.h"

/* invariant:  currentgraph contains the current graph */

/* These are what gets plotted as points when you specify point plots */
static char pointchars[128];
#define DEFPOINTCHARS   "oxabcdefhgijklmnpqrstuvwyz"

#ifdef __STDC__
extern void plottrace(struct dvec*,struct dvec*,int);
static void do_plot_trace(struct dvec*,struct dvec*,int);
static void plotinterval(struct dvec*,double,double,double*,int,bool);
static GRDATA copydveclist(GRDATA);
static void freedveclist(GRDATA);
static void initdvecs(GRAPH*);
static void drawlegend(int,struct dvec*);
static void resize(GRAPH*);
static void redraw(GRAPH*);
static void restoretext(GRAPH*);
#else
extern void plottrace();
static void do_plot_trace();
static void plotinterval();
static GRDATA copydveclist();
static void freedveclist();
static void initdvecs();
static void drawlegend();
static void resize();
static void redraw();
static void restoretext();
#endif


/* ARGSUSED */
GRAPH *
gr_init(dl0,grp)

/*
 *  Start of a new graph.
 *  Fill in the data that gets displayed.
 *  Difference from old gr_init:
 *    we don't try to determine the look of the screen from here
 *    leave to lower level routines
 *
 */
struct dvlist *dl0;             /* List of dvecs to plot. */
char *grp;
{
    struct grinit *gr = (struct grinit*)grp;
    GRAPH *graph;
    int b;
    char *comb_title;
    struct dvec *d;
    extern char *kw_polydegree, *kw_pointchars, *kw_ticmarks;

    if (!gr->xlims || !gr->ylims) {
        internalerror("gr_init:  no range specified");
        return (NULL);
    }

    if (!(graph = NewGraph())) {
        return (NULL);
    }

    graph->graphtype = GR_PLOT;
    graph->onevalue = (gr->xname ? false : true);

    /* communicate filename to hardcopy driver */
    if (gr->hcopy) {
        graph->devdep = gr->hcopy;
    }

    if (!cp_getvar(kw_polydegree, VT_NUM, (char *) &graph->degree))
        graph->degree = 1;

    if (!cp_getvar(kw_pointchars, VT_STRING, pointchars))
        (void) strcpy(pointchars, DEFPOINTCHARS);

    if (!cp_getvar(kw_ticmarks, VT_NUM, (char *) &graph->ticmarks)) {
        if (cp_getvar(kw_ticmarks, VT_BOOL, (char *) &b))
            graph->ticmarks = 10;
        else
            graph->ticmarks = 0;
    }

    /* set upper and lower limits */
    graph->datawindow.xmin = gr->xlims[0];
    graph->datawindow.xmax = gr->xlims[1];
    graph->datawindow.ymin = gr->ylims[0];
    graph->datawindow.ymax = gr->ylims[1];

    graph->grid.gridtype = gr->gridtype;
    graph->plottype = gr->plottype;
    graph->grid.xdatatype = gr->xtype;
    graph->grid.ydatatype = gr->ytype;
    graph->grid.xdelta = gr->xdelta;
    graph->grid.ydelta = gr->ydelta;

    if (!graph->onevalue) {
        if (gr->xlabel) {
            graph->grid.xlabel = copy(gr->xlabel);
        }
        else {
            graph->grid.xlabel = copy(gr->xname);
        }
        if (gr->ylabel) {
            graph->grid.ylabel = copy(gr->ylabel);
        }
    }
    else {
        if (gr->xlabel) {
            graph->grid.xlabel = copy(gr->xlabel);
        }
        else {
            graph->grid.xlabel = copy("real");
        }
        if (gr->ylabel) {
            graph->grid.ylabel = copy(gr->ylabel);
        }
        else {
            graph->grid.ylabel = copy("imag");
        }
    }

    if (!gr->pname)
        gr->pname = "(unknown)";
    if (!gr->plotname)
        gr->plotname = "(unknown)";
    comb_title = tmalloc(strlen(gr->plotname) + strlen(gr->pname) + 3);
    sprintf(comb_title, "%s: %s", gr->pname, gr->plotname);
    graph->plotname = comb_title;
    graph->commandline = copy(gr->commandline);

    graph->destroy = freedveclist;
    graph->copydata = copydveclist;
    graph->redraw = redraw;

    graph->plotdata = (GRDATA)dl0;
    initdvecs(graph);

    if (DevNewViewport(graph) == 1) {
        fprintf(cp_err, "Can't open viewport for graphics.\n");
        DestroyGraph(graph->graphid);
        return (NULL);
    }

    if (!dispdev->windows) {    
        redraw(graph);
    }
    return (graph);
}


void
gr_end()

/* end a plot */
{
    char buf[BSIZE_SP];
    char *text = "Enter p for hardcopy, return to continue";
    char *text1 = "Hit p for hardcopy, any other key to continue";
    char c;
    int x, y;
    REQUEST request;
    RESPONSE response;
    extern char *kw_device;

    buf[0] = 0;
    DevUpdate();
    if (!dispdev->windows && !dispdev->hardcopy) {

        if (cp_getvar(kw_device, VT_STRING, buf)
                && prefix("/dev/tty", buf)) {
            fprintf(cp_err, "%s", text);
            fflush(cp_err);
            c = getchar();
        }
        else {
            x = (currentgraph->absolute.width -
                strlen(text1)*currentgraph->fontwidth)/2;
            if (x < 0) x = 0;
            y = currentgraph->absolute.height - currentgraph->fontheight - 1;
            DevSetColor(8);
            DevText(text1,x,y);
            request.option = char_option;
            request.fp = (FILE*)0;
            DevInput(&request, &response);
            c = response.reply.ch;
            DevSetColor(0);
            DevBox(x,y,x + strlen(text1)*currentgraph->fontwidth,
                y + currentgraph->fontheight);
            DevUpdate();
        }
        if (c == 'p') {
            ft_hardcopy(NULL,currentgraph,true,NULL,false);
        }
    }
}


void
plottrace(v, xs, plotno)

struct dvec *v, *xs;
int plotno;
{
    int i, blsize, nblks, size, len;
    char *sdata, *vdata;

    /* Take care of multi-dimensional vectors here.  We know that the
     * vectors have been coerced to the same data length.  The scale
     * sets the block length, and we plot according to the dimensionality
     * of the vector, block by block.
     */

    if (!xs || xs->v_numdims <= 1) {
        /* if the scale has unit dimension, ignore dimensionality
         * of vector.
         */
        do_plot_trace(v,xs,plotno);
        return;
    }

    blsize = xs->v_dims[xs->v_numdims-1];
    nblks = 1;
    for (i = 0; i < v->v_numdims - 1; i++)
        nblks *= v->v_dims[i];

    sdata = (isreal(xs) ? (char*)xs->v_realdata : (char*)xs->v_compdata);
    vdata = (isreal(v) ? (char*)v->v_realdata : (char*)v->v_compdata);
    len = xs->v_length;
    xs->v_length = blsize;
    v->v_length = blsize;

    size = 0;
    for (i = 0; i < nblks; i++) {
        size += blsize;
        if (size > len) {
            /* unfinished block */
            blsize = len - (size - blsize);
            xs->v_length = blsize;
            v->v_length = blsize;
            do_plot_trace(v,xs,plotno);
            break;
        }
        do_plot_trace(v,xs,plotno);
        
        if (isreal(xs))
            xs->v_realdata += blsize;
        else
            xs->v_compdata += blsize;
        if (isreal(v))
            v->v_realdata += blsize;
        else
            v->v_compdata += blsize;
    }
    xs->v_length = len;
    if (isreal(xs))
        xs->v_realdata = (double*) sdata;
    else
        xs->v_compdata = (complex*) sdata;
    v->v_length = len;
    if (isreal(v))
        v->v_realdata = (double*) vdata;
    else
        v->v_compdata = (complex*) vdata;
}


static void
do_plot_trace(v, xs, plotno)

/* Plot the vector v, with scale xs.  If we are doing curve-fitting, then
 * do some tricky stuff.
 */
struct dvec *v, *xs;
int plotno;
{
    int degree, gridsize, length;
    int i, j, l;
    double *scratch, *result, *gridbuf, *mm;
    double *xdata, *ydata;
    bool rot, increasing;
    double dx, dy, lx = 0, ly = 0;
    extern char *kw_gridsize;

    degree = currentgraph->degree;

    drawlegend(plotno, v);

    if (degree > v->v_length)
        degree = v->v_length;
    if (degree < 1) {
        fprintf(cp_err, "Error: polydegree is %d, can't plot...\n",
                degree);
        return;
    }

    if (!cp_getvar(kw_gridsize, VT_NUM, (char *) &gridsize))
        gridsize = 0;
    if ((gridsize < 0) || (gridsize > 10000)) {
        fprintf(cp_err, "Error: bad grid size %d\n", gridsize);
        return;
    }
    if (gridsize && xs) {
        if (isreal(xs)) {
            increasing = (xs->v_realdata[0] < xs->v_realdata[1]);
        for (i = 0; i < xs->v_length - 1; i++)
            if (increasing != (xs->v_realdata[i] < 
                    xs->v_realdata[i + 1])) {
                fprintf(cp_err, 
        "Warning: scale not monotonic, gridsize not relevant.\n");
                gridsize = 0;
                break;
            }
        }
        else {
            increasing = (realpart (&xs->v_compdata[0]) < 
                realpart( &xs->v_compdata[1]));
        for (i = 0; i < xs->v_length - 1; i++)
            if (increasing != (realpart( &xs->v_compdata[i]) < 
                            realpart( &xs->v_compdata[i + 1]))) {
                fprintf(cp_err, 
        "Warning: scale not monotonic, gridsize not relevant.\n");
                gridsize = 0;
                break;
            }
        }
    }

    /* Do the one value case */

    if (!xs) {
        for (i = 0; i < v->v_length; i++) {

            /* We should do the one - point case too!
             *    Important for polar plots of pole-zero for example
             */
            if (v->v_length == 1) {
                j = 0;
            }
            else {
                j = i-1;
                if (i == 0)
                    continue;
            }

            if (isreal(v)) {
                gr_point(v, v->v_realdata[i],
                    0.0,
                    v->v_realdata[j],
                    0.0, (j==i ? 1 : i));
            }
            else {
                gr_point(v, realpart(&v->v_compdata[i]),
                    imagpart(&v->v_compdata[i]),
                    realpart(&v->v_compdata[j]),
                    imagpart(&v->v_compdata[j]), (j==i ? 1 : i));
            }
        }
        DevUpdate();
        return;
    }

    xs->v_flags |= VF_PERMANENT;

    /* First check the simple case, where we don't have to do any 
     * interpolation.
     */
    if ((degree == 1) && (gridsize == 0)) {
        for (i = 0, j = v->v_length; i < j; i++) {
            dx = isreal(xs) ? xs->v_realdata[i] : 
                    realpart(&xs->v_compdata[i]);
            dy = isreal(v) ? v->v_realdata[i] : 
                    realpart(&v->v_compdata[i]);
            gr_point(v, dx, dy, lx, ly, i);
            lx = dx;
            ly = dy;
        }
        if (v->v_length == 1)
            gr_point(v, dx, dy, lx, ly, 1);
        DevUpdate();
        return;
    }

    if (gridsize < degree + 1)
        gridsize = 0;

    if (gridsize) {
        /* This is done quite differently from what we do below... */
        gridbuf = (double *) tmalloc(gridsize * sizeof (double));
        result = (double *) tmalloc(gridsize * sizeof (double));
        if (isreal(v))
            ydata = v->v_realdata;
        else {
            ydata = (double *) tmalloc(v->v_length * 
                    sizeof (double));
            for (i = 0; i < v->v_length; i++)
                ydata[i] = realpart(&v->v_compdata[i]);
        }
        if (isreal(xs))
            xdata = xs->v_realdata;
        else {
            xdata = (double *) tmalloc(xs->v_length * 
                    sizeof (double));
            for (i = 0; i < xs->v_length; i++)
                xdata[i] = realpart(&xs->v_compdata[i]);
        }
        
        mm = ft_minmax(xs, true);
        dx = (mm[1] - mm[0]) / gridsize;
        if (increasing)
            for (i = 0, dy = mm[0]; i < gridsize; i++, dy += dx)
                gridbuf[i] = dy;
        else
            for (i = 0, dy = mm[1]; i < gridsize; i++, dy -= dx)
                gridbuf[i] = dy;
        if (!ft_interpolate(ydata, result, xdata, v->v_length, gridbuf,
                gridsize, degree)) {
            fprintf(cp_err, "Error: can't put %s on gridsize %d\n",
                    v->v_name, gridsize);
            return;
        }
        /* Now this is a problem.  There's no way that we can
         * figure out where to put the tic marks to correspond with
         * the actual data...
         */
        for (i = 0; i < gridsize; i++)
            gr_point(v, gridbuf[i], result[i], gridbuf[i ? (i - 1)
                    : i], result[i ? (i - 1) : i], -1);
        DevUpdate();
        tfree(gridbuf);
        tfree(result);
        if (!isreal(v))
            tfree(ydata);
        if (!isreal(xs))
            tfree(xdata);
        return;
    }

    /* We need to do curve fitting now. First get some scratch 
     * space
     */
    scratch = (double *) tmalloc((degree + 1) * (degree + 2) * 
            sizeof (double));
    result = (double *) tmalloc((degree + 1) * sizeof (double));
    xdata = (double *) tmalloc((degree + 1) * sizeof (double));
    ydata = (double *) tmalloc((degree + 1) * sizeof (double));


    /* Plot the first degree segments... */
    if (isreal(v)) {
        DCOPY(v->v_realdata, ydata, degree + 1);
    }
    else
        for (i = 0; i <= degree; i++)
            ydata[i] = realpart(&v->v_compdata[i]);
    if (isreal(xs)) {
        DCOPY(xs->v_realdata, xdata, degree + 1);
    }
    else
        for (i = 0; i <= degree; i++)
            xdata[i] = realpart(&xs->v_compdata[i]);

    rot = false;
    while (!ft_polyfit(xdata, ydata, result, degree, scratch)) {
        /* Rotate the coordinate system 90 degrees and try again.
         * If it doesn't work this time, bump the interpolation
         * degree down by one...
         */
        if (ft_polyfit(ydata, xdata, result, degree, scratch)) {
            rot = true;
            break;
        }
        if (--degree == 0) {
            fprintf(cp_err, "plotcurve: Internal Error: ack...\n");
            return;
        }
    }

    /* Plot this part of the curve... */
    for (i = 0; i < degree; i++)
        if (rot)
            plotinterval(v, ydata[i], ydata[i + 1], result, degree,
                    true);
        else
            plotinterval(v, xdata[i], xdata[i + 1], result, degree,
                    false);

    /* Now plot the rest, piece by piece... l is the 
     * last element under consideration.
     */
    length = v->v_length;
    for (l = degree + 1; l < length; l++) {

        /* Shift the old stuff by one and get another value. */
        for (i = 0; i < degree; i++) {
            xdata[i] = xdata[i + 1];
            ydata[i] = ydata[i + 1];
        }
        if (isreal(v))
            ydata[i] = v->v_realdata[l];
        else
            ydata[i] = realpart(&v->v_compdata[l]);
        if (isreal(xs))
            xdata[i] = xs->v_realdata[l];
        else
            xdata[i] = realpart(&xs->v_compdata[l]);

        rot = false;
        while (!ft_polyfit(xdata, ydata, result, degree, scratch)) {
            if (ft_polyfit(ydata, xdata, result, degree, scratch)) {
                rot = true;
                break;
            }
            if (--degree == 0) {
                fprintf(cp_err, 
                    "plotcurve: Internal Error: ack...\n");
                return;
            }
        }
        if (rot)
            plotinterval(v, ydata[degree - 1], ydata[degree], 
                    result, degree, true);
        else
            plotinterval(v, xdata[degree - 1], xdata[degree],
                    result, degree, false);
    }
    tfree(scratch);
    tfree(xdata);
    tfree(ydata);
    tfree(result);
    DevUpdate();
    return;
}


#define GRANULARITY 10

static void
plotinterval(v, lo, hi, coeffs, degree, rotated)

struct dvec *v;
double lo, hi;
double *coeffs;
int degree;
bool rotated;
{
    double incr, dx, dy, lx, ly;
    register int i;
    int steps;
    extern char *kw_polysteps;

    /*
    fprintf(cp_err, "plotinterval(%s, %G, %G, [ ", v->v_name, lo, hi);
    for (i = 0; i <= degree; i++)
        fprintf(cp_err, "%G ", coeffs[i]);
    fprintf(cp_err, "], %d, %s)\n\r", degree, rotated ? "true" : "false");
    */

    /* This is a problem -- how do we know what granularity to use?  If
     * the guy cares about this he will use gridsize.
     */
    if (!cp_getvar(kw_polysteps, VT_NUM, (char *) &steps))
        steps = GRANULARITY;

    incr = (hi - lo) / (double) (steps + 1);
    dx = lo + incr;
    lx = lo;
    ly = ft_peval(lo, coeffs, degree);
    for (i = 0; i <= steps; i++, dx += incr) {
        dy = ft_peval(dx, coeffs, degree);
        if (rotated)
            gr_point(v, dy, dx, ly, lx, -1);
        else
            gr_point(v, dx, dy, lx, ly, -1);
        lx = dx;
        ly = dy;
        /* fprintf(cp_err, "plot (%G, %G)\n\r", dx, dy); */
    }
    return;
}


void
gr_point(dv, newx, newy, oldx, oldy, np)

/*
 *  Add a point to the curve we're currently drawing.
 */
struct dvec *dv;
double newx, newy, oldx, oldy;
int np;
{
    int oldtox, oldtoy;     /* value before clipping */

    char pointc[2];

    int fromx, fromy, tox, toy;
    int ymin, dummy;

    /* If it's a linear plot, ignore first point since we don't want
        to connect with oldx and oldy. */
    if (dv->v_plottype == PLOT_LIN && !np) return;

    DevDatatoScreen(currentgraph, oldx, oldy, &fromx, &fromy);
    DevDatatoScreen(currentgraph, newx, newy, &tox, &toy);

/* note: we do not particularly want to clip here */
    oldtox = tox; oldtoy = toy;
    if (!currentgraph->grid.circular) {
        if (clip_line(&fromx, &fromy, &tox, &toy,
            currentgraph->viewportxoff, currentgraph->viewportyoff,
            currentgraph->viewport.width + currentgraph->viewportxoff,
            currentgraph->viewport.height + currentgraph->viewportyoff))
            return;
    }
    else {
        if (clip_to_circle(&fromx, &fromy, &tox, &toy,
            currentgraph->grid.xaxis.circular.center,
            currentgraph->grid.yaxis.circular.center,
            currentgraph->grid.xaxis.circular.radius))
            return;
    }

    if (currentgraph->plottype != PLOT_POINT) {
        DevSetLinestyle(dv->v_linestyle);
    }
    else {
        /* if PLOT_POINT,
           don't want to plot an endpoint which have been clipped */
        if (tox != oldtox || toy != oldtoy)
        return;
    }
    DevSetColor(dv->v_color);

    switch (currentgraph->plottype) {
        case PLOT_LIN:
            DevLine(fromx, fromy, tox, toy);
            if ((currentgraph->ticmarks >0) && (np > -1)
                && (np % currentgraph->ticmarks == 0)) {
                /* Draw an 'x' */
                DevText("x", (int) (tox - currentgraph->fontwidth / 2),
                    (int) (toy - currentgraph->fontheight / 2));
                SaveText(currentgraph, "x",
                    (int) (tox - currentgraph->fontwidth / 2),
                    (int) (toy - currentgraph->fontheight / 2));
            }
            break;
        case PLOT_COMB:
            DevDatatoScreen(currentgraph,
                (double) 0, currentgraph->datawindow.ymin,
                &dummy, &ymin);
            DevLine(tox, ymin, tox, toy);
            break;
        case PLOT_POINT:
            /* Here, gi_linestyle is the character used for the point.  */
            pointc[0] = dv->v_linestyle;
            pointc[1] = '\0';
            DevText(pointc, (int) (tox - currentgraph->fontwidth / 2),
                (int) (toy - currentgraph->fontheight / 2));
            /* redraw will redraw this w/o our having to save it */
            /* SaveText(currentgraph, pointc,
                (int) (tox - currentgraph->fontwidth / 2),
                (int) (toy - currentgraph->fontheight / 2)); */
        default:
            break;
    }
}


static GRDATA
copydveclist(linkp)

/* copy dvecs */
GRDATA linkp;
{

    struct dvlist *dl, *dn;
    struct dvec *new;

    if (linkp == NULL)
        return (NULL);

    dl = (struct dvlist*)linkp;
    dn = alloc(struct dvlist);
    linkp = (GRDATA)dn;

    for ( ; dl; dl = dl->dl_next) {

        /* the copy is not linked into any plot */
        new = vec_copy(dl->dl_dvec);
        /* vec_copy doesn't set v_color or v_linestyle */
        new->v_color = dl->dl_dvec->v_color;
        new->v_linestyle = dl->dl_dvec->v_linestyle;
        if (new->v_scale)
            new->v_scale = vec_copy(new->v_scale);
        dn->dl_dvec = new;
        if (dl->dl_next) {
            dn->dl_next = alloc(struct dvlist);
            dn = dn->dl_next;
        }
    }
    return (linkp);
}


static void
freedveclist(linkp)

/* de-allocate dveclist */
GRDATA linkp;
{
    struct dvlist *dl, *dn;
    struct dvec *d;

    for (dl = (struct dvlist*)linkp; dl; dl = dn) {
        dn = dl->dl_next;
        d = dl->dl_dvec;
        txfree(d->v_name);
        if (isreal(d)) {
            txfree((char*)d->v_realdata);
        }
        else {
            txfree((char*)d->v_compdata);
        }
        if (d->v_scale) {
            txfree(d->v_scale->v_name);
            if (isreal(d->v_scale)) {
                txfree((char*)d->v_scale->v_realdata);
            }
            else {
                txfree((char*)d->v_scale->v_compdata);
            }
            txfree((char*)d->v_scale);
        }
        txfree((char*)d);
        txfree((char*)dl);
    }
}


static void
initdvecs(graph)

GRAPH *graph;
{
    char *s;
    int curlst;
    int curcolor;
    int curptchar = 1;
    struct dvec *v;
    struct dvlist *dl, *dl0 = (struct dvlist *)graph->plotdata;;

    if (graph->graphtype != GR_PLOT)
        return;

    /* Set up colors and line styles. */
    if (dispdev->numlinestyles == 1)
        curlst = 0; /* Use the same one all the time. */
    else
        curlst = 1;
    if (dispdev->numcolors > 2)
        curcolor = (graph->grid.gridtype == GRID_SMITH) ? 3 : 1;
    else
        curcolor = 1;

    for (dl = dl0; dl; dl = dl->dl_next) {
        v = dl->dl_dvec;

        /* Do something special with poles and zeros.  Poles are 'x's, and
         * zeros are 'o's.
         */
        s = ft_typenames(v->v_type);
        if (eq(s, "pole")) {
            v->v_linestyle = 'x';
            continue;
        }
        else if (eq(s, "zero")) {
            v->v_linestyle = 'o';
            continue;
        }

        /* Find a (hopefully) new line style and color. */
        if (graph->plottype == PLOT_POINT) {
            if (pointchars[curptchar + 1])
                curptchar++;
            else
                curptchar = 2;
        }
        else if ((curlst > 0) && (++curlst == dispdev->numlinestyles))
            curlst = 2;
        if ((curcolor > 0) && (++curcolor == dispdev->numcolors))
            curcolor = ((graph->grid.gridtype == GRID_SMITH &&
                (dispdev->numcolors > 3)) ? 4 : 2);

        if (graph->plottype == PLOT_POINT)
            v->v_linestyle = pointchars[curptchar];
        else {
            if (dispdev->numcolors > 2)
                /* use solid lines in color */
                v->v_linestyle = 0;
            else
                v->v_linestyle = curlst;
        }
        v->v_color = curcolor;
    }
}


static void
drawlegend(plotno, dv)

int plotno;
struct dvec *dv;
{

    int x, y, i;
    char buf[16];

    x = ((plotno % 2) ? (currentgraph->viewport.width / 2) :
        currentgraph->viewportxoff);
    y = currentgraph->absolute.height - currentgraph->fontheight
            - ((plotno + 2) / 2) * (currentgraph->fontheight);
    i = y + currentgraph->fontheight / 2 + 1;
    DevSetColor(dv->v_color);
    if (currentgraph->plottype == PLOT_POINT) {
        (void) sprintf(buf, "%c : ", dv->v_linestyle);
        DevText(buf, x + currentgraph->viewport.width / 20
                - 3 * currentgraph->fontwidth, y);
    }
    else {
        DevSetLinestyle(dv->v_linestyle);
        DevLine(x, i, x + currentgraph->viewport.width / 20, i);
    }
    DevSetColor(1);
    DevText(dv->v_name, x + currentgraph->viewport.width / 20
            + currentgraph->fontwidth, y);
}


static void
resize(graph)

/* call this routine after viewport size changes */
GRAPH *graph;
{

    double oldxratio, oldyratio;
    double scalex, scaley;
    struct _keyed *k;

    oldxratio = graph->aspectratiox;
    oldyratio = graph->aspectratioy;
    graph->viewportxoff = graph->fontwidth * 8; /* 8 lines on left */
    graph->viewportyoff = graph->fontheight * 4;    /* 4 on bottom */

    graph->viewport.width = graph->absolute.width -
            2 * graph->viewportxoff;
    graph->viewport.height = graph->absolute.height -
            2 * graph->viewportyoff;

    /* shrink datawindow slightly to avoid numerical strangeness */
    if (graph->grid.gridtype != GRID_LOGLOG &&
            graph->grid.gridtype != GRID_XLOG &&
            graph->grid.gridtype != GRID_SMITH) {
        scalex = 1e-3*(graph->datawindow.xmax - graph->datawindow.xmin);
        graph->datawindow.xmin += scalex;
        graph->datawindow.xmax -= scalex;
    }
    if (graph->grid.gridtype != GRID_LOGLOG &&
            graph->grid.gridtype != GRID_YLOG &&
            graph->grid.gridtype != GRID_SMITH) {
        scaley = 1e-3*(graph->datawindow.ymax - graph->datawindow.ymin);
        graph->datawindow.ymin += scaley;
        graph->datawindow.ymax -= scaley;
    }
    
    gr_fixgrid(graph, graph->grid.xdelta, graph->grid.ydelta,
            graph->grid.xdatatype, graph->grid.ydatatype);

    /* cache width and height info to make DevDatatoScreen go fast */
    /* note: XXX see if this is actually used anywhere */
    graph->datawindow.width = graph->datawindow.xmax -
                    graph->datawindow.xmin;
    graph->datawindow.height = graph->datawindow.ymax -
                    graph->datawindow.ymin;

    /* cache (datawindow size) / (viewport size) */
    graph->aspectratiox = graph->datawindow.width / graph->viewport.width;
    graph->aspectratioy = graph->datawindow.height / graph->viewport.height;

    /* scale keyed text */
    scalex = oldxratio / graph->aspectratiox;
    scaley = oldyratio / graph->aspectratioy;
    for (k = graph->keyed; k; k = k->next) {
        k->x = (k->x - graph->viewportxoff) * scalex + graph->viewportxoff;
        k->y = (k->y - graph->viewportyoff) * scaley + graph->viewportyoff;
    }
}


static void
redraw(graph)

/* redraw everything in struct graph */
GRAPH *graph;
{
    struct dvlist *link;
    int plotno;

    /* establish current graph so default graphic calls will work right */
    PushGraphContext(graph);

    DevClear();

    resize(graph);

    /* redraw grid */
    gr_redrawgrid(graph);

    for (link = (struct dvlist*)graph->plotdata, plotno = 0; link;
            link = link->dl_next, plotno++) {

        /* replot data
         * if onevalue, pass it a NULL scale
         * otherwise, if vec has its own scale, pass that
         *     else pass vec's plot's scale
         */
        plottrace(link->dl_dvec,
            graph->onevalue ? (struct dvec *) NULL :
                    (link->dl_dvec->v_scale ?
                    link->dl_dvec->v_scale :
                    link->dl_dvec->v_plot->pl_scale), plotno);
    }

    restoretext(graph);

    PopGraphContext();
}


static void
restoretext(graph)
GRAPH *graph;
{
    struct _keyed *k;

    /* restore text */
    for (k = graph->keyed; k; k = k->next) {
        DevSetColor(k->colorindex);
        DevText(k->text, k->x, k->y);
    }
}
