/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jeffrey M. Hsu
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * MFB driver
 */

#include "spice.h"

#ifdef HAVE_MFB

#include "mfb.h"
#include "plotdefs.h"

extern char ErrorMessage[];

static int mfb_wincount;

char *MFBRCD;


#ifdef __STDC__
static void mfbsetcolors(void);
static void mfbsetlinestyles(void);
#else
static void mfbsetcolors();
static void mfbsetlinestyles();
#endif


int
Mfb_Init()
{
    char buf[BSIZE_SP], dev[BSIZE_SP], *term;
    static  char old_term[BSIZE_SP] = "";
    static  char old_dev[BSIZE_SP] = "";
    static  char success = false;
    int err;
    extern char *kw_term, *kw_device;

    /* The mfb.rc file location defaults to the Spice Library */
    MFBRCD = Spice_Lib_Dir;

    /* Try to identify the terminal type -- check variable,
     * environment, then ask the user.
     */
    if (cp_getvar(kw_term, VT_STRING, buf)) {
        term = buf;
    }
    else {
        term = getenv("TERM");
        if (term == NULL) { 
            /* prompt user w/ input() */
            fprintf(cp_out, "Terminal type = ");
            (void) fflush(cp_out);
            (void) fscanf(cp_in, " %s", buf);
            term = copy(buf);
        }
    }
    if (!cp_getvar(kw_device, VT_STRING, dev))
        dev[0] = '\0';

/* We only re-open if things have changed */
    if( (!success) 
    || (strcmp(old_term, term) != 0) || (strcmp(old_dev,dev) != 0)) {
        (void) MFBOpen( term, (dev[0] ? dev : (char *) NULL), &err);
        strcpy( old_term, term); strcpy( old_dev, dev );
        if (err == MFBOK) {
            success = true;
        }
        else {
            success = false;
            sprintf(ErrorMessage, "MFB error: %s", MFBError(err));
            externalerror(ErrorMessage);
            return (1);
        }
    }

    dispdev->numlinestyles = MFBInfo(MAXLINESTYLES);
    dispdev->numcolors = MFBInfo(MAXCOLORS);
    dispdev->width = MFBInfo(MAXX) + 1;
    dispdev->height = MFBInfo(MAXY) + 1;

#ifdef DEFCOLORS
    if (dispdev->numcolors > 2 && MFBInfo(DEFCOLORS))
        mfbsetcolors();
#endif
    if (dispdev->numlinestyles > 2 && MFBInfo(DEFLINEPATTERN))
        mfbsetlinestyles();

    (void)MFBSetTextMode(1);      /* text overstrikes */

    /* get control back */
    if (MFBHalt() != MFBOK) {
        externalerror("Can't restore control of tty.");
        return (1);
    }
    return (0);
}


int
Mfb_NewViewport(graph)

GRAPH *graph;
{
    if (MFBInitialize() != MFBOK) {
        externalerror("Can't initialize MFB.");
        return (1);
    }

    graph->fontwidth = MFBInfo(FONTWIDTH);
    graph->fontheight = MFBInfo(FONTHEIGHT);

    graph->viewport.height = dispdev->height;
    graph->viewport.width = dispdev->width;

    graph->absolute.xpos = 0;
    graph->absolute.ypos = 0;
    graph->absolute.width = dispdev->width;
    graph->absolute.height = dispdev->height;

    mfb_wincount++;
    return (0);
}


int
Mfb_Close()
{
    (void)MFBClose();
    mfb_wincount = 0;
    return (0);
}


int
Mfb_Halt()
{
    if (!mfb_wincount)
        return (0);
    mfb_wincount--;
    if (!mfb_wincount)
        (void)MFBHalt();
    return (0);
}


int
Mfb_Clear()
{
    (void)MFBSetColor(0);
    (void)MFBFlood();
    return (0);
}


Mfb_Pixel(x, y)

int x, y;
{
    (void)MFBPixel(x, y);
    return (0);
}


Mfb_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    (void)MFBLine(x1, y1, x2, y2);
    return (0);
}


Mfb_Box(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    (void)MFBBox(x1, y1, x2, y2);
    return (0);
}


Mfb_Arc(x0, y0, radius, theta1, theta2)
int x0, y0, radius;
double theta1, theta2;
{
    double theta3, theta4;

/* note: figure some better value for nsides as a fn of radius
        and delta theta */
    if( (theta1 != theta2)
         &&  sin(fabs( theta1 - theta2 )/2.0) < 0.5  ) {
        theta3 = ceil((double) 180.0/M_PI * theta1);
        theta4 = floor((double) 180.0/M_PI * theta2);

        if ( theta3 < theta4  && theta3 < 360.0 ) {
            MFBArc(x0, y0, radius, (int)theta3, (int)theta4, 50);

            theta3 *= M_PI/180.0;
            theta4 *= M_PI/180.0;
        }
        else {
            theta3 = theta2;
            theta4 = theta1;
        }

        (void)MFBLine((int)(x0+radius*cos(theta1)),(int)(y0+radius*sin(theta1)), 
           (int)(x0+radius*cos(theta3)),(int)(y0+radius*sin(theta3)));
        (void)MFBLine((int)(x0+radius*cos(theta2)),(int)(y0+radius*sin(theta2)), 
           (int)(x0+radius*cos(theta4)),(int)(y0+radius*sin(theta4)));
    }
    else {
        (void)MFBArc(x0, y0, radius, 
        rnd((double) 180.0/M_PI * theta1),
                rnd((double) 180.0/M_PI * theta2), 50);
    }
    return (0);
}


int
Mfb_Polygon(p)

POLYGON *p;
{
    (void)MFBPolygon(p);
    return (0);
}


int
Mfb_Text(text, x, y)

char *text;
int x, y;
{
    (void)MFBText(text, x, y, 0);
    return (0);
}


int
Mfb_SetGhost(callback,x,y)

#ifdef __STDC__
void (*callback)(int,int,int,int);
#else
void (*callback)();
#endif
int x,y;
{
    (void)MFBSetGhost(callback,x,y);
    return (0);
}


int
Mfb_DefineColor(colorid, red, green, blue)

int colorid;
double red, green, blue;
{
    (void)MFBDefineColor(colorid,
        rnd(red*1000),rnd(green*1000),rnd(blue*1000));
    return (0);
}


int
Mfb_DefineLinestyle(linestyleid, mask)

int linestyleid;
int mask;
{
    (void)MFBDefineLineStyle(linestyleid, mask);
    return (0);
}


int
Mfb_SetLinestyle(linestyle)

int linestyle;
{
    if (MFBSetLineStyle(linestyle) != MFBOK) {
        externalerror("Bad linestyle");
    }
    return (0);
}


int
Mfb_SetColor(colorid)

int colorid;
{
    if (MFBSetColor(colorid) != MFBOK) {
        externalerror("bad color");
    }
    return (0);
}


int
Mfb_Update()
{
    MFBUpdate();
    return (0);
}


int
Mfb_MakeMenu()
{
    return (0);
}


int
Mfb_MakeDialog()
{
    return (0);
}


int
Mfb_Input(request, response)
REQUEST *request;
RESPONSE *response;
{
    int button = 0,key = 0;
    int x, y;

    switch (request->option) {
        case char_option:
            if (request->fp)
                response->reply.ch = cp_inchar(request->fp);
            else
                response->reply.ch = MFBGetchar();
            response->option = request->option;
            break;
        case button_option:
            while (button == 0)
                MFBPoint(&x,&y,&key,&button);
            if (response) {
                response->x = x;
                response->y = y;
                response->reply.button = button;
                response->option = request->option;
            }
            break;
        case point_option:
            while (button == 0 && key == 0)
                MFBPoint(&x,&y,&key,&button);
            if (response) {
                response->x = x;
                response->y = y;
                if (button) {
                    response->option = button_option;
                    response->reply.button = button;
                }
                else {
                    response->option = char_option;
                    response->reply.ch = key;
                }
            }
            break;

        case click_option:
        case checkup_option:
            break;

        default:
            /* just ignore, since we don't want a million error messages */
            if (response)
                response->option = error_option;
            break;
    }
    return (0);
}

/* Set colors properly. Assume that we won't need more than 20 colors. 
 * Also we should try to set the foreground and background colors.
 * These color values are just off of the top of my head...
 * Make the grid color (color 1) blue.
 */

#ifdef DEFCOLORS

static int reds[] = { 999,   0, 999,   0,   0, 999, 999,   0, 999, 500,
              500, 999, 999, 500, 999, 999, 600, 300, 600, 300 } ;
static int grns[] = { 999,   0,   0, 999,   0, 999,   0, 999, 500, 999,
              500, 999, 500, 999, 600, 300, 999, 999, 300, 600 } ;
static int blus[] = { 999,   0,   0,   0, 999,   0, 999, 999, 500, 500,
              999, 500, 999, 999, 300, 600, 300, 600, 999, 999 } ;

static void
mfbsetcolors()
{
    int i;
    int red, green, blue;

    MFBDefineColor(1, 0, 0, 999);
    for (i = 2; (i < dispdev->numcolors) && (i < 22); i++) {
        red = reds[i - 2];
        green = grns[i - 2];
        blue = blus[i - 2];
        MFBDefineColor(i, red, green, blue);
    }
    return;
}

#endif


/* Set line styles if necessary. Assume we won't use more than 20 line
 * styles. These line styles also are pretty random. Make sure that
 * line style 0 is solid and line style 1 is dotted.
 */

static int styles[] = { 0333, 0347, 0360, 0071, 0216, 0366, 0155, 0011,
            0123, 0317, 0330, 0145, 0275, 0016, 0315, 0101,
            0222, 0174, 0251, 0376 } ;

static void
mfbsetlinestyles()
{
    int i, j;

    MFBDefineLineStyle(0, 0377);    /* Solid. */
    MFBDefineLineStyle(1, 0125);    /* Dotted. */
    for (i = 2; (i < dispdev->numlinestyles) && (i < 22); i++) {
        j = styles[i - 2];
        MFBDefineLineStyle(i, j);
    }
    return;
}

#endif

