/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jeffrey M. Hsu
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Postscript driver
 */

#include "spice.h"
#include "plotdev.h"

static FILE *plotfile;

#define SOLID 0
static char *linestyle[] = {
                "[]",           /* solid */
                "[1 2]",        /* dotted */
                "[7 7]",        /* longdashed */
                "[3 3]",        /* shortdashed */
                "[3 5 1 5]" };  /* dotdashed */

typedef struct {
    int lastlinestyle;      /* initial invalid value */
    int strokecnt;          /* count line elements, avoid overflow */
} PSdevdep;

#define DEVDEP(g) (*((PSdevdep *) (g)->devdep))

static int Xoff, Yoff;

/* line memory */
static int lastx ,lasty;


int
PS_Init()
{
    dispdev->numlinestyles = 4;
    dispdev->numcolors = 2;

    /* 8" X 10.5" drawable area */
    dispdev->width = 576;       /* 8 * 72 */
    dispdev->height = 756;      /* 10.5 * 72 */

    return (0);
}


int
PS_NewViewport(graph)

/* devdep initially contains name of output file */
GRAPH *graph;
{

    plotfile = fopen((char*)graph->devdep, "w");

    if (!plotfile) {
        perror(graph->devdep);
        graph->devdep = (char *) NULL;
        return (1);
    }

    switch (graph->graphtype) {
        default:
        case GR_PLOT:
        case GR_GRAF:
        case GR_MPLT:
            /* square plotting area */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->width;
            Xoff = 18;
            Yoff = 108;
            break;
        case GR_SCED:
            /* full page plot */
            graph->absolute.width = dispdev->width;
            graph->absolute.height = dispdev->height;
            Xoff = 18;
            Yoff = 18;
            break;
    }

    /* start file off with a % */
    fprintf(plotfile, "%%! nutmeg plot\n");

    /* reasonable values, used in gr_ for placement */
    graph->fontwidth = 5;
    graph->fontheight = 10;

    /* set up a reasonable font */
    fprintf(plotfile, "/Helvetica findfont 10 scalefont setfont\n");

    graph->devdep = tmalloc(sizeof(PSdevdep));
    DEVDEP(graph).lastlinestyle = -1;
    DEVDEP(graph).strokecnt = 0;
    graph->linestyle = -1;
    lastx = -1;
    lasty = -1;

    return (0);
}


int
PS_Close()

{
    return (0);
}


int
PS_Halt()
{

    if (plotfile) {
        if (DEVDEP(currentgraph).lastlinestyle != -1) {
            /* haven't stroked last path */
            fprintf(plotfile, "stroke\n");
        }
        fprintf(plotfile, "showpage\n");
        fclose(plotfile);
    }
    return (0);
}


/* ARGSUSED */
int
PS_Pixel(x, y)

int x, y;
{
    int savedlstyle;

    savedlstyle = currentgraph->linestyle;
    PS_SetLinestyle(SOLID);
    if (DEVDEP(currentgraph).strokecnt > 499) {
        if (DEVDEP(currentgraph).lastlinestyle == -1)
            fprintf(plotfile, "newpath\n");
        else
            fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }
    fprintf(plotfile, "%d %d moveto\n", x+Xoff, y+Yoff);
    fprintf(plotfile, "%d %d lineto\n", x+Xoff, y+Yoff+1);
    (DEVDEP(currentgraph).strokecnt)++;
    PS_SetLinestyle(savedlstyle);
    return (0);
}


int
PS_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    /* clip out unchanged points */
    if (x1 == x2 && y1 == y2 && x1 == lastx && y1 == lasty)
        return (0);

    if (currentgraph->linestyle == -1)
        PS_SetLinestyle(SOLID);
    if (DEVDEP(currentgraph).lastlinestyle != currentgraph->linestyle ||
            DEVDEP(currentgraph).strokecnt > 499) {
        if (DEVDEP(currentgraph).lastlinestyle == -1)
            fprintf(plotfile, "newpath\n");
        else
            fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }
    if ((x1 != lastx) || (y1 != lasty) || !DEVDEP(currentgraph).strokecnt)
        fprintf(plotfile, "%d %d moveto\n", x1+Xoff, y1+Yoff);
    if ((x2 != x1) || (y2 != y1))
        fprintf(plotfile, "%d %d lineto\n", x2+Xoff, y2+Yoff);

    /* validate cache */
    lastx = x2;
    lasty = y2;
    DEVDEP(currentgraph).lastlinestyle = currentgraph->linestyle;
    (DEVDEP(currentgraph).strokecnt)++;
    return (0);
}


int
PS_Box(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
    PS_Line(x1,y1,x1,y2);
    PS_Line(x1,y2,x2,y2);
    PS_Line(x2,y2,x2,y1);
    PS_Line(x2,y1,x1,y1);
    return (0);
}


/* ARGSUSED */
int
PS_Arc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    int a1, a2;

    if (theta1 >= theta2)
        theta2 = 2 * M_PI + theta2;

    a1 = (180.0 / M_PI) * theta1;
    a2 = (180.0 / M_PI) * theta2;

    if (currentgraph->linestyle == -1)
        PS_SetLinestyle(SOLID);
    if (DEVDEP(currentgraph).strokecnt) {
        fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }
    else if (DEVDEP(currentgraph).lastlinestyle == -1) {
        fprintf(plotfile, "newpath\n");
        DEVDEP(currentgraph).lastlinestyle = currentgraph->linestyle;
    }

    fprintf(plotfile, "%d %d %d %d %d arc\n",
        x0+Xoff,y0+Yoff,radius,a1,a2);
    fprintf(plotfile, "stroke\n");
    lastx = -1;
    lasty = -1;
    return (0);
}


int
PS_Polygon(p)
POLYGON *p;
{
    /* This is currently used only to print dots for sced.  The
     * polygon code below looks bad and causes errors if the dots
     * are too small, thus we use the specialized routine below.
     */

    int *xy, x, y, savedlstyle;

    xy = (int*)p->xy;
    if (xy[0] == xy[8]) {
        x = xy[0];
        y = xy[5];
    }
    else {
        x = xy[4];
        y = xy[1];
    }
    if (DEVDEP(currentgraph).strokecnt) {
        fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }
    savedlstyle = currentgraph->linestyle;
    PS_SetLinestyle(SOLID);
    fprintf(plotfile,"%d %d moveto\n",x+1+Xoff,y+Yoff);
    fprintf(plotfile,"%d %d 3 0 359 arc closepath fill\n",x+Xoff,y+Yoff);
    PS_SetLinestyle(savedlstyle);
    lastx = -1;
    lasty = -1;
    return (0);


/*
    int *xy, nvert, savedlstyle;

    xy = (int*)p->xy;
    nvert = p->nvertices;
    savedlstyle = currentgraph->linestyle;
    PS_SetLinestyle(SOLID);

    fprintf(plotfile, "%d %d moveto ", *xy+Xoff, *(xy+1)+Yoff);
    xy += 2;
    nvert--;
    while (nvert--) {
        fprintf(plotfile, "%d %d lineto\n", *xy+Xoff, *(xy+1)+Yoff);
        xy += 2;
    }
    fprintf(plotfile,"closepath\nfill\n");
    PS_SetLinestyle(savedlstyle);
    return (0);
*/
}


int
PS_Text(text, x, y)

char *text;
int x, y;
{

    int savedlstyle;
    char tbuf[BSIZE_SP], *s;

    if (text == NULL)
        return (0);

    s = tbuf;
    while (*text) {
        if (*text == '(' || *text == ')' || *text == '\\')
            *s++ = '\\';
        *s++ = *text++;
    }
    *s = '\0';

    if (DEVDEP(currentgraph).strokecnt) {
        fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }
    /* set linestyle to solid
     * or may get funny color text on some plotters
     */
    savedlstyle = currentgraph->linestyle;
    PS_SetLinestyle(SOLID);

    /* move to (x, y) */
    fprintf(plotfile, "%d %d moveto\n", x+Xoff, y+Yoff);
    fprintf(plotfile, "(%s) show\n", tbuf);

    /* restore old linestyle */
    PS_SetLinestyle(savedlstyle);
    lastx = -1;
    lasty = -1;
    return (0);
}


int
PS_SetLinestyle(linestyleid)

int linestyleid;
{

    /* special case
        get it when PS_Text restores a -1 linestyle */
    if (linestyleid == -1) {
        currentgraph->linestyle = -1;
        return;
    }

    if (linestyleid < 0) {
        internalerror("bad linestyleid");
        return;
    }

    if (linestyleid > dispdev->numlinestyles)
        linestyleid = (linestyleid % dispdev->numlinestyles) + 1;

    if (currentgraph->linestyle != linestyleid) {
        if (DEVDEP(currentgraph).strokecnt) {
            fprintf(plotfile, "stroke\n");
            DEVDEP(currentgraph).strokecnt = 0;
            lastx = -1;
            lasty = -1;
        }
        fprintf(plotfile, "%s 0 setdash\n", linestyle[linestyleid]);
        currentgraph->linestyle = linestyleid;
    }
    return (0);
}


/* ARGSUSED */
int
PS_SetColor(colorid)
{
    /* do nothing */
    return (0);
}


int
PS_Update()

{
    fflush(plotfile);
    return (0);
}


/* transformed text for sced hardcopy */

int
PS_ScaledText(text,x,y,Xform)

char *text;
int x, y, Xform;
{
    /* Xform code:
     * 0x8: mirror x
     * 0x4: mirror y
     * 0x3: 0-no rotation, 1-90, 2-180, 3-270
     */
    int savedlstyle, deg;
    char tbuf[BSIZE_SP], *s;

    if (text == NULL)
        return (0);

    if (Xform == (char)0) {
        PS_Text(text,x,y);
        return;
    }

    s = tbuf;
    while (*text) {
        if (*text == '(' || *text == ')' || *text == '\\')
            *s++ = '\\';
        *s++ = *text++;
    }
    *s = '\0';

    if (DEVDEP(currentgraph).strokecnt) {
        fprintf(plotfile, "stroke\n");
        DEVDEP(currentgraph).strokecnt = 0;
    }

    /* set linestyle to solid
     * or may get funny color text on some plotters
     */
    savedlstyle = currentgraph->linestyle;
    PS_SetLinestyle(SOLID);

    fprintf(plotfile,"gsave\n");
    fprintf(plotfile,"%d %d translate\n",x+Xoff,y+Yoff);
    if (Xform & 12)
        fprintf(plotfile,"%d %d scale\n",
            (Xform & 8) ? -1 : 1, (Xform & 4) ? -1 : 1);
    deg = (Xform & 3) * 90;
    if (deg)
        fprintf(plotfile,"%d rotate\n",deg);
    fprintf(plotfile,"0 0 moveto\n");
    fprintf(plotfile,"(%s) show\n",tbuf);
    fprintf(plotfile,"grestore\n");

    /* restore old linestyle */
    PS_SetLinestyle(savedlstyle);
    lastx = -1;
    lasty = -1;
    return (0);
}
