/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1988 Jeffrey M. Hsu
         1995 Stephen R. Whiteley
****************************************************************************/

/*
 * X11 drivers.
 */

#include "spice.h"

#ifdef HAVE_X11

#include "plotdefs.h"
#include "newgraf.h"
#include "scedio.h"
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Form.h>
#include <X11/Shell.h>
#include <X11/keysym.h>
#include "x11util.h"

/* for export */
Display *Xdisplay;

typedef struct {
    Window window;
    Bool isopen;
    GC gc;
    GC xorgc;
    widget_bag widgets;
    XFontStruct *font;
    int lastlinestyle;
    int lastx, lasty;
    int refx, refy;
    int firstghost;
    Bool noclear;
#ifdef __STDC__
    void (*drawghost)(int,int,int,int);
#else
    void (*drawghost)();
#endif
} X11devdep;

#ifdef __STDC__
static int  errorhandler(Display*,XErrorEvent*);
static void CRaction(Widget,XEvent*,String*,Cardinal*);
static void QUaction(Widget,XEvent*,String*,Cardinal*);
static void resize(Widget,XtPointer,XEvent*,Boolean*);
static void redraw(Widget,XtPointer,XEvent*,Boolean*);
static void quit(Widget,XtPointer,XtPointer);
static void help(Widget,XtPointer,XtPointer);
static void hardcopy(Widget,XtPointer,XtPointer);
static void save(Widget,XtPointer,XtPointer);
static void DoSave(Widget,XtPointer,XtPointer);
static void motion(Widget,XtPointer,XEvent*,Boolean*);
static void keypress(Widget,XtPointer,XEvent*,Boolean*);
static void grpl_button(Widget,XtPointer,XEvent*,Boolean*);
static void graf_addplot(Widget,XtPointer,XtPointer);
static void graf_points(Widget,XtPointer,XtPointer);
static void graf_x_y(Widget,XtPointer,XtPointer);
static void graf_separate(Widget,XtPointer,XtPointer);
static void graf_marker(Widget,XtPointer,XtPointer);
static void graf_scale1(Widget,XtPointer,XtPointer);
static void graf_scale2(Widget,XtPointer,XtPointer);
static void initcolors(GRAPH*);
static void slopelocation(GRAPH*,int,int);
static void zoomin(GRAPH*);
static void X_ScreentoData(GRAPH*,int,int,double*,double*);
#else
static int  errorhandler();
static void CRaction();
static void QUaction();
static void resize();
static void redraw();
static void quit();
static void help();
static void hardcopy();
static void save();
static void DoSave();
static void motion();
static void keypress();
static void grpl_button();
static void graf_addplot();
static void graf_points();
static void graf_x_y();
static void graf_separate();
static void graf_marker();
static void graf_scale1();
static void graf_scale2();
static void initcolors();
static void slopelocation();
static void zoomin();
static void X_ScreentoData();
#endif

/* export to Help */
#ifdef __STDC__
extern void ToTopOfSced(Widget,XtPointer,XEvent*,Boolean*);
#else
extern void ToTopOfSced();
#endif

/* X dependent default parameters */
#define DEF_FONT "6x10"
#define NUMLINESTYLES 8
#define NXPLANES 5
#define BOXSIZE 30      /* initial size of bounding box for zoomin */
#define NumPrivateColors 16

static unsigned long x11_priv_colors[NumPrivateColors];

/* If there are (NumPrivateColors + 8) or
 * more available colors, NumPrivateColors are allocated to the
 * display for use by sced, and the rest are allocated by name.
 * Otherwise, all colors available up to NumPrivateColors are
 * allocated by name, and shared with sced if sced is used. The variable
 * dispdev->numcolors is set to the total number of allocated colors.
 */

#define NUMCOLORTAB 20
static char *colortable[NUMCOLORTAB];

#define Resource(str, class, ofst, what) \
   {str, class, XtRString, sizeof(char*), ofst*sizeof(char*), XtRString, what}

static XtResource resources[] = {
    Resource("color0", "Color0", 0, "white"),
    Resource("color1", "Color1", 1, "black"),
    Resource("color2", "Color2", 2, "red"),
    Resource("color3", "Color3", 3, "green"),
    Resource("color4", "Color4", 4, "blue"),
    Resource("color5", "Color5", 5, "magenta"),
    Resource("color6", "Color6", 6, "cyan"),
    Resource("color7", "Color7", 7, "violet"),
    Resource("color8", "Color8", 8, "sienna"),
    Resource("color9", "Color9", 9, "orange"),
    Resource("color10", "Color10", 10, "orchid"),
    Resource("color11", "Color11", 11, "violet"),
    Resource("color12", "Color12", 12, "maroon"),
    Resource("color13", "Color13", 13, "turquoise"),
    Resource("color14", "Color14", 14, "pink"),
    Resource("color15", "Color15", 15, "coral"),
    Resource("color16", "Color16", 16, "khaki"),
    Resource("color17", "Color17", 17, "yellow"),
    Resource("color18", "Color18", 18, "gold"),
    Resource("color19", "Color19", 19, "plum")
};

#undef Offset
#undef Resource

static String fallback_resources[] = {
    "Jspice3.Form.background: lightgrey",
    "Jspice3.Form.buttonbox.background: lightblue",
    "Jspice3.Form.buttonbox.Command.background: yellow",
    "Jspice3.Form.buttonbox.Toggle.background: yellow",

    "Jspice3.help.background: lightgrey",
    "Jspice3.help.Box.background: lightblue",
    "Jspice3.help.Box.Command.background: yellow",
    "Jspice3.help.Box.Toggle.background: yellow",
    "Jspice3.help.Box.Label.background: pink",
    "Jspice3.help.Label.background: pink",

    "Jspice3.xeditor.background: lightgrey",
    "Jspice3.xeditor.buttonbox.background: lightblue",
    "Jspice3.xeditor.buttonbox.Command.background: yellow",

    "Jspice3.popup.form.background: lightblue",
    "Jspice3.popup.form.Command.background: yellow",
    "Jspice3.popup.form.Label.background: pink",

    "Jspice3.popup_m.form_m.background: pink",
    "Jspice3.popup_m.form_m.Command.background: yellow",
    "Jspice3.popup_e.form_e.background: pink",
    "Jspice3.popup_e.form_e.Command.background: yellow",
    NULL
};


static char xlinestyles[NUMLINESTYLES][10] = {
    {0,4,1,1,1,1,0,0,0,0},    /* solid */
    {0,4,1,3,1,3,0,0,0,0},    /* spaced dots */
    {0,4,1,2,1,2,0,0,0,0},    /* dots */
    {0,4,3,3,3,3,0,0,0,0},    /* shortdash */
    {0,4,1,4,6,4,0,0,0,0},    /* dots longdash */
    {0,4,7,7,7,7,0,0,0,0},    /* longdash */
    {0,4,1,4,4,4,0,0,0,0},    /* dots shortdash */
    {0,4,3,3,7,3,0,0,0,0},    /* short/longdash */
};

/* Small cross-hair cursor */
#define cursor_width 16
#define cursor_height 16
#define cursor_x_hot 7
#define cursor_y_hot 7
static char cursorCross_bits[] = {
        0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
        0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x7e, 0xfc,
        0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
        0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00 };
static char cursorCross_mask[] = {
        0x80, 0x03, 0x80, 0x03, 0x80, 0x03, 0x80, 0x03,
        0x80, 0x03, 0x80, 0x03, 0x7e, 0xfc, 0x7e, 0xfc,
        0x7e, 0xfc, 0x80, 0x03, 0x80, 0x03, 0x80, 0x03,
        0x80, 0x03, 0x80, 0x03, 0x80, 0x03, 0x00, 0x00 };


static XtActionsRec actions[] = {
{"cr_action", CRaction},
{"quit_action", QUaction}
};

static char *sens_list[] = { "save", 0 };
static char *wmTranslations = "<Message>WM_PROTOCOLS: quit_action()";
static XtAppContext app_con;
static Widget sced_shell;
static Widget sced_view;

static Colormap sced_colormap;
static XVisualInfo *sced_visual_info;


int
X11_Init()
{
    char buf[BSIZE_SP];
    int numdispplanes;
    XGCValues gcvalues;
    char *argv[3];
    int argc;
    extern bool hlp_usex;

    argc = 1;
    argv[0] = cp_program;
    if (cp_display) {
        argv[1] = "-display";
        argv[2] = cp_display;
        argc = 3;
    }

    /* initialize X toolkit */
    XtToolkitInitialize();
    app_con = XtCreateApplicationContext();
    XtAppSetFallbackResources(app_con, fallback_resources);
    Xdisplay = XtOpenDisplay(app_con, NULL, NULL, "Jspice3", NULL, 0,
        &argc, argv);
    if (!Xdisplay) {
        internalerror("Can't open X display");
        return (1);
    }
    XSetErrorHandler(errorhandler);
    XtAppAddActions(app_con, actions, XtNumber(actions));

    hlp_usex = true;

    /* set correct information */
    numdispplanes = DisplayPlanes(Xdisplay, DefaultScreen(Xdisplay));
    if (numdispplanes < NXPLANES)
        dispdev->numcolors = 1 << numdispplanes;
    else
        dispdev->numcolors = 1 << NXPLANES;
    if (dispdev->numcolors > NUMCOLORS)
        dispdev->numcolors = NUMCOLORS;
    if (dispdev->numcolors > NumPrivateColors + NUMCOLORTAB)
        dispdev->numcolors = NumPrivateColors + NUMCOLORTAB;
    dispdev->numlinestyles = NUMLINESTYLES;

    dispdev->width = DisplayWidth(Xdisplay, DefaultScreen(Xdisplay));
    dispdev->height = DisplayHeight(Xdisplay, DefaultScreen(Xdisplay));

    if (dispdev->numcolors - NumPrivateColors >= 8) {
        if (!XAllocColorCells(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)), 0, NULL, 0,
            x11_priv_colors, NumPrivateColors)) {
            /* no private colors */
            dispdev->numcolors = NumPrivateColors;
        }
    }
    else
        if (dispdev->numcolors > NumPrivateColors)
            dispdev->numcolors = NumPrivateColors;

    return (0);
}


static int
errorhandler(display, errorev)

Display *display;
XErrorEvent *errorev;
{
    XGetErrorText(display, errorev->error_code, ErrorMessage, 1024);
    externalerror(ErrorMessage);
    return (0);
}


static void
CRaction(caller, call_data, params, numparams)

Widget caller;
XEvent *call_data;
String *params;
Cardinal *numparams;
{
    /* simulate a button press when CR entered in text of popup */
    Widget action = XtNameToWidget(XtParent(caller), "action");

    if (action)
        XtCallCallbacks(action, XtNcallback, NULL);
}


static void
QUaction(caller, call_data, params, numparams)

Widget caller;
XEvent *call_data;
String *params;
Cardinal *numparams;
{
    /* simulate a Quit button press when DeleteWindow requested */
    Widget quit;

    quit = XtNameToWidget(caller, "help.buttonbox.delete");
    if (!quit)
        quit = XtNameToWidget(caller, "*.buttonbox.quit");
    if (!quit)
        quit = XtNameToWidget(caller, "form.cancel");
    if (!quit)
        quit = XtNameToWidget(caller, "form_m.cancel_m");
    if (!quit)
        quit = XtNameToWidget(caller, "form_e.cancel_e");

    if (quit)
        XtCallCallbacks(quit, XtNcallback, NULL);
}


int
X11_NewViewport(graph)

/* NewViewport is responsible for filling in graph->viewport */
GRAPH *graph;
{

    Cursor cursor;
    XSetWindowAttributes w_attrs;
    XGCValues gcvalues;
    Widget button;
    X11devdep *xd;
    widget_bag *w;
    Position xpos, ypos;
    Dimension width, height;
    char *formname;
    char fontname[64];
    struct gplot *graf;
    int i, trys, options;
    int numdispplanes = DisplayPlanes(Xdisplay, DefaultScreen(Xdisplay));
    extern char *kw_scedfont, *kw_xfont;

    *fontname = '\0';

    switch (graph->graphtype) {
    case GR_PLOT:
        formname = "plot";
        break;
    case GR_GRAF:
        formname = "graf";
        break;
    case GR_MPLT:
        formname = "mplot";
        break;
    case GR_SCED:
        formname = "sced";
        break;
    default:
        sprintf(ErrorMessage, "unknown graph type");
        internalerror(ErrorMessage);
        graph->devdep = NULL;
        return (1);
    }
    graph->devdep = calloc(1, sizeof(X11devdep));
    xd = (X11devdep*)graph->devdep;
    w = &xd->widgets;
    w->popup_sens_list = sens_list;

    /* set up new shell */
    w->shell = XtVaAppCreateShell(NULL, "Jspice3", applicationShellWidgetClass,
        Xdisplay, NULL);
    XtOverrideTranslations(w->shell, XtParseTranslationTable(wmTranslations));
    XtGetApplicationResources(w->shell, (XtPointer) colortable,
        resources, XtNumber(resources), NULL, 0);

    /* set up form widget */
    w->form = XtVaCreateManagedWidget(formname, formWidgetClass, w->shell,
        NULL);

    /* set up viewport */
    w->view = XtVaCreateManagedWidget("viewport", widgetClass, w->form,
        XtNresizable, True,
        XtNright, XtChainRight,
        NULL);
    XtAddEventHandler(w->view, StructureNotifyMask, FALSE, resize, graph);
    XtAddEventHandler(w->view, ExposureMask, FALSE, redraw, graph);
    initcolors(graph);

    switch (graph->graphtype) {
    case GR_PLOT:
        if (SCEDactive())
            XtAddEventHandler(w->shell, VisibilityChangeMask,
                FALSE, ToTopOfSced, NULL);

        XtVaSetValues(w->view, XtNwidth, 300, XtNheight, 300, NULL);
        XtAddEventHandler(w->view, ButtonPressMask, FALSE, grpl_button, graph);
        XtAddEventHandler(w->view, KeyPressMask, FALSE, keypress, graph);

        /* set up button box */
        w->butbox = XtVaCreateManagedWidget("buttonbox",boxWidgetClass,
            w->form,
            XtNfromHoriz, w->view,
            XtNleft, XtChainRight,
            XtNright, XtChainRight,
            XtNtop, XtChainTop,
            XtNbottom, XtChainTop,
            XtNheight, 150,
            NULL);

        /* set up buttons */
        button = XtVaCreateManagedWidget("quit", commandWidgetClass,
            w->butbox,
            XtNlabel, "  quit  ",
            NULL);
        XtAddCallback(button, XtNcallback, quit, graph);

        button = XtVaCreateManagedWidget("help", commandWidgetClass,
            w->butbox,
            XtNlabel, "  help  ",
            NULL);
        XtAddCallback(button, XtNcallback, help, "plot");

        button = XtVaCreateManagedWidget("hardcopy", commandWidgetClass,
            w->butbox,
            XtNlabel, "hardcopy",
            NULL);
        XtAddCallback(button, XtNcallback, hardcopy, graph);

        button = XtVaCreateManagedWidget("save", commandWidgetClass,
            w->butbox,
            XtNlabel, "  save  ",
            NULL);
        XtAddCallback(button, XtNcallback, save, graph);

        /* set up slopelocation readout area */
        (void)XtVaCreateManagedWidget("slopetext", asciiTextWidgetClass,
            w->form,
            XtNfromVert, w->butbox,
            XtNfromHoriz, w->view,
            XtNwidth, 90,
            XtNheight, 200,
            XtNstring, "",
            XtNdisplayCaret, False,
            XtNborderWidth, 0,
            NULL);
        break;

    case GR_GRAF:
        graf = (struct gplot *)graph->plotdata;
        options = graf->opt;

        if (SCEDactive())
            XtAddEventHandler(w->shell, VisibilityChangeMask,
                FALSE, ToTopOfSced, NULL);

        XtVaSetValues(w->view, XtNwidth, 400, XtNheight, 300, NULL);
        XtAddEventHandler(w->view, PointerMotionMask, FALSE, motion, graph);
        XtAddEventHandler(w->view, ButtonPressMask, FALSE, grpl_button, graph);
        XtAddEventHandler(w->view, KeyPressMask, FALSE, keypress, graph);

        /* set up button box */
        w->butbox = XtVaCreateManagedWidget("buttonbox", boxWidgetClass,
            w->form,
            XtNfromHoriz,w ->view,
            XtNleft, XtChainRight,
            XtNright, XtChainRight,
            XtNtop, XtChainTop,
            XtNbottom, XtChainTop,
            NULL);

        /* set up buttons */
        button = XtVaCreateManagedWidget("quit", commandWidgetClass,
            w->butbox,
            XtNlabel, "  quit   ",
            NULL);
        XtAddCallback(button, XtNcallback, quit, graph);

        button = XtVaCreateManagedWidget("help", commandWidgetClass,
            w->butbox,
            XtNlabel, "  help   ",
            NULL);
        XtAddCallback(button, XtNcallback, help, "graf");

        button = XtVaCreateManagedWidget("hardcopy", commandWidgetClass,
            w->butbox,
            XtNlabel, "hardcopy ",
            NULL);
        XtAddCallback(button, XtNcallback, hardcopy, graph);
                                                                               
        button = XtVaCreateManagedWidget("save", commandWidgetClass,
            w->butbox,
            XtNlabel, "  save   ",
            NULL);
        XtAddCallback(button, XtNcallback, save, graph);

/* obsolete
        button = XtVaCreateManagedWidget("addplot", commandWidgetClass,
            w->butbox,
            XtNlabel, "addplot  ",
            NULL);
        XtAddCallback(button, XtNcallback, graf_addplot, graph);
*/

        button = XtVaCreateManagedWidget("points", toggleWidgetClass,
            w->butbox,
            XtNlabel, " points  ",
            XtNstate, (options & OPT_p) ? True : False,
            NULL);
        XtAddCallback(button, XtNcallback, graf_points, graph);

        if (graf->numtr > 1) {
            button = XtVaCreateManagedWidget("x-y", toggleWidgetClass,
                w->butbox,
                XtNlabel, "   x-y   ",
                XtNstate, (options & OPT_b) ? True : False,
                NULL);
            XtAddCallback(button, XtNcallback, graf_x_y, graph);

            button = XtVaCreateManagedWidget("separate", toggleWidgetClass,
                w->butbox,
                XtNlabel, "separate ",
                XtNstate, (options & OPT_t) ? True : False,
                NULL);
            XtAddCallback(button, XtNcallback, graf_separate, graph);
        }

        button = XtVaCreateManagedWidget("marker", toggleWidgetClass,
            w->butbox,
            XtNlabel, " marker  ",
            NULL);
        XtAddCallback(button, XtNcallback, graf_marker, graph);

        for (i = 1; i < graf->numtr; i++) {
            if (graf->tdata[0].type != graf->tdata[i].type) {
                i = 0;
                break;
            }
        }
        if (i > 1) {
            /* more than one trace, same type */
            button = XtVaCreateManagedWidget("onescale", toggleWidgetClass,
                w->butbox,
                XtNlabel, "one scale",
                XtNstate, (options & OPT_N) ? True : False,
                NULL);
            XtAddCallback(button, XtNcallback, graf_scale1, graph);
        }
        else if (!i) {
            /* more than one trace type */
            button = XtVaCreateManagedWidget("onescale", toggleWidgetClass,
                w->butbox,
                XtNlabel, "one scale",
                XtNstate, (options & OPT_N) ? True : False,
                NULL);
            XtAddCallback(button, XtNcallback, graf_scale2, graph);
            button = XtVaCreateManagedWidget("grpscale", toggleWidgetClass,
                w->butbox,
                XtNlabel, "grp scale",
                XtNradioGroup, button,
                XtNstate, (options & OPT_n) ? True : False,
                NULL);
            XtAddCallback(button, XtNcallback, graf_scale2, graph);
        }
        break;

    case GR_MPLT:
        if (SCEDactive())
            XtAddEventHandler(w->shell, VisibilityChangeMask,
                FALSE, ToTopOfSced, graph);

        XtVaSetValues(w->view, XtNwidth, 300, XtNheight, 300, NULL);
        XtAddEventHandler(w->view, KeyPressMask, FALSE, keypress, graph);

        /* set up button box */
        w->butbox = XtVaCreateManagedWidget("buttonbox", boxWidgetClass,
            w->form,
            XtNfromHoriz, w->view,
            XtNleft, XtChainRight,
            XtNright, XtChainRight,
            XtNtop, XtChainTop,
            XtNbottom, XtChainTop,
            NULL);

        /* set up buttons */
        button = XtVaCreateManagedWidget("quit", commandWidgetClass,
            w->butbox,
            XtNlabel, "  quit  ",
            NULL);
        XtAddCallback(button, XtNcallback, quit, graph);

        button = XtVaCreateManagedWidget("help", commandWidgetClass,
            w->butbox,
            XtNlabel, "  help  ",
            NULL);
        XtAddCallback(button, XtNcallback, help, "mplot");

        button = XtVaCreateManagedWidget("hardcopy", commandWidgetClass,
            w->butbox,
            XtNlabel, "hardcopy",
            NULL);
        XtAddCallback(button, XtNcallback, hardcopy, graph);

        button = XtVaCreateManagedWidget("save", commandWidgetClass,
            w->butbox,
            XtNlabel, "  save  ",
            NULL);
        XtAddCallback(button, XtNcallback, save, graph);
        break;

    case GR_SCED:
        sced_shell = w->shell;
        sced_view = w->view;
        XtVaSetValues(w->view,
            XtNwidth, dispdev->width - 30,
            XtNheight, dispdev->height - 50,
            XtNbackground, BlackPixel(Xdisplay, DefaultScreen(Xdisplay)),
            NULL);
        XtAddEventHandler(w->view, PointerMotionMask, FALSE, motion, graph);
        /* don't trap these events */
        XtOverrideTranslations(w->view,
            XtParseTranslationTable("<BtnUp>:\n<BtnDown>:\n<Key>:"));

        XtVaSetValues(w->shell, XtNx, 10, XtNy, 10, NULL);

        if (!cp_getvar(kw_scedfont, VT_STRING, fontname)) {
            (void) strcpy(fontname, "9x15");
        }
    }

    /* set up fonts */
    if (!*fontname) {
        if (!cp_getvar(kw_xfont, VT_STRING, fontname)) {
            (void) strcpy(fontname, DEF_FONT);
        }
    }

    trys = 1;
    while (!(xd->font = XLoadQueryFont(Xdisplay, fontname))) {
        sprintf(ErrorMessage, "can't open font %s", fontname);
        strcpy(fontname, "fixed");
        if (trys > 1) {
            internalerror(ErrorMessage);
            return (1);
        }
        trys += 1;
    }

    graph->fontwidth = xd->font->max_bounds.width;
    graph->fontheight = xd->font->max_bounds.ascent +
        xd->font->max_bounds.descent;

    XtRealizeWidget(w->shell);
    w->wm_delete = XInternAtom(Xdisplay, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(Xdisplay, XtWindow(w->shell), &w->wm_delete, 1);

    xd->window = XtWindow(w->view);
    w_attrs.bit_gravity = ForgetGravity;
    XChangeWindowAttributes(Xdisplay, xd->window, CWBitGravity, &w_attrs);

    gcvalues.cap_style = CapNotLast;
    xd->gc = XCreateGC(Xdisplay, xd->window, GCCapStyle, &gcvalues);

    gcvalues.function = GXxor;
    xd->xorgc = XCreateGC(Xdisplay, xd->window,
        GCFunction | GCCapStyle, &gcvalues);

    XSetFont(Xdisplay,xd->gc, xd->font->fid);
    XSetFont(Xdisplay,xd->xorgc, xd->font->fid);


    /* set up cursor */
    if (graph->graphtype == GR_SCED) {
        Pixmap cursor_data, cursor_mask;
        XColor backg, foreg;

        backg.pixel = BlackPixel(Xdisplay, DefaultScreen(Xdisplay));
        foreg.pixel = WhitePixel(Xdisplay, DefaultScreen(Xdisplay));
        XQueryColor(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)), &backg);
        XQueryColor(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)), &foreg);
        cursor_data = XCreateBitmapFromData(Xdisplay, xd->window,
            cursorCross_bits, cursor_width, cursor_height);
        cursor_mask = XCreateBitmapFromData(Xdisplay, xd->window,
            cursorCross_mask, cursor_width, cursor_height);
        cursor = XCreatePixmapCursor(Xdisplay, cursor_data,
            cursor_mask, &foreg, &backg, cursor_x_hot, cursor_y_hot);
        XDefineCursor(Xdisplay, xd->window, cursor);
        XFreePixmap(Xdisplay, cursor_data);
        XFreePixmap(Xdisplay, cursor_mask);
    }
    else {
        cursor = XCreateFontCursor(Xdisplay, XC_left_ptr);
        XDefineCursor(Xdisplay, xd->window, cursor);
    }

    /* set up initial xor color */
    /* offset 1 is assumed to be the highlighting color */
    if (graph->graphtype == GR_SCED &&
            dispdev->numcolors - NumPrivateColors >= 8) {
        XSetForeground(Xdisplay, xd->xorgc,
            graph->colors[NUMCOLORS - NumPrivateColors] ^
            graph->colors[NUMCOLORS - NumPrivateColors + 1]);
    }
    else {
        XSetForeground(Xdisplay,xd->xorgc,
            graph->colors[0] ^ graph->colors[1]);
    }

    XtVaGetValues(w->view,
        XtNx, &xpos,
        XtNy, &ypos,
        XtNwidth, &width,
        XtNheight, &height,
        NULL);
    graph->absolute.xpos = xpos;
    graph->absolute.ypos = ypos;
    graph->absolute.width = width;
    graph->absolute.height = height;

    return (0);
}


/* ARGSUSED */
static void
resize(caller, client_data, event, ctd)

Widget caller;
XtPointer client_data;
XEvent *event;
Boolean *ctd;
{
    GRAPH *graph = (GRAPH *) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    XEvent ev;
    Dimension width, height;

    /* 
     *  Get rid of other StructureNotify events on this window.
     */
    while (XCheckWindowEvent(Xdisplay, xd->window,
        (long) StructureNotifyMask, &ev)) ;

    XClearWindow(Xdisplay, xd->window);
    XtVaGetValues(caller, XtNwidth, &width, XtNheight, &height, NULL);
    graph->absolute.width = width;
    graph->absolute.height = height;
}


/* ARGSUSED */
static void
redraw(caller, client_data, event, ctd)

Widget caller;
XtPointer client_data;
XEvent *event;
Boolean *ctd;
{

    GRAPH *graph = (GRAPH *) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    XExposeEvent *pev = (XExposeEvent *) event;
    XEvent ev;
    XRectangle rects[30];
    int n = 1;
    Dimension width, height;

    /* hack for Motif */
    if (pev->window != xd->window)
        return;

    XtVaGetValues(caller, XtNwidth, &width, XtNheight, &height, NULL);
    graph->absolute.width = width;
    graph->absolute.height = height;
    xd->isopen = True;

    rects[0].x = pev->x;
    rects[0].y = pev->y;
    rects[0].width = pev->width;
    rects[0].height = pev->height;

    /* pull out all other expose regions that need to be redrawn */
    while (n < 30 && XCheckWindowEvent(Xdisplay, xd->window,
            (long) ExposureMask, &ev)) {
        pev = (XExposeEvent *) &ev;
        rects[n].x = pev->x;
        rects[n].y = pev->y;
        rects[n].width = pev->width;
        rects[n].height = pev->height;
        n++;
    }
    XSetClipRectangles(Xdisplay, xd->gc, 0, 0, rects, n, Unsorted);

    xd->noclear = True;
    if (graph->redraw)
        (*graph->redraw)(graph);
    xd->noclear = False;
    while (XCheckWindowEvent(Xdisplay, xd->window,
        (long) StructureNotifyMask, &ev)) ;
    XSetClipMask(Xdisplay, xd->gc, None);
}


/* ARGSUSED */
static void
quit(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH *) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;

    xd->isopen = False;
    XtDestroyWidget(xd->widgets.shell);
    DestroyGraph(graph->graphid);
}


/* ARGSUSED */
static void
help(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    wordlist wl;

    wl.wl_prev = wl.wl_next = NULL;
    wl.wl_word = (char*)client_data;
    com_ghelp(&wl);
}


/* ARGSUSED */
static void
hardcopy(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH *) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    char buf[BSIZE_SP];

    ft_hardcopy(NULL, graph, true, buf, false);
    PopUpMessage(buf, &xd->widgets);
}


static void
save(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH *) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    
    xd->widgets.client_data = (XtPointer)graph;
    PopUpInput("", "save", DoSave, &xd->widgets);
}


/* ARGSUSED */
static void
DoSave(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    widget_bag *w = (widget_bag*)client_data;
    wordlist wl;
    char *fname;
    char buf[BSIZE_SP];

    XtVaGetValues(w->popup_text, XtNstring, &fname, NULL);
    if (fname && *fname) {
        strcpy(buf, fname);
        cp_pathfix(buf);
        fname = buf;
    }
    if (CheckFile(fname, W_OK, w) == NOGO)
        return;
    wl.wl_prev = wl.wl_next = NULL;
    wl.wl_word = fname;
    ft_hardcopy(&wl, (GRAPH*)w->client_data, true, buf, true);
    PopDownInput(caller, (XtPointer)w, call_data);
    PopUpMessage("Plot saved to file", w);
}


/* ARGSUSED */
static void
motion(caller, clientdata, event, ctd)

Widget caller;
XtPointer clientdata;
XEvent *event;
Boolean *ctd;
{
    GRAPH *graph = (GRAPH*)clientdata;
    X11devdep *xd = (X11devdep*)graph->devdep;
    XMotionEvent *motionev = (XMotionEvent *) event;
    XEvent nextEvent, newEvent;
    GC tempgc;

    if (xd->drawghost) {
        while (XEventsQueued(motionev->display, QueuedAfterReading) > 0) {
            XPeekEvent(motionev->display, &nextEvent);
            if (nextEvent.type != MotionNotify)
                break;
            XNextEvent(motionev->display, &newEvent);
            motionev = &newEvent.xmotion;
        }
        PushGraphContext(graph);
        tempgc = xd->gc;
        xd->gc = xd->xorgc;
        XGrabServer(Xdisplay);
        if (xd->firstghost)
            xd->firstghost = false;
        else {
            (*xd->drawghost)(xd->lastx, xd->lasty, xd->refx, xd->refy);
        }
        xd->lastx = motionev->x;
        xd->lasty = graph->absolute.height - motionev->y - 1;
        (*xd->drawghost)(xd->lastx, xd->lasty, xd->refx, xd->refy);
        XUngrabServer(Xdisplay);
        xd->gc = tempgc;
        PopGraphContext();
    }
}


/* ARGSUSED */
static void
keypress(caller, clientdata, event, ctd)

Widget caller;
XtPointer clientdata;
XEvent *event;
Boolean *ctd;
{
    GRAPH *graph = (GRAPH *) clientdata;
    X11devdep *xd = (X11devdep*)graph->devdep;
    XKeyEvent *keyev = (XKeyPressedEvent *) event;
    char text[4];
    int nbytes;

    nbytes = XLookupString(keyev, text, 4, NULL, NULL);
    if (!nbytes) return;

    /* write it */
    PushGraphContext(graph);
    text[nbytes] = '\0';
    XSetForeground(Xdisplay, xd->gc, graph->colors[1]);
    X11_Text(text, keyev->x, graph->absolute.height - keyev->y - 1);
    /* save it */
    SaveText(graph, text, keyev->x, graph->absolute.height - keyev->y - 1);
    /* warp mouse so user can type in sequence */
    XWarpPointer(Xdisplay, None, xd->window, 0, 0, 0, 0,
        keyev->x + XTextWidth(xd->font, text, nbytes),
        keyev->y);
    PopGraphContext();
}


/* ARGSUSED */
void
ToTopOfSced(caller, clientdata, event, ctd)

Widget caller;
XtPointer clientdata;
XEvent *event;
Boolean *ctd;
{
    /* in SCED, prevent windows from disappearing */
    XVisibilityEvent *vev = (XVisibilityEvent*)event;
    XWindowChanges xv;

    if (!SCEDactive())
        return;

    if (vev->state == VisibilityFullyObscured) {
        xv.sibling = XtWindow(sced_shell);
        xv.stack_mode = Above;
        XReconfigureWMWindow(vev->display, vev->window,
            DefaultScreen(vev->display), CWSibling|CWStackMode, &xv);
    }
}


/* ARGSUSED */
static void
grpl_button(caller, clientdata, event, ctd)

Widget caller;
XtPointer clientdata;
XEvent *event;
Boolean *ctd;
{
    XButtonEvent *buttonev = (XButtonEvent *) event;

    switch (buttonev->button) {
        case Button1:
            slopelocation((GRAPH *) clientdata, buttonev->x, buttonev->y);
            break;
        case Button3:
            zoomin((GRAPH *) clientdata);
            break;
    }
}


/* ARGSUSED */
/* obsolete
static void
graf_addplot(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt |= OPT_add;
    if (graph->redraw)
        graph->redraw(graph);
    PopUpMessage("Plot data appended to file \"plotdt\"", &xd->widgets);
}
*/


/* ARGSUSED */
static void
graf_points(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt ^= OPT_p;
    if (graph->redraw)
        graph->redraw(graph);
}


/* ARGSUSED */
static void
graf_x_y(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt ^= OPT_b;
    if (graph->redraw)
        graph->redraw(graph);
}


/* ARGSUSED */
static void
graf_separate(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt ^= OPT_t;
    if (graph->redraw)
        graph->redraw(graph);
}


/* ARGSUSED */
static void
graf_marker(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt ^= OPT_mark;
    if (graph->redraw)
        graph->redraw(graph);
}


/* ARGSUSED */
static void
graf_scale1(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    struct gplot *graf = (struct gplot *)graph->plotdata;

    graf->opt ^= OPT_N;
    if (graph->redraw)
        graph->redraw(graph);
}


/* ARGSUSED */
static void
graf_scale2(caller, client_data, call_data)

Widget caller;
XtPointer client_data, call_data;
{
    GRAPH *graph = (GRAPH*) client_data;
    X11devdep *xd = (X11devdep*)graph->devdep;
    struct gplot *graf = (struct gplot *)graph->plotdata;
    char *s;

    s = (char*)XawToggleGetCurrent(
        XtNameToWidget(xd->widgets.butbox, "onescale"));

    if (s == NULL) {
        graf->opt &= ~OPT_n;
        graf->opt &= ~OPT_N;
    }
    else if (!strcmp(s,"onescale")) {
        graf->opt |= OPT_N;
        if (graf->opt & OPT_n) {
            graf->opt &= ~OPT_n;
            return;
        }
    }
    else {
        graf->opt |= OPT_n;
        if (graf->opt & OPT_N) {
            graf->opt &= ~OPT_N;
            return;
        }
    }
    if (graph->redraw)
        graph->redraw(graph);
}


/* Return information about the user's graphics mode
 */
static XVisualInfo *
find_visual(Display *display, int screen)
{
    XVisualInfo vt, *vl;
    int i, vmatch;
    vt.screen = screen;
    /* This sucks, there must be an easier way to get info on the
     * default visual */
    vl = XGetVisualInfo(display, VisualScreenMask, &vt, &vmatch);
    for (i = 0; i < vmatch; i++)
        if (vl[i].visual == DefaultVisual(display, screen))
            break;
    if (vl[i].class == PseudoColor && vl[i].depth >= 8)
        return (vl + i);
    if (vl[i].class == TrueColor)
        return (vl + i);
    return (0);
}


static void
initcolors(graph)

GRAPH *graph;
{
    X11devdep *xd = (X11devdep*)graph->devdep;
    XColor visual, exact;
    char buf[16], colorstring[BSIZE_SP];
    int numcolors;
    Pixel b_pixel, f_pixel;
    int i;

    // Make sure that the necessary visual is available
    sced_visual_info = find_visual(Xdisplay, DefaultScreen(Xdisplay));
    if (!sced_visual_info) {
        fprintf(stderr, "Uh-oh: Graphics mode unsupported.\n");
        return;
    }
    /*
    printf("Found %s visual, %d planes.\n",
        sced_visual_info->c_class == PseudoColor ? "PseudoColor" : "TrueColor",
        sced_visual_info->depth);
    if (sced_visual_info->c_class == TrueColor) {
        ColorAlloc.no_alloc = True;
        return (False);
    }
int ncolors = vl->colormap_size;
    */

    XtVaGetValues(xd->widgets.view, XtNbackground, &b_pixel, NULL);

    if (!cp_getvar("color0", VT_STRING, colorstring))
        strcpy(colorstring, colortable[0]);
    if (!XAllocNamedColor(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)),
            colorstring, &visual, &exact)) {
        (void) sprintf(ErrorMessage, "can't get color %s\n", colorstring);
        externalerror(ErrorMessage);
        graph->colors[0] = b_pixel;
    }
    else {
        XtVaSetValues(xd->widgets.view, XtNbackground, visual.pixel, NULL);
        graph->colors[0] = visual.pixel;
    }

    if (b_pixel == WhitePixel(Xdisplay, DefaultScreen(Xdisplay)))
        f_pixel = BlackPixel(Xdisplay, DefaultScreen(Xdisplay));
    else
        f_pixel = WhitePixel(Xdisplay, DefaultScreen(Xdisplay));
    if (dispdev->numcolors == 2) {
        /* black and white */
        graph->colors[1] = f_pixel;
        return;
    }
    if (dispdev->numcolors - NumPrivateColors >= 8) {
        numcolors = dispdev->numcolors - NumPrivateColors;
        /* top of map is used for private colors */
        for (i = 0; i < NumPrivateColors; i++)
            graph->colors[i + NUMCOLORS - NumPrivateColors] =
                x11_priv_colors[i];
    }
    else
        numcolors = dispdev->numcolors;

    for (i = 1; i < numcolors; i++) {
        (void) sprintf(buf, "color%d", i);
        if (!cp_getvar(buf, VT_STRING, colorstring))
            strcpy(colorstring, colortable[i]);
        if (!XAllocNamedColor(Xdisplay,
                DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)),
                colorstring, &visual, &exact)) {
            (void) sprintf(ErrorMessage,
                "can't get color %s\n", colorstring);
            externalerror(ErrorMessage);
            graph->colors[i] = f_pixel;
            continue;
        }
        if (visual.pixel == b_pixel)
            graph->colors[i] = f_pixel;
        else
            graph->colors[i] = visual.pixel;
    }
}


/* This routine closes the X connection.
 * It is not to be called for finishing a graph.
 */
int
X11_Close()
{
    XCloseDisplay(Xdisplay);
    return (0);
}


/* This routine closes the window.
 * It can be called for finishing a graph, but remember to free the
 * graph structure also.
 */
int
X11_Halt()
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    xd->isopen = 0;
    XtDestroyWidget(xd->widgets.shell);
    return (0);
}


int
X11_Title(wintitle, icontitle)

char *wintitle, *icontitle;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    XtVaSetValues(xd->widgets.shell,
        XtNtitle, wintitle,
        XtNiconName, icontitle,
        NULL);
    return (0);
}


int
X11_Pixel(x, y)

int x, y;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    if (xd->isopen)
        XDrawPoint(Xdisplay, xd->window, xd->gc, x,
            currentgraph->absolute.height - y - 1);
    return (0);
}


int
X11_Line(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    if (xd->isopen)
        XDrawLine(Xdisplay, xd->window, xd->gc,
            x1, currentgraph->absolute.height - y1 - 1,
            x2, currentgraph->absolute.height - y2 - 1);
    return (0);
}


int
X11_Box(x1, y1, x2, y2)

int x1, y1, x2, y2;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    int temp;

    if (x1 > x2) {
        temp = x1;
        x1 = x2;
        x2 = temp;
    }
    if (y1 > y2) {
        temp = y1;
        y1 = y2;
        y2 = temp;
    }
    if (xd->isopen)
        XFillRectangle(Xdisplay, xd->window, xd->gc, x1,
            currentgraph->absolute.height - y2 - 1, x2-x1+1, y2-y1+1);
    return (0);
}


/*ARGSUSED*/
int
X11_Arc(x0, y0, radius, theta1, theta2)

int x0, y0, radius;
double theta1, theta2;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    int t1, t2;

    if (xd->isopen) {
    if (theta1 >= theta2)
        theta2 = 2 * M_PI + theta2;
    t1 = 64 * (180.0 / M_PI) * theta1;
    t2 = 64 * (180.0 / M_PI) * theta2 - t1;
    if (t2 == 0)
        return (0);
    XDrawArc(Xdisplay, xd->window, xd->gc, x0 - radius,
        currentgraph->absolute.height - radius - y0 - 1,
        2 * radius, 2 * radius, t1, t2);
    }
    return (0);
}


int
X11_Polygon(p)

POLYGON *p;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    int index;
    XPoint *pts;

    if (xd->isopen) {
        pts = (XPoint *) tmalloc( (unsigned) (sizeof(XPoint) *
            (p->nvertices+1)) );
        
        for (index = 0;  index < p->nvertices;  index++) {
            pts[index].x = p->xy[index*2];
            pts[index].y = currentgraph->absolute.height - p->xy[index*2+1] - 1;
        }

        XFillPolygon(Xdisplay, xd->window, xd->gc, pts, index,
            Complex, CoordModeOrigin);

        txfree((char *)pts);
    }
    return (0);
}


/* note: x and y are the LOWER left corner of text */
int
X11_Text(text, x, y)

char *text;
int x, y;
{
    /* We specify text position by lower left corner, so have to adjust for
     * X11's font nonsense.
     */
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    if (xd->isopen)
        XDrawString(Xdisplay, xd->window, xd->gc, x,
            currentgraph->absolute.height - (y + xd->font->max_bounds.descent),
            text, strlen(text));
    return (0);
}


int
X11_SetGhost(callback,x,y)

#ifdef __STDC__
void (*callback)(int,int,int,int);
#else
void (*callback)();
#endif
int x, y;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    GC tempgc;

    if (callback) {
        xd->drawghost = callback;
        xd->refx = x;
        xd->refy = y;
        xd->lastx = x;
        xd->lasty = y;
        xd->firstghost = true;
        return (0);
    }
    if (xd->drawghost) {
        if (!xd->firstghost) {
            /* undraw last */
            tempgc = xd->gc;
            xd->gc = xd->xorgc;
            (*xd->drawghost)(xd->lastx, xd->lasty, xd->refx, xd->refy);
            xd->gc = tempgc;
        }
        xd->drawghost = NULL;
    }
    return (0);
}


/*ARGSUSED*/
int
X11_DefineColor(colorid, red, green, blue)

int colorid;
double red, green, blue;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    XColor newcolor;

    if (currentgraph->graphtype != GR_SCED)
        /* used for SCED only, return */
        return (0);

    if (sced_visual_info->class == PseudoColor) {

        if (dispdev->numcolors - NumPrivateColors < 8)
            /* no private color cells, just return */
            return (0);

        if (colorid < 0 || colorid >= NumPrivateColors)
            /* bad colorid, return */
            return (0);

        /* offset into private area */
        colorid += NUMCOLORS - NumPrivateColors;

        newcolor.red   = (red   * 65535);
        newcolor.green = (green * 65535);
        newcolor.blue  = (blue  * 65535);
        newcolor.flags = -1;
        newcolor.pixel = currentgraph->colors[colorid];
        XStoreColor(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)), &newcolor);
        XSetForeground(Xdisplay, xd->gc, currentgraph->colors[colorid]);
        currentgraph->currentcolor = colorid;

        colorid -= NUMCOLORS - NumPrivateColors;

        if (colorid == 0) {
            XSetWindowBackground(Xdisplay, xd->window,
                currentgraph->colors[NUMCOLORS-NumPrivateColors]);
            XSetBackground(Xdisplay, xd->gc,
                currentgraph->colors[NUMCOLORS-NumPrivateColors]);
        }
        if (colorid <= 1) {
            XSetForeground(Xdisplay, xd->xorgc,
                currentgraph->colors[NUMCOLORS - NumPrivateColors] ^
                currentgraph->colors[NUMCOLORS - NumPrivateColors + 1]);
        }
    }
    else if (sced_visual_info->class == TrueColor) {
        if (colorid < 0 || colorid >= NumPrivateColors)
            return (0);
        newcolor.red   = (red   * 65535);
        newcolor.green = (green * 65535);
        newcolor.blue  = (blue  * 65535);
        newcolor.flags = -1;
        newcolor.pixel = 0;
        XAllocColor(Xdisplay,
            DefaultColormap(Xdisplay, DefaultScreen(Xdisplay)), &newcolor);
        XSetForeground(Xdisplay, xd->gc, newcolor.pixel);
        currentgraph->colors[colorid] = newcolor.pixel;

        if (colorid == 0) {
            XSetWindowBackground(Xdisplay, xd->window,
                currentgraph->colors[0]);
            XSetBackground(Xdisplay, xd->gc,
                currentgraph->colors[0]);
        }
        if (colorid <= 1) {
            XSetForeground(Xdisplay, xd->xorgc,
                currentgraph->colors[0] ^ currentgraph->colors[1]);
        }
    }
    return (0);
}


/*ARGSUSED*/
int
X11_DefineLinestyle(linestyleid, mask)

int linestyleid;
int mask;
{
    int imask = 0x80;
    int n = 0, off = 0;
    char *dash;

    if (linestyleid < 0 || linestyleid >= NUMLINESTYLES)
        return (0);

    if (linestyleid == 0 || (mask & 0xff) == 0xff) {
        X11_SetLinestyle(0);
        return (0);
    }

    if ((mask & 0xff) == 0)
        return (0);

    dash = xlinestyles[linestyleid]+2;

    while (!(imask & mask)) {
        imask >>= 1;
        off++;
    }
    for (;;) {
        dash[n] = 0;
        while (imask && (imask & mask)) {
            dash[n]++;
            imask >>= 1;
        }
        n++;
        if (!imask) break;
        dash[n] = 0;
        while (imask && !(imask & mask)) {
            dash[n]++;
            imask >>= 1;
        }
        n++;
        if (!imask) break;
    }

    if (off) {
        if (n & 1)
            dash[n++] = off;
        else
            dash[n-1] += off;
    }

    xlinestyles[linestyleid][0] = (char)0;
    xlinestyles[linestyleid][1] = (char)n;
    X11_SetLinestyle(linestyleid);
    
    return (0);
}


int
X11_SetLinestyle(linestyleid)

int linestyleid;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;
    XGCValues values;

    if (linestyleid < 0 || linestyleid >= NUMLINESTYLES)
        return (0);

    if ((linestyleid == 0) ||
        (DisplayPlanes(Xdisplay, DefaultScreen(Xdisplay)) > 1 &&
        linestyleid != 1)) {
        /* solid if linestyle 0 or if has color,
         * allow only one dashed linestyle
         */
        values.line_style = LineSolid;
    }
    else {
        values.line_style = LineOnOffDash;
    }
    XChangeGC(Xdisplay, xd->gc, GCLineStyle, &values);

    currentgraph->linestyle = linestyleid;
    XSetDashes(Xdisplay, xd->gc, xlinestyles[linestyleid][0],
        &xlinestyles[linestyleid][2], xlinestyles[linestyleid][1]);

    return (0);
}


int
X11_SetColor(colorid)

int colorid;
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    if (xd->gc == xd->xorgc)
        return (0);
    if (currentgraph->graphtype == GR_SCED &&
            dispdev->numcolors - NumPrivateColors >= 8) {
        /* offset into private area */
        if (colorid < 0 || colorid >= NumPrivateColors)
            /* bad colorid, return */
            return (0);
        colorid += NUMCOLORS - NumPrivateColors;
    }
    else {
        if (dispdev->numcolors - NumPrivateColors >= 8) {
            if (colorid < 0 ||
                    colorid >= dispdev->numcolors - NumPrivateColors)
                /* bad colorid, return */
                return (0);
        }
        else {
            if (colorid < 0 || colorid >= dispdev->numcolors)
                /* bad colorid, return */
                return (0);
        }
    }
    currentgraph->currentcolor = colorid;
    XSetForeground(Xdisplay, xd->gc, currentgraph->colors[colorid]);
    return (0);
}


int
X11_Update()
{
    XSync(Xdisplay, 0);
    return (0);
}


int
X11_Clear()
{
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    if (!xd->noclear)
        /* hack so exposures look like they're handled nicely */
        XClearWindow(Xdisplay, xd->window);
    return (0);
}


int
X11_Input(request, response)

REQUEST *request;
RESPONSE *response;
{
    X11devdep *xd;
    XEvent ev;
    int nfds;
    fd_set readfds;
    int nbytes;
    char text[4], c;
    KeySym key;
    GC tempgc;
    extern int ft_intrpt;

    if (currentgraph)
        xd = (X11devdep*)currentgraph->devdep;
    else
        xd = NULL;

    switch (request->option) {
        case char_option:

            if (request->fp) {
                nfds = ConnectionNumber(Xdisplay) > fileno(request->fp) ?
                    ConnectionNumber(Xdisplay) :
                    fileno(request->fp);
            }
            else
                nfds = ConnectionNumber(Xdisplay);

            for (;;) {

                /* first read off the queue before doing the select */
                while (XtAppPending(app_con)) {
                    XtAppNextEvent(app_con, &ev);
                    XtDispatchEvent(&ev);
                }

                FD_ZERO(&readfds);
                FD_SET(ConnectionNumber(Xdisplay), &readfds);
                if (request->fp)
                    FD_SET(fileno(request->fp), &readfds);

                /* block on ConnectionNumber and request->fp */
                select(nfds + 1, &readfds, NULL, NULL, NULL);

                /* handle X events first */
                if (FD_ISSET(ConnectionNumber(Xdisplay), &readfds)) {
                    /* handle ALL X events */
                    while (XtAppPending(app_con)) {
                        XtAppNextEvent(app_con, &ev);
                        if (ev.type == KeyPress && xd &&
                                ev.xkey.window == xd->window) {
                            nbytes = XLookupString((XKeyEvent *)&ev, text, 4,
                                &key, NULL);
                            if (!nbytes) {
                                response->reply.ch = (int)key;
                                /* ignore all but arrow keys */
                                if (response->reply.ch != 0xff51 &&
                                    response->reply.ch != 0xff52 &&
                                    response->reply.ch != 0xff53 &&
                                    response->reply.ch != 0xff54)
                                    continue;
                            }
                            else {
                                response->reply.ch = text[0];
                            }
                            response->x = ev.xkey.x;
                            response->y = currentgraph->absolute.height -
                                ev.xkey.y - 1;
                            response->option = char_option;
                            return (0);
                        }
                        XtDispatchEvent(&ev);
                    }
                }

                if (request->fp) {
                        if (FD_ISSET(fileno(request->fp), &readfds)) {
                        response->reply.ch = cp_inchar(request->fp);
                        goto out;
                    }
                }
            }
            break;

        case button_option:
            /* sit and handle events until get a button selection */
            internalerror("button_option not implemented");
            response->option = error_option;
            return (0);

        case point_option:
            for (;;) {
                XtAppNextEvent(app_con, &ev);
                if (ev.type == ButtonPress && xd &&
                        ev.xbutton.window == xd->window) {
                    switch (ev.xbutton.button) {
                        default:
                        case Button1:
                            response->reply.button = 1;
                            break;
                        case Button2:
                            response->reply.button = 2;
                            break;
                        case Button3:
                            response->reply.button = 4;
                            break;
                        case Button4:
                            response->reply.button = 8;
                            break;
                    }
                    response->x = ev.xbutton.x;
                    response->y = currentgraph->absolute.height -
                        ev.xbutton.y - 1;
                    response->option = button_option;
                    break;
                }
                if (ev.type == KeyPress && xd &&
                        ev.xkey.window == xd->window) {
                    nbytes = XLookupString((XKeyEvent *)&ev, text, 4,
                        &key, NULL);
                    if (!nbytes) {
                        response->reply.ch = (int)key;
                        /* ignore all but arrow keys */
                        if (response->reply.ch != 0xff51 &&
                            response->reply.ch != 0xff52 &&
                            response->reply.ch != 0xff53 &&
                            response->reply.ch != 0xff54)
                            continue;
                    }
                    else {
                        response->reply.ch = text[0];
                    }
                    response->x = ev.xkey.x;
                    response->y = currentgraph->absolute.height -
                        ev.xkey.y - 1;
                    response->option = char_option;
                    break;
                }
                XtDispatchEvent(&ev);
            }
            if (xd && xd->drawghost && xd->firstghost == false) {
                tempgc = xd->gc;
                xd->gc = xd->xorgc;
                (*xd->drawghost)(xd->lastx, xd->lasty, xd->refx, xd->refy);
                xd->gc = tempgc;
                xd->firstghost = true;
            }
            return (0);

        case click_option:
            break;

        case checkup_option:
            while (XtAppPending(app_con)) {
                XtAppNextEvent(app_con, &ev);
                /* trap ^C for pause */
                if (ev.type == KeyPress) {
                    nbytes =
                        XLookupString((XKeyEvent *)&ev, text, 4, &key, NULL);
                    if (!nbytes)
                        c = (int)key;
                    else
                        c = *text;
                    if (c == '\003' /* ^C */) {
                        ft_intrpt = true;;
                        continue;
                    }
                }
                XtDispatchEvent(&ev);
            }
            break;

        default:
            internalerror("unrecognized input type");
            response->option = error_option;
            return (0);
    }

out:
    if (response)
        response->option = request->option;
    return (0);
}


static void
slopelocation(graph, x0, y0)

GRAPH *graph;
int x0, y0;         /* initial position of mouse */
{
    X11devdep *xd = (X11devdep*)graph->devdep;
    Window window = xd->window;
    GC xorgc = xd->xorgc;
    Window rootwindow, childwindow;
    Widget stw;
    int rootx, rooty;
    int x1, y1;
    int x, y;
    unsigned int state;
    double fx0, fx1, fy0, fy1;
    char buf[BSIZE_SP];

    /* hack! hack! */
    if (dispdev->hardcopy) return;
    if (graph->graphtype == GR_GRAF) {
        struct gplot *graf = (struct gplot *)graph->plotdata;
        if (graf) {
            if (graf->opt & (OPT_help | OPT_b))
                return;
            if (!(graf->opt & OPT_mark))
                return;
            if (graf->ref.set) {
                XDrawLine(Xdisplay, window, xorgc, graf->ref.x, graf->scr.YL,
                    graf->ref.x, graf->scr.YU);
                graf->ref.set = false;
            }
            graf_slopelocation(graph, x0, y0, 0, 0);
            if (graf->ref.set) {
                XDrawLine(Xdisplay, window, xorgc, graf->ref.x, graf->scr.YL,
                    graf->ref.x, graf->scr.YU);
            }
        }
        return;
    }

    x1 = x0;
    y1 = y0;
    XQueryPointer(Xdisplay, window, &rootwindow, &childwindow,
        &rootx, &rooty, &x, &y, &state);
    XDrawLine(Xdisplay, window, xorgc, x0, y0, x0, y1-1);
    XDrawLine(Xdisplay, window, xorgc, x0, y1, x1, y1);
    while (state & Button1Mask) {
        if (x != x1 || y != y1) {
            XDrawLine(Xdisplay, window, xorgc, x0, y0, x0, y1-1);
            XDrawLine(Xdisplay, window, xorgc, x0, y1, x1, y1);
            x1 = x;
            y1 = y;
            XDrawLine(Xdisplay, window, xorgc, x0, y0, x0, y1-1);
            XDrawLine(Xdisplay, window, xorgc, x0, y1, x1, y1);
        }
        XQueryPointer(Xdisplay, window, &rootwindow,
            &childwindow, &rootx, &rooty, &x, &y, &state);
    }
    XDrawLine(Xdisplay, window, xorgc, x0, y0, x0, y1-1);
    XDrawLine(Xdisplay, window, xorgc, x0, y1, x1, y1);

    X_ScreentoData(graph, x0, y0, &fx0, &fy0);
    X_ScreentoData(graph, x1, y1, &fx1, &fy1);

    stw = XtNameToWidget(xd->widgets.form, "slopetext");
    if (stw) {
        /* print it out */
        if (x1 == x0 && y1 == y0) {
            /* only one location */
            sprintf(buf, "x0,y0\n%g\n%g", fx0, fy0);
        }
        else {
            /* need to print info about two points */
            sprintf(buf, "x0,y0\n%g\n%g\n\nx1,y1\n%g\n%g\n",
                fx0, fy0, fx1, fy1);
            sprintf(buf + strlen(buf),
                "\ndx,dy\n%g\n%g\n", fx1-fx0, fy1 - fy0);
            if (x1 != x0 && y1 != y0) {
                /* add slope info if both dx and dy are nonzero,
                 * because otherwise either dy/dx or dx/dy is zero,
                 * which is uninteresting
                 */
                sprintf(buf + strlen(buf), "\ndy/dx,dx/dy\n%g\n%g\n",
                (fy1-fy0)/(fx1-fx0), (fx1-fx0)/(fy1-fy0));
            }
        }
        XtVaSetValues(stw, XtNstring, buf, NULL);
    }
}


static void
zoomin(graph)

GRAPH *graph;
{
    X11devdep *xd = (X11devdep*)graph->devdep;
    int x0, y0, x1, y1;
    double fx0, fx1, fy0, fy1, ftemp;
    char buf[BSIZE_SP];
    char buf2[128];
    char *t;
    wordlist *wl;
    int dummy;
    Window rootwindow, childwindow;
    int rootx, rooty;
    unsigned int state;
    int x, y, upperx, uppery, lowerx, lowery;
    Window window = xd->window;
    GC xorgc = xd->xorgc;


    /* open box and get area to zoom in on */

    /* hack! hack! */
    if (graph->graphtype == GR_GRAF) {
        struct gplot *graf = (struct gplot *)graph->plotdata;
        if (graf) {
            if (graf->opt & (OPT_help | OPT_b))
                /* don't want to zoom in if not a plot */
                /* can't handle x-y mode yet */
                return;
        }
    }

    XQueryPointer(Xdisplay, window, &rootwindow,
            &childwindow, &rootx, &rooty, &x0, &y0, &state);

    x = lowerx = x1 = x0 + BOXSIZE;
    y = lowery = y1 = y0 + BOXSIZE;
    upperx = x0;
    uppery = y0;

    XDrawRectangle(Xdisplay, window, xorgc,
        upperx, uppery, lowerx - upperx, lowery - uppery);

    XWarpPointer(Xdisplay, None, xd->window, 0, 0, 0, 0, x1, y1);

    while (state & Button3Mask) {
        if (x != x1 || y != y1) {
            XDrawRectangle(Xdisplay, window, xorgc,
                upperx, uppery, lowerx - upperx, lowery - uppery);
            x1 = x;
            y1 = y;
            /* figure out upper left corner */
            /* remember X11's (and X10's) demented coordinate system */
            if (y0 < y1) {
                uppery = y0;
                lowery = y1;
            }
            else {
                uppery = y1;
                lowery = y0;
            }
            if (x0 < x1) {
                upperx = x0;
                lowerx = x1;
            }
            else {
                upperx = x1;
                lowerx = x0;
            }
            XDrawRectangle(Xdisplay, window, xorgc,
                upperx, uppery, lowerx - upperx, lowery - uppery);
        }
        XQueryPointer(Xdisplay, window, &rootwindow,
            &childwindow, &rootx, &rooty, &x, &y, &state);
    }
    XDrawRectangle(Xdisplay, window, xorgc,
        upperx, uppery, lowerx - upperx, lowery - uppery);

    if (graph->graphtype == GR_GRAF) {
        graf_newdisplay(graph,x0,x1);
        return;
    }

    X_ScreentoData(graph, x0, y0, &fx0, &fy0);
    X_ScreentoData(graph, x1, y1, &fx1, &fy1);

    if (fx0 > fx1) {
        ftemp = fx0;
        fx0 = fx1;
        fx1 = ftemp;
    }
    if (fy0 > fy1) {
        ftemp = fy0;
        fy0 = fy1;
        fy1 = ftemp;
    }

    strncpy(buf2, graph->plotname, sizeof(buf2));
    if (t = strchr(buf2, ':'))
        *t = 0;

    if (!eq(plot_cur->pl_typename, buf2)) {
        (void) sprintf(buf,
   "setplot %s ; %s xlimit %lg %lg ylimit %lg %lg ; setplot $curplot\n",
        buf2, graph->commandline, fx0, fx1, fy0, fy1);
    }
    else {
        (void) sprintf(buf, "%s xlimit %lg %lg ylimit %lg %lg\n",
            graph->commandline, fx0, fx1, fy0, fy1);
    }

    /* hack for Gordon Jacobs */
    /* add to history list if plothistory is set */
    if (cp_getvar("plothistory", VT_BOOL, (char *) &dummy)) {
        wl = cp_parse(buf);
        (void) cp_addhistent(cp_event++, wl);
        wl_free(wl);
    }

    (void) cp_evloop(buf);
}


static void
X_ScreentoData(graph, x, y, fx, fy)

GRAPH *graph;
int x,y;
double *fx, *fy;
{
    /* note: figure something out for log grids XXX */

    *fx = (x - graph->viewportxoff) * graph->aspectratiox
        + graph->datawindow.xmin;
    *fy = ((graph->absolute.height - y - 1) - graph->viewportyoff)
        * graph->aspectratioy
        + graph->datawindow.ymin;
}


void
com_setrdb(wl)

wordlist *wl;
{
    char *str = wl_flatten(wl);
    XrmDatabase db = XtDatabase(Xdisplay);

    XrmPutLineResource(&db, str);
    txfree(str);
}


int
PopUpErrMessage(fp)

FILE *fp;
{
    int cnt = 0;
    char buf[BSIZE_SP];
    String s, t;
    X11devdep *xd = (X11devdep*)currentgraph->devdep;

    while (fgets(buf, BSIZE_SP, fp))
        cnt += strlen(buf);
    rewind(fp);
    s = XtMalloc(cnt + 1);
    t = s;
    cnt = 0;
    while (fgets(buf, BSIZE_SP, fp)) {
        strcpy(t+cnt, buf);
        cnt += strlen(buf);
    }
    PopUpErr(s, &xd->widgets);
    XtFree(s);
    return (True);
}


void
ScedESC()

{
    /* send an ESC keypress to sced */
    static XKeyEvent ev;

    if (SCEDactive() == SCED_inhelp) {
        ev.type = KeyPress;
        ev.display = Xdisplay;
        ev.window = XtWindow(sced_view);
        ev.keycode = XKeysymToKeycode(Xdisplay, XK_Escape);
        XSendEvent(Xdisplay, XtWindow(sced_view), False, (long)KeyPressMask,
            (XEvent*)&ev);
    }
}

#else /* not HAVE_X11 */

/* ARGSUSED */
void
com_setrdb(wl)

wordlist *wl;
{
/* be silent */
}

int
PopUpErrMessage(fp)

File *fp;
{
    return (0);
}

void
ScedESC()
{
    return;
}

#endif /* HAVE_X11 */
