/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED library access routines.
 */

#include "spice.h"
#include "sced.h"
#include <string.h>

/* linked list to keep track of models actually referenced */
struct wl {
    char *wl_name;
    struct wl *wl_next;
};

/* list of library elements */
struct lib {
    char *lib_name;
    long lib_offset;
    struct wl *lib_info;
    struct lib *lib_next;
};



/***********************************************************************
 *
 * Routines to access the device library.
 *
 ***********************************************************************/

static struct lib *Lib;
static char *LibName;

#ifdef __STDC__
static struct lib *allocate_lib(char*,long,struct wl*);
#else
static struct lib *allocate_lib();
#endif


void
LibraryOpen(name)

/* Create a list of library devices with file offsets. */
char *name;
{
    FILE *fp;
    char buf[512],*c,*d,*fgets();
    struct lib *l;
    int InSymbol = False;
    long ftold, ftell();
    char *nam;

    if (Lib != NULL) return;
    fp = POpen(name,"r",&nam);
    if (fp == NULL) return;

    (void) fgets(buf,512,fp);
    if (buf[0] != 'S' || buf[1] != 'C' || buf[2] != 'E' || buf[3] != 'D') {
        fclose(fp);
        return;
    }
    LibName = nam;

    ftold = ftell(fp);
    while (fgets(buf,512,fp) != NULL) {
        if (*buf == '9' && isspace(*(buf+1)) && !InSymbol) {
            c = buf+2;
            while (isspace(*c)) c++;
            for (d = c+1; d; d++) {
                if (isspace(*d) || *d == ';') {
                    *d = '\0';
                    break;
                }
            }
            if (Lib == NULL)
                l = Lib = allocate_lib(c,ftold,(struct wl *)NULL);
            else {
                l->lib_next = allocate_lib(c,ftold,(struct wl *)NULL);
                l = l->lib_next;
            }
            InSymbol = True;
        }
        if (*buf == 'E' && (isspace(*(buf+1)) || *(buf+1) == '\0'))
            InSymbol = False;
        ftold = ftell(fp);
    }
    fclose(fp);
}


FILE *
OpenDevice(name)

/* Return a FILE pointer to the library at the
 * correct position for the device.
 */
char *name;
{
    struct lib *l;
    FILE *fp;

    for (l = Lib; l; l = l->lib_next) {
        if (!strcmp(name,l->lib_name)) {
            fp = fopen(LibName,"r");
            if (fp == NULL) return (NULL);
            fseek(fp,l->lib_offset,0);
            return (fp);
        }
    }
    return (NULL);
}


int
IsCellInLib(name)

/* Return True is name is found in the device library. */
char *name;
{
    struct lib *l;
    FILE *fp;

    for (l = Lib; l; l = l->lib_next)
        if (!strcmp(name,l->lib_name))
            return (True);
    return (False);
}


void
SetLibraryChoices()

/* Create a menu of the devices in the device library */
{
    struct lib *w;
    int i;

    for (i = 0,w = Lib;  w; w = w->lib_next) {
        DeviceMenu[i].mEntry = w->lib_name;
        i++;
        if (i == 79) break;
    }
    DeviceMenu[i].mEntry = NULL;
    /*
    FixMenuPrefix(DeviceMenu);
    */
}


static struct lib *
allocate_lib(Name,Offset,Info)

char *Name;
long Offset;
struct wl *Info;
{
    struct lib *l;

    l = alloc(lib);
    if (l == NULL) MallocFailed();
    l->lib_next = NULL;
    l->lib_name = CopyString(Name);
    l->lib_offset = Offset;
    l->lib_info = Info;
    return (l);
}



/***********************************************************************
 *
 * Routines to access the spice model library.
 *
 ***********************************************************************/

static struct lib *Mod;
static char *ModLibName;

struct ml {
    char *ml_name;
    struct ml *ml_next;
    int ml_level;
};

/* referenced model list head */
static struct ml *Models;

static int Mlevel;

#ifdef __STDC__
static struct wl *allocate_wl(char*);
static struct ml *allocate_ml(char*,int);
#else
static struct wl *allocate_wl();
static struct ml *allocate_ml();
#endif


void
ModelLibraryOpen(name)

/* Create a list of model library entries and file offsets. */
char *name;
{
    FILE *fp;
    char buf[512],*c,*d,*fgets();
    struct lib *l;
    long ftold, ftell();
    char *nam;

    if (Mod != NULL) return;
    fp = POpen(name,"r",&nam);
    if (fp == NULL) return;
    (void) fgets(buf,512,fp);
    if (buf[0] != 'S' || buf[1] != 'C' || buf[2] != 'E' || buf[3] != 'D') {
        fclose(fp);
        return;
    }
    ModLibName = nam;

    ftold = ftell(fp);
    while (fgets(buf,512,fp) != NULL) {
        if (ciprefix(".model",buf)) {
            c = buf+7;
            while (isspace(*c)) c++;
            for (d = c+1; d; d++) {
                if (isspace(*d) || ispunct(*d)) {
                    *d = '\0';
                    break;
                }
            }
            if (Mod == NULL)
                l = Mod = allocate_lib(c,ftold,(struct wl *)NULL);
            else {
                l->lib_next = allocate_lib(c,ftold,(struct wl *)NULL); 
                l = l->lib_next;
            }
        }
        ftold = ftell(fp);
    }
}


void
QueueModel(name)

/* Put model in referenced list if it is not there already.
 * If name is NULL, increment the "stack" pointer.
 */
char *name;
{
    struct ml *m;

    if (name) {
        while (isspace(*name)) name++;
        if (!*name)
            return;
        for (m = Models; m; m = m->ml_next)
            if (cieq(name,m->ml_name)) return;
        m = allocate_ml(name,Mlevel);
        m->ml_next = Models;
        Models = m;
    }
    else
        Mlevel++;
}


struct line *
PrintModels()

/* Return the spice lines for the models referenced in the current
 * context.
 */
{
    struct ml *m;
    struct line *d,*d1,*d0;

    d0 = NULL;
    while (Models && Models->ml_level == Mlevel) {
        d1 = GetModelText(Models->ml_name);
        if (d1 != NULL) {
            if (d0 == NULL)
                d = d0 = d1;
            else
                d->li_next = d1;
            for (; d->li_next; d = d->li_next) ;
        }
        m = Models;
        Models = Models->ml_next;
        txfree(m->ml_name);
        tfree(m);
    }
    Mlevel--;
    return (d0);
}


struct line *
GetModelText(name)

/* Return the text for the model. */
char *name;
{
    FILE *ip;
    struct lib *l;
    struct line *d,*d0 = NULL;
    struct wl *w,*w0;
    char buf[512],*c;

    for  (l = Mod; l;  l = l->lib_next) {
        if (cieq(name,l->lib_name)) {
            if (l->lib_info == NULL) {
                ip = fopen(ModLibName,"r");
                if (ip == NULL) return (NULL);
                fseek(ip,l->lib_offset,0);
                if (fgets(buf,512,ip) == NULL) return (NULL);
                if ((c = strchr(buf,'\n')) != NULL) *c = '\0';
                w = w0 = allocate_wl(buf);
                while (fgets(buf,512,ip) != NULL) {
                    if (ciprefix(".model",buf)) break;
                    if ((c = strchr(buf,'\n')) != NULL) *c = '\0';
                    w->wl_next = allocate_wl(buf);
                    w = w->wl_next;
                }
                fclose(ip);
                l->lib_info = w0;
            }
            d = d0 = AllocateLine(l->lib_info->wl_name);
            for (w = l->lib_info->wl_next; w; w = w->wl_next) {
                d->li_next = AllocateLine(w->wl_name);
                d = d->li_next;
            }
            return (d0);
        }
    }
    return (NULL);
}


static struct wl *
allocate_wl(string)

char *string;
{
    struct wl *w;

    w = alloc(wl);
    if (w == NULL) MallocFailed();
    w->wl_next = NULL;
    w->wl_name = CopyString(string);
    return (w);
}


static struct ml *
allocate_ml(string,level)

char *string;
int level;
{
    struct ml *m;

    m = alloc(ml);
    if (m == NULL) MallocFailed();
    m->ml_next = NULL;
    m->ml_name = CopyString(string);
    m->ml_level = level;
    return (m);
}



/***********************************************************************
 *
 * Routine to delete libraries.
 *
 ***********************************************************************/


void
DestroySCEDlibs()

{
    struct lib *l;
    struct wl *w;
    struct ml *m;

    for (l = Lib; l; l = Lib) {
        Lib = l->lib_next;
        txfree(l->lib_name);
        tfree(l);
    }
    Lib = NULL;
    tfree(LibName);

    for (l = Mod; l; l = Mod) {
        Mod = l->lib_next;
        txfree(l->lib_name);
        for (w = l->lib_info; w; w = l->lib_info) {
            l->lib_info = w->wl_next;
            txfree(w->wl_name);
            tfree(w);
        }
        tfree(l);
    }
    Mod = NULL;
    tfree(ModLibName);

    for (m = Models; m; m = Models) {
        Models = m->ml_next;
        txfree(m->ml_name);
        tfree(m);
    }
    Models = NULL;
    Mlevel = 0;
}


void
LibraryClose()

{
    struct lib *l;

    for (l = Lib; l; l = Lib) {
        Lib = l->lib_next;
        txfree(l->lib_name);
        tfree(l);
    }
    Lib = NULL;
    tfree(LibName);
}


void
ModelLibraryClose()

{
    struct wl *w;
    struct ml *m;
    struct lib *l;

    for (l = Mod; l; l = Mod) {
        Mod = l->lib_next;
        txfree(l->lib_name);
        for (w = l->lib_info; w; w = l->lib_info) {
            l->lib_info = w->wl_next;
            txfree(w->wl_name);
            tfree(w);
        }
        tfree(l);
    }
    Mod = NULL;
    tfree(ModLibName);

    for (m = Models; m; m = Models) {
        Models = m->ml_next;
        txfree(m->ml_name);
        tfree(m);
    }
    Models = NULL;
    Mlevel = 0;
}
