/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED code for handling mutual inductors.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"



/***********************************************************************
 *
 * Routines to keep track of mutual inductors.
 *
 ***********************************************************************/

extern char *MenuMUTUL;

#ifdef __STDC__
static int mut_kb(struct prpty*,struct o*,struct o*,int*);
static int add_mutual(int*);
static struct o *select_inductor(long,long,struct s*);
static struct o *get_mut(long,long,struct s*);
#else
static int mut_kb();
static int add_mutual();
static struct o *select_inductor();
static struct o *get_mut();
#endif


void
ShowMutual(LookedAhead)

/* Show the mutual inductors, and allow additions or deletions */
int *LookedAhead;
{
    struct prpty *PDesc,*PNext;
    struct o *Pointer1,*Pointer2;
    char *TypeIn;
    long L1X,L2X,L1Y,L2Y;
    int i, mutfound = False;

    MenuSelect(MenuMUTUL);

again:
    for (PDesc = Parameters.kpCellDesc->sPrptyList; PDesc; PDesc = PNext) {
        PNext = PDesc->prpty_Succ;
        if (PDesc->prpty_Value != P_MUT) continue;
        /* Referenced by lower left corner */
        L1X = PDesc->prpty_Data->p_mut.x1;
        L1Y = PDesc->prpty_Data->p_mut.y1;
        L2X = PDesc->prpty_Data->p_mut.x2;
        L2Y = PDesc->prpty_Data->p_mut.y2;
        Pointer1 = get_mut(L1X,L1Y,Parameters.kpCellDesc);
        Pointer2 = get_mut(L2X,L2Y,Parameters.kpCellDesc);
        if (Pointer1 == NULL || Pointer2 == NULL)
            continue;
        mutfound = True;
        ShowCurrentObject(Pointer1,DISPLAY);
        ShowCurrentObject(Pointer2,DISPLAY);
        sprintf(TypeOut,"k = %g   (a add, d delete) > ",
            PDesc->prpty_Data->p_mut.coeff);
        ShowPrompt(TypeOut);
        i = mut_kb(PDesc,Pointer1,Pointer2,LookedAhead);
        if (i < 0)
            goto quit;
        if (i > 0)
            goto again;
    }
    if (mutfound == False) {
        ShowPrompt(
            "No mutual inductors found.  Enter 'a' to add mutual inductor > ");
        if (mut_kb(NULL,NULL,NULL,LookedAhead) > 0)
            goto again;
    }
quit:
    ShowCurrentObject((struct o *)NULL,ERASE);
    MenuDeselect(MenuMUTUL);
    ErasePrompt();
}


static int
mut_kb(PDesc,Pointer1,Pointer2,LookedAhead)

/* Process keyboard input */
struct prpty  *PDesc;
struct o *Pointer1,*Pointer2;
int *LookedAhead;
{
    struct prpty *pd;
    char InChar;

    InChar = FBGetchar(DISPLAY);

    if (Pointer1 != NULL)
        ShowCurrentObject(Pointer1,ERASE);
    if (Pointer2 != NULL)
        ShowCurrentObject(Pointer2,ERASE);

    if (InChar == ESCAPE)
        return (-1);
    if (InChar == '\0' || InChar == '\n' || InChar == '\r')
        return (0);
    if (InChar == 'a' || InChar == 'A')
        return (add_mutual(LookedAhead));
    if (InChar == 'd' || InChar == 'D') {
        if (PDesc == NULL) {
            ShowPromptAndWait("No mutual inductor to delete.");
            return (0);
        }
        pd = Parameters.kpCellDesc->sPrptyList;
        if (pd == PDesc)
            Parameters.kpCellDesc->sPrptyList = PDesc->prpty_Succ;
        else {
            /* must be in the list */
            for (; pd && pd->prpty_Succ != PDesc; pd = pd->prpty_Succ) ;
            pd->prpty_Succ = PDesc->prpty_Succ;
        }
        free(PDesc->prpty_Data);
        tfree(PDesc);
        Parameters.kpModified = True;
        ShowPromptAndWait("Mutual inductance deleted.");
    }
    return (0);
}


static int
add_mutual(LookedAhead)

/* Add a mutual inductor */
int *LookedAhead;
{
    struct o *Pointer1,*Pointer2;
    struct prpty *PDesc;
    char value[32],*TypeIn;
    double kv;
    int ret = True;

top:        
    ShowPrompt("Point to first coupled inductor.");
    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_UND:
            case PL_CMD:
            case PL_ESC:
                ret = False;
                goto quit;
            case PL_PCW:
                if ((Pointer1 =
                    select_inductor(SCursor.kcRawX,SCursor.kcRawY,
                    Parameters.kpCellDesc)) != NULL) {
                    ShowCurrentObject(Pointer1,DISPLAY);
                    break;
                }
                continue;
        }
        break;
    }

    ShowPrompt("Point to second coupled inductor.");
    loop {
        switch (PointLoop(LookedAhead)) {
            case PL_UND:
                goto top;
            case PL_ESC:
            case PL_CMD:
                ret = False;
                goto quit;
            case PL_PCW:
                if ((Pointer2 =
                    select_inductor(SCursor.kcRawX,SCursor.kcRawY,
                    Parameters.kpCellDesc)) != NULL &&
                    Pointer2 != Pointer1) {
                    ShowCurrentObject(Pointer2,DISPLAY);
                    break;
                }
                continue;
        }
        break;
    }

    do {
        ShowPrompt("Enter SPICE coupling factor: k = ");
        if ((TypeIn = FBEdit(NULL)) == NULL) {
            ret = False;
            goto quit;
        }
        strcpy(value,TypeIn);
    }
    while (sscanf(value,"%le",&kv) < 1 || kv < 0 || kv > 1);

    /* see if this mutual already exists */
    PDesc = Parameters.kpCellDesc->sPrptyList;
    for (; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_MUT) continue;
        if ((Pointer1->oLeft == PDesc->prpty_Data->p_mut.x1 &&
            Pointer1->oBottom == PDesc->prpty_Data->p_mut.y1 &&
            Pointer2->oLeft == PDesc->prpty_Data->p_mut.x2 &&
            Pointer2->oBottom == PDesc->prpty_Data->p_mut.y2) ||
            (Pointer2->oLeft == PDesc->prpty_Data->p_mut.x1 &&
            Pointer2->oBottom == PDesc->prpty_Data->p_mut.y1 &&
            Pointer1->oLeft == PDesc->prpty_Data->p_mut.x2 &&
            Pointer1->oBottom == PDesc->prpty_Data->p_mut.y2)) {
            PDesc->prpty_Data->p_mut.coeff = kv;
            ShowPromptAndWait("Existing mutual inductor updated");
            goto quit;
        }
    }

    sprintf(TypeOut,"%ld %ld %ld %ld %s",
        Pointer1->oLeft,Pointer1->oBottom,
        Pointer2->oLeft,Pointer2->oBottom,
        value);
    CDAddProperty(Parameters.kpCellDesc,(struct o*)NULL,P_MUT,TypeOut);

quit:
    ShowCurrentObject((struct o *)NULL,ERASE);
    return (ret);
}


static struct o *
select_inductor(X,Y,CellDesc)

long X,Y;
struct s *CellDesc;
{
    struct g *GenDesc;
    struct o *Pointer;
    char *name;

    if (Not CDInitGen(CellDesc,0,
        X,Y,X,Y,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        name = ((struct c *)Pointer->oRep)->cMaster->mName;
        if (!cieq(name,"ind")) continue;
        /* folks, we have an inductor */
        tfree(GenDesc);
        return (Pointer);
    }
    return (NULL);
}


static struct o *
get_mut(X,Y,CellDesc)

/* Return the inductor with lower left corner at X,Y */
long X,Y;
struct s *CellDesc;
{
    struct g *GenDesc;
    struct o *Pointer;
    char *name;

    if (Not CDInitGen(CellDesc,0,
        X+100,Y+100,X+100,Y+100,&GenDesc)) MallocFailed();

    loop {
        CDGen(CellDesc,GenDesc,&Pointer);
        if (Pointer == NULL) break;
        if (Pointer->oInfo == SQ_GONE) continue;

        name = ((struct c *)Pointer->oRep)->cMaster->mName;
        if (!cieq(name,"ind")) continue;
        if (X == Pointer->oLeft && Y == Pointer->oBottom) {
            tfree(GenDesc);
            return (Pointer);
        }
    }
    return (NULL);
}


struct line *
PrintMutual(CellDesc)

/* Return the SPICE lines for the mutual inductors */
struct s *CellDesc;
{
    struct prpty *PDesc,*pd1,*pd2;
    struct o *Pointer1,*Pointer2;
    struct line *d,*d0 = NULL;
    char name1[8],name2[8],value[32];
    long L1X,L2X,L1Y,L2Y;
    int count = 0;

    PDesc = CellDesc->sPrptyList;
    for (; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_MUT) continue;
        L1X = PDesc->prpty_Data->p_mut.x1;
        L1Y = PDesc->prpty_Data->p_mut.y1;
        L2X = PDesc->prpty_Data->p_mut.x2;
        L2Y = PDesc->prpty_Data->p_mut.y2;
        Pointer1 = get_mut(L1X,L1Y,CellDesc);
        Pointer2 = get_mut(L2X,L2Y,CellDesc);
        if (Pointer1 == NULL || Pointer2 == NULL)
            continue;

        strcpy(name1,GetName(Pointer1->oPrptyList));
        strcpy(name2,GetName(Pointer2->oPrptyList));
        count++;
        sprintf(TypeOut,"K%d %s %s %g",count,name1,name2,
            PDesc->prpty_Data->p_mut.coeff);
        if (d0 == NULL)
            d = d0 = AllocateLine(TypeOut);
        else {
            d->li_next = AllocateLine(TypeOut);
            d = d->li_next;
        }
    }
    return (d0);
}


char *
GetName(PDesc)

/* Return the device name, as known to SPICE */
struct prpty *PDesc;
{
    static char nbuf[8];

    for (; PDesc; PDesc = PDesc->prpty_Succ) {
        if (PDesc->prpty_Value != P_NAME) continue;
        sprintf(nbuf,"%s%d",PDesc->prpty_Data->p_name.name,
            PDesc->prpty_Data->p_name.num);
        return (nbuf);
    }
    return (NULL);
}

