/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Fast CIF file parser.
 * 
 * It is easy to use it to write programs that traverse a CIF file. 
 *
 * Examples follows.
 * 1.    A program that translates CIF to STREAM.
 * 2.    A pattern factoring program for PG tape generation.
 *       See TI's PFP.
 * 3.    A statistics generation program. 
 * 4.    A program that extracts each symbol, generates a name for it, and
 *       writes it to a file.  See the CD package.
 * 5.    A program that builds a binary representation of the CIF file that can
 *       then be edited via procedure calls and then written back to the CIF
 *       file.  See the CD package.
 * 
 * Here's how you use it.
 * 
 * First, you include the files actions.c and parser.h.
 * Then you invoke PCIF(CIFFileName,StatusString,StatusInt). 
 * Each time the parser recognizes a CIF command, it invokes an action routine.
 * You should fill in the action routines in the file Actions.c.
 * Each action routine name is prefixed by A.
 * 
 * Each parser routine and global variable name is prefixed by P.
 * StatusInt == PFAILED if the parse failed and an error message along with
 * about where in the CIF file the error is in the StatusString.
 * Else StatusInt == PSUCCEEDED and StatusString == "".
 * 
 */

#include "cddefs.h"
#include "cdparser.h"

FILE *PCIFFileDesc;
char PStatus[80*3+1];
long PInt;
int  PChar;
char PString[PSTRINGSIZE];
int  PReturned;

#ifdef __STDC__
static int  p_primitive_cmd(void);
static int  p_end(void);
static int  p_symbol(void);
static int  p_delete_symbol(void);
static int  p_call(void);
static int  p_polygon(void);
static int  p_box(void);
static int  p_round_flash(void);
static int  p_wire(void);
static int  p_path(struct p**);
static int  p_point(long*,long*);
static int  p_layer(void);
static int  p_user_extension(void);
static int  p_comment(void);
static void p_error(char*);
static void p_error_EOF(void);
static void p_error_nosemi(void);
static void p_error_badlayer(int,char*);
static void p_error_CD(void);
static int  p_white_space(int,int);
static int  p_white_space_1(int);
static int  p_look_ahead(int,int);
static int  p_look_for_semi(void);
static int  p_integer(int);
#else
static int  p_primitive_cmd();
static int  p_end();
static int  p_symbol();
static int  p_delete_symbol();
static int  p_call();
static int  p_polygon();
static int  p_box();
static int  p_round_flash();
static int  p_wire();
static int  p_path();
static int  p_point();
static int  p_layer();
static int  p_user_extension();
static int  p_comment();
static void p_error();
static void p_error_EOF();
static void p_error_nosemi();
static void p_error_badlayer();
static void p_error_CD();
static int  p_white_space();
static int  p_white_space_1();
static int  p_white_space_2();
static int  p_white_space_1();
static int  p_look_ahead();
static int  p_look_for_semi();
static int  p_integer();
#endif

void
PCIF(CIFFileName,StatusString,StatusInt)
char *CIFFileName,**StatusString;
int *StatusInt;
{
    PStatus[0] = EOS;
    *StatusString = PStatus;
    if((PCIFFileDesc = POpen(CIFFileName,"r",(char **)NULL))
        == NULL){
        PChar = EOF;
        p_error("Can't open CIF file.");
        *StatusInt = PFAILED;
        return;
    }
    loop{
        PReturned = PCharacter(PSTRIPWHITESPACE2,PDONTFAILONEOF);
        if(PReturned == PFAILED){
            fclose(PCIFFileDesc);
            *StatusInt = PFAILED;
            return;
        }
        elif(PChar == 'D'){
            PReturned = PCharacter(PSTRIPWHITESPACE2,PFAILONEOF);
            if(PReturned == PFAILED){
                fclose(PCIFFileDesc);
                *StatusInt = PFAILED;
                return;
            }
            elif(PChar == 'S'){
                if(p_symbol() == PFAILED){
                    fclose(PCIFFileDesc);
                    *StatusInt = PFAILED;
                    return;
                }
            }
            elif(PChar == 'D'){
                if(p_delete_symbol() == PFAILED){
                    fclose(PCIFFileDesc);
                    *StatusInt = PFAILED;
                    return;
                }
            }
        }
        elif(PChar == 'E'){
            *StatusInt = p_end();
            return;
        }
        elif((PReturned = p_primitive_cmd()) == PFAILED){
            fclose(PCIFFileDesc);
            *StatusInt = PFAILED;
            return;
        }
        elif(PReturned == PNOTAPPLICABLE){
            p_error("Can't understand next command.");
            fclose(PCIFFileDesc);
            *StatusInt = PFAILED;
            return;
        }
    }
}

static int
p_primitive_cmd()
{
    if(PChar == 'P')
        return(p_polygon());                
    elif(PChar == 'B')
        return(p_box());
    elif(PChar == 'W')
        return(p_wire());
    elif(PChar == 'L')
        return(p_layer());
    elif(PChar == 'C')
        return(p_call());
    elif(PChar == 'R')
        return(p_round_flash());
    elif('0' <= PChar And PChar <= '9')
        return(p_user_extension());
    elif(PChar == '(')
        return(p_comment());
    elif(PChar == ';')
        return(PSUCCEEDED);
    else
        return(PNOTAPPLICABLE);
}

static int
p_end()
{
    AEnd();
    fclose(PCIFFileDesc);
    return(PSUCCEEDED);
}

static int
p_symbol()
{
    char For;
    int SymbolNum;
    long A,B;

#ifdef TRACE
fprintf(stderr," ENTERING PSYMBOL\n");
#endif
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    SymbolNum = PInt;
    A = B = 1;
    /* look for scaling factor */
    for(For = '0'; For <= '9'; ++For){
        PReturned = p_look_ahead(PSTRIPWHITESPACE3,For);
        if(PReturned == PFAILED)
            return(PFAILED);
        if(PChar == For){
            ungetc(For,PCIFFileDesc);
            if(p_point(&A,&B) == PFAILED)
                return(PFAILED);
            break;
        }
    }
    PReturned = p_look_for_semi();
    if(PReturned == PFAILED)
        return(PFAILED);
    if(ABeginSymbol((int)SymbolNum,(int)A,(int)B) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    if(PReturned == PFAILED)
        return(PFAILED);
    loop{
        PReturned = PCharacter(PSTRIPWHITESPACE1,PFAILONEOF);
        if(PReturned == PFAILED)
            return(PFAILED);
        if((PReturned = p_primitive_cmd()) == PFAILED)
            return(PFAILED);
        elif(PReturned == PNOTAPPLICABLE)
            break;
    }
    if(PChar != 'D')
        return(PFAILED);
    PReturned = PCharacter(PSTRIPWHITESPACE2,PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    if(PChar != 'F')
        return(PFAILED);
    PReturned = p_look_for_semi();
    if(PReturned == PFAILED)
        return(PFAILED);
    AEndSymbol();
    return(PSUCCEEDED);
}

static int
p_delete_symbol()
{
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    else{
        PReturned = p_look_for_semi();
        if(PReturned == PFAILED)
            return(PFAILED);
        ADeleteSymbol(PInt);
        return(PSUCCEEDED);
    }
}

static int
p_call()
{
    int SymbolNum;
    long X,Y;

    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    SymbolNum = PInt;
    if(ABeginCall(SymbolNum) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    loop{
        PReturned = PCharacter(PSTRIPWHITESPACE2,PFAILONEOF);
        if(PReturned == PFAILED)
            return(PFAILED);
        elif(PChar == 'T'){
            if(p_point(&X,&Y) == PFAILED){
                p_error("Can't parse translation transform.");
                return(PFAILED);
                }
            elif(AT(CDTRANSLATE,X,Y) == PFAILED){
                p_error_CD();
                return(PFAILED);
            }
        }
        elif(PChar == 'M'){
            PReturned = PCharacter(PSTRIPWHITESPACE2,PFAILONEOF);
            if(PReturned == PFAILED){
                p_error_EOF();
                return(PFAILED);
            }
            elif(PChar == 'X'){
                if(AT(CDMIRRORX,X,Y) == PFAILED){
                    p_error_CD();
                    return(PFAILED);
                }
            }
            elif(PChar == 'Y'){
                if(AT(CDMIRRORY,X,Y) == PFAILED){
                    p_error_CD();
                    return(PFAILED);
                }
            }
            else{
                p_error("Can't parse mirror transform.");
                return(PFAILED);
            }
        }
        elif(PChar == 'R'){
            if(p_point(&X,&Y) == PFAILED){
                p_error("Can't parse rotation transform.");
                return(PFAILED);
            }
            if(AT(CDROTATE,X,Y) == PFAILED){
                p_error_CD();
                return(PFAILED);
            }
        }
        elif(PChar == ';')
            break;
        else{
            p_error("Can't parse transformation.");
            return(PFAILED);
        }
    } 
    if(AEndCall() == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    return(PSUCCEEDED);
}

static int
p_polygon()
{
    struct p *Path, *Pt;

#ifdef TRACE
fprintf(stderr," ENTERING PPOLYGON\n");
#endif
    if(p_path(&Path) == PFAILED)
        return(PFAILED);
#ifdef TRACE
fprintf(stderr," LEAVING PPATH\n");
#endif

/* ignore polygons found in input

    if(APolygon(Path) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
*/
    for (Pt = Path; Pt; Pt = Path) {
        Path = Pt->pSucc;
        txfree(Pt);
    }


#ifdef TRACE
fprintf(stderr," LEAVING PPOLYGON\n");
#endif
    return(PSUCCEEDED);
}

static int
p_box()
{
    long Length,Width,X,Y;
    long XDirection,YDirection;

    XDirection = 1;
    YDirection = 0;
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    Length = PInt;
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    Width = PInt;
    if(p_point(&X,&Y) == PFAILED)
        return(PFAILED);
    PReturned = p_look_for_semi();
    if(PReturned == PFAILED)
        return(PFAILED);
    elif(PChar != ';'){
        if(p_point(&XDirection,&YDirection) == PFAILED)
            return(PFAILED);
        PReturned = p_look_for_semi();
        if(PReturned == PFAILED Or PChar != ';'){
            p_error_nosemi();
            return(PFAILED);
        }
    }
    if(ABox(Length,Width,X,Y,(int)XDirection,(int)YDirection) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    return(PSUCCEEDED);
}

static int
p_round_flash()
{
    long Width,X,Y;

    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    Width = PInt;
    if(p_point(&X,&Y) == PFAILED)
        return(PFAILED);
    PReturned = p_look_for_semi();
    if(PReturned == PFAILED)
        return(PFAILED);
    elif(PChar != ';'){
        p_error_nosemi();
        return(PFAILED);
    }
    if(ARoundFlash(Width,X,Y) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    return(PSUCCEEDED);
}

static int
p_wire()
{
    long Width;
    struct p *Path;

#ifdef TRACE
fprintf(stderr," ENTERING PWIRE\n");
#endif
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    Width = PInt;
#ifdef TRACE
fprintf(stderr," ENTERING PPATH\n");
#endif
    if(p_path(&Path) == PFAILED)
        return(PFAILED);
    if(AWire(Width,Path) == PFAILED){
        p_error_CD();
        return(PFAILED);
    }
    return(PSUCCEEDED);
}

static int
p_path(Path)
struct p **Path;
{
    long X,Y;
    struct p *Pair;

    *Path = NULL;
    loop{
        PReturned = p_look_for_semi();
#ifdef TRACE
if(PReturned == PFAILED)
    fprintf(stderr," PPATH FAILED AFTER PLOOKAHEAD\n");
#endif
        if(PReturned == PFAILED)
            return(PFAILED);
        elif(PChar == ';')
            break; 
        else{
            if(p_point(&X,&Y) == PFAILED)
                return(PFAILED);
#ifdef TRACE
fprintf(stderr," PPATH POINT %ld, %ld\n",X,Y);
#endif
            if((Pair = alloc(p)) == NULL){
                p_error("Out of memory.");
                return(PFAILED);
            }
            Pair->pSucc = *Path;
            Pair->pX = X;
            Pair->pY = Y;
            *Path = Pair;
        }
    }
    return(PSUCCEEDED);
}

static int
p_point(X,Y)
long *X,*Y;
{
    /* it is assumed that a LookAhead is done prior to calling p_point */
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    *X = PInt;
    PReturned = p_look_for_semi();
    if(PReturned == PFAILED)
        return(PFAILED);
    elif(PChar == ';'){
        p_error("Bad X,Y path element.");
        return(PFAILED);
    }
    PReturned = p_integer(PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    *Y = PInt;
    return(PSUCCEEDED);
}

static int
p_layer()
{
    char Technology,Mask[4];
    int i;

    Mask[0] = Mask[1] = Mask[2] = ' '; Mask[3] = '\0';
    PReturned = PCharacter(PSTRIPWHITESPACE2,PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    if(PChar == ';'){
        p_error("At least one character expected after L in layer command.");
        return(PFAILED);
    }
    Technology = PChar;
    for(i=0; i<3; ++i){
        if((PChar = getc(PCIFFileDesc)) == EOF){
            p_error_EOF();
            return(PFAILED);
        }
        elif(PChar == ';'){
            if(ALayer(Technology,Mask) == PSUCCEEDED)
                return(PSUCCEEDED);
            else{
                p_error_badlayer(Technology,Mask);
                return(PFAILED);
            }
        }
        Mask[i] = PChar;
        /* check for valid CIF layer name character */
        if(Not ((PChar >= '0' And PChar <= '9') Or
            (PChar >= 'A' And PChar <= 'Z'))){
            p_error_badlayer(Technology,Mask);
            return(PFAILED);
        }
    }
    /* clear the semicolon */
    PReturned = PCharacter(PSTRIPWHITESPACE1,PFAILONEOF);
    if(PReturned == PFAILED)
        return(PFAILED);
    elif(PChar == ';'){
        if(ALayer(Technology,Mask) == PSUCCEEDED)
            return(PSUCCEEDED);
        else{
            p_error_badlayer(Technology,Mask);
            return(PFAILED);
        }
    }
    p_error("Illegal CIF layer name with > 4 characters discovered.");
    return(PFAILED);
}

static int
p_user_extension()
{
    char Digit;
    int Int1;

    Digit = PChar; 
    Int1 = 0;
    loop{
        PReturned = PCharacter(PLEAVEWHITESPACE,PFAILONEOF);
        if(PReturned == PFAILED)
            return(PFAILED);
        elif(PChar == ';'){
            PString[Int1] = EOS;
            if(AUserExtension(Digit,PString) == PFAILED){
                p_error_CD();
                return(PFAILED);
            }
#ifdef TRACE
fprintf(stderr,"PString = %s\n",PString);
#endif
            return(PSUCCEEDED);
        }
        elif(Int1 == PSTRINGSIZE){
            p_error("User extension command longer than 1920 characters.");
            return(PFAILED);
        }
        else PString[Int1++] = PChar;
    }
}

static int
p_comment()
{
    int Int1;

    Int1 = 0;
    loop{
        PReturned = PCharacter(PLEAVEWHITESPACE,PFAILONEOF);
        if(PReturned == PFAILED)
            return(PFAILED); 
        elif(PChar == ')'){
            PReturned = PCharacter(PSTRIPWHITESPACE1,PFAILONEOF);
            if(PReturned == PFAILED)
                return(PFAILED); 
            elif(PChar == ';'){
                PString[Int1] = EOS;
                AComment(PString);
                return(PSUCCEEDED);
            }
            elif(Int1 == PSTRINGSIZE){
                p_error("Comment command longer than 1920 characters.");
                return(PFAILED);
            }
            else{
                p_error_nosemi();
                return(PFAILED);
            }
        }
        else PString[Int1++] = PChar;
    }
}

static void
p_error(PErrorMessage)
char *PErrorMessage;
{
    int Int1;
    
    for(Int1 = 0;Int1 < 20;++Int1){
        if(PChar == EOF)
            break;
        PReturned = PCharacter(PLEAVEWHITESPACE,PFAILONEOF);
        if(PReturned == PFAILED) 
            break;
        PString[Int1] = PChar;
    }
    PString[Int1] = EOS;
    sprintf(PStatus,"%s  Failed at around %s.",PErrorMessage,PString);
#ifdef TRACE
fprintf(stderr,"%s\n",PErrorMessage);
#endif
}

static void
p_error_EOF()
{
    p_error("Early EOF.");
}

static void
p_error_nosemi()
{
    p_error("; expected and not found.");
}

static void
p_error_badlayer(Tech,Mask)
char Tech,*Mask;
{
    char buf[35];
    sprintf(buf,"Undefined layer: %c%s. ",Tech,Mask);
    p_error(buf);
}

static void
p_error_CD()
{
    p_error(CDStatusString);
}


/* the following functions used to be macros */

int
PCharacter(WhiteSpaceControl,EOFControl)
int WhiteSpaceControl,EOFControl;
{
    int Returned;

    Returned = p_white_space(WhiteSpaceControl,EOFControl);
    if(Returned != PFAILED){
        PChar = getc(PCIFFileDesc);
        if(PChar == EOF){
            if(EOFControl != PDONTFAILONEOF){
                p_error_EOF();
                Returned = PFAILED;
            }
        }
    }
    return Returned;
}

static int
p_white_space(WhiteSpaceControl,EOFControl)
int WhiteSpaceControl,EOFControl;
{
    int Returned = PSUCCEEDED;

    if(WhiteSpaceControl == PSTRIPWHITESPACE1){
        while((PChar = getc(PCIFFileDesc)) == ' ' Or PChar == '\t' Or
            PChar == '\n' Or PChar == ','){
            if(PChar == EOF){
                if(EOFControl != PDONTFAILONEOF){
                    p_error_EOF();
                    Returned = PFAILED;
                }
                break;
            }
         }
         ungetc((char)PChar,PCIFFileDesc);
    }
    elif(WhiteSpaceControl == PSTRIPWHITESPACE2){
        while((PChar = getc(PCIFFileDesc)) != '-' And PChar != '('
            And PChar != ')' And (PChar < 'A' Or PChar > 'Z')
                And (PChar < '0' Or PChar > '9') And PChar != ';'){
            if(PChar == EOF){
                if(EOFControl != PDONTFAILONEOF) {
                    p_error_EOF();
                    Returned = PFAILED;
                }
                break;
            }
        }
        ungetc((char)PChar,PCIFFileDesc);
    }
    elif(WhiteSpaceControl == PSTRIPWHITESPACE3){
        while((PChar = getc(PCIFFileDesc)) != '-' And PChar != '('
            And PChar != ')' And (PChar< '0' Or PChar > '9') And
                PChar != ';'){
            if(PChar == EOF){
                if(EOFControl != PDONTFAILONEOF){
                    p_error_EOF();
                    Returned = PFAILED;
                }
                break;
            }
        }
        ungetc((char)PChar,PCIFFileDesc);
    }
    return Returned;
}

static int
p_white_space_1(EOFControl)
int EOFControl;
{
    int Returned = PSUCCEEDED;

    while((PChar = getc(PCIFFileDesc)) != '-' And PChar != '('
        And PChar != ')' And (PChar< '0' Or PChar > '9') And
            PChar != ';'){
        if(PChar == EOF){
            if(EOFControl != PDONTFAILONEOF){
                p_error_EOF();
                Returned = PFAILED;
            }
            break;
        }
    }
    ungetc((char)PChar,PCIFFileDesc);
    return Returned;
}

static int
p_look_ahead(WhiteSpaceControl, For)
int WhiteSpaceControl, For;
{
    int Returned;

    Returned = PCharacter(WhiteSpaceControl,PFAILONEOF);
    if(PChar != For && Returned != PFAILED)
        ungetc((char)PChar,PCIFFileDesc);
    return Returned;
}

static int
p_look_for_semi()
{
    int Returned;

    Returned = p_white_space_1(PFAILONEOF);
    if(Returned != PFAILED){
        PChar = getc(PCIFFileDesc);
        if(PChar == EOF){
            p_error_EOF();
            Returned = PFAILED;
        }
        elif(PChar != ';')
            ungetc((char)PChar,PCIFFileDesc);
    }
    return Returned;
}

static int
p_integer(EOFControl)
int EOFControl;
{
    int Returned;

    loop{
        PReturned = p_white_space_1(EOFControl);
        if(PReturned == PFAILED){
            Returned = PFAILED;
            break;
        }
        if((PChar >= '0' And PChar <= '9') Or
            (PChar == '-' Or PChar == '+')) {
            /* read integer */
            if((fscanf(PCIFFileDesc,"%ld",&PInt)) == EOF){
                if(EOFControl != PDONTFAILONEOF){
                    PChar = EOF;
                    p_error_EOF();
                    Returned = PFAILED;
                }
            }
            Returned = PSUCCEEDED;
            break;
        }
        /* drop unwanted characters */
        PReturned = PCharacter(PSTRIPWHITESPACE1,EOFControl);
        if(PReturned == PFAILED){
            Returned = PFAILED;
            break;
        }
    }
    return Returned;
}

