/***************************************************************************
SCED - Schematic Capture Editor
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1981 Giles C. Billingsley  (parts of KIC layout editor)
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * SCED graphical and keyboard input.
 */

#include "spice.h"
#include "sced.h"
#include "scedmacs.h"


#ifdef __STDC__
static void new_fine_window(long,long,long,long);
static int  control_at(long,long);
static unsigned long actiontime(void);
static void point_input(long*,long*,char*,int*);
#else
static void new_fine_window();
static int  control_at();
static unsigned long actiontime();
static void point_input();
#endif

#define DEBOUNCETIME 100

/* button1: basic point operation
 * button2: pan operation
 * button3: window operation
 * button4: return coords, no operation
 */

#define BUTTON1 1
#define BUTTON2 2
#define BUTTON3 4
#define BUTTON4 8

/* Point() command characters */
#define ESC 27
#define BSP '\b'
#define DEL 127
#define NEWL 13
#define SPA  32
#define CTRL_A 1
#define CTRL_F 6
#define CTRL_G 7
#define CTRL_N 14
#define CTRL_T 20
#define CTRL_U 21
#define CTRL_V 22
#define CTRL_X 24

static int OldButton3;  /* remember last button 3 */
static long OldButton3X,OldButton3Y;

/* we keep track of the time between pointing events to debounce the cursor */
static unsigned long LastPointTime = 0;

/* disable screen modifying functions such as scaling, pan. */
int LockOut;

/* return from Point() if esc entered */
static int EscReturn;


void
Point()

{
    /*
     * When user has typed Condition, do Action.
     *
     * Condition
     *     shortest unique prefix of Menu
     * Action
     *     Return with command selected stored in Parameters.kpCommand.
     *
     * Condition
     *     ESC
     * Action
     *     Forget remembered type in, return if EscReturn == True.
     *
     * Condition
     *     ctrl-a
     * Action
     *     Execute abort routine.
     *
     * Condition
     *     point key or tablet stylus button Z 
     * Action
     *     If user is pointing at a layer in the layer table viewport,
     *     change the current layer.
     *     If user is pointing at a command menu selection, return with
     *     command selected stored as Parameters.kpCommand.
     *     If user is pointing inside a layout viewport--coarse or fine--
     *     return with Parameters.kpCommand[0] == EOS and cursor
     *     descriptor up-to-date.
     *
     * Condition
     *     ctrl-f or tablet stylus button 1
     * Action
     *     Wait for user to point.
     *     Redisplay in fine viewport around where he pointed
     *     if fine positioning is enabled, or pan otherwise.
     *
     * Condition
     *     ctrl-g
     * Action
     *     Change scale of magnifying glass, or otherwise window, using
     *     next two point actions.
     *
     * Condition
     *     ctrl-n
     * Action
     *     Save the present view context in a list.
     *
     * Condition
     *     ctrl-t or ctrl-v
     * Action
     *     Toggle position of magnifying glass, bottom or right.
     *
     */
    char *TypeIn;
    MENU *Menu;
    unsigned long newtime;
    int NumCommand,Buttons,Int1,Int2,Int3;
    int Layer;
    long X,Y,centerX,centerY;
    char Key;
    extern char *MenuABORT;

    /*
     * The best way to handle interrupts reliably is to 
     * initialize the service routines as frequently as
     * possible.  Therefore, . . .
     */
    InitSignals();
    Menu = GetCurrentMenu();
    Parameters.kpCommand[NumCommand = 0] = EOS;
    Parameters.kpPointCoarseWindow = False;
    if (OldButton3) {
        OldButton3 = False;
        FBSetRubberBanding(0);
    }
    LastPointTime = actiontime();

    loop {
        loop {
            point_input(&X,&Y,&Key,&Buttons);

            if (Key)
                break;

            if ((Buttons == BUTTON1 Or
                Buttons == BUTTON2 Or
                Buttons == BUTTON3 Or
                Buttons == BUTTON4) And
                (X < gi_maxx And X > 0 And Y < gi_maxy And Y > 0))
                break;
            
        }

        newtime = actiontime();
        if (newtime > LastPointTime && newtime - LastPointTime < DEBOUNCETIME)
            continue;
        LastPointTime = newtime;

        switch (Key) {
 
            case 0:
                break;
 
            case BSP:
            case DEL:
                if (NumCommand) NumCommand--;
                Parameters.kpCommand[NumCommand] = EOS;
                continue;
 
            case ESC:
                Parameters.kpCommand[NumCommand = 0] = EOS;
                /* SRW ** so we know if esc was entered */
                Parameters.kpCommand[1] = ESC;
                if (EscReturn) return;
                continue;
 
            case NEWL:
                Parameters.kpCommand[NumCommand = 0] = EOS;
                if (control_at(X,Y)) return;
                continue;

            case '!':
                /* shell command */
                if (LockOut) continue; /* ignore */
                ShowPrompt("! ");
                TypeIn = FBEdit(NULL);
                if (TypeIn != NULL)    {
                    ShowProcess(TypeIn);
                    ErasePrompt();
                }
                Parameters.kpCommand[NumCommand = 0] = EOS;
                continue;

            case CTRL_A:
                /* SRW ** abort gracefully */
                strcpy(Parameters.kpCommand,MenuABORT);
                NumCommand = 0;
                return;
 
            case CTRL_F:
                if (LockOut) continue;  /* ignore ^F */
                Parameters.kpCommand[NumCommand = 0] = EOS;
                SaveLastView();
                /* SRW ** pan if coarse viewport only */
                if (Parameters.kpRedisplayControl == COARSEVIEWPORTONLY) {
                    if (!InBox(X,Y,View->kvCoarseViewport)) {
                        NotPointingAtLayout();
                        continue;
                    }
                    PToL(View->kvCoarseWindow,&X,&Y);
                    InitCoarseWindow(X,Y,(long)View->kvCoarseWindow->kaWidth);
                    InitFineWindow(X,Y);
                    EraseLargeCoarseViewport();
                    Redisplay(View->kvCoarseWindow);
                }
                else
                    FinePosition(X,Y,Key);
                /*
                 * This is necessary for debouncing.
                 * It takes time to redisplay, so we set the time of the last
                 * pointing event when the redisplay is finished.
                 */
                LastPointTime = actiontime();
                continue;

            case CTRL_G:
                if (LockOut) continue; /* ignore ^G */
                Parameters.kpCommand[NumCommand = 0] = EOS;
                if (Parameters.kpCellName[0] == EOS)
                    return;
                loop {
                    ShowPrompt("Point to diagonal of area to be magnified.");
                    point_input(&X,&Y,&Key,&Buttons);
                    if (Key == ESC) goto skip;
                    if (Buttons == BUTTON1) {
                        if (InBox(X,Y,View->kvCoarseViewport))
                            PToL(View->kvCoarseWindow,&X,&Y);
                        elif (InBox(X,Y,View->kvFineViewport))
                            PToL(View->kvFineWindow,&X,&Y);
                        else {
                            NotPointingAtLayout();
                            continue;
                        }
                        OldButton3X = X;
                        OldButton3Y = Y;
                        FBSetRubberBanding('R');
                        break;
                    }
                }
                loop {
                    ShowPrompt("Point to second endpoint.");
                    point_input(&X,&Y,&Key,&Buttons);
                    if (Key == ESC) {
                        FBSetRubberBanding(0);
                        goto skip;
                    }
                    if (Buttons == BUTTON1) {
                        if (InBox(X,Y,View->kvCoarseViewport))
                            PToL(View->kvCoarseWindow,&X,&Y);
                        elif (InBox(X,Y,View->kvFineViewport))
                            PToL(View->kvFineWindow,&X,&Y);
                        else {
                            NotPointingAtLayout();
                            continue;
                        }
                        FBSetRubberBanding(0);
                        break;
                    }
                }
                new_fine_window(OldButton3X,OldButton3Y,X,Y);
skip:
                ErasePrompt();
                continue;

            case CTRL_N:
                Parameters.kpCommand[NumCommand = 0] = EOS;
                SaveViewOnStack();
                continue;
 
            case CTRL_U:
            case CTRL_X:
            case SPA:
                NumCommand = 0;
                Parameters.kpCommand[0] = EOS;
                continue;
 
            case CTRL_T:
            case CTRL_V:
                if (LockOut) continue;  /* ignore */
                Parameters.kpCommand[NumCommand = 0] = EOS;
                View->kvFineViewportOnBottom ^= 1;
                InitViewport();
                if (Parameters.kpRedisplayControl == COARSEVIEWPORTONLY)
                    continue;
                SetPositioning();
                EraseLargeCoarseViewport();
                Redisplay(View->kvCoarseWindow);
                continue;
 
            default:
                Parameters.kpCommand[NumCommand++] = Key;
                if (NumCommand > 80) NumCommand = 80;
                Parameters.kpCommand[NumCommand] = EOS;
                /*
                 * Test for shortest unique prefix, or prefix matching upper
                 * case part of menu.  Stupid search is plenty fast.
                 */
                Int3 = -1;
                Int2 = 0;
                for (Int1 = 0; Menu[Int1].mEntry != NULL; Int1++) {
                    for (Int2 = 0; Int2 < NumCommand; Int2++) {
                        char c = Menu[Int1].mPrefix[Int2];
                        if (isupper(c)) c = tolower(c);
                        if (Parameters.kpCommand[Int2] != c) break;
                    }
                    if (Parameters.kpCommand[Int2] == EOS And Int2 > 0) {
                        if (!Menu[Int1].mPrefix[Int2]) {
                            /* found a match */
                            if (Int3 >= 0) {
                                /* oops, more than 1 match */
                                Int2 = -1;
                                break;
                            }
                            Int3 = Int1;
                        }
                    }
                }
                if (Int3 >= 0 && Int2 >= 0) {
                    strcpy(Parameters.kpCommand,Menu[Int3].mEntry);
                    return;
                }
                continue;
        }
        NumCommand = 0;
        if (ButtonPress(Buttons,X,Y))
            return;
    }
}
 

int
ButtonPress(Buttons,X,Y)

int Buttons;
long X,Y;
{

    Parameters.kpCommand[0] = EOS;
    if (Buttons == 0) {
        /* shouldn't get here unless null from keyboard */
        return (False);
    }
    if (Buttons == BUTTON1) {
        if (control_at(X,Y)) return (True);
        return (False);
    }
    if (Buttons == BUTTON2) {
        if (LockOut) {              /* treat like button 0 */
            Buttons = BUTTON1;
            if (control_at(X,Y)) return (True);
            return (False);
        }
        SaveLastView();
        /* pan if coarse viewport only */
        if (Parameters.kpRedisplayControl == COARSEVIEWPORTONLY) {
            if (!InBox(X,Y,View->kvCoarseViewport)) {
                return (False);
            }
            PToL(View->kvCoarseWindow,&X,&Y);
            InitCoarseWindow(X,Y,(long)View->kvCoarseWindow->kaWidth);
            InitFineWindow(X,Y);
            EraseLargeCoarseViewport();
            Redisplay(View->kvCoarseWindow);
        }
        else
            FinePosition(X,Y,(char)0);
        LastPointTime = actiontime();
        return (False);
    }
    if (Buttons == BUTTON3) {
        if (Parameters.kpCellName[0] == EOS) return (False);
        if (LockOut) {  /* treat like button 0 */
            Buttons = BUTTON1;
            if (control_at(X,Y)) return (True);
            return (False);
        }
        if (Not OldButton3) {
            if (InBox(X,Y,View->kvCoarseViewport))
                PToL(View->kvCoarseWindow,&X,&Y);
            elif (InBox(X,Y,View->kvFineViewport))
                PToL(View->kvFineWindow,&X,&Y);
            else {
                return (False);
            }
            OldButton3X = SCursor.kcRawX;
            OldButton3Y = SCursor.kcRawY;
            SCursor.kcRawX = X;
            SCursor.kcRawY = Y;
            FBSetRubberBanding('R');
            SCursor.kcRawX = OldButton3X;
            SCursor.kcRawY = OldButton3Y;
            OldButton3X = X;
            OldButton3Y = Y;
            OldButton3 = True;
        }
        else {
            if (InBox(X,Y,View->kvCoarseViewport))
                PToL(View->kvCoarseWindow,&X,&Y);
            elif (InBox(X,Y,View->kvFineViewport))
                PToL(View->kvFineWindow,&X,&Y);
            else {
                FBSetRubberBanding(0);
                return (False);
            }
            FBSetRubberBanding(0);
            OldButton3 = False;
            new_fine_window(OldButton3X,OldButton3Y,X,Y);
        }
        return (False);
    }
    if (Buttons == BUTTON4) {
        return (False);
    }
    return (False);
}


static void
new_fine_window(OldX,OldY,X,Y)

long OldX,OldY,X,Y;
{
    long NewWindowWidth, Hei, Wid, Tmp, CenterX, CenterY;

    if (Parameters.kpRedisplayControl == COARSEVIEWPORTONLY) {

        SaveLastView();
        Wid = X - OldX;
        Hei = Y - OldY;
        if (Wid < 0) Wid = -Wid;
        if (Hei < 0) Hei = -Hei;
        Tmp = Hei*
            (View->kvCoarseViewport->kaWidth/
            View->kvCoarseViewport->kaHeight);
        NewWindowWidth = Max(Wid,Tmp);
        if (NewWindowWidth <= 0)
            NewWindowWidth = RESOLUTION;
        Wid /= 2;
        Hei /= 2;
        InitCoarseWindow(Min(X,OldX)+Wid,
            Min(Y,OldY)+Hei,NewWindowWidth);
        InitFineWindow(Min(X,OldX)+Wid,
            Min(Y,OldY)+Hei);
        EraseLargeCoarseViewport();
        Redisplay(View->kvCoarseWindow);
    }
    else {
        if (X > OldX)
            SwapInts(X,OldX);
        if ((OldX - X) < 2) { /* two lambda minimum width */
            ShowPrompt("Magnifying glass width too small.");
            return;
        }
        SaveLastView();
        if (Y > OldY)
            SwapInts(Y,OldY);
        CenterX = (OldX - X)/2 + X;
        CenterY = (OldY - Y)/2 + Y;
        View->kvFineWindow->kaWidth = OldX - X;
        View->kvFineWindow->kaHeight =
            View->kvFineWindow->kaWidth*
            (View->kvFineViewport->kaHeight/
            View->kvFineViewport->kaWidth);
        EraseFineViewport();
        InitFineWindow(CenterX,CenterY);
        ShowFineViewport();
    }
}


int
PointLoop(LookedAhead)

/* Loop until UNDO, a "non-safe" command, point to coarse window, or
 * ESC is entered.  Return value used for dispatching in non-safe (i.e.,
 * cell modifying) commands.
 */
int *LookedAhead;
{
    extern char *MenuUNDO;

    loop {
        if (*LookedAhead == False) {
            EscReturn = True;
            Point();
            EscReturn = False;
        }
        else
            *LookedAhead = False;
        if (Parameters.kpCommand[1] == ESC)
            return (PL_ESC);
        if (Parameters.kpCommand[0] != EOS) {
            if (SafeCmds(LookedAhead))
                continue;
            if (Matching(MenuUNDO)) return (PL_UND);
            *LookedAhead = True;
            return (PL_CMD);
        }
        if (Parameters.kpPointCoarseWindow)
            return (PL_PCW);
        NotPointingAtLayout();
    }
}


int
PointLoopSafe(LookedAhead)

/* Loop until UNDO, any  command, point to coarse window, or
 * ESC is entered.  Return value used for dispatching in safe
 * (i.e. non cell modifying) commands.
 */
int *LookedAhead;
{

    loop {
        EscReturn = True;
        Point();
        EscReturn = False;
        if (Parameters.kpCommand[1] == ESC)
            return (PL_ESC);
        if (Parameters.kpCommand[0] != EOS) {
            *LookedAhead = True;
            return (PL_CMD);
        }
        if (Parameters.kpPointCoarseWindow)
            return (PL_PCW);
        NotPointingAtLayout();
    }
}


int
PointColor(LookedAhead)

/* Loop until UNDO, a non color setting command, point to coarse window, or
 * ESC is entered.
 */
int *LookedAhead;
{
    extern char *MenuPLUSR,*MenuMINSR;
    extern char *MenuPLUSG,*MenuMINSG;
    extern char *MenuPLUSB,*MenuMINSB;
    extern char *MenuRGB;
    extern char *MenuUNDO;

    loop {
        if (*LookedAhead == False) {
            EscReturn = True;
            Point();
            EscReturn = False;
        }
        else
            *LookedAhead = False;
        if (Parameters.kpCommand[1] == ESC)
            return (PL_ESC);
        if (Parameters.kpCommand[0] != EOS) {

            if (Matching(MenuMINSB)) { AlterColor('b','-');  continue; }
            if (Matching(MenuMINSG)) { AlterColor('g','-');  continue; }
            if (Matching(MenuMINSR)) { AlterColor('r','-');  continue; }
            if (Matching(MenuPLUSB)) { AlterColor('b','+');  continue; }
            if (Matching(MenuPLUSG)) { AlterColor('g','+');  continue; }
            if (Matching(MenuPLUSR)) { AlterColor('r','+');  continue; }
            if (Matching(MenuRGB))   { RGB();                continue; }

            if (Matching(MenuUNDO))
                return (PL_UND);
            *LookedAhead = True;
            return (PL_CMD);
        }
        if (Parameters.kpPointCoarseWindow)
            return (PL_PCW);
        NotPointingAtLayout();
    }
}


void
NotPointingAtLayout()

{
    ShowPrompt("You aren't pointing in the layout viewport.");
}


void
FinePosition(X,Y,Key)

long X,Y;
char Key;
{
    int Buttons;

    if (Parameters.kpCellName[0] == EOS)
        return;
    if (Parameters.kpRedisplayControl == COARSEVIEWPORTONLY) {
        ShowPrompt("Fine positioning isn't required.");
        Parameters.kpCommand[0] = EOS;
        return;
    } 
    if (Key != EOS) {
        ShowPrompt("Point to center of area you want magnified.");    
        loop {
            point_input(&X,&Y,&Key,&Buttons);
            if (Key == EOS Or Key == NEWL)
                break;
        }
    }
    if (InBox(X,Y,View->kvCoarseViewport))
        PToL(View->kvCoarseWindow,&X,&Y);
    elif (InBox(X,Y,View->kvFineViewport))
        PToL(View->kvFineWindow,&X,&Y);
    else {
        NotPointingAtLayout();
        return;
    }
    EraseFineViewport();
    InitFineWindow(X,Y);
    ShowFineViewport();
}


static int
control_at(X,Y)

long X,Y;
{
    int i, Row, Column;
    MENU *Menu;
    extern char *MenuPLACE;

    Parameters.kpPointCoarseWindow = False;
    if (InBox(X,Y,View->kvCoarseViewport)) {
        SCursor.kcInFine = False;
        SCursor.kcPredX = SCursor.kcX;
        SCursor.kcPredY = SCursor.kcY;
        SCursor.kcX = X; 
        SCursor.kcY = Y; 
        SCursor.kcRawX = X; 
        SCursor.kcRawY = Y; 
        PToL(View->kvCoarseWindow,&SCursor.kcX,&SCursor.kcY);
        PToL(View->kvCoarseWindow,&SCursor.kcRawX,&SCursor.kcRawY);
        ClipToGridPoint(&SCursor.kcX,&SCursor.kcY);
        SCursor.kcDX = SCursor.kcX-SCursor.kcPredX;
        SCursor.kcDY = SCursor.kcY-SCursor.kcPredY;
        Parameters.kpCommand[0] = '\0';
        Parameters.kpPointCoarseWindow = True;
        return (True);
    }
    if (InBox(X,Y,View->kvFineViewport)) {
        SCursor.kcInFine = True;
        SCursor.kcPredX = SCursor.kcX;
        SCursor.kcPredY = SCursor.kcY;
        SCursor.kcX = X; 
        SCursor.kcY = Y; 
        SCursor.kcRawX = X; 
        SCursor.kcRawY = Y; 
        PToL(View->kvFineWindow,&SCursor.kcX,&SCursor.kcY);
        PToL(View->kvFineWindow,&SCursor.kcRawX,&SCursor.kcRawY);
        ClipToGridPoint(&SCursor.kcX,&SCursor.kcY);
        SCursor.kcDX = SCursor.kcX-SCursor.kcPredX;
        SCursor.kcDY = SCursor.kcY-SCursor.kcPredY;
        Parameters.kpCommand[0] = '\0';
        Parameters.kpPointCoarseWindow = True;
        return (True);
    }
    Menu = GetCurrentMenu();
    Row = (gi_maxy-Y-3)/gi_fntheight+1;
    Column = X/gi_fntwidth+1;
    if (InBox((long)Column,(long)Row,&MenuViewport)) {
        if (Column > 5)
            Row += MenuViewport.kaY;
        if (Menu == BasicMenu) {
            for  (i = 0; ; i++)
                if (Menu[i].mEntry == NULL) break;

            if (i > Row - 1) {
                strcpy(Parameters.kpCommand,Menu[Row-1].mEntry);
                return (True);
            }
            if (Column <= 5)
                return (False);
            if (i < MenuViewport.kaY)
                i = MenuViewport.kaY;
            Row -= i;
            for  (i = 0; ; i++)
                if (DeviceMenu[i].mEntry == NULL) break;
            if (i > Row - 1) {
                /* hide the cell name after a null */
                sprintf(Parameters.kpCommand,"%s %s",MenuPLACE,
                    DeviceMenu[Row-1].mEntry);
                Parameters.kpCommand[strlen(MenuPLACE)] = '\0';
                return (True);
            }
        }
        else {
            for  (i = 0; ; i++)
                if (Menu[i].mEntry == NULL) break;
            if (i > Row - 1) {
                strcpy(Parameters.kpCommand,Menu[Row-1].mEntry);
                return (True);
            }
        }
    }
    return (False);
}


static unsigned long
actiontime()

{
    return ((unsigned long)1000*seconds());
}


static void
point_input(x,y,key,but)

long *x, *y;
char *key;
int *but;
{
    REQUEST request;
    RESPONSE response;

    request.option = point_option;
    DevInput(&request,&response);
    *x = response.x;
    *y = response.y;
    *key = 0;
    *but = 0;
    if (response.option == char_option) {
        *key = (isupper(response.reply.ch) ?
            tolower(response.reply.ch) : response.reply.ch);
    }
    else {
        *but = response.reply.button;
    }
    return;
}
