/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Wayne A. Christopher
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Code to deal with breakpoints and tracing.
 */

#include "spice.h"
#include "ftedefs.h"
#include "ftedebug.h"
#include "spfteext.h"
#include "scedio.h"

#ifdef __STDC__
static void printdbg(struct dbcomm*);
static bool print_trace(struct dbcomm*,struct plot*,bool*);
static char *spec(char**);
static void printcond(struct dbcomm*,FILE*);
static int  hackplot(void);
static void unhackplot(int);
#else
static void printdbg();
static bool print_trace();
static char *spec();
static void printcond();
static int  hackplot();
static void unhackplot();
#endif

static int debugcnt = 1;
struct dbcomm *dbiplot = NULL;      /* export for iplot */
static struct dbcomm *dbtrace = NULL;
static struct dbcomm *dbsave = NULL;
static struct dbcomm *dbstop = NULL;
static int howmanysteps = 0;
static int steps = 0;


/* Set a breakpoint. Possible commands are:
 *  stop after n
 *  stop when var cond val
 *
 * If more than one is given on a command line, then this is a conjunction.
 */

void
com_stop(wl)

wordlist *wl;
{
    struct dbcomm *d, *thisone = NULL;
    char *s, *t, buf[64];
    int i;
    double *val;
    wordlist *wx;

    while (wl) {
        if (thisone == NULL)
            thisone = d = alloc(struct dbcomm);
        else {
            d->db_also = alloc(struct dbcomm);
            d = d->db_also;
        }

        /* Figure out what the first condition is. */
        if (eq(wl->wl_word, "after")) {
            d->db_type = DB_STOPAFTER;
            i = 0;
            wl = wl->wl_next;
            if (wl) {
                if (wl->wl_word) {
                    for (s = wl->wl_word; *s; s++)
                        if (!isdigit(*s))
                            goto bad;
                    i = atoi(wl->wl_word);
                }
                wl = wl->wl_next;
            }
            d->db_iteration = i;
            continue;
        }
        if (eq(wl->wl_word, "when")) {
            d->db_type = DB_STOPWHEN;
            wl = wl->wl_next;
            if (!wl)
                goto bad;
            for (i = 0,wx = wl; wx; wx = wx->wl_next) {
                if (eq(wx->wl_word,"after") || eq(wx->wl_word,"when"))
                    break;
                i += strlen(wx->wl_word) + 1;
            }
            i++;
            s = d->db_nodename1 = tmalloc(i*sizeof(char));
            strcpy(s,wl->wl_word);
            s += strlen(s);
            for (wl = wl->wl_next; wl; wl = wl->wl_next) {
                if (eq(wl->wl_word,"after") || eq(wl->wl_word,"when"))
                    break;
                t = wl->wl_word;
                *s++ = ' ';
                while (*t)
                    *s++ = *t++;
            }
            *s = '\0';
            cp_unquote(d->db_nodename1);
        }
        else
            goto bad;
    }
    if (thisone) {
        thisone->db_number = debugcnt++;
        if (dbstop) {
            for (d = dbstop; d->db_next; d = d->db_next)
                ;
            d->db_next = thisone;
        }
        else
            dbstop = thisone;
    }
    return;

bad:
    fprintf(cp_err, "Syntax error.\n");
    return;
}


/* Trace a node. Usage is "trace node ..."*/

void
com_trce(wl)

wordlist *wl;
{
    struct dbcomm *d = NULL, *td, *ld;
    char *s;

    while (wl) {
        s = copy(wl->wl_word);
        cp_unquote(s);
        d = alloc(struct dbcomm);
        d->db_type = DB_TRACE;
        d->db_nodename1 = s;
        d->db_number = debugcnt++;
        if (dbtrace) {
            ld = NULL;
            for (td = dbtrace; td; ld = td, td = td->db_next)
                if (eq(td->db_nodename1,d->db_nodename1)) {
                    fprintf(cp_err, "Already tracing %s, ignored.\n",
                        td->db_nodename1);
                    dbfree(d);
                    debugcnt--;
                    goto bot;
                }
            ld->db_next = d;
        }
        else
            dbtrace = d;
bot:
        wl = wl->wl_next;
    }
}


/* Incrementally plot a value. This is just like trace. */

void
com_iplot(wl)

wordlist *wl;
{
    struct dbcomm *d, *td, *currentdb = NULL;
    char *s;
    char *which = NULL;

    /* We use a modified ad-hoc algorithm here where db_also denotes
     * vectors on the same command line and db_next denotes
     * separate iplot commands.
     */

    /* The user can choose between the two graphics packages by
     * entering "plot" or "graf" on the command line.
     */

    for ( ; wl; wl = wl->wl_next) {
        s = copy(wl->wl_word);
        cp_unquote(s);

        if (eq(s,"plot") || eq(s,"graf")) {
            which = s;
            continue;
        }
        if (currentdb == NULL) {
            d = currentdb = alloc(struct dbcomm);
        }
        else {
            d->db_also = alloc(struct dbcomm);
            d = d->db_also;
        }
        d->db_type = DB_IPLOT;
        d->db_nodename1 = s;
    }

    if (currentdb) {
        currentdb->db_number = debugcnt++;
        currentdb->db_nodename2 = which;
        if (currentdb->db_nodename2 == NULL)
            /* the default package */
            currentdb->db_nodename2 = copy("plot");
        if (dbiplot) {
            for (td = dbiplot; td->db_next; td = td->db_next)
                ;
            td->db_next = currentdb;
        }
        else
            dbiplot = currentdb;
    }
}


/* Save a vector. */

void
com_save(wl)

wordlist *wl;
{
    struct dbcomm *d = NULL, *td, *ld;
    char *s;

    while (wl) {
        s = copy(wl->wl_word);
        cp_unquote(s);
        d = alloc(struct dbcomm);
        d->db_type = DB_SAVE;
        d->db_nodename1 = s;
        d->db_number = debugcnt++;
        if (dbsave) {
            ld = NULL;
            for (td = dbsave; td; ld = td,td = td->db_next)
                if (eq(td->db_nodename1,d->db_nodename1)) {
                    fprintf(cp_err, "Already saving %s, ignored.\n",
                        td->db_nodename1);
                    dbfree(d);
                    debugcnt--;
                    goto bot;
                }
            ld->db_next = d;
        }
        else
            dbsave = d;
bot:
        wl = wl->wl_next;
    }
}


/* Step a number of iterations. */

void
com_step(wl)

wordlist *wl;
{
    if (wl)
        steps = howmanysteps = atoi(wl->wl_word);
    else
        steps = howmanysteps = 1;
    com_resume((wordlist *) NULL);
    return;
}


/* Print out the currently active breakpoints and traces. If we are printing
 * to a file, assume that the file will be used for a later source and leave
 * off the event numbers (with UNIX, that is).  -- I don't like this.
 */

/* ARGSUSED */
void
com_sttus(wl)

wordlist *wl;
{
    struct dbcomm *d;

    for (d = dbstop; d; d = d->db_next)
        printdbg(d);
    for (d = dbtrace; d; d = d->db_next)
        printdbg(d);
    for (d = dbiplot; d; d = d->db_next)
        printdbg(d);
    for (d = dbsave; d; d = d->db_next)
        printdbg(d);
}


static void
printdbg(d)

struct dbcomm *d;
{
    char *s1, *s2;
    char *all = "all",*trace = "trace",*iplot = "iplot",*save = "save";
    struct dbcomm *dc;


    switch (d->db_type) {
        case DB_SAVE:
            s1 = save;
            s2 = d->db_nodename1;
            break;

        case DB_TRACE:
            s1 = trace;
            s2 = d->db_nodename1;
            break;

        case DB_STOPWHEN:
        case DB_STOPAFTER:
            out_printf("%-4d stop", d->db_number);
            printcond(d, cp_out);
            out_send("\n");
            return;

        case DB_IPLOT:
            out_printf("%-4d %s (%s) %s", d->db_number, iplot,
                d->db_nodename2, d->db_nodename1);
            for (dc = d->db_also; dc; dc = dc->db_also)
                out_printf(" %s", dc->db_nodename1);
            out_send("\n");
            return;

        case DB_DEADIPLOT:
            out_printf("%-4d exiting iplot (%s) %s", d->db_number,
                d->db_nodename2,dc->db_nodename1);
            for (dc = d->db_also; dc; dc = dc->db_also)
                out_printf(" %s", dc->db_nodename1);
            out_send("\n");
            return;

        default:
            fprintf(cp_err, 
                "com_sttus: Internal Error: bad db %d...\n",d->db_type);
            return;
    }
    out_printf("%-4d %s %s\n", d->db_number, s1, s2);
}


void
dbfree(db)

struct dbcomm *db;
{
    struct dbcomm *dd, *dn;

    for (dd = db; dd; dd = dn) {
        dn = dd->db_also;
        tfree(dd->db_nodename1);
        tfree(dd->db_nodename2);
        tfree(dd);
    }
    return;
}


void
com_delete(wl)

/* Delete breakpoints and traces. Usage is delete [number ...] */
wordlist *wl;
{
    int i;
    char *s, buf[64];
    struct dbcomm *d, *dt, *dx;

    if (!wl) {
        d = dbstop;
        if (!d) d = dbtrace;
        if (!d) d = dbiplot;
        if (!d) d = dbsave;
        if (!d) {
            fprintf(cp_err, "Error: no debugs in effect\n");
            return;
        }
        printdbg(d);
        (void)SCEDfgets(buf,4,cp_in,"delete? [y] ");
        if (*buf != 'y' && *buf != 'Y' && *buf != '\n')
            return;
        if (d == dbstop)
            dbstop = d->db_next;
        else if (d == dbtrace)
            dbtrace = d->db_next;
        else if (d == dbiplot)
            dbiplot = d->db_next;
        else if (d == dbsave)
            dbsave = d->db_next;
        dbfree(d);
        return;
    }

    while (wl) {
        if (eq(wl->wl_word, "trace") || eq(wl->wl_word,"all")) {
            for (dt = dbtrace; dt; dt = d) {
                d = dt->db_next;
                dbfree(dt);
            }
            dbtrace = NULL;
            if (eq(wl->wl_word,"trace")) {
                wl = wl->wl_next;
                continue;
            }
        }
        if (eq(wl->wl_word, "save") || eq(wl->wl_word,"all")) {
            for (dt = dbsave; dt; dt = d) {
                d = dt->db_next;
                dbfree(dt);
            }
            dbsave = NULL;
            if (eq(wl->wl_word,"save")) {
                wl = wl->wl_next;
                continue;
            }
        }
        if (eq(wl->wl_word, "iplot") || eq(wl->wl_word,"all")) {
            for (d = dbiplot; d; d = dx) {
                dx = d->db_next;
                dbfree(d);
            }
            dbiplot = NULL;
            if (eq(wl->wl_word,"iplot")) {
                wl = wl->wl_next;
                continue;
            }
        }
        if (eq(wl->wl_word, "stop") || eq(wl->wl_word,"all")) {
            for (dt = dbstop; dt; dt = d) {
                d = dt->db_next;
                dbfree(dt);
            }
            dbstop = NULL;
            wl = wl->wl_next;
            continue;
        }

        i = 0;
        for (s = wl->wl_word; *s; s++) {
            if (!isdigit(*s)) {
                fprintf(cp_err, "Error: %s isn't a number.\n",
                    wl->wl_word);
                break;
            }
        }
        if (*s) {
            wl = wl->wl_next;
            continue;
        }
        i = atoi(wl->wl_word);
        for (dt = NULL,d = dbtrace; d; dt = d,d = d->db_next) {
            if (d->db_number != i) continue;
            if (dt == NULL)
                dbtrace = d->db_next;
            else
                dt->db_next = d->db_next;
            dbfree(d);
            goto foundit;
        }
        for (dt = NULL,d = dbiplot; d; dt = d,d = d->db_next) {
            if (d->db_number != i) continue;
            if (dt == NULL)
                dbiplot = d->db_next;
            else
                dt->db_next = d->db_next;
            dbfree(d);
            goto foundit;
        }
        for (dt = NULL,d = dbsave; d; dt = d,d = d->db_next) {
            if (d->db_number != i) continue;
            if (dt == NULL)
                dbsave = d->db_next;
            else
                dt->db_next = d->db_next;
            dbfree(d);
            goto foundit;
        }
        for (dt = NULL,d = dbstop; d; dt = d,d = d->db_next) {
            if (d->db_number != i) continue;
            if (dt == NULL)
                dbstop = d->db_next;
            else
                dt->db_next = d->db_next;
            dbfree(d);
        }
foundit:
        wl = wl->wl_next;
    }
}


/* Writedata calls this routine to see if it should keep going. If it
 * returns true, then the run should resume.
 */

bool
ft_bpcheck(runplot, iteration)

struct plot *runplot;
int iteration;
{
    struct dbcomm *d, *dt;
    bool tflag = true;
    bool after = true;
    bool when  = true;
    bool nohalt = true;
    wordlist wl;
    int len;

    if (dbtrace || dbstop) {
        len = hackplot();
        for (d = dbtrace; d; d = d->db_next) {
            if (d->db_type == DB_TRACE)
                print_trace(d,runplot,&tflag);
        }
        for (d = dbstop; d; d = d->db_next) {
            for (dt = d; dt; dt = dt->db_also) {
                if (dt->db_type == DB_STOPAFTER) {
                    if (iteration != dt->db_iteration)
                        after = false;
                }
                else if (dt->db_type == DB_STOPWHEN) {
                    wl.wl_prev = wl.wl_next = NULL;
                    wl.wl_word = dt->db_nodename1;
                    if (!cp_istrue(&wl))
                        when = false;
                }
            }
            if (when && after) {
                fprintf(cp_err, "%-2d: condition met: stop ",d->db_number);
                printcond(d, cp_err);
                (void) putc('\n', cp_err);
                nohalt = false;
            }
        }
        unhackplot(len);
    }

    if ((howmanysteps > 0) && (--howmanysteps == 0)) {
        if (steps > 1)
            fprintf(cp_err, "Stopped after %d steps.\n", steps);
        return (false);
    }
    return (nohalt);
}


/* print the node voltages for trace */

static bool
print_trace(d, plot, flag)

struct dbcomm *d;
struct plot *plot;
bool *flag;
{

    struct dvec *v1;
    struct pnode *pn;
    double d1, d2;
    char *s, *t, buf[128];
    wordlist wl;
    struct dvlist *dvl, *dv, dv0;

    if (*flag) {
        v1 = plot->pl_scale;
        if (v1) {
            if (isreal(v1))
                fprintf(cp_out,"%s = %s:\n",v1->v_name,
                    printnum(v1->v_realdata[0]));
            else
                fprintf(cp_out,"%s = %s:\n",v1->v_name,
                    printnum(realpart(&(v1->v_compdata[0]))));
        }
        *flag = false;
    }

    if (d->db_nodename1) {

        wl.wl_next = wl.wl_prev = NULL;
        wl.wl_word = d->db_nodename1;
        pn = ft_getpnames(&wl, false);
        if (!pn) {
            fprintf(cp_err, "Error: can't find %s\n", d->db_nodename1);
            return (false);
        }
        v1 = ft_evaluate(pn);
        inp_pnfree(pn);

        if (v1 == NULL) {
            fprintf(cp_err, "Error: can't find %s\n", d->db_nodename1);
            return (false);
        }
        if (eq(d->db_nodename1,"all"))
            dvl = v1->v_link2;
        else {
            dv0.dl_dvec = v1;
            dv0.dl_next = NULL;
            dvl = &dv0;
        }
        for (dv = dvl; dv; dv = dvl) {
            dvl = dv->dl_next;
            v1 = dv->dl_dvec;
            if (v1 == plot->pl_scale)
                continue;
            if (isreal(v1))
                fprintf(cp_out,"  %-10s %s\n",v1->v_name,
                    printnum(v1->v_realdata[0]));

            else
                fprintf(cp_out,"  %-10s %s,%s\n",v1->v_name,
                    printnum(realpart(&(v1->v_compdata[0]))),
                    printnum(imagpart(&(v1->v_compdata[0]))));
        }
    }
    return (true);
}


/* Writedata calls this before it starts a run, to set the proper flags
 * on the dvecs. If a change is made during a break, then the routine
 * wrd_chtrace is used from these routines. We have to be clever with save:
 * if there was no save given, then save everything. Unfortunately you
 * can't stop in the middle, do a save, and have the rest then discarded.
 */

void
ft_trquery()
{
    return;
}


int
ft_getSaves(savesp)

char ***savesp;
{
    struct dbcomm *d, *dd;
    int count = 0, icount = 0, i = 0;
    char *s, *var;
    char **array;

    /* In addition to the saves, we hunt through all of the debugs
     * for special variables that should be saved as well.
     */

    for (d = dbtrace; d; d = d->db_next) {
        s = d->db_nodename1;
        while ((var = spec(&s)) != NULL) {
            icount++;
            free(var);
        }
    }
    for (d = dbiplot; d; d = d->db_next) {
        for (dd = d; dd; dd = dd->db_also) {
            s = dd->db_nodename1;
            while ((var = spec(&s)) != NULL) {
                icount++;
                free(var);
            }
        }
    }
    for (d = dbstop; d; d = d->db_next) {
        for (dd = d; dd; dd = dd->db_also) {
            s = dd->db_nodename1;
            while ((var = spec(&s)) != NULL) {
                icount++;
                free(var);
            }
        }
    }
    for (d = dbsave; d; d = d->db_next)
        count++;
    icount += count;

    if (!icount)
        return (0);
    if (!count)
        icount++;
    
    *savesp = array = (char **)tmalloc(sizeof(char *) * icount);

    if (!count)
        array[i++] = copy("all");
    else {
        for (d = dbsave; d; d = d->db_next)
            array[i++] = copy(d->db_nodename1);
    }
    
    for (d = dbtrace; d; d = d->db_next) {
        s = d->db_nodename1;
        while ((var = spec(&s)) != NULL)
            array[i++] = var;
    }
    for (d = dbiplot; d; d = d->db_next) {
        s = d->db_nodename1;
        while ((var = spec(&s)) != NULL)
            array[i++] = var;
        for (dd = d->db_also; dd; dd = dd->db_also) {
            s = dd->db_nodename1;
            while ((var = spec(&s)) != NULL)
                array[i++] = var;
        }
    }
    for (d = dbstop; d; d = d->db_next) {
        s = d->db_nodename1;
        while ((var = spec(&s)) != NULL)
            array[i++] = var;
        s = d->db_nodename2;
        while ((var = spec(&s)) != NULL)
            array[i++] = var;
    }
    return (icount);
}


static char *
spec(str)

char **str;
{
    /* yank out the 'special' vectors: those keyed with '@' */
    char *t, *tt, *tok, c; 

    if (!str || !*str)
        return (NULL);
    t = strchr(*str, '@');
    if (t) {
        tt = t+1;
        while (*tt && (isalpha(*tt) || isdigit(*tt) || *tt == '_')) tt++;
        if (*tt && *tt == '[') {
            tt++;
            while (*tt && *tt != ']') tt++;
            if (*tt = ']') tt++;
        }
        c = *tt;
        *tt = '\0';
        tok = copy(t);
        *tt = c;
        *str = tt;
        return (tok);
    }
    return (NULL);
}


static void
printcond(d, fp)

struct dbcomm *d;
FILE *fp;
{
    struct dbcomm *dt;
    char buf[BSIZE_SP];

    for (dt = d; dt; dt = dt->db_also) {
        if (dt->db_type == DB_STOPAFTER)
            sprintf(buf, " after %ld",dt->db_iteration);
        else
            sprintf(buf, " when %s",dt->db_nodename1);
        if (fp == cp_out)
            (void)out_send(buf);
        else
            (void)fputs(buf, fp);
    }
}


/* Temporarily convert all the permanent vectors in the plot to
 * unit length, pointing to the latest value.  Then we can use
 * the vector expression parser efficiently in the debugs.
 */

static int
hackplot()

{
    struct dvec *d;
    int length;
    wordlist *wl0, *wl;

    if (plot_cur->pl_scale)
        length = plot_cur->pl_scale->v_length;
    else
        length = plot_cur->pl_dvecs->v_length;
    if (length <= 1)
        return (0);
    length--;
    wl0 = (wordlist*)htab_list(plot_cur->pl_hashtab);
    for (wl = wl0; wl; wl = wl->wl_next) {
        d = (struct dvec*)((char**)wl->wl_word)[1];
        d->v_length = 1;
        if (d->v_flags & VF_COMPLEX)
            d->v_compdata += length;
        else
            d->v_realdata += length;
    }
    return (length);
}


static void
unhackplot(length)

int length;
{
    struct dvec *d;
    wordlist *wl0, *wl;

    if (length <= 0)
        return;
    wl0 = (wordlist*)htab_list(plot_cur->pl_hashtab);
    for (wl = wl0; wl; wl = wl->wl_next) {
        d = (struct dvec*)((char**)wl->wl_word)[1];
        d->v_length = length + 1;
        if (d->v_flags & VF_COMPLEX)
            d->v_compdata -= length;
        else
            d->v_realdata -= length;
    }
}
