/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Routines for operating range and related analysis
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"
#include "spfteext.h"
#include "ftegraph.h"
#include "plotext.h"
#include "outdata.h"
#include "util.h"
#include "iferrmsg.h"


static FILE *Op;            /* output file pointer          */
static int graphid;         /* graph id if plotting         */
static bool batchmode;      /* if true, no user prompts     */
static bool no_output;      /* supress output recording     */
static struct sOUTcontrol *OUTcntrl;
                            /* structure to control output  */

struct control {
    double delta1;
    double delta2;
    int step1;
    int step2;
    double val1;
    double val2;
    int iterno;    /* number iterations for extrema seeking */
};

struct sLOOPprms {
    double start[2];      /* starting parameter value */
    double stop[2];       /* ending parameter value */
    double step[2];       /* step size */
    double state[2];      /* internal values saved */
    int dims[3];          /* dimensions of output vector */
    int nestLevel;        /* number of levels of nesting */
    int nestSave;         /* iteration state during pause */
    char *rundesc;        /* save rundesc */
    int inprogress;       /* not done yet */
};

#ifdef __STDC__
static bool parse_var(struct sLOOPprms *,wordlist**);
static void pr_parms(void);
static void checkloop(wordlist*);
static void set_opvec(int);
static int  mainloop(struct control*);
static int  do_it(int,int,double,double);
static int  findext(int,double*,double,double);
static void check_end(void);
static bool read_check_control_file(FILE*,char*);
static bool check_evaluate(void);
static void execblock(char*);
static void modify(double,double);
static void setvec(double,double);
static FILE *df_open(void);
static void df_close(FILE*);
#else
static bool parse_var();
static void pr_parms();
static void checkloop();
static void set_opvec();
static int  mainloop();
static int  do_it();
static int  findext();
static void check_end();
static bool read_check_control_file();
static bool check_evaluate();
static void execblock();
static void modify();
static void setvec();
static FILE *df_open();
static void df_close();
#endif


void
com_check(wl)

wordlist *wl;
{
    int i, step1, step2;
    struct dvec *d;
    char buf[BSIZE_SP];
    char *c, *hdr;
    wordlist *ww, *wn;
    struct circ *ct;
    struct plot *pl;
    extern char *kw_mplot_cur;

    batchmode = true;

    if (!ft_curckt || !ft_curckt->ci_ckt) {
        (void) fprintf(cp_err,"Error: no circuit loaded.\n");
        return;
    }
    if (!ft_curckt->ci_contblk) {
        (void) fprintf(cp_err,"Error: no bound codeblock.\n");
        return;
    }
    if (!cp_isblockdef(ft_curckt->ci_contblk)) {
        (void) fprintf(cp_err,"Error: codeblock %s not found.\n",
            ft_curckt->ci_contblk);
        return;
    }
    for (ct = ft_circuits; ct; ct = ct->ci_next)
        ct->ci_inprogress = false;
    ft_intrpt = false;

    pl = plot_alloc("range");
    plot_new(pl);
    plot_setcur(pl->pl_typename);

    /* margin analysis creates two codeblocks:
     * #_name and ##name.  ##name is the header block, which is
     * executed after the plot is created.  #_name is the bound
     * codeblock.
     */

    hdr = copy(ft_curckt->ci_contblk);
    hdr[1] = '#';
    if (cp_isblockdef(hdr))
        execblock(hdr);
    txfree(hdr);

    ww = cp_lexer("checkFAIL = 0");
    com_let(ww);
    wl_free(ww);

    if (wl) {
        wl = wl_copy(wl);
        for (ww = wl; wl; wl = wn) {
            wn = wl->wl_next;
            if (wl->wl_word[0] == '-' && wl->wl_word[1] == 'v') {
                if (wl->wl_prev)
                    wl->wl_prev->wl_next = wl->wl_next;
                if (wl->wl_next)
                    wl->wl_next->wl_prev = wl->wl_prev;
                if (ww == wl)
                    ww = wl->wl_next;
                txfree(wl->wl_word);
                txfree((char*)wl);
                batchmode = false;
            }
        }
        wl = ww;
    }

    /* analysis specification */
    if (!wl || !wl->wl_word || *wl->wl_word == '\0') {
        wn = GetAnalysisFromDeck();
        if (wn == NULL) {
            (void) fprintf(cp_err,
                "Error: no or more than one analysis specified\n");
            wl_free(wl);
            com_destroy(NULL);
            return;
        }
        wl_free(wl);
        wl = wn;
    }
    for (c = wl->wl_word; *c; c++)
        if (isupper(*c)) *c = tolower(*c);

    Op = (FILE*)NULL;
    step1 = step2 = 0;
    if ((d = vec_get("checkSTP1")) != NULL)
        step1 = (int) *d->v_realdata;
    if ((d = vec_get("checkSTP2")) != NULL)
        step2 = (int) *d->v_realdata;
    if (step1 || step2) {
        if ((Op = df_open()) == NULL) {
            (void) fprintf(cp_err,"Error: Can't open data output file\n");
            wl_free(wl);
            com_destroy(NULL);
            return;
        }
    }
    if ((d = vec_get("checkPNTS")) == NULL) {
        (void) fprintf(cp_err,"Error: check variable checkPNTS not set\n");
        com_destroy(NULL);
        return;
    }

    OUTcntrl = (struct sOUTcontrol *)OUTcntrlInit();
    OUTcntrl->out_evaluate = check_evaluate;
    OUTcntrl->out_end      = check_end;
    OUTcntrl->out_max      = d->v_length;
    OUTcntrl->out_points   = d->v_realdata;

    /* make sure an "exact" point is recognized */
    for (i = 0; i < d->v_length; i++)
        d->v_realdata[i] *= .9999;

    if (!batchmode)
        pr_parms();

    checkloop(wl);

    df_close(Op);
    if (OUTcntrl->out_destroy)
        (*OUTcntrl->out_destroy)();
    wl_free(wl);
    mp_done(graphid);

    *buf = '\0';
    (void)cp_getvar(kw_mplot_cur,VT_STRING,buf);
    if (!batchmode && Op && *buf)
        (void) fprintf(cp_out,
        "Operating range analysis complete.  Data in file %s.\n",buf);
    Op = NULL;
}


void
com_loop(wl)

wordlist *wl;
{
    static struct sLOOPprms loop;
    double tt;
    int i;
    int error;
    int count;
    bool firsttime;
    char *c, *task;


    if (ft_curckt->ci_inprogress && loop.inprogress) {
        i = loop.nestSave;
        firsttime = false;
        OUTcntrl->out_rundesc = loop.rundesc;
        OUTcntrl->out_keepplot = true;
        loop.inprogress = false;
        task = ft_curckt->ci_curTask;
        goto didpause;
    }
    else {
        /* reset static variables */
        error = parse_var(&loop,&wl);
        if (error) {
            fprintf(cp_err,"Syntax error\n");
            return;
        }
        /* analysis specification */
        if (!wl || !wl->wl_word || *wl->wl_word == '\0') {
            wl = GetAnalysisFromDeck();
            if (wl == NULL) {
                (void) fprintf(cp_err,
                    "Error: no or more than one analysis specified\n");
                return;
            }
        }
        for (c = wl->wl_word; *c; c++)
            if (isupper(*c)) *c = tolower(*c);

        OUTcntrl = (struct sOUTcontrol *)OUTcntrlInit();
        OUTcntrl->out_keepplot = true;
        for (i = 0; i <= loop.nestLevel; i++)
            loop.state[i] = loop.start[i];
        i = 0;
        firsttime = true;
    }

    for (;;) {

        tt = loop.state[i] - loop.stop[i];

        if ((loop.step[i] > 0 && tt > 1e-8) ||
            (loop.step[i] < 0 && tt < 1e-8) ||
            loop.step[i] == 0) {

            i++; 
            if (i > loop.nestLevel)
                break;
            loop.dims[0]++;
            if (loop.dims[2] <= 1) {
                OUTsetDims(OUTcntrl->out_rundesc,loop.dims,2);
            }
        }
        else {
            if (!loop.dims[0])
                loop.dims[1]++;
        
            while (i > 0) { 
                i--; 
                loop.state[i] = loop.start[i];
            }

            modify(loop.state[0],loop.state[1]);
            if (ft_curckt->ci_specTask) {
                task = ft_curckt->ci_specTask;
                ft_curckt->ci_specTask = NULL;
                com_rset((wordlist*)NULL);
                ft_curckt->ci_specTask = task;
            }
            else {
                com_rset((wordlist*)NULL);
                task = ft_curckt->ci_defTask;

            }
            ft_curckt->ci_curTask = task;
didpause:

            if (firsttime) {
                count = 0;
                error = if_run(wl->wl_word, wl->wl_next);
                firsttime = false;
            }
            else {
                count = plot_cur->pl_scale->v_length;
                error = (*(ft_sim->doAnalyses))
                    (ft_curckt->ci_ckt, 1, task);
            }
            if (error) {
                if (error == E_PAUSE) {
                    loop.nestSave = i;
                    loop.rundesc = OUTcntrl->out_rundesc;
                    OUTcntrl->out_rundesc = NULL;
                    loop.inprogress = true;
                }
                gr_end_iplot();
                if (OUTcntrl->out_destroy)
                    (*OUTcntrl->out_destroy)();
                return;
            }

            /* store block size in dims[] */
            loop.dims[2] = plot_cur->pl_scale->v_length - count;

            if (loop.dims[2] > 1) {

                if (loop.dims[0])
                    OUTsetDims(OUTcntrl->out_rundesc,loop.dims,3);
                else
                    OUTsetDims(OUTcntrl->out_rundesc,loop.dims+1,2);
            }
        }
        loop.state[i] += loop.step[i];
    }

    gr_end_iplot();
    ft_curckt->ci_inprogress = false;
    if (OUTcntrl->out_destroy)
        (*OUTcntrl->out_destroy)();

    return;
}


static bool
parse_var(loop,wl)

struct sLOOPprms *loop;
wordlist **wl;
{
    char *line;
    int error;
    double tmp;
    double INPevaluate();

    loop->start[0]   = 0;
    loop->stop[0]    = 0;
    loop->step[0]    = 0;
    loop->start[1]   = 0;
    loop->stop[1]    = 0;
    loop->step[1]    = 0;
    loop->nestLevel  = 0;
    loop->dims[0]    = 0;
    loop->dims[1]    = 0;
    loop->dims[2]    = 0;
    loop->nestSave   = 0;
    loop->inprogress = 0;

    if (wl == NULL || *wl == NULL)
        return (true);
    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* start1 */
    if (error == 0)
        loop->start[0] = tmp;
    else
        return (error);
    loop->nestLevel = 0;

    *wl = (*wl)->wl_next;
    if (!*wl) {
        loop->stop[0] = loop->start[0];
        loop->step[0] = 0;
        return (false);
    }

    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* stop1 */
    if (error == 0)
        loop->stop[0] = tmp;
    else {
        loop->stop[0] = loop->start[0];
        loop->step[0] = 0;
        return (false);
    }

    (*wl) = (*wl)->wl_next;
    if (!*wl) {
        loop->step[0] = loop->stop[0] - loop->start[0];
        return (false);
    }

    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* step1 */
    if (error == 0)
        loop->step[0] = tmp;
    else {
        loop->step[0] = loop->stop[0] - loop->start[0];
        return (false);
    }

    (*wl) = (*wl)->wl_next;
    if (!*wl) {
        return (false);
    }

    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* start2 */
    if (error == 0)
        loop->start[1] = tmp;
    else {
        return (false);
    }
    loop->nestLevel = 1;

    (*wl) = (*wl)->wl_next;
    if (!*wl) {
        loop->stop[1] = loop->start[1];
        loop->step[1] = 0;
        return (false);
    }

    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* stop2 */
    if (error == 0)
        loop->start[1] = tmp;
    else {
        loop->stop[1] = loop->start[1];
        loop->step[1] = 0;
        return (false);
    }

    (*wl) = (*wl)->wl_next;
    if (!*wl) {
        loop->step[1] = loop->stop[1] - loop->start[1];
        return (false);
    }

    line = (*wl)->wl_word;
    tmp = INPevaluate(&line,&error,1);  /* step2 */
    if (error == 0)
        loop->start[1] = tmp;
    else {
        loop->step[1] = loop->stop[1] - loop->start[1];
        return (false);
    }
    (*wl) = (*wl)->wl_next;
    return (false);
}


void
ft_check(wl,infile)

wordlist *wl;
FILE *infile;
{
    FILE *fp;
    struct line *deck;
    wordlist *ww;
    char *filename = NULL;
    char *cblock;

    if (currentgraph && currentgraph->graphtype == GR_SCED) {
        ShowPrompt("Exit sced to run operating range analysis.");
        return;
    }

    out_init(); /* this might not have been called before, needed
                 * for inp_dodeck()
                 */

    if (infile != NULL) {
        /* file sourced */
        fp = infile;
        if (wl && wl->wl_word && *wl->wl_word) {
            filename = wl->wl_word;
            wl = wl->wl_next;
        }
    }

    else {
        if (!wl || !wl->wl_word || *wl->wl_word == '\0') {
            fp = cp_in;
        }
        else {
            filename = wl->wl_word;
            wl = wl->wl_next;
            if ((fp = fopen(filename,"r")) == NULL) {
                (void) fprintf(cp_err,"Error: can't open source file\n");
                return;
            }
        }
    }

    if (filename && *filename) {
        cblock = tmalloc(strlen(filename) + 3);
        sprintf(cblock,"#_%s",filename);
    }
    else
        cblock = copy("#_");


    if (read_check_control_file(fp,cblock)) {
        (void) fclose(fp);
        return;
    }
    inp_readall(fp,&deck,NULL);
    if (fp != cp_in)
        (void) fclose(fp);

    ww = cp_lexer("set value1=0.0 value2=0.0");
    com_set(ww);
    wl_free(ww);
    inp_spdeck(deck,filename);
    ft_curckt->ci_contblk = cblock;

    if (ft_batchmode) {
        /* run the analysis and free */
        com_check((wordlist*)NULL);
        if_cktclear();
    }
}


static void
pr_parms()

{
    int step1, step2, iterno;
    double val1, delta1, val2, delta2;
    struct dvec *d;
    extern char *kw_checkiterate;

    val1 = 0;
    step1 = 0;
    delta1 = 0;
    val2 = 0;
    step2 = 0;
    delta2 = 0;
    iterno = 0;

    if ((d = vec_get("checkDEL1")) != NULL)
        delta1 = *d->v_realdata;
    if ((d = vec_get("checkVAL1")) != NULL)
        val1 = *d->v_realdata;
    if ((d = vec_get("checkDEL2")) != NULL)
        delta2 = *d->v_realdata;
    if ((d = vec_get("checkVAL2")) != NULL)
        val2 = *d->v_realdata;
    if ((d = vec_get("checkSTP1")) != NULL)
        step1 = (int) *d->v_realdata;
    if ((d = vec_get("checkSTP2")) != NULL)
        step2 = (int) *d->v_realdata;
    (void) cp_getvar(kw_checkiterate,VT_NUM,(char*)&iterno);

    out_printf(
"Substitution for value1:\n\
    value: %g\n\
    delta: %g\n\
    steps: %d\n\n",val1,delta1,step1);

    out_printf(
"Substitution for value2:\n\
    value: %g\n\
    delta: %g\n\
    steps: %d\n\n",val2,delta2,step2);

    out_printf("checkiterate is set to %d\n\n",iterno);
}


static void
checkloop(cmdwl)

wordlist *cmdwl;
{
    struct dvec *d;
    struct control cntrl;
    double value1, value2;
    char *apology = "Sorry, margin analysis can't be resumed in this version.\n";
    extern char *kw_checkiterate;

    OUTcntrl->out_check = true;
    OUTcntrl->out_keepplot = true;
    OUTcntrl->out_usecurplot = true;

    cntrl.val1 = 0;
    cntrl.step1 = 0;
    cntrl.delta1 = 0;
    cntrl.val2 = 0;
    cntrl.step2 = 0;
    cntrl.delta2 = 0;
    cntrl.iterno = 0;

    if ((d = vec_get("checkDEL1")) != NULL)
        cntrl.delta1 = *d->v_realdata;
    if ((d = vec_get("checkVAL1")) != NULL)
        cntrl.val1 = *d->v_realdata;
    if ((d = vec_get("checkDEL2")) != NULL)
        cntrl.delta2 = *d->v_realdata;
    if ((d = vec_get("checkVAL2")) != NULL)
        cntrl.val2 = *d->v_realdata;
    if ((d = vec_get("checkSTP1")) != NULL)
        cntrl.step1 = (int) *d->v_realdata;
    if ((d = vec_get("checkSTP2")) != NULL)
        cntrl.step2 = (int) *d->v_realdata;

    value1 = cntrl.val1 - cntrl.step1 * cntrl.delta1;
    value2 = cntrl.val2 - cntrl.step2 * cntrl.delta2;
    modify(value1,value2);
    com_rset((wordlist*)NULL);
    graphid = (mp_init(2*cntrl.step1+1,2*cntrl.step2+1,
        value1, cntrl.val1+cntrl.step1*cntrl.delta1,
        value2, cntrl.val2+cntrl.step2*cntrl.delta2));

    if (!graphid && !batchmode)
        (void) fprintf(cp_out,"%d %d %g %g\n",
            -cntrl.step1,-cntrl.step2,value1,value2);

    if (Op)
        (void) fprintf(Op,"%d %d %g %g\n",
            -cntrl.step1,-cntrl.step2,value1,value2);

    mp_where(graphid,-cntrl.step1,-cntrl.step2);


    /* run the analysis */
    (void)cp_getvar(kw_checkiterate,VT_NUM,(char*)&cntrl.iterno);
    if (cntrl.iterno < 0 || cntrl.iterno > 10) {
        cntrl.iterno = 0;
        (void)fprintf(cp_err,
            "Warning: bad value for checkiterate, ignored.\n");
    }
    set_opvec(cntrl.step2);

    (void) if_run(cmdwl->wl_word, cmdwl->wl_next);

    if (OUTcntrl->out_check == 0) { /* user interrupt */
        (void) fprintf(cp_err,apology);
        return;
    }

    if (mainloop(&cntrl))
        (void) fprintf(cp_err,apology);
}


static void
set_opvec(n)

/* create vectors for operating range exterma (value 2) */
int n;
{
    char buf[BSIZE_SP];
    bool temp;

    temp = cp_interactive;
    cp_interactive = false;
    cp_pushcontrol();
    sprintf(buf,"unlet oplo; let oplo[%d]=0",2*n);
    cp_evloop(buf);
    sprintf(buf,"unlet ophi; let ophi[%d]=0",2*n);
    cp_evloop(buf);
    cp_popcontrol();
    cp_interactive = temp;
}


static int
mainloop(cntrl)

struct control *cntrl;
{
    int i, j, last;
    int num1, num2;
    double value1, value2, delta;
    char *flags, *rowflags;
    struct dvec *d;

    if (cntrl->step1 == 0 && cntrl->step2 == 0) {
        /* new feature - find the operating range of value1 */

        if (cntrl->iterno == 0)
            return (false);
        if (OUTcntrl->out_fail) {
            /* center point bad */
            (void) fprintf(cp_err,
                "Error: central point bad, range not found.\n");
            return (false);
        }
        value2 = cntrl->val2;
        value1 = cntrl->val1;
        delta = .5*value1;
        for (i = 0; i < 10; i++) {
            value1 += delta;
            no_output = true;
            do_it(0,0,value1,value2);
            no_output = false;
            if (OUTcntrl->out_fail) {
                value1 -= delta;
                if (findext(cntrl->iterno,&value1,value2,-delta))
                    goto quit;
                if ((d = vec_get("ophi")) != NULL)
                    d->v_realdata[0] = value1;
                break;
            }
        }
        if (i == 10) {
            (void) fprintf(cp_err,"Warning: could not find upper limit.\n");
            if ((d = vec_get("ophi")) != NULL)
                d->v_realdata[0] = value1;
        }
        value1 = cntrl->val1;
        delta = .49*value1;
        for (i = 0; i < 2; i++) {
            value1 -= delta;
            no_output = true;
            do_it(0,0,value1,value2);
            no_output = false;
            if (OUTcntrl->out_fail) {
                value1 += delta;
                if (findext(cntrl->iterno,&value1,value2,delta))
                    goto quit;
                if ((d = vec_get("oplo")) != NULL)
                    d->v_realdata[0] = value1;
                break;
            }
        }
        if (i == 2) {
            (void) fprintf(cp_err,"Warning: could not find lower limit.\n");
            if ((d = vec_get("oplo")) != NULL)
                d->v_realdata[0] = value1;
        }
        return (false);
    }

    num1 = 2*cntrl->step1 + 1;
    num2 = 2*cntrl->step2 + 1;
    flags = tmalloc(num1*num2);
    /* fill in the first point */
    *flags = 1 + OUTcntrl->out_fail;

    /* Find the range along the rows. */
    for (j = -cntrl->step2; j <= cntrl->step2; j++) {
        value2 = cntrl->val2 + j*cntrl->delta2;

        rowflags = flags + (j + cntrl->step2)*num1;
        for (last = i = -cntrl->step1; i <= cntrl->step1; i++,rowflags++) {
            if (*rowflags == 2) continue;
            if (*rowflags == 1) break;
            value1 = cntrl->val1 + i*cntrl->delta1;
            *rowflags = do_it(i,j,value1,value2);
            if (OUTcntrl->out_check == 0) /* user interrupt */
                goto quit;
            last = i;
            if (!OUTcntrl->out_fail) break;
        }
        if (!OUTcntrl->out_fail && i != -cntrl->step1) {
            if (findext(cntrl->iterno,&value1,value2,cntrl->delta1))
                goto quit;
            if ((d = vec_get("oplo")) != NULL)
                d->v_realdata[j+cntrl->step2] = value1;
        }

        rowflags = flags + (j + cntrl->step2 + 1)*num1 - 1;
        for (i = cntrl->step1; i > last; i--,rowflags--) {
            if (*rowflags == 2) continue;
            if (*rowflags == 1) break;
            value1 = cntrl->val1 + i*cntrl->delta1;
            *rowflags = do_it(i,j,value1,value2);
            if (OUTcntrl->out_check == 0) /* user interrupt */
                goto quit;
            if (!OUTcntrl->out_fail) break;
        }
        if (!OUTcntrl->out_fail && i != cntrl->step1) {
            if (findext(cntrl->iterno,&value1,value2,-cntrl->delta1))
                goto quit;
            if ((d = vec_get("ophi")) != NULL)
                d->v_realdata[j+cntrl->step2] = value1;
        }
    }

    /* Now check the columns, fill in any missing points.
     * Do this for column height > 3 only.
     */
    if (cntrl->step2 > 1) {
        for (i = -cntrl->step1; i <= cntrl->step1; i++) {
            value1 = cntrl->val1 + i*cntrl->delta1;

            rowflags = flags + i + cntrl->step1;
            for (last = j = -cntrl->step2; j <= cntrl->step2;
                    j++,rowflags += num1) {
                if (*rowflags == 2) continue;
                if (*rowflags == 1) break;
                value2 = cntrl->val2 + j*cntrl->delta2;
                *rowflags = do_it(i,j,value1,value2);
                if (OUTcntrl->out_check == 0) /* user interrupt */
                    goto quit;
                last = j;
                if (!OUTcntrl->out_fail) break;
            }

            rowflags = flags + i + cntrl->step1 + 2*cntrl->step2*num1;
            for (j = cntrl->step2; j > last; j--,rowflags -= num1) {
                if (*rowflags == 2) continue;
                if (*rowflags == 1) break;
                value2 = cntrl->val2 + j*cntrl->delta2;
                *rowflags = do_it(i,j,value1,value2);
                if (OUTcntrl->out_check == 0) /* user interrupt */
                    goto quit;
                if (!OUTcntrl->out_fail) break;
            }
        }
    }
    txfree(flags);
    return (false);
quit:
    txfree(flags);
    return (true);
}


static int
do_it(i,j,value1,value2)

int i,j;
double value1,value2;
{
    char *task;

    modify(value1,value2);
    if (!no_output) {
        if (mp_where(graphid,i,j) && !batchmode)
            (void) fprintf(cp_out,"%d %d %g %g\n",i,j,value1,value2);
        if (Op)
            (void) fprintf(Op,"%d %d %g %g\n",i,j,value1,value2);
    }
    if (ft_curckt->ci_specTask) {
        task = ft_curckt->ci_specTask;
        ft_curckt->ci_specTask = NULL;
        com_rset((wordlist*)NULL);
        ft_curckt->ci_specTask = task;
    }
    else {
        com_rset((wordlist*)NULL);
        task = ft_curckt->ci_defTask;
    }

    ft_curckt->ci_curTask = task;
    (void)(*(ft_sim->doAnalyses))(ft_curckt->ci_ckt, 1,task);
    return (OUTcntrl->out_fail + 1);
}


static int
findext(iterno,value1,value2,delta)

/* Iterate to the edge of the operating region along row */
int iterno;
double *value1,value2,delta;
{
    char *task;

    delta *= .5;
    *value1 -= delta;

    no_output = true;
    while (iterno--) {
        modify(*value1,value2);

        if (ft_curckt->ci_specTask) {
            task = ft_curckt->ci_specTask;
            ft_curckt->ci_specTask = NULL;
            com_rset((wordlist*)NULL);
            ft_curckt->ci_specTask = task;
        }
        else {
            com_rset((wordlist*)NULL);
            task = ft_curckt->ci_defTask;
        }

        ft_curckt->ci_curTask = task;
        (void) (*(ft_sim->doAnalyses))(ft_curckt->ci_ckt, 1, task);

        if (OUTcntrl->out_check == 0) { /* user interrupt */
            no_output = false;
            return (true);
        }

        delta *= .5;
        if (!OUTcntrl->out_fail)
            *value1 -= delta;
        else
            *value1 += delta;
    }
    no_output = false;
    return (false);
}


static void
check_end()
{
    if (no_output)
        return;
    if (mp_mark(graphid,!OUTcntrl->out_fail) && !batchmode)
        (void) fprintf(cp_out,OUTcntrl->out_fail ? " fail\n\n" : " pass\n\n");
    if (Op)
        (void) fprintf(Op,OUTcntrl->out_fail ? "\tfail\n" : "\tpass\n");
}


static int
read_check_control_file(fp,blname)

FILE *fp;
char *blname;
{
    char *buf;
    wordlist *chklst = NULL, *ckl;
    wordlist *hdrlst = NULL, *hdl;
    bool begin = false;

    while ((buf = readline(fp)) != NULL) {
        if (*buf == '#') {
            txfree(buf);
            continue;
        }
        if (prefix(".control",buf)) {
            begin = true;
            txfree(buf);
            continue;
        }
        if (prefix(".endc",buf)) {
            if (!begin)
                (void) fprintf(cp_err,"Warning: no .control line found\n");
            txfree(buf);
            break;
        }
        *strchr(buf,'\n') = '\0';
        if (!*buf) {
            txfree(buf);
            continue;
        }
        if (!begin) {
            if (!hdrlst)
                hdrlst = hdl = alloc(wordlist);
            else {
                hdl->wl_next = alloc(wordlist);
                hdl = hdl->wl_next;
            }
            hdl->wl_word = buf;
            continue;
        }
        if (!chklst)
            chklst = ckl = alloc(wordlist);
        else {
            ckl->wl_next = alloc(wordlist);
            ckl = ckl->wl_next;
        }
        ckl->wl_word = buf;
    }
    cp_addblock(blname,chklst);
    blname[1] = '#';
    cp_addblock(blname,hdrlst);
    blname[1] = '_';
    wl_free(chklst);
    wl_free(hdrlst);
    return (false);
}


static bool
check_evaluate()
{
    struct dvec *d;
    char *blname;

    blname = ft_curckt->ci_contblk;
    if (!blname)
        return (true);

    execblock(blname);
    d = vec_get("checkFAIL");

    /* checkFAIL is true if failed */
    return ((int) *d->v_realdata ? true : false);
}


static void
execblock(blname)

char *blname;
{
    bool temp;
    FILE *lastin, *lastout, *lasterr;

    temp = cp_interactive;
    cp_interactive = false;
    lastin = cp_curin;
    lastout = cp_curout;
    lasterr = cp_curerr;
    cp_curin = cp_in;
    cp_curout = cp_out;
    cp_curerr = cp_err;

    cp_execcontrol(blname);

    cp_curin = lastin;
    cp_curout = lastout;
    cp_curerr = lasterr;
    cp_interactive = temp;
}


static void
modify(value1,value2)

double value1,value2;
{
    char buf[BSIZE_SP];
    wordlist *wl;

    sprintf(buf,"set value1=%g value2=%g",value1,value2);
    wl = cp_lexer(buf);
    com_set(wl);
    wl_free(wl);
    setvec(value1,value2);
}


static void
setvec(v1,v2)

double v1,v2;
{
    struct dvec *d, *n1, *n2;
    int i, len;

    d = vec_get("value");
    if (d == NULL)
        return;
    if (!isreal(d))
        return;
    n1 = vec_get("checkN1");
    n2 = vec_get("checkN2");
    if (n1 == NULL && n2 == NULL)
        return;
    len = d->v_length;
    if (n1) {
        i = (int)n1->v_realdata[0];
        if (i < 0 || i >= len)
            return;
        d->v_realdata[i] = v1;
    }
    if (n2) {
        i = (int)n2->v_realdata[0];
        if (i < 0 || i >= len)
            return;
        d->v_realdata[i] = v2;
    }
}



static FILE *
df_open()

/* open the data output file for writing.  name: xxxxx.dnn , where
 * xxxxx is the base of the circuit file name, or "check" if no
 * current file name.  nn is 00 - 99.
 * If successful, set the variable "mplot_cur" to the file name.
 */
{
    FILE *fp, *fopen();
    char buf[128], buf1[128], *s;
    wordlist *wl;
    int i;

    if (ft_servermode) {
        wl = cp_lexer("set mplot_cur=mplot");
        com_set(wl);
        wl_free(wl);
        return (cp_out);
    }
        
    if (ft_curckt->ci_filename)
        sprintf(buf,ft_curckt->ci_filename);
    else
        strcpy(buf,"check");
    if ((s = strrchr(buf,DIR_TERM)) != NULL)
        s++;
    else
        s = buf;
    strcpy(buf1,s);

    if ((s = strrchr(buf1,'.')) != NULL)
        *s = '\0';
    strcat(buf1,".d00");
    s = strchr(buf1,'.') + 2;
    for (i = 1; i < 100; i++) {
        if (access(buf1,0)) break;
        *s = i/10 + '0';
        *(s+1) = i%10 + '0';
    }
    if (i == 100)
        return (NULL);
    fp = fopen(buf1,"w");
    if (!fp) {
        perror(buf1);
        return (NULL);
    }

    sprintf(buf,"set mplot_cur=%s",buf1);
    wl = cp_lexer(buf);
    com_set(wl);
    wl_free(wl);

    return (fp);
}


static void
df_close(fp)

FILE *fp;
{
    if (fp && fp != cp_out)
        (void) fclose(fp);
}
