/***************************************************************************
JSPICE3 adaptation of Spice3e2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Wayne A. Christopher, Norbert Jeske
         1992 Stephen R. Whiteley
****************************************************************************/

/*
 * Expand subcircuits. This is very spice-dependent.
 */

#include "spice.h"
#include "ftedefs.h"
#include "fteinp.h"
#include "kwords.h"
#include "spfteext.h"


#ifdef __STDC__
static struct line *doit(struct line*);
static bool translate(struct line*,char*,char*,char*,char*,char*);
static void finishLine(char*,char*,char*);
static bool settrans(char*,char*,char*,char*);
static bool aredups(void);
static char *gettrans(char*);
static void numrefs(char*,int*,int*);
static bool modtranslate(struct line*,char*);
static void devmodtranslate(struct line*,char*);
static void looksubmod(char**,char*,char*);
#else
static struct line *doit();
static bool translate();
static void finishLine();
static bool settrans();
static bool aredups();
static char *gettrans();
static void numrefs();
static bool modtranslate();
static void devmodtranslate();
static void looksubmod();
#endif

struct subs {
    char *su_name;          /* The name. */
    char *su_args;          /* The arguments, space seperated. */
    int su_numargs;
    struct line *su_def;    /* The deck that is to be substituted. */
    struct subs *su_next;
} ;

static struct tab {
    char *t_old;
    char *t_new;
} table[512];   /* That had better be enough. */


/* Expand all subcircuits in the deck. This handles imbedded .subckt
 * definitions. The variables substart, subend, and subinvoke can be used
 * to redefine the controls used. The syntax is invariant though.
 * NOTE: the deck must be passed without the title card.
 * What we do is as follows: first make one pass through the circuit
 * and collect all of the subcircuits. Then, whenever a card that starts
 * with 'x' is found, copy the subcircuit associated with that name and
 * splice it in. A few of the problems: the nodes in the spliced-in
 * stuff must be unique, so when we copy it, append "subcktname:" to
 * each node. If we are in a nested subcircuit, use foo:bar:...:node.
 * Then we have to systematically change all references to the renamed
 * nodes. On top of that, we have to know how many args BJT's have,
 * so we have to keep track of model names.
 */

static wordlist *modnames, *submod;
static struct subs *subs;
static bool nobjthack = false;

static char start[32], sbend[32], invoke[32], model[32];


struct line *
inp_subcktexpand(deck)

struct line *deck;
{
    struct line *ll, *c;
    char *s;
    extern char *kw_substart, *kw_subend, *kw_subinvoke, *kw_modelcard;
    extern char *kw_nobjthack;

    if (!cp_getvar(kw_substart, VT_STRING, start))
        (void) strcpy(start, ".subckt");
    if (!cp_getvar(kw_subend, VT_STRING, sbend))
        (void) strcpy(sbend, ".ends");
    if (!cp_getvar(kw_subinvoke, VT_STRING, invoke))
        (void) strcpy(invoke, "X");
    if (!cp_getvar(kw_modelcard, VT_STRING, model))
        (void) strcpy(model, ".model");
    (void) cp_getvar(kw_nobjthack, VT_BOOL, (char *) &nobjthack);

    /* Let's do a few cleanup things first... Get rid of ( ) around node
     * lists...
     */
    for (c = deck; c; c = c->li_next) {
        if (prefix(start, c->li_line)) {
            for (s = c->li_line; *s && (*s != '('); s++)
                ;
            if (*s) {
                while (s[0] && (s[1] != ')')) {
                    s[0] = s[1];
                    s++;
                }
                while (s[1]) {
                    s[0] = s[2];
                    s++;
                }
            }
        }
        else {
            for (s = c->li_line; *s && !isspace(*s); s++)
                ;
            while (isspace(*s))
                s++;
            if (*s == '(') {
                while (s[0] && (s[1] != ')')) {
                    s[0] = s[1];
                    s++;
                }
                while (s[1]) {
                    s[0] = s[2];
                    s++;
                }
            }
        }
    }
    
    ll = doit(deck);

    /* Now check to see if there are still subckt instances undefined... */
    if (ll) {
        for (c = ll; c; c = c->li_next)
            if (ciprefix(invoke, c->li_line)) {
                fprintf(cp_err, "Error: unknown subckt: %s\n",
                        c->li_line);
                ll = NULL;
            }
    }

    return (ll);
}


#define MAXNEST 21

static struct line *
doit(deck)

struct line *deck;
{
    struct line *c, *last, *lc, *lcc;
    struct subs *sss;
    char *s, *t, *tt, *scname, *subname;
    int nest, numpasses = MAXNEST, i;
    bool gotone;
    wordlist *wl;
    wordlist *tmodnames = modnames;
    wordlist *tsubmod = submod;
    struct subs *ts = subs;

    /* Save all the old stuff... */
    modnames = NULL;
    subs = NULL;
    submod = NULL;

    /* Extract all the .subckts */
    for (last = deck, lc = NULL; last; ) {
        if (prefix(sbend, last->li_line)) {
            fprintf(cp_err, "Error: misplaced %s card: %s\n", sbend,
                last->li_line);
            deck = NULL;
            goto cleanup;
        }
        else if (prefix(start, last->li_line)) {
            if (last->li_next == NULL) {
                fprintf(cp_err, "Error: no %s card.\n", sbend);
                deck = NULL;
                goto cleanup;
            }
            gotone = false;
            lcc = NULL;
            for (nest = 0, c = last->li_next; c; lcc = c,c = c->li_next) {
                if (prefix(sbend, c->li_line)) {
                    if (!nest)
                        break;
                    else {
                        nest--;
                        continue;
                    }
                }
                else if (prefix(start, c->li_line))
                    nest++;
            }
            if (!c) {
                fprintf(cp_err, "Error: no %s card.\n", sbend);
                deck = NULL;
                goto cleanup;
            }
            if (lcc)
                lcc->li_next = NULL;
            else
                last->li_next = NULL;

            sss = alloc(struct subs);
            if (lc)
                lc->li_next = c->li_next;
            else
                deck = c->li_next;

            sss->su_def = last->li_next;
            s = last->li_line;
            advtok(&s);
            sss->su_name = gettok(&s);
            sss->su_args = copy(s);
            for (sss->su_numargs = 0, i = 0; s[i]; ) {
                while (isspace(s[i]))
                    i++;
                if (s[i]) {
                    sss->su_numargs++;
                    while (s[i] && !isspace(s[i]))
                        i++;
                }
            }
            sss->su_next = subs;
            subs = sss;

            last->li_next = NULL;
            inp_deckfree(last);

            last = c->li_next;

            c->li_next = NULL;
            inp_deckfree(c);
        }
        else {
            lc = last;
            last = last->li_next;
        }
    }

    /* Expand sub-subcircuits. */
    for (sss = subs; sss; sss = sss->su_next)
        if (sss->su_def) {
            if (!(sss->su_def = doit(sss->su_def))) {
                deck = NULL;
                goto cleanup;
            }
        }
    
    /* Get all the model names so we can deal with BJT's. */
    for (c = deck; c; c = c->li_next)
        if (prefix(model, c->li_line)) {
            s = c->li_line;
            advtok(&s);
            wl = alloc(struct wordlist);
            wl->wl_next = modnames;
            if (modnames)
                modnames->wl_prev = wl;
            modnames = wl;
            wl->wl_word = gettok(&s);
        }

    /* Now do the replacements. */
    do {
        gotone = false;
        for (c = deck, lc = NULL; c; ) {
            if (ciprefix(invoke, c->li_line)) {
                char *tmpscname;
                gotone = true;
                t = s = copy(c->li_line);
                tmpscname = scname = gettok(&s);
                scname += strlen(invoke);
                while ((*scname == ' ') || (*scname == '\t') ||
                        (*scname == ':'))
                    scname++;
                while(*s)
                    s++;
                s--;
                while ((*s == ' ') || (*s == '\t'))
                    *s-- = '\0';
                while ((*s != ' ') && (*s != '\t'))
                    s--;
                s++;
                for (sss = subs; sss; sss = sss->su_next)
                    if (eq(sss->su_name, s))
                        break;

                /* Don't complain -- this might be an
                 * instance of a subckt that is defined above.
                 */
                if (!sss) {
                    lc = c;
                    c = c->li_next;
                    txfree(tmpscname);
                    txfree(t);
                    continue;
                }

                if (!sss->su_def) {
                    /* null subccircuit, just clip out
                     * invocation and ignore
                     */
                    if (lc)
                        lc->li_next = c->li_next;
                    else
                        deck = c->li_next;

                    c->li_next = NULL;
                    inp_deckfree(c);

                    if (lc)
                        c = lc->li_next;
                    else
                        c = deck;

                    txfree(tmpscname);
                    txfree(t);
                    fprintf(cp_err,
                        "Warning: empty subcircuit %s referenced, ignored\n",
                        sss->su_name);
                    continue;
                }

                /* Now we have to replace this card with the
                 * macro definition.
                 */
                subname = copy(sss->su_name);

                lcc = inp_deckcopy(sss->su_def);

                /* Change the names of the models... */
                if (modtranslate(lcc, scname))
                    devmodtranslate(lcc, scname);

                s = sss->su_args;
                tt = t;
                advtok(&t); /* Throw out the name. */

                if (translate(lcc, s, t, scname, subname, c->li_line)) {
                    txfree(tmpscname);
                    tfree(subname);
                    txfree(tt);
                    deck = NULL;
                    goto cleanup;
                }

                txfree(tmpscname);
                tfree(subname);
                txfree(tt);

                /* Now splice the decks together. */
                if (lc)
                    lc->li_next = lcc;
                else
                    deck = lcc;
                while (lcc->li_next != NULL)
                    lcc = lcc->li_next;
                lcc->li_next = c->li_next;

                c->li_next = NULL;
                inp_deckfree(c);
                c = lcc->li_next;
                lc = lcc;
            }
            else {
                lc = c;
                c = c->li_next;
            }
        }
    } while (numpasses-- && gotone);

    if (!numpasses) {
        fprintf(cp_err, "Error: infinite subckt recursion\n");
        deck = NULL;
    }

cleanup:
    wl_free(modnames);
    wl_free(submod);
    for (sss = subs; sss; sss = subs) {
        subs = sss->su_next;
        txfree(sss->su_name);
        txfree(sss->su_args);
        inp_deckfree(sss->su_def);
        txfree((char*)sss);
    }
    subs = ts;
    modnames = tmodnames;
    submod = tsubmod;
    return (deck);
}


static bool
translate(deck, formal, actual, scname, subname, instr)

/* Translate all of the device names and node names in the deck. They are
 * pre-pended with subname:, unless they are in the formal list, in which
 * case they are replaced with the corresponding entry in the actual list.
 * The one special case is node 0 -- this is always ground and we don't
 * touch it.
 */
struct line *deck;
char *formal, *actual, *scname, *subname, *instr;
{
    struct line *c;
    char *buffer, *name, *s, *t, ch;
    int nnodes, ndevs, i;
    char *tmpname;
    bool retval = false;

    if (settrans(formal, actual, subname, instr)) {
        retval = true;
        goto cleanup;
    }

    for (c = deck; c; c = c->li_next) {
        /* Rename the device. */
        ch = *c->li_line;
        if (ch == '.' || ch == '*' || ch == '\0')
            /* Nothing any good here. */
            continue;

        s = c->li_line;
        tmpname = name = gettok(&s);
        if (!name)
            continue;
        if (!*name) {
            txfree(name);
            continue;
        }
        ch = *name;
        buffer = tmalloc(1000);
        name++;
        if (*name == ':')
            name++;
        if (*name)
            (void) sprintf(buffer, "%c:%s:%s ", ch, scname, name);
        else
            (void) sprintf(buffer, "%c:%s ", ch, scname);
        txfree(tmpname);

        numrefs(c->li_line,&nnodes,&ndevs);
        while (nnodes-- > 0) {
            name = gettok(&s);
            if (name == NULL) {
               fprintf(cp_err, "Warning: too few nodes: %s\n", c->li_line);
                retval = true;
                txfree(name);
                goto cleanup;
            }
            t = gettrans(name);
            if (t)
                (void) sprintf(buffer + strlen(buffer), "%s ", t);
            else
                (void) sprintf(buffer + strlen(buffer),
                        "%s:%s ", scname, name);
            txfree(name);
        }    
        while (ndevs-- > 0) {
            tmpname = name = gettok(&s);
            if (name == NULL) {
                fprintf(cp_err, "Warning: too few devs: %s\n", c->li_line);
                retval = true;
                goto cleanup;
            }
            ch = *name;
            name++;
            if (*name == ':')
                name++;
            if (*name)
                (void) sprintf(buffer + strlen(buffer),
                        "%c:%s:%s ", ch, scname, name);
            else
                (void) sprintf(buffer + strlen(buffer),
                        "%c:%s ", ch, scname);
            txfree(tmpname);
        }
        /* Now scan through the line for v(something) and
         * i(something)...
         */
        finishLine(buffer + strlen(buffer), s, scname);
        txfree(c->li_line);
        c->li_line = copy(buffer);
        txfree(buffer);
    }
cleanup:
    for (i = 0; i < 512; i++) {
        if (table[i].t_old == NULL) {
            tfree(table[i].t_new);
            break;
        }
        tfree(table[i].t_old);
        tfree(table[i].t_new);
    }
    return retval;
}


static void
finishLine(dst, src, scname)

char *dst;
char *src;
char *scname;
{
    char buf[4 * BSIZE_SP], which;
    char *s;
    int i;
    int lastwasalpha;

    lastwasalpha = 0;
    while (*src) {
        /* Find the next instance of "<non-alpha>[vi]<opt spaces>(" in
         * this string.
         */
        if (((*src != 'v') && (*src != 'V') &&
            (*src != 'i') && (*src != 'I')) ||
            lastwasalpha) {
            lastwasalpha = isalpha(*src);
            *dst++ = *src++;
            continue;
        }
        for (s = src + 1; *s && isspace(*s); s++)
            ;
        if (!*s || (*s != '(')) {
            lastwasalpha = isalpha(*src);
            *dst++ = *src++;
            continue;
        }
        lastwasalpha = 0;
        which = *dst++ = *src;
        src = s;
        *dst++ = *src++;
        while (isspace(*src))
            src++;
        for (i = 0;
            *src && !isspace(*src) && *src != ',' && (*src != ')'); i++) {
            buf[i] = *src++;
        }
        buf[i] = '\0';

        if ((which == 'v') || (which == 'V'))
            s = gettrans(buf);
        else
            s = NULL;

        if (s) {
            while (*s)
                *dst++ = *s++;
        }
        else {
            /*
             * i(vname) -> i(v:subckt:name)
             * i(v:other:name) -> i(v:subckt:other:name)
             */
            if (buf[0] == 'v' || buf[0] == 'V') {
                *dst++ = buf[0];
                *dst++ = ':';
                i = 1;
            }
            else {
                i = 0;
            }
            for (s = scname; *s; )
                *dst++ = *s++;
            *dst++ = ':';
            for (s = buf + i; *s; )
                *dst++ = *s++;
        }

        /* translate the reference node, as in the "2" in "v(4,2)" */

        if ((which == 'v') || (which == 'V')) {
            while (*src && (isspace(*src) || *src == ',')) {
                src++;
            }
            if (*src && *src != ')') {
                for (i = 0; *src && !isspace(*src) && (*src != ')'); i++)
                    buf[i] = *src++;
                buf[i] = '\0';
                s = gettrans(buf);
                *dst++ = ',';
                if (s) {
                    while (*s)
                    *dst++ = *s++;
                }
                else {
                    for (s = scname; *s; )
                        *dst++ = *s++;
                    *dst++ = ':';
                    for (s = buf; *s; )
                        *dst++ = *s++;
                }
            }
        }
    }
    return;
}


static bool
settrans(formal, actual, subname, instr)

char *formal, *actual, *subname, *instr;
{
    int i;

    for (i = 0; ; i++) {
        table[i].t_old = gettok(&formal);
        table[i].t_new = gettok(&actual);

        if (table[i].t_new == NULL) {
            fprintf(cp_err, "Error: too many params: %s\n", instr);
            return (true);        /* Too few formal / too many actual */
        }
        if (table[i].t_old == NULL) {
            if (eq(table[i].t_new, subname))
                break;
            else {
                fprintf(cp_err, "Error: too few params: %s\n", instr);
                return (true);    /* Too few actual / too many formal */
            }
        }
    }
    if (aredups()) {
        fprintf(cp_err,
            "Error: duplicate nodes given in subckt %s\n",subname);
        return (true);
    }
    return (false);
}


static bool
aredups()

{
    int i, j;
    char *s;

    for (i = 0; table[i].t_old; i++) {
        s = table[i].t_old;
        for (j = i+1; table[j].t_old; j++) {
            if (eq(s,table[j].t_old))
                return (true);
        }
    }
    return (false);
}


static char *
gettrans(name)

char *name;
{
    int i;

    if (eq(name, "0"))
        return (name);
    for (i = 0; table[i].t_old; i++)
        if (eq(table[i].t_old, name))
            return (table[i].t_new);
    return (NULL);
}


static void
numrefs(name,terms,devs)

char *name;
int *terms, *devs;
{
    char c;
    struct subs *sss;
    char *s, *t, buf[4 * BSIZE_SP];
    wordlist *wl;
    int n, i, soft;

    while (*name && isspace(*name))
    name++;

    c = (isupper(*name) ? tolower(*name) : *name);

    (void) strncpy(buf, name, sizeof(buf));
    s = buf;
    if (c == 'x') {     /* Handle this ourselves. */
        while(*s)
            s++;
        s--;
        while ((*s == ' ') || (*s == '\t'))
            *s-- = '\0';
        while ((*s != ' ') && (*s != '\t'))
            s--;
        s++;
        for (sss = subs; sss; sss = sss->su_next)
            if (eq(sss->su_name, s))
                break;
        if (!sss) {
            *terms = 0;
            *devs = 0;
            fprintf(cp_err, "Error: no such subcircuit: %s\n", s);
            return;
        }
        *terms = sss->su_numargs;
        *devs = 0;
        return;
    }

    (void)INPnumRefs(c,&n,devs,&soft,NULL);
    if (isalpha(c) && n == 0 && *devs == 0)
        fprintf(cp_err, "Warning: unknown device key: %c\n", c);

    if (!soft) {
        *terms = n;
        return;
    }
    for (s = buf, i = 0; *s && (i < n); i++)
        advtok(&s);
    if (i == n-1) {
        *terms = n-1;
        return;
    }
    else if (i < n) {
        fprintf(cp_err, "Error: too few nodes for %s\n", name);
        *terms = 0;
        return;
    }
    /* Now, is this a model? */
    t = gettok(&s);
    for (wl = modnames; wl; wl = wl->wl_next) {
        if (eq(t, wl->wl_word)) {
            txfree(t);
            *terms = n-1;
            return;
        }
    }
    txfree(t);
    *terms = n;
    return;
}


static bool
modtranslate(deck, subname)

struct line *deck;
char *subname;
{
    struct line *c;
    char *buffer, *name, *t;
    wordlist *wl, *wlsub;
    bool gotone;

    gotone = false;
    for (c = deck; c; c = c->li_next) {
        if (prefix(model, c->li_line)) {
            gotone = true;
            t = c->li_line;
            name = gettok(&t);
            buffer = tmalloc(strlen(name) + strlen(t) +
                    strlen(subname) + 4);
            (void) sprintf(buffer, "%s ",name);
            txfree(name);
            name = gettok(&t);
            wlsub = alloc(struct wordlist);
            wlsub->wl_next = submod;
            if (submod)
                submod->wl_prev = wlsub;
            submod = wlsub;
            wlsub->wl_word = name;
            (void) sprintf(buffer + strlen(buffer), "%s:%s ",
                    subname, name);
            (void) strcat(buffer, t);
            txfree(c->li_line);
            c->li_line = buffer;
            t = c->li_line;
            advtok(&t);
            wl = alloc(struct wordlist);
            wl->wl_next = modnames;
            if (modnames) 
                modnames->wl_prev = wl;
            modnames = wl;
            wl->wl_word = gettok(&t);
        }
    }
    return (gotone);
}


static void
devmodtranslate(deck, subname)

struct line *deck;
char *subname;
{
    struct line *s;
    char *buffer, *name, *t, c;
    int terms, devs, soft, hasmod;

    for (s = deck; s; s = s->li_next) {
        t = s->li_line;
        while (*t && isspace(*t))
            t++;
        c = isupper(*t) ? tolower(*t) : *t;

        INPnumRefs(c,&terms,&devs,&soft,&hasmod);
        if (!hasmod)
            continue;

        buffer = tmalloc(strlen(t) + strlen(subname) + 4);

        terms += devs + 1;
        if (soft) terms--;
        *buffer = '\0';
        while (terms--)
            copytok1(buffer + strlen(buffer),&t);

        looksubmod(&t,buffer,subname);
        if (soft)
            looksubmod(&t,buffer,subname);
        (void) strcat(buffer, t);
        txfree(s->li_line);
        s->li_line = buffer;
    }
}


static void
looksubmod(t,buffer,subname)

char **t, *buffer, *subname;
{
    wordlist *wlsub;
    char *name;
    bool found = false;;

    if (**t == '\0') return;
    name = gettok(t);
    /* Now, is this a subcircuit model? */
    for (wlsub = submod; wlsub; wlsub = wlsub->wl_next)
        if (eq(name, wlsub->wl_word)) {
            (void) sprintf(buffer + strlen(buffer), "%s:%s ", subname, name);
            found = true;
            break;
        }
    if (!found)
        (void) sprintf(buffer + strlen(buffer), "%s ", name);
    txfree(name);
}







